//___________________________________________________________________
//
//                          Class BoxSizer
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/sizer.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/slider.h>
#include <wx/textctrl.h>
#include <wx/listbox.h>
#include <wx/listctrl.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"
#include "Pair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Float.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "Frame.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "BoxSizer.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Point.h"
#include "Graph.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BoxSizer::BoxSizer(BoxSizerNodeP boxSizerNode_, int orient_):
	wxBoxSizer(orient_),
	boxSizerNode(boxSizerNode_)
{}

BoxSizer::~BoxSizer()
{
	if (boxSizerNode)
		boxSizerNode->boxSizer=0;
}

wxString BoxSizer::wxstring()
{
	wxString str;
	str << "BoxSizer";
	return str;
}

wxDataInputStream &BoxSizer::read(wxDataInputStream &istr) 
{	
	return istr;
}
wxDataOutputStream &BoxSizer::write(wxDataOutputStream &ostr) 
{
	return ostr;
}

BoxSizerNode::BoxSizerNode(): 
	Node(BOXSIZER), boxSizer(0), children(0) {}

void BoxSizerNode::free() 
{	
	if (boxSizer) 
	{	boxSizer->boxSizerNode=0;
		boxSizer=0;
	}
}

void BoxSizerNode::inc_gc() {fifo << children;}
int BoxSizerNode::youngs() {return gmark(children);}
int BoxSizerNode::youngsPtr() {return gmarkPtr(children);}
void BoxSizerNode::check_oldyoungs() {check_oldyoung(children);}

wxDataInputStream &BoxSizerNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	istr >> orient >> (int&)children;
		boxSizer=new BoxSizer(BoxSizerNodeP(this), orient);
		boxSizer->read(istr);
	}
	return istr;
}
wxDataOutputStream &BoxSizerNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (boxSizer)
	{	ostr << 1 << orient << GC_::ptrToInt->find(children);
		boxSizer->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void BoxSizerNode::setPtrs() 
{
	children = PairP((Pair*)GC_::intToPtr->find(int(children)));
}

void BoxSizerNode::aftersetPtrs2() 
{
	PairP list=children;
	while (list)
	{	wxWindow *win=0;
		wxSizer *sizer=0;
		bigSwitch(list->first(), win, sizer);
		next(list);
		int proportion=IntP(list->first())->int_;
		next(list);
		int    flag   =IntP(list->first())->int_;
		next(list);
		int   border  =IntP(list->first())->int_;
		next(list);
		if (sizer)
			boxSizer->Add(sizer, proportion, flag, border);
		else if (win)
			boxSizer->Add(win, proportion, flag, border);
		else
			win=win;	// dummy line for debugging
	}
}

StringP BoxSizerNode::string()
{
	if (boxSizer)
		return new String(boxSizer->wxstring());
	return _Null_Word->name();
}

wxString BoxSizerNode::wxstring()
{
	if (boxSizer)
		return boxSizer->wxstring();
	return wxEmptyString;
}

void BoxSizerNode::bigSwitch(NodeP arg, wxWindow *&win, wxSizer *&sizer)
{
	switch (arg->type)
	{
	case SLIDER:
		{	SliderNodeP s=SliderNodeP(arg);
			win=s->slider;									break;
		}
	case GAUGE:
		{	GaugeNodeP g=GaugeNodeP(arg);
			win=g->gauge;									break;
		}
	case INTCONTROL:
		{	IntControlNodeP i=IntControlNodeP(arg);
			win=i->intControl;							break;
		}
	case FLOATCONTROL:
		{	FloatControlNodeP f=FloatControlNodeP(arg);
			win=f->floatControl;							break;
		}
	case BUTTON:
		{	ButtonNodeP b=ButtonNodeP(arg);
			win=b->button;									break;
		}
	case TOGGLEBUTTON:
		{	ToggleButtonNodeP t=ToggleButtonNodeP(arg);
			win=t->toggleButton;							break;
		}
	case CHECKBOX:
		{	CheckBoxNodeP c=CheckBoxNodeP(arg);
			win=c->checkBox;								break;
		}
	case RADIOBUTTON:
		{	RadioButtonNodeP b=RadioButtonNodeP(arg);
			win=b->radioButton;							break;
		}
	case STATICTEXT:
		{	StaticTextNodeP s=StaticTextNodeP(arg);
			win=s->staticText;							break;
		}
	case TEXTCONTROL:
		{	TextControlNodeP tc=TextControlNodeP(arg);
			win=tc->textControl;							break;
		}
	case LISTBOX:
		{	ListBoxNodeP lb=ListBoxNodeP(arg);
			win=lb->listBox;								break;
		}
	case CHOICEBOX:
		{	ChoiceBoxNodeP cb=ChoiceBoxNodeP(arg);
			win=cb->choiceBox;							break;
		}
	case COMBOBOX:
		{	ComboBoxNodeP cb=ComboBoxNodeP(arg);
			win=cb->comboBox;								break;
		}
	case LISTCONTROL:
		{	ListControlNodeP lc=ListControlNodeP(arg);
			win=lc->listControl;							break;
		}
	case FRAME:
		{	FrameNodeP f=FrameNodeP(arg);
			win=f->frame;									break;
		}
	case GRAPH:
		{	GraphNodeP g=GraphNodeP(arg);
			win=g->graph;									break;
		}
	case BOXSIZER:
		{	BoxSizerNodeP b=BoxSizerNodeP(arg);
			sizer=b->boxSizer;							break;
		}
	default:
		{	error(BAD_DATA, arg, 2);
		}
	}
}

NodeP BoxSizerNode::bigArgSwitch(NodeP arg, 
											wxWindow *&win, wxSizer *&sizer)
{
	switch (arg->type)
	{
	case SLIDER:
		{	SliderNodeP s=arg->a_slider(2);
			if throwing return Unbound;
			win=s->slider;									break;
		}
	case GAUGE:
		{	GaugeNodeP g=arg->a_gauge(2);
			if throwing return Unbound;
			win=g->gauge;									break;
		}
	case INTCONTROL:
		{	IntControlNodeP i=arg->a_intControl(2);
			if throwing return Unbound;
			win=i->intControl;							break;
		}
	case FLOATCONTROL:
		{	FloatControlNodeP f=arg->a_floatControl(2);
			if throwing return Unbound;
			win=f->floatControl;							break;
		}
	case BUTTON:
		{	ButtonNodeP b=arg->a_button(2);
			if throwing return Unbound;
			win=b->button;									break;
		}
	case TOGGLEBUTTON:
		{	ToggleButtonNodeP t=arg->a_toggleButton(2);
			if throwing return Unbound;
			win=t->toggleButton;							break;
		}
	case CHECKBOX:
		{	CheckBoxNodeP c=arg->a_checkBox(2);
			if throwing return Unbound;
			win=c->checkBox;								break;
		}
	case RADIOBUTTON:
		{	RadioButtonNodeP b=arg->a_radioButton(2);
			if throwing return Unbound;
			win=b->radioButton;							break;
		}
	case STATICTEXT:
		{	StaticTextNodeP s=arg->a_staticText(2);
			if throwing return Unbound;
			win=s->staticText;							break;
		}
	case TEXTCONTROL:
		{	TextControlNodeP tc=arg->a_textControl(2);
			if throwing return Unbound;
			win=tc->textControl;							break;
		}
	case LISTBOX:
		{	ListBoxNodeP lb=arg->a_listBox(2);
			if throwing return Unbound;
			win=lb->listBox;								break;
		}
	case CHOICEBOX:
		{	ChoiceBoxNodeP cb=arg->a_choiceBox(2);
			if throwing return Unbound;
			win=cb->choiceBox;							break;
		}
	case COMBOBOX:
		{	ComboBoxNodeP cb=arg->a_comboBox(2);
			if throwing return Unbound;
			win=cb->comboBox;								break;
		}
	case LISTCONTROL:
		{	ListControlNodeP lc=arg->a_listControl(2);
			if throwing return Unbound;
			win=lc->listControl;							break;
		}
	case FRAME:
		{	FrameNodeP f=arg->a_frame(2);
			if throwing return Unbound;
			win=f->frame;									break;
		}
	case GRAPH:
		{	GraphNodeP g=arg->a_graph(2);
			if throwing return Unbound;
			win=g->graph;									break;
		}
	case BOXSIZER:
		{	BoxSizerNodeP b=arg->a_boxSizer(2);
			if throwing return Unbound;
			sizer=b->boxSizer;							break;
		}
	default:
		{	error(BAD_DATA, arg, 2);
		}
	}
	return Unbound;
}

NodeP lBoxSizer(PairPP arg)
{
	int orient = wxHORIZONTAL;
	{	int o=arg[1]->a_int(1);
		if throwing return Unbound;
		orient=o;	
	}
	BoxSizerNodeP boxSizerNode=new BoxSizerNode();
	BoxSizerP boxSizer=new BoxSizer(boxSizerNode, orient);
	boxSizerNode->boxSizer=boxSizer;
	boxSizerNode->orient=orient;
	return boxSizerNode;
}

NodeP lBoxSizerAdd(PairPP arg)
{
	BoxSizerNodeP boxSizer=arg[1]->a_boxSizer(1);
	if throwing return Unbound;
	if (!arg[2])
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	wxWindow *win=0;
	wxSizer *sizer=0;
	boxSizer->bigArgSwitch(arg[2], win, sizer);
	if throwing return Unbound;
	int proportion=arg[3]->a_int(3);
	if throwing return Unbound;
	int flag=arg[4]->a_int(4);
	if throwing return Unbound;
	int border=arg[5]->a_int(5);
	if throwing return Unbound;
	if (sizer)
		boxSizer->boxSizer->Add(sizer, proportion, flag, border);
	else
		boxSizer->boxSizer->Add(win, proportion, flag, border);
	PairP &c=boxSizer->children;
	if (!c)
		c=new Pair(arg[2], 0);
	else
		c=PairP(c->lPut(arg[2]));
	c=PairP(c->lPut(new Int(proportion)));
	c=PairP(c->lPut(new Int(flag)));
	c=PairP(c->lPut(new Int(border)));
	return Unbound;
}
