//___________________________________________________________________
//
//                           Class  Button
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/button.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Color.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "Button.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Point.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Button, wxButton)
	EVT_BUTTON	(-1, Button::OnClick	)
END_EVENT_TABLE()

Button::Button(ButtonNodeP buttonNode_):
	buttonNode(buttonNode_)
{
	wxWindow *parent=console;
	if (buttonNode->parentFrame)
		parent=buttonNode->parentFrame->frame;
	else if (buttonNode->parentGraph)
		parent=buttonNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			"Button");
}

Button::Button(ButtonNodeP buttonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxButton(parent, id, label, 
			pos, size, style, 
			validator, name),
onClick(onClick_),
buttonNode(buttonNode_)
{}

Button::~Button()
{
	if (buttonNode)
		buttonNode->button=0;
}

void Button::OnClick(wxCommandEvent &event)
{
	Event::fifo << new EventButton(this);
}

wxString Button::wxstring()
{
	wxString str;
	str << "Button [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << onClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] ";
	return str;
}

wxDataInputStream &Button::read(wxDataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style;
	
	istr >> label >> name >> (int&)onClick
		>> pos.x >> pos.y >> size.x >> size.y >> style;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	return istr;
}
wxDataOutputStream &Button::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(&*onClick)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle());
	return ostr;
}

PairP ButtonNode::onClick()
{
	if (button)
		return button->onClick;
	return 0;
}

void ButtonNode::setonClick(NodeP nd) 
{
	if (button)
	{	button->onClick = PairP(nd);  check_oldyoung(nd);
	}
}

void ButtonNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ButtonNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ButtonNode::ButtonNode(): 
	Node(BUTTON), button(0), 	
	parentFrame(0), 
	parentGraph(0) 
{}

ButtonNode::ButtonNode(FrameNodeP parentFrame_, 
							  GraphNodeP parentGraph_): 
	Node(BUTTON), button(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ButtonNode::free() 
{	
	if (button) 
	{	button->buttonNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		button->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tButton:
				if (((EventButton*)ev)->button==button)
					((EventButton*)ev)->button=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		button=0;
	}
}
void ButtonNode::inc_gc() {fifo << onClick();}
int ButtonNode::youngs() {return gmark(onClick());}
int ButtonNode::youngsPtr() {return gmarkPtr(onClick());}
void ButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

wxDataInputStream &ButtonNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode;
		button=new Button(ButtonNodeP(this));
		button->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &ButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (button)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph);
		button->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ButtonNode::setPtrs()
{
	parentFrame=FrameNodeP((FrameNode*)GC_::intToPtr->find(int(parentFrame)));
	parentGraph=GraphNodeP((GraphNode*)GC_::intToPtr->find(int(parentGraph)));
	if (button)
	{	setonClick(PairP((Pair*)GC_::intToPtr->find(int(button->onClick))));
	}
}

void ButtonNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ButtonP b=button, b2=0;
	wxPoint pos=b->GetPosition();
	b2=new Button(ButtonNodeP(this),
			p, -1, b->GetLabel(), 
			b->onClick, pos, b->GetSize(), b->GetWindowStyle(), 
			wxDefaultValidator, b->GetName());
	button=b2;
	setonClick(b->onClick);
	delete b;
	button=b2;
}

StringP ButtonNode::string()
{
	if (button)
		return new String(button->wxstring());
	return _Null_Word->name();
}

wxString ButtonNode::wxstring()
{
	if (button)
		return button->wxstring();
	return wxEmptyString;
}

NodeP lButton(PairPP arg)
{
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=label;
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
	ButtonNodeP buttonNode=new ButtonNode(
		parentFrameNode, 
		parentGraphNode);
	ButtonP button=new Button(buttonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	buttonNode->button=button;
	return buttonNode;
}

NodeP lButtonOnClick(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	button->setonClick(function);
	return Unbound;
}

NodeP lButtonEnable(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	button->button->Enable(state!=0);
	return Unbound;
}

NodeP lButtonDestroy(PairPP arg)
{
	ButtonNodeP button=arg[1]->a_button(1);
	if throwing return Unbound;
	{	FrameNodeP p=button->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(button));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=button->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(button));
			p->check_oldyoung(p->children);
		}
	}
	button->button->Destroy();
	return Unbound;
}
