//___________________________________________________________________
//
//                           Class ChoiceBox
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/listbox.h>
#include <wx/choice.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "ChoiceBox.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyChoiceBox : public wxChoice
{
public:
	MyChoiceBox(wxWindow* parent, wxWindowID id, 
		const wxArrayString& choices, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = "ChoiceBox");

	// MSW sends extra kill focus event
	virtual ~MyChoiceBox()
	{
		if (parent) 
			parent->choiceBox = 0;
		parent=0;
	}

	ChoiceBox *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to ChoiceBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void Set(wxArrayString &choices);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyChoiceBox,wxChoice)
	EVT_CHAR		(MyChoiceBox::OnChar		)
	EVT_KEY_DOWN(MyChoiceBox::OnKeyDown	)
	EVT_KEY_UP	(MyChoiceBox::OnKeyUp		)
	EVT_CHOICE	(-1,	MyChoiceBox::OnSelect	)
END_EVENT_TABLE()

MyChoiceBox::MyChoiceBox(wxWindow *parent_, wxWindowID id,
							  const wxArrayString &choices_,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxChoice(parent_, id, pos, size, choices_, style, validator, name)
{
	parent = (ChoiceBox*)parent_;
}

void MyChoiceBox::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyChoiceBox::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyChoiceBox::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyChoiceBox::OnSelect(wxCommandEvent &event)
{
	if (parent) 
		parent->OnSelect(event);
}

void MyChoiceBox::Set(wxArrayString &choices)
{
	Clear();
	int nr=choices.GetCount();
	for (int i=0; i<nr; i++)
		Append(choices[i]);
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ChoiceBox, wxControl)
	EVT_CHAR				(		ChoiceBox::OnChar		)
	EVT_KEY_DOWN		(		ChoiceBox::OnKeyDown	)
	EVT_KEY_UP			(		ChoiceBox::OnKeyUp		)
	EVT_CHOICE			(-1,	ChoiceBox::OnSelect		)
END_EVENT_TABLE()

ChoiceBox::ChoiceBox(ChoiceBoxNodeP choiceBoxNode_, int style):
label(0), choiceBox(0), bestSize(-999, -999),
choices(0),
onSelect(0), onChar(0), onKeyUp(0), onKeyDown(0), 
choiceBoxNode(choiceBoxNode_)
{
	wxWindow *parent=console;
	if (choiceBoxNode->parentFrame)
		parent=choiceBoxNode->parentFrame->frame;
	else if (choiceBoxNode->parentGraph)
		parent=choiceBoxNode->parentGraph->graph;
	create(parent, -1, wxArrayString(), 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, "ChoiceBox");
}

ChoiceBox::ChoiceBox(ChoiceBoxNodeP choiceBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			PairP onSelect_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), choiceBox(0), bestSize(-999, -999), choices(choices_),
onSelect(onSelect_), onChar(0), onKeyUp(0), onKeyDown(0), 
choiceBoxNode(choiceBoxNode_)
{
	bestSize = wxSize(-999, -999);
	choiceBoxNode->choiceBox=this;
	choiceBoxNode->style=style;
	create(parent, id, choices, pos, size, style, validator, name);
}

bool ChoiceBox::create(wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		choiceBox = new MyChoiceBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		choiceBox = new MyChoiceBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	choices.Clear();

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void ChoiceBox::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (choiceBox)   
		choiceBox->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ChoiceBox::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxChoice *t=new wxChoice((wxWindow*)this, wxID_ANY, 
			wxPoint(0,0), wxDefaultSize, choices, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ChoiceBox::~ChoiceBox()
{
	if (choiceBoxNode)
		choiceBoxNode->choiceBox=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (choiceBox)
	{	delete choiceBox;
		choiceBox=0;
	}

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::ChoiceBoxSelect:
		case Event::ChoiceBoxChar:
		case Event::ChoiceBoxKeyDown:
		case Event::ChoiceBoxKeyUp:
			if (((EventChoiceBox*)ev)->choiceBox==this)
				((EventChoiceBox*)ev)->choiceBox=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void ChoiceBox::Set(wxArrayString &choices)
{
	if (choiceBox)
		choiceBox->Set(choices);
}

#define LB_ON(NAME)																\
void ChoiceBox::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventChoiceBox##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ChoiceBox::OnSelect(wxCommandEvent &event)
{
	Event::fifo << new EventChoiceBoxSelect(this);
}

wxString ChoiceBox::wxstring()
{
	wxString str;
	str << "ChoiceBox " 
		<< '[' << GetName() << "] ";
	if (onSelect)
		str << onSelect->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &ChoiceBox::read(wxDataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name 
		>> (int&)onSelect 
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int nr;
	istr >> nr;
	choices.Clear();
	for (int i=0; i<nr; i++)
	{	wxString item;
		istr >> item;
		choices.Add(item);
	}
	choiceBox->Set(choices);

	istr >> nr;
	choiceBox->SetSelection(nr);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ChoiceBox::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetName() 
		<< GC_::ptrToInt->find(onSelect)
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=choiceBox->GetForegroundColour(), 
			b=choiceBox->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=choiceBox->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int nr=choiceBox->GetCount();
	ostr << nr;
	for (int i=0; i<nr; i++)
		ostr << choiceBox->GetString(i);

	wxArrayInt sel;
	nr=choiceBox->GetSelection();
	ostr << nr;
	return ostr;
}

bool ChoiceBox::SetBackgroundColour(const wxColour &c)
{
	if (choiceBox)
		return choiceBox->SetBackgroundColour(c);
	return false;
}

bool ChoiceBox::SetForegroundColour(const wxColour &c)
{
	if (choiceBox)
		return choiceBox->SetForegroundColour(c);
	return false;
}

bool ChoiceBox::SetFont(const wxFont &f)
{
	if (choiceBox)
		return choiceBox->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)							\
PairP ChoiceBoxNode::on##x()					\
{													\
	if (choiceBox)								\
		return choiceBox->on##x;				\
	return 0;									\
}													\
													\
void ChoiceBoxNode::seton##x(NodeP nd)	\
{													\
	if (choiceBox)								\
	{	choiceBox->on##x = PairP(nd);		\
		check_oldyoung(nd);					\
	}												\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Select)

void ChoiceBoxNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ChoiceBoxNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ChoiceBoxNode::ChoiceBoxNode(): 
	Node(CHOICEBOX), choiceBox(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

ChoiceBoxNode::ChoiceBoxNode(FrameNodeP parentFrame_, 
											GraphNodeP parentGraph_): 
	Node(CHOICEBOX), choiceBox(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ChoiceBoxNode::free() 
{	
	if (choiceBox) 
	{	choiceBox->choiceBoxNode=0;
		choiceBox->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::ChoiceBoxSelect:
			case Event::ChoiceBoxChar:
			case Event::ChoiceBoxKeyDown:
			case Event::ChoiceBoxKeyUp:
				if (((EventChoiceBox*)ev)->choiceBox==choiceBox)
					((EventChoiceBox*)ev)->choiceBox=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		choiceBox=0;
	}
}
void ChoiceBoxNode::inc_gc() 
{	fifo << onSelect()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ChoiceBoxNode::youngs() 
{	return gmark(onSelect())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ChoiceBoxNode::youngsPtr() 
{	return gmarkPtr(onSelect())
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp());
}
void ChoiceBoxNode::check_oldyoungs() 
{	check_oldyoung(onSelect());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

wxDataInputStream &ChoiceBoxNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		choiceBox=new ChoiceBox(ChoiceBoxNodeP(this), style);
		choiceBox->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &ChoiceBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (choiceBox)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph)
			<< style;
		choiceBox->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ChoiceBoxNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (choiceBox)
	{	ChoiceBoxP t=choiceBox;
		setonSelect ((Pair*)GC_::intToPtr->find(int(t->onSelect )));
		setonChar   ((Pair*)GC_::intToPtr->find(int(t->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(t->onKeyUp  )));
	}
}

void ChoiceBoxNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ChoiceBoxP t=choiceBox, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ChoiceBox(ChoiceBoxNodeP(this),
			p, -1, t->choices, 
			t->onSelect, pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	choiceBox=t2;
	wxArrayInt sel;
	int nr=t->choiceBox->GetSelection();
	t2->choiceBox->SetSelection(nr);
	setonSelect (t->onSelect );
	setonChar   (t->onChar   );
	setonKeyDown(t->onKeyDown);
	setonKeyUp  (t->onKeyUp  );
	t2->choiceBox->SetForegroundColour(t->choiceBox->GetForegroundColour());
	t2->choiceBox->SetBackgroundColour(t->choiceBox->GetBackgroundColour());
	wxFont f=t->choiceBox->GetFont();
	t2->choiceBox->SetFont(f);
	delete t;
	choiceBox=t2;
}

StringP ChoiceBoxNode::string()
{
	if (choiceBox)
		return new String(choiceBox->wxstring());
	return _Null_Word->name();
}

wxString ChoiceBoxNode::wxstring()
{
	if (choiceBox)
		return choiceBox->wxstring();
	return wxEmptyString;
}

NodeP lChoiceBox(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("ChoiceBox");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxArrayString choices;
	{	PairP list=arg[3]->a_list(3);
		if throwing return Unbound;
		while (list)
		{	StringP s=list->first()->a_string(3); 
			if throwing return Unbound;
			choices.Add(s->ptr());
			next(list);
		}
	}
	PairP onSelect=0;
	if (exist(4))
	{	PairP os=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onSelect=os;
	}
	long style = 0;
	if (exist(5))
	{	int s=arg[5]->a_int(5);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(6))
	{	PairP p=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(7))
	{	PairP s=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	ChoiceBoxNodeP choiceBoxNode=new ChoiceBoxNode(
		parentFrameNode, 
		parentGraphNode);
	new ChoiceBox(choiceBoxNode,
			parent, id, choices, 
			onSelect, 
			pos, size, style, wxDefaultValidator, name);
	return choiceBoxNode;
}

NodeP lChoiceBoxSelection(PairPP arg)
{
	ChoiceBox *choiceBox=0;
	if (exist(1))
	{	ChoiceBoxNodeP choiceBoxNode=arg[1]->a_choiceBox(1);
		if throwing return Unbound;
		choiceBox=choiceBoxNode->choiceBox;
	}
	else
	{	choiceBox=Event::choiceBoxValue;
	}
	if (!choiceBox) return 0;
	wxArrayInt sel;
	int nr=choiceBox->choiceBox->GetSelection();
	return new Int(nr);
}

NodeP lChoiceBoxSetSelection(PairPP arg)
{
	ChoiceBoxNodeP choiceBoxNode=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	ChoiceBox *choiceBox=choiceBoxNode->choiceBox;
	if (!choiceBox) return 0;
	int sel=arg[2]->a_int(2);
	if throwing return Unbound;
	int nr=choiceBox->choiceBox->GetCount();
	if (sel < 0 || sel >= nr)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	choiceBox->choiceBox->SetSelection(sel);
	return Unbound;
}

NodeP lChoiceBoxSetChoices(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	wxArrayString choices;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	while (list)
	{	StringP s=list->first()->a_string(2);
		if throwing return Unbound;
		choices.Add(s->ptr());
		next(list);
	}
	choiceBox->choiceBox->choiceBox->Set(choices);
	return Unbound;
}

NodeP lChoiceBoxAppend(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	choiceBox->choiceBox->choiceBox->Append(value->ptr());
	return Unbound;
}

NodeP lChoiceBoxSetItem(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= choiceBox->choiceBox->choiceBox->GetCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	StringP str=arg[3]->a_string(3);
	if throwing return Unbound;
	choiceBox->choiceBox->choiceBox->SetString(n, str->ptr());
	return Unbound;
}

NodeP lChoiceBoxRemoveItem(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= choiceBox->choiceBox->choiceBox->GetCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	choiceBox->choiceBox->choiceBox->Delete(n);
	return Unbound;
}

NodeP lChoiceBoxCount(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int count=choiceBox->choiceBox->choiceBox->GetCount();
	return new Int(count);
}

#define _TC_ON(NAME)									\
NodeP lChoiceBoxOn##NAME(PairPP arg)			\
{															\
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	choiceBox->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Select)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lChoiceBoxSetBackgroundColor(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);

	choiceBox->choiceBox->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lChoiceBoxSetColor(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	choiceBox->choiceBox->SetForegroundColour(wc);
	return Unbound;
}

NodeP lChoiceBoxSetFontSize(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetPointSize(s);
	choiceBox->choiceBox->SetFont(f);
	return Unbound;
}

NodeP lChoiceBoxSetFontName(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetFaceName(s->ptr());
	choiceBox->choiceBox->SetFont(f);
	return Unbound;
}

NodeP lChoiceBoxSetFontStyle(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetStyle(s);
	choiceBox->choiceBox->SetFont(f);
	return Unbound;
}

NodeP lChoiceBoxSetFontWeight(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=choiceBox->choiceBox->GetFont();
	f.SetWeight(s);
	choiceBox->choiceBox->SetFont(f);
	return Unbound;
}

NodeP lChoiceBoxEnable(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	choiceBox->choiceBox->Enable(state!=0);
	return Unbound;
}

NodeP lChoiceBoxDestroy(PairPP arg)
{
	ChoiceBoxNodeP choiceBox=arg[1]->a_choiceBox(1);
	if throwing return Unbound;
	{	FrameNodeP p=choiceBox->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(choiceBox));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=choiceBox->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(choiceBox));
			p->check_oldyoung(p->children);
		}
	}
	choiceBox->choiceBox->Destroy();
	return Unbound;
}
