//___________________________________________________________________
//
//                         ComboBox Header
//___________________________________________________________________


extern Function 
	lComboBox, lComboBoxDestroy,
	lComboBoxSelection, lComboBoxSetSelection,
	lComboBoxSetChoices, 
	lComboBoxAppend, 
	lComboBoxSetItem, lComboBoxRemoveItem, 
	lComboBoxCount,
	lComboBoxValue, lComboBoxSetValue,
	lComboBoxSetBackgroundColor, lComboBoxSetColor,
	lComboBoxSetFontSize, lComboBoxSetFontName,
	lComboBoxSetFontStyle, lComboBoxSetFontWeight,
	lComboBoxOnChar, lComboBoxOnKeyDown, lComboBoxOnKeyUp,
	lComboBoxOnSelect, lComboBoxOnChange, lComboBoxOnEnter,
	lComboBoxEnable;

class wxStaticText;
class MyComboBox;

class ComboBox: public wxControl
{
	public:
		wxStaticText *label;
		MyComboBox *comboBox;
		wxSize bestSize;
		wxString value;
		wxArrayString choices;

		PairP onSelect, onChange, onEnter, onChar, onKeyUp, onKeyDown;
		ComboBoxNodeP comboBoxNode;

		ComboBox(ComboBoxNodeP comboBoxNode_, int style);
		ComboBox(ComboBoxNodeP comboBoxNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			PairP onSelect_=PairP(0),
			PairP onChange_=PairP(0),
			PairP onEnter_=PairP(0),
			const wxPoint &pos=wxDefaultPosition, 
			const wxSize &size=wxDefaultSize, 
			long style = 0, 
			const wxValidator &validator = wxDefaultValidator, 
			const wxString &name = "ComboBox");

		bool create(wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			const wxPoint &pos, 
			const wxSize &size, 
			long style, 
			const wxValidator &validator, 
			const wxString &name);

		~ComboBox();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		virtual void Set(wxArrayString &choices);
		virtual void SetValue(const wxString &value);
		virtual wxString GetValue() const;

		void 
			OnClose	(wxCloseEvent &event),
			OnSelect	(wxCommandEvent &event),
			OnChange	(wxCommandEvent &event),
			OnEnter	(wxCommandEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef ComboBox *ComboBoxP;

class ComboBoxNode: public Node
{
	public:
		ComboBoxP comboBox;
		FrameNodeP parentFrame;
		GraphNodeP parentGraph;
		int style;

		static const ComboBoxNode aComboBoxNode;

		operator_new;
		PairP onSelect(), onChange(), onEnter(), 
			onChar(), onKeyDown(), onKeyUp();
		void
			setonSelect	(NodeP function),
			setonChange	(NodeP function),
			setonEnter	(NodeP function),
			setonChar	(NodeP function), 
			setonKeyDown(NodeP function),
			setonKeyUp	(NodeP function);

		ComboBoxNode();
		ComboBoxNode(FrameNodeP parentFrame_, GraphNodeP parentGraph_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();
		
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		StringP _typeof();

		void setParentFrame(FrameNodeP frame);
		void setParentGraph(GraphNodeP graph);
};
