//___________________________________________________________________
//
//                       Evaluator Header
//___________________________________________________________________

#ifndef Node_included
	#include "Node.h"
#endif

#ifdef _DEBUG
//	#define PROC_EVAL
#endif

#define Eval_Labels(x)					\
	x( fetch_command				)		\
	x( begin_line					)		\
	x( begin_sequence				)		\
	x( end_line						)		\
	x( tail_dispatch				)		\
	x( dispatch						)		\
	x( proc_with_args				)		\
	x( arg_loop						)		\
	x( accumulate_arg				)		\
	x( args_done					)		\
	x( apply_dispatch				)		\
	x( compound_apply				)		\
												\
		x( lambda_apply			)		\
			x( begin_formal_loop	)		\
				x( set_args_continue	)	\
			x( set_args_continue2)		\
		x( end_formal_loop		)		\
		x( after_formals			)		\
												\
	x( sequence						)		\
	x( non_tail_eval				)		\
	x( sequence_continue			)		\
	x( compound_apply_continue	)		\
	x( macro_return				)		\
												\
	x( runresult_continuation	)		\
	x( runresult_followup		)		\
	x( repeat_continuation		)		\
	x( repeat_again				)		\
	x( repeat_followup			)		\
	x( catch_continuation		)		\
	x( catch_followup				)		\
	x( goto_continuation			)		\
												\
	x( begin_apply					)		\
	x( questionmark_continue	)		\
	x( questionmark_failed		)		\
	x( after_lambda				)		\
	x( end							)		\

#define create_enum(x) x,

enum EvalCommand
{
	Eval_Labels(create_enum)
	eval_labels
};

class __single_inheritance Evaluator;

typedef (Evaluator::*Do)();

struct ToDo
{
#ifdef PROC_EVAL
	#ifdef _DEBUG
		EvalCommand mnemonic;
	#endif
		union 
		{	Do instruction;
			int i;
		};
#else
		EvalCommand mnemonic;
#endif
};

#ifdef _DEBUG
	#undef new
	#define new new
#endif


#ifdef PTR_DEBUG
	class ToDoNode;	typedef NPtr<ToDoNode> ToDoNodeP;
#else
	class ToDoNode;	typedef ToDoNode* ToDoNodeP;
#endif


class ToDoNode: public Node		// is a Do Evaluator function and a ToDoStack pointer
{
	NodeP _tail;

	public:
		ToDo toDo;
		static const ToDoNode ToDoNode::aToDoNode;
		operator_new;

		NodeP tail() {return _tail;}
		void settail(NodeP nd) {_tail=nd;  check_oldyoung(nd);}

		ToDoNode(): Node(TODO) {}
		ToDoNode(ToDo toDo_, NodeP tail_): Node(TODO)
			{toDo = toDo_;   settail(tail_);}

		void inc_gc() {fifo << tail();}
		int youngs() {return gmark(tail());}
		void check_oldyoungs() {check_oldyoung(tail());}
};

#ifdef _DEBUG
	#undef new
#endif

enum TailCall
{	tailcall_arg=-1,
	in_sequence,  // 0
	tailcall_tail,  // 1
};

enum ValStatus 
{	no_value_allowed,  // body of command = 0
	value_required,    // argument = 1
	output_ok,         // body of function = 2
	dont_care,       // function inside catch = 3
	no_value_in_macro,  // repeat = 4
	value_maybe_ok_in_macro  // catch = 5
};

class NodePStack;
class IntStack;
class ToDoStack;
class VarStack;
class VarPStack;
class PairPPStack;
class ErrorStack;

class Evaluator
{
	bool tracing;  // are we tracing the current procedure? 
	int old_ift_iff;
	TailCall oldtailcall;  // in case of reentrant use of Evaluator 

	NodeP		// Registers 
		val,     // the value of the last expression 
		catch_tag;

	PairP
		nodeList,    // flat list of expressions to evaluate
		unev,       // list of unevaluated expressions
		exp;       // the current expression 

	NodeP proc; // the procedure definition

	PairP
		formals,  // list of formal parameters 
		parm,    // the current formal 
		arg;    // the current actual 

	public:
		Evaluator(PairP nodeList_, ToDo where);

#ifdef PROC_EVAL
	#define create_function(x)	_##x(),

	int Eval_Labels(create_function)		dummy;
#endif
		static bool dont_fix_ift;
		static int ift_iff_flag,
			repcount, user_repcount,	// counters for repeat 
			trace_level;	// indentation level when tracing

		static TailCall tailcall;
		static ValStatus val_status;

		static   NodeP last_call;	// the last proc called
		static ObjectP last_ufun;	// the function that called this one 
		static   PairP last_line;	// the line that called this one 
		static    VarP       var;	// frame pointer into varStack
		static ErrorStructP didnt_get_output;	// the name of the proc that wanted the OP

		static PairPP asp;		// argStack pointer
		static   VarP vsp;	  // varStack pointer

		static  NodePStack      stack;	// eval register stack
		static    IntStack   intStack;	// integer stack
		static   ToDoStack  toDoStack;	// instruction stack 
		static    VarStack   varStack;	// stack for storing global objects while local vars are bound
		static   VarPStack  varPStack;	// stack for pointers into varStack
		static PairPPStack  argPStack;	// stack for pointers into argStack
		static  ErrorStack errorStack;	// error stack

		static const char commandName[][eval_labels];

	friend class Pair;
};

#ifdef PROC_EVAL
	#ifdef _DEBUG
		#define create_ToDo(x)	__##x={x, Evaluator::_##x},
	#else
		#define create_ToDo(x)	__##x={Evaluator::_##x},
	#endif
#else
	#define create_ToDo(x)	__##x={x},
#endif

const ToDo Eval_Labels(create_ToDo)__;

extern NodeP didnt_output_name;
extern PairP output_node;

extern ObjectP fun, ufun;
extern   PairP this_line;

extern PairPP qmList;