#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/utils.h>
#include <wx/app.h>
#include <wx/textctrl.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
	#include <process.h>
	#include <direct.h>
#else
	#include <unistd.h>
#endif
	
#include "Main.h"

#include "String.h"
#include "File.h"
#include "List.h"
#include "Tree.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Console.h"
#include "LogoApp.h"

#ifdef WIN32
	#include "Piper.h"
#endif
	
#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

NodeP lignore(PairPP)
{
	return Unbound;
}

NodeP lisCaseIgnored(PairPP)
{
	return Node::boolnode(String::caseIgnored);
}

NodeP lsetCaseIgnored(PairPP arg)
{
	int ci=arg[1]->a_bool(1);
	if (!throwing)
	{	String::caseIgnored=(ci==1);
	}
	return Unbound;
}

NodeP lstop(PairPP) 
{
	if (!throwing) stoppingFlag=STOP;
	return Unbound;
}

NodeP loutput(PairPP arg) 
{
	if (!throwing) {stoppingFlag=OUTPUT; output_node=arg[1];}
	return Unbound;
}

NodeP lrun(PairPP arg) // macroized
{    
	NodeP instructions = arg[1]->a_runnable(1);
	return instructions;
}

NodeP lrunResult(PairPP arg) 
{	
	return new ToDoNode(__runresult_continuation, lrun(arg));
}
NodeP lapply(PairPP arg)	// The logo word APPLY. 
{
	return new ToDoNode(__begin_apply, new Pair(arg[1], arg[2]));
}
NodeP lgoTo(PairPP arg) 
{
	return new ToDoNode(__goto_continuation, arg[1]);
}
NodeP ltag(PairPP) {return Unbound;}

NodeP lcatch(PairPP arg) 
{	
	return new ToDoNode(__catch_continuation, new Pair(arg[1], lrun(arg+1)));
}

NodeP lthrow(PairPP arg) 
{
	if (!throwing)
	{
		if (arg[1]->equal(_Error)) 
		{	if (exist(2))
				error(USER_ERR, arg[2]);
			else
				error(USER_ERR, Unbound);
		}else 
		{	stoppingFlag=THROWING;
			throw_node=arg[1];
		
			if (exist(2))
				output_node=arg[2];
			else
				output_node=(PairP)Unbound;
		}
	}
	return Unbound;
}

NodeP lquestionmark(PairPP arg)		// The logo word ?
{
	if (noarg) 
		if (qmList>=0) return argStack.base[qmList+1];
		else return error(BAD_DATA_UNREC, fun);
	else
	{	int argnum = arg[1]->a_pos_int(1)->int_;

		if (throwing) return Unbound;

		if (qmList<0)
			return error(BAD_DATA, new Int(argnum));

		PairPP np = PairPP(&argStack.base[qmList+argnum]);
//		PairPP np = &qmList[1];
//		int i=argnum;
//		while (--i > 0  &&  np!=argStack  &&  *++np)
//			;
		if (!argnum || unsigned(np)>=unsigned(argStack.index) || !*np)
			return error(BAD_DATA_UNREC, new Int(argnum));

		return *np;
	}
}

NodeP ltest(PairPP arg) 
{
	int b = arg[1]->a_bool(1);

	if (throwing || Evaluator::tailcall != in_sequence) 
		return Unbound;

	Evaluator::ift_iff_flag=b;
	Evaluator::dont_fix_ift=true;
	return Unbound;
}

NodeP lifTrue(PairPP arg) 
{
	if (Evaluator::ift_iff_flag < 0)
		return error(NO_TEST, 0);
	else
		if (Evaluator::ift_iff_flag > 0)
			return lrun(arg);
		else
			return 0;
}

NodeP lifFalse(PairPP arg) 
{
	if (Evaluator::ift_iff_flag < 0)
		return error(NO_TEST, 0);
	else
		if (Evaluator::ift_iff_flag==0)
			return lrun(arg);
		else
			return 0;
}

NodeP lifElse(PairPP arg)  // macroized
{    
	int cond = arg[1]->a_bool(1);
	NodeP
		yes= arg[2]->a_runnable(2),
		no = arg[3]->a_runnable(3);

	if (throwing) return Unbound;
	return cond? yes : no;
}

NodeP lif(PairPP arg) 	// macroized
{
	if (exist(3)) return lifElse(arg);
	
	 int pred = arg[1]->a_bool(1);
	NodeP yes = arg[2]->a_runnable(2);

	if throwing return Unbound;
	return pred? yes : 0;
}

NodeP lcase(PairPP arg)
{
	NodeP value=arg[1];
	PairP conds=arg[2]->a_list(2);
	if (!conds)
		return 0;
	if throwing return Unbound;

	Evaluator e;
	e.val_status=value_maybe_ok_in_macro;

	for (PairP l=conds;  l;  next(l))
	{	if (l->first())
		{	if (l->first()->isnolist())
			{	error(BAD_DATA, l, 2);
				return Unbound;
			}
			PairP cond=l->first()->first();
			if (!cond)
			{	error(BAD_DATA, l, 2);
				return Unbound;
			}
			if (cond->equal(_Else))
				return l->first()->tail();

			cond=cond->a_runnable(2);
			if throwing return Unbound;
			NodeP caseval;
			caseval=e.eval(cond, __begin_sequence);


			if (caseval && caseval->equal(value))
				return l->first()->tail();
		}
		else if (value==NodeP(0))
			return l->tail();
	}
	return 0;
}

NodeP lforever(PairPP arg) 
{
	NodeP instructionlist = arg[1]->a_runnable(1);
	if throwing return 0;
	return new ToDoNode
	(	__repeat_continuation, 
		new Pair( new Int(-1), instructionlist)
	);
}

NodeP lrepeat(PairPP arg) 
{
	IntP count = arg[1]->a_pos_int(1);
	PairP instructionlist = arg[2]->a_runnable(2);

	if (throwing || !count->int_) return 0;
	return new ToDoNode
	(	__repeat_continuation, 
		new Pair(count, instructionlist)
	);
}

NodeP lrepeatCount(PairPP)
{
	return new Int(Evaluator::user_repcount);
}

NodeP lwhile(PairPP arg) 
{
	PairP         cond = arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);

	return new ToDoNode
	(	__while_continuation, 
		new Pair(cond, instructionlist)
	);
}

NodeP luntil(PairPP arg) 
{
	PairP         cond = arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);

	return new ToDoNode
	(	__until_continuation, 
		new Pair(cond, instructionlist)
	);
}

NodeP ldo_while(PairPP arg) 
{
	PairP
		instructionlist = arg[1]->a_runnable(1),
					  cond = arg[2]->a_runnable(2);

	return new ToDoNode
	(	__do_while_continuation, 
		new Pair(cond, instructionlist)
	);
}

NodeP ldo_until(PairPP arg) 
{
	PairP
		instructionlist = arg[1]->a_runnable(1),
					  cond = arg[2]->a_runnable(2);

	return new ToDoNode
	(	__do_until_continuation, 
		new Pair(cond, instructionlist)
	);
}

NodeP lfor(PairPP arg)	// What a messy primitive! But faster than for_old
{
	PairP l=arg[1]->a_runnable(1),
		instructionlist = arg[2]->a_runnable(2);

	if throwing return 0;
	if (!l 
	|| !l->first() || !l->tail() 
	|| !l->second() || !l->tail()->tail()
	|| !l->third())
	{	error(BAD_DATA, l);
		return 0;
	}
	NodeP name=l->first()->a_name(1);
	if throwing return 0;

	PairP start=l->second(),
		end=l->third(),
		step=0;

	if (start->isnolist())
		start=new Line(start, 0);

	if (end->isnolist())
		end=new Line(end, 0);

	double dstart, dend, dstep;

	Evaluator e;
	NodeP startval, endval, stepval;
	e.val_status=value_maybe_ok_in_macro;

	startval=e.eval(start, __begin_sequence);
	dstart=startval->a_float(1);
	if throwing return 0;

	  endval=e.eval( end , __begin_sequence);
	dend=endval->a_float(1);
	if throwing return 0;

	if (l->tail()->tail()->tail() && l->third())
	{	step=l->fourth();
		if (step->isnolist())
			step=new Line(step, 0);
		stepval=e.eval(step, __begin_sequence);
		if throwing return 0;
		dstep=stepval->a_float(1);
	}
	else
		dstep= dend-dstart >= 0 ? 1 : -1;

	if ((dend-dstart > 0  &&  dstep < 0)
	||  (dend-dstart < 0  &&  dstep > 0))
		return 0;

	if (dstart==int(dstart) && dend==int(dend) && dstep==int(dstep))
		return new ToDoNode
		(	__ifor_continuation,

			new Pair(name,
			new Pair(new Int(int(dstart)), 
			new Pair(new Int(int(dend)), 
			new Pair(new Int(int(dstep)), 
			new Pair(instructionlist, 0)))))
		);

	return new ToDoNode
	(	__for_continuation, 
		new Pair(name,
		new Pair(new Float(dstart), 
		new Pair(new Float(dend), 
		new Pair(new Float(dstep), 
		new Pair(instructionlist, 0)))))
	);
}

NodeP lforeach(PairPP arg)
{
	PairP list=0;
	NodeP instructions=*argStack.index;	//->a_runnable(2);
	if (!instructions)
		return 0;

	NodePP i=argStack.index-1;
	if (!*i)
		return 0;

	while (i != NodePP(arg))
		push(list, *i--);

	return new ToDoNode
	(	__foreach_continuation,
		new Pair(list, instructions)
	);
}

NodeP lquestionmarkRest(PairPP arg)
// equivalent to:
// to ?rest [:which 1]
// output bf item :which :template_lists
// end
{
	PairP var=PairP(_Template_Lists->var());
	if (noarg)
	{	if (var 
		&& var!=Unbound
		&& var->first())
			return var->first()->butFirst();
	}else
	{	int i=arg[1]->a_int(1);
		if throwing return Unbound;
		if (var
		&& var!=Unbound
		&& var->item(i))
			return PairP(var->item(i))->butFirst();
	}
	return 0;
}

NodeP lbreak(PairPP)
{
	stoppingFlag=BREAK;
	return Unbound;
}

NodeP lcontinueLoop(PairPP)
{
	stoppingFlag=CONTINUE;
	return Unbound;
}

NodeP lbye(PairPP) 
{
	if (!ufun  &&  loader.file()==stdin) 
		c_out << "Thank you for using Logo." << endL
			  << "Have a nice day." << endL;

	theApp->action=EXIT_LOGO;
	return Unbound;
}

NodeP lchangeDir(PairPP arg)
{
	StringP path=arg[1]->a_string(1);
	if (!throwing)
	{
#ifdef WIN32
		if (_chdir(path->ptr()))
#else
		if (chdir(path->ptr()))
#endif

		{	error(FILE_ERROR, path);
			return Unbound;
		}
	}
	return Unbound;
}

NodeP lLogoVersion(PairPP)
{
	StringP os=0;
	switch (theApp->osVersion)
	{
case wxCURSES:		os=new String("Curses"); break;				//  Text-only CURSES 
case wxXVIEW_X:		os=new String("XView X"); break;				//  Sun's XView OpenLOOK toolkit 
case wxMOTIF_X:		os=new String("Motif X"); break;				//  OSF Motif 1.x.x 
case wxCOSE_X:		os=new String("COSE"); break;					//  OSF Common Desktop Environment 
case wxNEXTSTEP:		os=new String("NeXTStep"); break;			//  NeXTStep 
case wxMAC:			os=new String("Mac OS 8/9/X with Mac paths"); break;	//  Apple Mac OS 8/9/X with Mac paths 
case wxMAC_DARWIN:	os=new String("Mac OS X with Unix paths"); break;		//  Apple Mac OS X with Unix paths 
case wxBEOS:			os=new String("BeOS"); break;					//  BeOS 
case wxGTK:			os=new String("GTK on X"); break;			//  GTK on X 
case wxGTK_WIN32:	os=new String("GTK on Win32"); break;		//  GTK on Win32 
case wxGTK_OS2:		os=new String("GTK on OS/2"); break;		//  GTK on OS/2 
case wxGTK_BEOS:		os=new String("GTK on BeOS"); break;		//  GTK on BeOS 
case wxGEOS:			os=new String("GEOS"); break;					//  GEOS 
case wxOS2_PM:		os=new String("OS/2 Workplace"); break;	//  OS/2 Workplace 
case wxWINDOWS:		os=new String("Windows or WfW"); break;	//  Windows or WfW 
case wxMICROWINDOWS:os=new String("MicroWindows"); break;		//  MicroWindows 
case wxPENWINDOWS:	os=new String("Pen Windows"); break;		//  Windows for Pen Computing 
case wxWINDOWS_NT:	os=new String("Windows NT"); break;			//  Windows NT 
case wxWIN32S:		os=new String("Windows 32S API"); break;	//  Windows 32S API 
case wxWIN95:			os=new String("Windows 95"); break;			//  Windows 95 
case wxWIN386:		os=new String("Watcom 32-bit supervisor modus"); break;	//  Watcom 32-bit supervisor modus 
case wxWINDOWS_CE:	os=new String("Windows CE (generic)"); break;				//  Windows CE (generic) 
case wxWINDOWS_POCKETPC: os=new String("Windows CE PocketPC"); break;			//  Windows CE PocketPC 
case wxWINDOWS_SMARTPHONE: os=new String("Windows CE Smartphone"); break;		//  Windows CE Smartphone 
case wxMGL_UNIX:		os=new String("MGL with direct hardware access"); break;	//  MGL with direct hardware access 
case wxMGL_X:			os=new String("MGL on X"); break;				//  MGL on X 
case wxMGL_WIN32:	os=new String("MGL on Win32"); break;			//  MGL on Win32 
case wxMGL_OS2:		os=new String("MGL on OS/2"); break;			//  MGL on OS/2 
case wxMGL_DOS:		os=new String("MGL on MS-DOS"); break;			//  MGL on MS-DOS 
case wxWINDOWS_OS2:	os=new String("Native OS/2 PM"); break;		//  Native OS/2 PM 
case wxUNIX:			os=new String("wxBase under Unix"); break;	//  wxBase under Unix 
case wxX11:			os=new String("Plain X11 and Universal widgets"); break;	//  Plain X11 and Universal widgets 
case wxPALMOS:		os=new String("PalmOS"); break;					//  PalmOS 
case wxDOS:			os=new String("wxBase under MS-DOS"); break;	//  wxBase under MS-DOS 
case wxUNKNOWN_PLATFORM: 
	default:			os=new String("Unknown Platform"); break;
	}
	wxString str;
	str << theApp->osMajor << "." << theApp->osMinor;
	double osVersionNr;
	str.ToDouble(&osVersionNr);
		 
	return new List(
		new String("aUCBLogo"), 
		new Float(4.692),
		os, 
		new Float(osVersionNr),
		_ListEnd);
}

NodeP lOSScreenSize(PairPP)
{
	return new List(
		new Int(theApp->max.x),
		new Int(theApp->max.y), _ListEnd);
}

NodeP lShellSpawn(PairPP arg) 
{
	char doscmd[0x100];
	
	while (arg!=argStack  &&  arg[1]->isnolist()  &&  !throwing) 
	{	arg[1] = error(BAD_DATA, arg[1]);
		arg++;
	}
	if (noarg) 
	{	c_out << "Type EXIT to return to Logo." << endL;

		if (!wxExecute(""))
		{	error
			(	FILE_ERROR, new String(
				"Could not open shell (probably due to low memory)")
			);
			return _False;
		}
	}else 
	{	File::stringlen=199;
		File::stringptr=doscmd;
		stringbuffer << arg[1];
		*File::stringptr='\0';

		if (!wxExecute(doscmd))
		{	error
			(	FILE_ERROR, new String(
				"Could not open shell (probably due to low memory)")
			);
			return _False;
		}
	}
	return _True;
}

NodeP lShell(PairPP arg) 
{
	while (arg!=argStack  &&  arg[1]->isnolist()  &&  !throwing) 
	{	arg[1] = error(BAD_DATA, arg[1]);
		arg++;
	}
	if (noarg) 
	{	error(NOT_YET_SUPPORTED, new String("Shell"));
		return Unbound;

		c_out << "Type EXIT to return to Logo." << endL;
#ifdef WIN32
		Piper piper(environment.comspec.c_str());
		return piper.output;
#else
		error(NOT_YET_SUPPORTED, new String("Shell"));
		return Unbound;
#endif
	}else 
	{
		if (!arg[1])
			return 0;
		wxString command;
#ifdef WIN32
		command << environment.comspec << wxT(" /C");
#endif
		command << arg[1]->string()->ptr();

		wxArrayString output, errors;
		int result = wxExecute(command, output, errors);

		PairP list=0;
		if(result != -1)
		{	bool wordmode = exist(2) ?  true : false;

			for (int i=output.GetCount()-1; i>=0; i--)
			{	if (wordmode)
				{	StringP str=new String(output[i]+'\n');
					push(list, str);
				}else
				{	StringP str=new String(output[i]);
					push(list, str->parser(false));
				}
			}
			if (errors.GetCount())
			{	c_out << wxT("Errors: ") << endL;
				for (unsigned int i=0; i < errors.GetCount(); i++)
					c_out << errors[i] << endL; 
			}
		}
		return list;

		//___________________here follows the old shell code___________
		char oscmd[0x100];
		File::stringlen=0xff;
		File::stringptr=oscmd;
		bool wordmode = exist(2) ?  true : false;

#ifdef WIN32
		stringbuffer 
			<< environment.comspec << " /C" << arg[1];
		*File::stringptr='\0';

		Piper piper(oscmd, wordmode);
		return piper.output;
#else
//		stringbuffer << environment.comspec << " " << arg[1];
		stringbuffer << arg[1];
		*File::stringptr='\0';

		FILE *strm;
		PairP head=0, tail=0;
		NodeP this_=0;
		int len;

		strm = popen(oscmd, "r");

		if (!strm)
		{	error(FILE_ERROR, new String("Could not open shell pipe"));
			return Unbound;
		}
		fgets(oscmd, 0x100, strm);
		while (!feof(strm))
		{	len = strlen(oscmd);
			if (oscmd[len-1] == '\n')
				oscmd[--len] = '\0';
			if (wordmode)
				this_ = String::make(oscmd, 0, len, STRING, String::copy);
			else
				this_ = (new String(oscmd))->parser(false);
			if (!head)
			{	tail = head = new Pair(this_, 0);
			} else
			{	tail->settail(new Pair(this_, 0));
				next(tail);
			}

			fgets(oscmd, 0x100, strm);
		}
		pclose(strm);
		return head;
#endif
	}
}
