//___________________________________________________________________
//
//                        Class Console
//___________________________________________________________________

class Mouse;

class Console: public wxTextCtrl
{
	friend Function lStandout;
	public:
		bool charAvailable, lineAvailable, charMode, insertMode, 
			mouseEvents, autosave;
		char ch;
		int chExt;
		wxPoint coord, Max;
		wxString readLine, oldReadLine, consoleFileName;
		Mouse *mouse;

		Console(wxWindow* parent, wxWindowID id, 
			const wxString& value = "", 
			const wxPoint& pos = wxDefaultPosition, 
			const wxSize& size = wxDefaultSize, 
			long style = 0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = wxTextCtrlNameStr,
			bool autosave_ = true,
			int caretpos = 0);
		~Console();

		inline bool checkBreak();
		void 
			OnChar		(wxKeyEvent& event),
			OnKeyDown	(wxKeyEvent& event),
			OnKeyUp		(wxKeyEvent& event),
			OnEnter		(wxKeyEvent &event),
			setMouse		(wxMouseEvent &event, Mouse *mouse),
			OnMouse		(wxMouseEvent& event),
			OnMouseLeftDown		(wxMouseEvent &event),
			OnMouseRightDown		(wxMouseEvent &event),
			OnMouseMiddleDown		(wxMouseEvent &event),
			OnMouseLeftUp  		(wxMouseEvent &event),
			OnMouseRightUp  		(wxMouseEvent &event),
			OnMouseMiddleUp  		(wxMouseEvent &event),
			OnMouseLeftDClick		(wxMouseEvent &event),
			OnMouseRightDClick	(wxMouseEvent &event),
			OnMouseMiddleDClick	(wxMouseEvent &event),
			OnMouseMotion			(wxMouseEvent &event),
			OnHelp		(wxHelpEvent& event),
			OnEditRepeat(wxCommandEvent &event),
			OnEditReset	(wxCommandEvent &event),
			updateGraph(),
			dispatchMessages(),
			print(FILEP strm, char ch),
			update_coords(char ch),
			finishLine(wxString line), flush(),
			setTextSelection(int x1, int y1, int x2, int y2),
			charmodeOn(), charmodeOff(), 
			setDefaultStyle(),
			modeBold(), modeItalic(), modeUnderline(), 
			modeStrikeout(), modePlain(),
			toggleInsertMode();
	
		bool kbhit();
		char getch(), 
			getc_(), putc_(char c, FILEP strm), read_getc(FILEP strm),
			charUnderCursor();
		int getchExt();
		wxString wordUnderCursor();

		NodeP _Cursor(), _setCursor(PairP args), _setMargins(PairP args), 
			  _setTextColor(PairPP arg),
			  _setTextSize(PairP args), 
			  _setTextFont(PairP args), 
			  _insertMode(), _overwriteMode(),
			  _standout(NodeP arg), _clearText(PairP);

	protected:
		bool underline;
		char standout, plainMode;
		int fontSize, fontFamily, fontStyle, fontWeight;
		unsigned int readIndex, writeBufferSize, tabSize;
		wxString writeBuffer, fontName;
		Color textColor, backgroundColor;

#define contextMenuNr 100
		wxMenu *contextMenu[contextMenuNr];

		struct MenuTree
		{	int level;
			charP str;
		};

		static MenuTree command[];

		void 
			OnDestroy(),
			OnSaveDocument(wxString pathname),
			createMenu(int &used, int &i, int level),
			OnContextMenu(wxContextMenuEvent &event),
			OnOpenFileAtCursor(wxCommandEvent &event),
			OnInsert(wxCommandEvent &event),
			Paste();

		DECLARE_EVENT_TABLE()
};
extern Console *console;

extern void print(FILEP strm, char ch);

extern Function 
	lsetTextColor, lsetTextSize, lsetTextFont, 
	linsertMode, loverwriteMode, 
	lboldTextMode, litalicTextMode, lunderlineTextMode, lplainTextMode,
	lclearText, lCursor, lsetMargins, lstandout,
	ldispatchMessages, lCharUnderCursor, lWordUnderCursor,
	lsetTextSelection,
	lenableTextMouseEvents, ldisableTextMouseEvents,
	lConsoleSetFocus;
