#ifdef _DEBUG
	#include <crtdbg.h>
#endif

#include <wx/object.h>
#include <wx/dynlib.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "Pair.h"
#include "List.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Float.h"
#include "Object.h"
#include "DL.h"
#include "Error.h"
#include "Init.h"

#include "IntToPtr.h"
#include "PtrToInt.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

DynamicLibrary::DynamicLibrary(StringP name_, wxDynamicLibrary *module_):
	Node(DYNAMIC_LIB), module(module_) 
{
	setname(name_);
}

void DynamicLibrary::free()
{
	if (module)
		delete module;
}

void DynamicLibrary::setname(StringP nd) {_name=nd;  check_oldyoung(nd);}

void DynamicLibrary::inc_gc() {fifo << name();}

int  DynamicLibrary::youngs() {return gmark(name());}
int  DynamicLibrary::youngsPtr() {return gmarkPtr(name());}
void DynamicLibrary::check_oldyoungs() {check_oldyoung(name());}

void DynamicLibrary::setPtrs()
{
	setname(StringP((String*)GC_::intToPtr->find(int(_name))));
	wxDynamicLibrary *dl=new wxDynamicLibrary;
	if (!dl->Load(_name->ptr()))
	{	error(FILE_ERROR, new String(
			wxString("Dynamic Library ")+_name->ptr()+" not found"));
		return;
	}
	module=dl;
}
wxDataInputStream &DynamicLibrary::read(wxDataInputStream &istr) 
{
	istr >> (int&)_name;
	return istr;
}
wxDataOutputStream &DynamicLibrary::write(wxDataOutputStream &ostr) 
{
	ostr << type << GC_::ptrToInt->find(name());
	return ostr;
}

StringP DynamicLibrary::string()
{
	wxString s;
	return new String(wxString("DynamicLibrary ")+name()->ptr());
}

NodeP lDynamicLibrary(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	wxDynamicLibrary *dl=new wxDynamicLibrary;
	if (!dl->Load(name->ptr()))
	{	error(FILE_ERROR, new List(
			new String("Dynamic Library "), 
			name,
			new String(" not found"), _ListEnd));
		return Unbound;
	}else
		return new DynamicLibrary(name, dl);
}

#ifdef WIN32
	#define PUSH(i)	__asm push i
#else
	#define PUSH(i)
#endif

NodeP lDynamicLibraryCall(PairPP arg)
{
	wxDynamicLibrary *dl=arg[1]->a_dynamicLibrary(1);
	if throwing return Unbound;
	StringP symbol=arg[2]->a_string(2);
	if throwing return Unbound;
	PairP list=arg[3]->a_list(3), l=list;
	if throwing return Unbound;
	if (!list)
	{	error(BAD_DATA, list, 3);
		return Unbound;
	}
	if (!dl->HasSymbol(symbol->ptr()))
	{	error(BAD_DATA, symbol, 2);
		return Unbound;
	}
	void *function=dl->GetSymbol(symbol->ptr());

	ObjectP ret=0;
	if (l->first())
		ret=l->first()->intern();
	else
		ret=_Void;
	if (!ret)
	{	error(BAD_DATA, NodeP(ret), 3); return Unbound;
	}
	next(l);
	if (l)
		l=PairP(l->reverse());
	NodeP t=0, n=0, third=0, n3=0;
	
	while (l && l->tail() && l->tail()->tail())
	{
		t=l->second()->intern(); n=t;
		third=l->first(); n3=third;
		int type;
		
		if (!n3)
			n3=_Null_Word;

		bool ptr=false;

		if (t==_Int)
		{	n3=n3->_int();			type=INT_;
		}else if (t==_Int16)
		{	n3=n3->toint16();		type=INT16;
		}else if (t==_Int8)
		{	n3=n3->toint8();		type=INT8;
		}else if (t==_UInt8)
		{	n3=n3->touint8();		type=UINT8;
		}else if (t==_Float)
		{	n3=n3->_float();		type=FLOAT_;
		}else if (t==_IntPtr)
		{	n3=n3->_int();			type=INT_;	ptr=true;
		}else if (t==_Int16Ptr)
		{	n3=n3->toint16();		type=INT16;	ptr=true;
		}else if (t==_Int8Ptr)
		{	n3=n3->toint8();		type=INT8;	ptr=true;
		}else if (t==_UInt8Ptr)
		{	n3=n3->touint8();		type=UINT8;	ptr=true;
		}else if (t==_FloatPtr)
		{	n3=n3->_float();		type=FLOAT_;ptr=true;
		}else if (t==_Word_)
		{	n3=n3->string();		type=STRING;
		}else if (t==_IntArray_)
		{	n3=n3->intarray();	type=INTARRAY;
		}else if (t==_Int16Array_)
		{	n3=n3->int16array();	type=INT16ARRAY;
		}else if (t==_FloatArray_)
		{	n3=n3->floatarray();	type=FLOATARRAY;
		}else if (t==_Bool)
		{	type=OBJECT;
		}else
		{	error(BAD_DATA, t, 3);	return Unbound;
		}
		if (n3->type != type && !(type==STRING && n3->is.string))
		{	error(BAD_DATA, third, 3); return Unbound;
		}
		int i;
		if (n3->type==INT_) 
		{	if (ptr)
			{	i=(int)&IntP(n3)->int_;				PUSH(i);
			}else
			{	i=IntP(n3)->int_;						PUSH(i);
			}
		}else if (n3->type==INT16) 
		{	if (ptr)
			{	i=(int)&Int16P(n3)->int_;			PUSH(i);
			}else
			{	i=Int16P(n3)->int_;					PUSH(i);
			}
		}else if (n3->type==INT8) 
		{	if (ptr)
			{	i=(int)&Int8P(n3)->int_;			PUSH(i);
			}else
			{	i=Int8P(n3)->int_;					PUSH(i);
			}
		}else if (n3->type==UINT8) 
		{	if (ptr)
			{	i=(int)&UInt8P(n3)->int_;			PUSH(i);
			}else
			{	i=UInt8P(n3)->int_;					PUSH(i);
			}
		}else if (n3->type==FLOAT_) 
		{	if (ptr)
			{	i=(int)&FloatP(n3)->float_;		PUSH(i);
			}else
			{	i=*intP(&FloatP(n3)->float_);		PUSH(i);
				i=*(intP(&FloatP(n3)->float_)+1);PUSH(i);
			}
		}else if (n3->is.string) 
		{	i=int(StringP(n3)->ptr());				PUSH(i);
		}else if (n3->type==INTARRAY) 
		{	i=int(IntArrayP(n3)->data());			PUSH(i);
		}else if (n3->type==INT16ARRAY) 
		{	i=int(Int16ArrayP(n3)->data());		PUSH(i);
		}else if (n3->type==FLOATARRAY) 
		{	i=int(FloatArrayP(n3)->data());		PUSH(i);
		}else if (n3->type==OBJECT) 
		{	if (n3==_True)
				i=1;
			else
				i=0;
															PUSH(i);
		}else
		{	error(BAD_DATA, third, 3);	return Unbound;
		}
		l=l->tail()->tail()->tail();
	}
	if (ret==_Int)
	{	int i = (*(int(__stdcall*)())function)();	
		return new Int(i);
	}else if (ret==_Int16)
	{	short i = (*(short(__stdcall*)())function)();
		return new Int16(i);
	}else if (ret==_Int8)
	{	char i = (*(char(__stdcall*)())function)();
		return new Int8(i);
	}else if (ret==_UInt8)
	{	unsigned char i = (*(unsigned char(__stdcall*)())function)();
		return new UInt8(i);
	}else if (ret==_Float)
	{	double i = (*(double(__stdcall*)())function)();
		return new Float(i);
	}else if (ret==_Word_)
	{	charP i = (*(charP(__stdcall*)())function)();
		if (!i)
			return _Null_Word;
		return new String(i);
	}else if (ret==_Void)
	{	(*(void(__stdcall*)())function)();
		return Unbound;
	}else if (ret==_Bool)
	{	bool i = (*(bool(__stdcall*)())function)();
		return Node::boolnode(i);
	}else
	{	error(BAD_DATA, ret, 3);	return Unbound;
	}
}
