//___________________________________________________________________
//
//                         Dialog Functions
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/utils.h>
#include <wx/textctrl.h>
#include <wx/dirdlg.h>
#include <wx/filedlg.h>
#include <wx/colordlg.h>
#include <wx/fontdlg.h>
#include <wx/choicdlg.h>
#include <wx/numdlg.h>
#include <wx/textdlg.h>
#include <wx/msgdlg.h>

#include "Main.h"
#include "UInt8.h"
#include "Pair.h"
#include "List.h"
#include "String.h"
#include "Int.h"
#include "Object.h"

#include "Error.h"
#include "Init.h"
#include "Eval.h"
#include "Stack.h"
#include "Color.h"
#include "Console.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

NodeP lbeginBusyCursor(PairPP)
{
	wxBeginBusyCursor();
	return Unbound;
}

NodeP lendBusyCursor(PairPP)
{
	wxEndBusyCursor();
	return Unbound;
}

NodeP lisBusyCursor(PairPP)
{
	return Node::boolnode(wxIsBusy());
}

NodeP lDirSelector(PairPP arg)
{
	wxString message=wxDirSelectorPromptStr;
	if (exist(1))
	{	StringP msg = arg[1]->a_string(1);
		if throwing return Unbound;
		message=msg->ptr();
	}
	wxString default_path="";
	if (exist(2))
	{	StringP dp = arg[2]->a_string(2);
		if throwing return Unbound;
		default_path=dp->ptr();
	}
	wxPoint pos=wxDefaultPosition;;
	if (exist(3))
	{	PairP p = arg[3]->a_int_vector(3);
		if throwing return Unbound;
		pos.x=int(p->first()->getnumber());
		pos.y=int(p->second()->getnumber());
	}
	wxString result = wxDirSelector(
		message, default_path, wxDD_DEFAULT_STYLE, pos, console);
	return new String(result);
}

NodeP lFileSelector(PairPP arg)
{
	wxString message="";
	StringP msg = arg[1]->a_string(1);
	if throwing return Unbound;
	message=msg->ptr();

	wxString default_path="";
	if (exist(2))
	{	StringP dp = arg[2]->a_string(2);
		if throwing return Unbound;
		default_path=dp->ptr();
	}
	wxString default_filename="";
	if (exist(3))
	{	StringP df = arg[3]->a_string(3);
		if throwing return Unbound;
		default_filename=df->ptr();
	}
	wxString default_extension="";
	if (exist(4))
	{	StringP de = arg[4]->a_string(4);
		if throwing return Unbound;
		default_extension=de->ptr();
	}
	wxString wildcard="*.*";
	if (exist(5))
	{	StringP wc = arg[5]->a_string(5);
		if throwing return Unbound;
		wildcard=wc->ptr();
	}

	int flags=0;
	if (exist(6))
	{	flags = arg[6]->a_int(6);
		if throwing return Unbound;
	}
	wxPoint pos=wxDefaultPosition;;
	if (exist(7))
	{	PairP p = arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=int(p->first()->getnumber());
		pos.y=int(p->second()->getnumber());
	}
	wxString result = wxFileSelector(
		message, default_path, default_filename, default_extension,
		wildcard, flags, console, pos.x, pos.y);
	return new String(result);
}

NodeP lgetColorFromUser(PairPP arg)
{
	wxColour colInit=(unsigned long)(0);
	if (exist(1))
	{	Color c = arg[1]->a_color(1);
		if throwing return Unbound;
		colInit=(unsigned long)(c);
	}
	wxColour result = wxGetColourFromUser(console, colInit);
	Color c;
	c.r=result.Red();
	c.g=result.Green();
	c.b=result.Blue();
	c.a=0xff;
	return new Int(c);
}

NodeP lgetFontFromUser(PairPP arg)
{
	wxFont fontInit;
	if (exist(1))
	{	StringP s = arg[1]->a_string(1);
		if throwing return Unbound;
		fontInit=wxFont(12, wxFONTFAMILY_ROMAN, wxFONTSTYLE_NORMAL,
			wxFONTWEIGHT_NORMAL, false, s->ptr());
	}
	wxFont r = wxGetFontFromUser(console, fontInit);
	if (r.Ok())
		return new List(
			new Int(r.GetPointSize()),
			new List(
				new Int(r.GetPixelSize().x),
				new Int(r.GetPixelSize().y), _ListEnd),
			new Int(r.GetFamily()),
			new Int(r.GetStyle()),
			new Int(r.GetWeight()),
			new String(r.GetFaceName()), _ListEnd);
	return 0;
}

NodeP lgetMultipleChoices(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption;
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	wxArrayString aChoices;
	{	PairP ac=arg[3]->a_list(3);
		if throwing return Unbound;
		while (ac)
		{	if (!ac->first())
			{	error(BAD_DATA, ac);
				return Unbound;
			}
			StringP s=ac->first()->string();
			aChoices.Add(s->ptr());
			next(ac);
		}
	}
	int x=-1, y=-1;
	if (exist(4))
	{	PairP pos = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		x=IntP(pos-> first())->int_;
		y=IntP(pos->second())->int_;
	}
	bool centre=true;
	if (exist(5))
	{	int c=arg[5]->a_bool(5);
		if throwing return Unbound;
		centre= c!=0;
	}
	int width=150, height=200;
	if (exist(6))
	{	PairP size = arg[6]->a_int_vector(6);
		if throwing return Unbound;
		width =IntP(size-> first())->int_;
		height=IntP(size->second())->int_;
	}
	wxArrayInt selections;
	wxGetMultipleChoices(
		selections,
		message, caption, aChoices, console,
		x, y, centre, width, height);

	int n = selections.GetCount()-1;
	PairP result=0;
	for (; n>=0; n--)
		result=new Pair(new Int(selections[n]+1), result);
	return result;
}

NodeP lgetNumberFromUser(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString prompt;
	{	StringP p=arg[2]->a_string(2);
		if throwing return Unbound;
		prompt=p->ptr();
	}
	wxString caption;
	{	StringP c=arg[3]->a_string(3);
		if throwing return Unbound;
		caption=c->ptr();
	}
	int value=0;
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		value=v;
	}
	int min=0;
	{	int m=arg[5]->a_int(5);
		if throwing return Unbound;
		min=m;
	}
	int max=100;
	{	int m=arg[6]->a_int(6);
		if throwing return Unbound;
		max=m;
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p = arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p-> first())->int_;
		pos.y=IntP(p->second())->int_;
	}
	int result = wxGetNumberFromUser( 
		message, prompt, caption, 
		value, min, max, console, pos);
	return new Int(result);
}

NodeP lgetPasswordFromUser(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption=_T("Input Text");
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	wxString default_value;
	{	StringP d=arg[3]->a_string(3);
		if throwing return Unbound;
		default_value=d->ptr();
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4))
	{	PairP p = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p-> first())->int_;
		pos.y=IntP(p->second())->int_;
	}
	bool centre=true;
	if (exist(5))
	{	int c=arg[5]->a_bool(5);
		if throwing return Unbound;
		centre= c!=0;
	}
	wxString result = wxGetPasswordFromUser( 
		message, caption, default_value,
		console, pos.x, pos.y, centre);
	return new String(result);
}

NodeP lgetTextFromUser(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption=_T("Input Text");
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	wxString default_value;
	{	StringP d=arg[3]->a_string(3);
		if throwing return Unbound;
		default_value=d->ptr();
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4))
	{	PairP p = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p-> first())->int_;
		pos.y=IntP(p->second())->int_;
	}
	bool centre=true;
	if (exist(5))
	{	int c=arg[5]->a_bool(5);
		if throwing return Unbound;
		centre= c!=0;
	}
	wxString result = wxGetTextFromUser( 
		message, caption, default_value,
		console, pos.x, pos.y, centre);
	return new String(result);
}

NodeP lgetSingleChoice(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption;
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	wxArrayString aChoices;
	{	PairP ac=arg[3]->a_list(3);
		if throwing return Unbound;
		while (ac)
		{	if (!ac->first())
			{	error(BAD_DATA, ac);
				return Unbound;
			}
			StringP s=ac->first()->string();
			aChoices.Add(s->ptr());
			next(ac);
		}
	}
	int x=-1, y=-1;
	if (exist(4))
	{	PairP pos = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		x=IntP(pos-> first())->int_;
		y=IntP(pos->second())->int_;
	}
	bool centre=true;
	if (exist(5))
	{	int c=arg[5]->a_bool(5);
		if throwing return Unbound;
		centre= c!=0;
	}
	int width=150, height=200;
	if (exist(6))
	{	PairP size = arg[6]->a_int_vector(6);
		if throwing return Unbound;
		width =IntP(size-> first())->int_;
		height=IntP(size->second())->int_;
	}
	wxString result = wxGetSingleChoice(
		message, caption, aChoices, console,
		x, y, centre, width, height);

	return new String(result);
}

NodeP lgetSingleChoiceIndex(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption;
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	wxArrayString aChoices;
	{	PairP ac=arg[3]->a_list(3);
		if throwing return Unbound;
		while (ac)
		{	if (!ac->first())
			{	error(BAD_DATA, ac);
				return Unbound;
			}
			StringP s=ac->first()->string();
			aChoices.Add(s->ptr());
			next(ac);
		}
	}
	int x=-1, y=-1;
	if (exist(4))
	{	PairP pos = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		x=IntP(pos-> first())->int_;
		y=IntP(pos->second())->int_;
	}
	bool centre=true;
	if (exist(5))
	{	int c=arg[5]->a_bool(5);
		if throwing return Unbound;
		centre= c!=0;
	}
	int width=150, height=200;
	if (exist(6))
	{	PairP size = arg[6]->a_int_vector(6);
		if throwing return Unbound;
		width =IntP(size-> first())->int_;
		height=IntP(size->second())->int_;
	}
	int result = wxGetSingleChoiceIndex(
		message, caption, aChoices, console,
		x, y, centre, width, height);

	return new Int(result);
}

NodeP lMessageBox(PairPP arg)
{
	wxString message;
	{	StringP m=arg[1]->a_string(1);
		if throwing return Unbound;
		message=m->ptr();
	}
	wxString caption=_T("Message");
	if (exist(2))
	{	StringP c=arg[2]->a_string(2);
		if throwing return Unbound;
		caption=c->ptr();
	}
	int style=wxOK;
	if (exist(3))
	{	int s = arg[3]->a_int(3);
		if throwing return Unbound;
		style=s;
	}
	int x=-1, y=-1;
	if (exist(4))
	{	PairP pos = arg[4]->a_int_vector(4);
		if throwing return Unbound;
		x=IntP(pos-> first())->int_;
		y=IntP(pos->second())->int_;
	}
	int result = wxMessageBox(
		message, caption, style, console, x, y);

	return new Int(result);
}
