//___________________________________________________________________
//
//                      aUCBLogo Editor module
//                         based on stctest
//            written by Otto Wyss and Andreas Micheler
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/filedlg.h>  // file dialog support
#include <wx/filename.h> // filename support
#include <wx/notebook.h> // notebook support
#include <wx/settings.h> // system settings
#include <wx/string.h>   // strings support
#include <wx/image.h>    // images support
#include <wx/html/helpctrl.h> // html help
#include <wx/glcanvas.h>

#include <fstream>

#include "Main.h"
#include "String.h"
#include "File.h"
#include "Pair.h"
#include "List.h"
#include "Quote.h"
#include "Line.h"
#include "Tree.h"
#include "Error.h"
#include "Eval.h"
#include "Object.h"
#include "Table.h"
#include "UInt8.h"
#include "Point.h"
#include "Procedur.h"

#include "Stack.h"
#include "Workspce.h"
#include "Files.h"
#include "Init.h"
#include "Color.h"

#include "EdDefsExt.h"     // Additional definitions
#include "Edit.h"        // Edit module
#include "EdPrefs.h"       // Prefs
#include "EdFrame.h"
#include "Console.h"
#include "Graph.h"
#include "LogoApp.h"
#include "AboutDlg.h"

// the application icon (under Windows and OS/2 it is in resources)
#if defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__) || defined(__WXMGL__) || defined(__WXX11__)
	#include "mondrian.xpm"
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________
//
//                           declarations
//___________________________________________________________________

#define APP_NAME _T("aUCBLogo")
#define APP_MAINT _T("Andreas Micheler")
#define APP_LICENCE _T("GNU GPL")

#define APP_VERSION _T("4.692")
#define APP_BUILD __DATE__

#define APP_WEBSITE _T("http://www.Physik.Uni-Augsburg.de/~micheler")
#define APP_MAIL _T("mailto:Andreas.Micheler@Physik.Uni-Augsburg.de")

#define NONAME _("noname.lg")

wxPrintData *g_printData=0;
wxPageSetupData *g_pageSetupData=0;

#ifdef WIN32
bool FileDropTarget::OnDropFiles(wxCoord WXUNUSED(x), wxCoord WXUNUSED(y),
											const wxArrayString& filenames)
{
	for (size_t i=0; i<filenames.Count(); i++)
		theApp->edFrame->FileOpen(filenames[i]);
	return true;
}
#endif

//___________________________________________________________________
//
//                            EdFrame
//___________________________________________________________________

BEGIN_EVENT_TABLE (EdFrame, wxFrame)
	// common
	EVT_CLOSE (                      EdFrame::OnClose)
	// file
	EVT_MENU (wxID_NEW,              EdFrame::OnFileNew)
	EVT_MENU (wxID_OPEN,             EdFrame::OnFileOpen)
	EVT_MENU (wxID_SAVE,             EdFrame::OnFileSave)
	EVT_MENU (wxID_SAVEAS,           EdFrame::OnFileSaveAs)
	EVT_MENU (wxID_CLOSE,            EdFrame::OnFileClose)
	// properties
	EVT_MENU (myID_PROPERTIES,       EdFrame::OnProperties)
	// print and exit
	EVT_MENU (wxID_PRINT_SETUP,      EdFrame::OnPrintSetup)
	EVT_MENU (wxID_PREVIEW,          EdFrame::OnPrintPreview)
	EVT_MENU (wxID_PRINT,            EdFrame::OnPrint)
	EVT_MENU (wxID_EXIT,             EdFrame::OnExit)
	// edit
	EVT_MENU (wxID_CLEAR,            EdFrame::OnEdit)
	EVT_MENU (wxID_CUT,              EdFrame::OnEdit)
	EVT_MENU (wxID_COPY,             EdFrame::OnEdit)
	EVT_MENU (wxID_PASTE,            EdFrame::OnEdit)
	EVT_MENU (myID_INDENTINC,        EdFrame::OnEdit)
	EVT_MENU (myID_INDENTRED,        EdFrame::OnEdit)
	EVT_MENU (wxID_SELECTALL,        EdFrame::OnEdit)
	EVT_MENU (myID_SELECTLINE,       EdFrame::OnEdit)
	EVT_MENU (wxID_REDO,             EdFrame::OnEdit)
	EVT_MENU (wxID_UNDO,             EdFrame::OnEdit)
	// find
	EVT_MENU (wxID_FIND,             EdFrame::OnEdit)
	EVT_MENU (myID_FINDNEXT,         EdFrame::OnEdit)
	EVT_MENU (myID_FINDPREVIOUS,     EdFrame::OnEdit)
	EVT_MENU (myID_REPLACE,          EdFrame::OnEdit)
	EVT_MENU (myID_REPLACENEXT,      EdFrame::OnEdit)
	EVT_MENU (myID_REPLACEALL,       EdFrame::OnEdit)
	EVT_MENU (myID_FINDWHOLEWORD,    EdFrame::OnEdit)
	EVT_MENU (myID_SKIP_COMMENTS,    EdFrame::OnEdit)
	EVT_MENU (myID_BRACEMATCH,       EdFrame::OnEdit)
	EVT_MENU (myID_GOTO,             EdFrame::OnEdit)
	EVT_MENU (myID_GOTODEF,          EdFrame::OnEdit)
	EVT_MENU (myID_GOTOPREV,         EdFrame::OnEdit)
	// view
	EVT_MENU_RANGE (myID_HILIGHTFIRST, myID_HILIGHTLAST,
								 EdFrame::OnEdit)
	EVT_MENU (myID_DISPLAYEOL,       EdFrame::OnEdit)
	EVT_MENU (myID_INDENTGUIDE,      EdFrame::OnEdit)
	EVT_MENU (myID_LINENUMBER,       EdFrame::OnEdit)
	EVT_MENU (myID_LONGLINEON,       EdFrame::OnEdit)
	EVT_MENU (myID_WHITESPACE,       EdFrame::OnEdit)
	EVT_MENU (myID_FOLDTOGGLE,       EdFrame::OnEdit)
	EVT_MENU (myID_FOLDTOGGLEALL,    EdFrame::OnEdit)
	EVT_MENU (myID_OVERTYPE,         EdFrame::OnEdit)
	EVT_MENU (myID_READONLY,         EdFrame::OnEdit)
	EVT_MENU (myID_WRAPMODEON,       EdFrame::OnEdit)
	// run
	EVT_MENU (myID_RUN,              EdFrame::callOnRun)
	EVT_MENU (myID_STEP,             EdFrame::callOnStep)
	EVT_MENU (myID_BREAKPOINT,       EdFrame::OnBreakpoint)
	EVT_MENU (myID_SETRUNDIR,        EdFrame::OnSetRunDir)

	// extra
	EVT_MENU (myID_BACKUP,           EdFrame::OnMakeBackupFiles)
	EVT_MENU (myID_CHANGEUPLOW,      EdFrame::OnEdit)
	EVT_MENU (myID_CHANGELOWER,      EdFrame::OnEdit)
	EVT_MENU (myID_CHANGEUPPER,      EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTCR,        EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTCRLF,      EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTLF,        EdFrame::OnEdit)
	EVT_MENU (myID_MAKETOASSIGN,     EdFrame::OnEdit)
	EVT_MENU (myID_CHARSETANSI,      EdFrame::OnEdit)
	EVT_MENU (myID_CHARSETMAC,       EdFrame::OnEdit)
	// window
	EVT_MENU (myID_PAGENEXT,       EdFrame::OnPageNext)
	EVT_MENU (myID_PAGEPREV,       EdFrame::OnPagePrev)

	// help
	EVT_MENU (wxID_HELP,             EdFrame::OnEdit)
	EVT_MENU (myID_SETHELPFILENAME,  EdFrame::OnSetHelpFileName)
	EVT_MENU (wxID_ABOUT,            EdFrame::OnAbout)
	EVT_NOTEBOOK_PAGE_CHANGED(myID_NOTEBOOKSEL, EdFrame::OnSelChange)
END_EVENT_TABLE ()

EdFrame::EdFrame (const wxString &title, wxPoint pos, wxSize size,
						  int argc, wxChar** argv)
        : wxFrame ((wxFrame *)NULL, wxID_ANY, title,
						  pos, size,
                    (wxDEFAULT_FRAME_STYLE | wxNO_FULL_REPAINT_ON_RESIZE) & ~wxCLOSE_BOX)
{
	// intitialize important variables
	g_printData = new wxPrintData;
	g_pageSetupData = new wxPageSetupDialogData;

	// set icon and background
	SetTitle(theApp->aUCBLogo);
#ifdef WIN32
	SetIcons(wxIconBundle(wxIcon(
		theApp->runDirectory+"/logo.ico", wxBITMAP_TYPE_ICO)));
#endif
	SetBackgroundColour(_T("WHITE"));

	// create menu
	m_menuBar = new wxMenuBar;
	CreateMenu();

	notebook = new wxNotebook(this, myID_NOTEBOOKSEL);

	if (argc <= 1)
		;//OnFileNew(wxCommandEvent());
	else
		for (int i=1; i<argc; i++)
			FileOpen(argv[i]);

#ifdef WIN32
	fileDropTarget=new FileDropTarget();
	SetDropTarget(fileDropTarget);
#endif
}

EdFrame::~EdFrame()
{
	if (g_printData) delete g_printData;
	if (g_pageSetupData) delete g_pageSetupData;
}

// common event handlers
void EdFrame::OnClose (wxCloseEvent &event)
{
	wxCommandEvent evt;
	OnFileClose (evt);
	for (size_t i=1; i < notebook->GetPageCount(); i++)
	{	Edit *edit = (Edit*)notebook->GetPage(i);
		if (edit && edit->Modified())
		{	if (event.CanVeto()) 
				event.Veto (true);
			if (!event.GetVeto())
				theApp->action=EXIT_LOGO;
			return;
		}
	}
	theApp->action=EXIT_LOGO;
}

void EdFrame::OnAbout (wxCommandEvent &WXUNUSED(event))
{
	AboutDlg aboutDlg;
	aboutDlg.ShowModal();
}

void EdFrame::OnExit (wxCommandEvent &WXUNUSED(event))
{
	theApp->action=EXIT_LOGO;
//	Close (true);
}

// file event handlers
void EdFrame::OnFileNew(wxCommandEvent &WXUNUSED(event))
{
	wxString fname=NONAME;
	Edit *edit = new Edit(notebook);
	edit->SetFilename(NONAME);
	notebook->AddPage(edit, NONAME, true);
	SetTitle(NONAME);
	edit->SetFocus();
}

void EdFrame::OnFileOpen (wxCommandEvent &WXUNUSED(event))
{
	wxString fname;
	wxFileDialog dlg (this, _T("Open file"), _T(""), _T(""), _T("Logo file (*.lg)|*.lg"),
					  wxOPEN | wxFILE_MUST_EXIST | wxCHANGE_DIR);
	if (dlg.ShowModal() != wxID_OK) return;
	fname = dlg.GetPath ();
	FileOpen (fname);
}

void EdFrame::OnFileSave (wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (!edit->Modified())
	 {   wxMessageBox (_("There is nothing to save!"), _("Save file"),
					  wxOK | wxICON_EXCLAMATION);
		return;
	}
	if (edit->GetFilename()==NONAME)
	{	wxCommandEvent cev;		
		OnFileSaveAs(cev);
		return;
	}
	edit->SaveFile ();
}

void EdFrame::OnFileSaveAs (wxCommandEvent &WXUNUSED(event))
{
	wxString filename = wxEmptyString;
	wxFileDialog dlg (this, _T("Save file"), _T(""), _T(""), _T("Logo file (*.lg)|*.lg"), wxSAVE|wxOVERWRITE_PROMPT);
	if (dlg.ShowModal() != wxID_OK) return;
	filename = dlg.GetPath();
	int page=notebook->GetSelection();
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->SaveFile (filename);
	edit->SetFilename(filename);
	wxFileName fn=filename;
	notebook->SetPageText(page, fn.GetFullName());
	SetTitle(edit->GetFilename());
}

void EdFrame::OnFileClose (wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page < 0)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (!edit)
		return;
	if (edit->Modified())
	{	if (wxMessageBox (_("Text is not saved, save before closing?"), _("Close"),
						  wxYES_NO | wxICON_QUESTION) == wxYES)
		{	edit->SaveFile();
		 if (edit->Modified())
			{	wxMessageBox (_("Text could not be saved!"), _("Close abort"),
							  wxOK | wxICON_EXCLAMATION);
				return;
			}
		}
	}
	notebook->DeletePage(page);
}

void EdFrame::callOnRun(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_RUN;
}

void EdFrame::OnRun(wxCommandEvent &WXUNUSED(event))
{
	root->contents(PROCS, NORMAL_OBJECT)->unBury( PROC_STEPPED );
	run();
}

void EdFrame::callOnStep(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_STEP;
}

void EdFrame::OnStep(wxCommandEvent &WXUNUSED(event))
{
	if (stepArmed)
		return;
	stepArmed=true;
	root->contents(PROCS, NORMAL_OBJECT)->bury( PROC_STEPPED );
	run();
}

void EdFrame::run()
{
	if (inpause)
	{	stoppingFlag=THROWING;
		throw_node=_Pause;
		console->readLine='\n';
		console->lineAvailable=true;
		return;
	}
	if (instep)
	{	console->readLine='\n';
		console->lineAvailable=true;
		return;
	}

	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (edit->Modified())
	{	if (edit->GetFilename()==NONAME)
		{	wxCommandEvent cev;
			OnFileSaveAs(cev);

		}
		else
			edit->SaveFile();
	}
	StringP myfn=0;
	{	wxString fn=edit->GetFilename();
		NodeP pn=new String(fn);
		leraseAll(0);

		PairPP asp=argStack;
		argStack << pn;
		lload(asp);
		argStack.pop_till(asp);

		if (!_Startup->var() || _Startup->var()==Unbound)
		{	int pos=fn.Find('.', true);
			fn.Remove(pos);
			pos=fn.Find(
		#ifdef WIN32
				'\\'
		#else
				'/'
		#endif
				, true);
			if (pos>=0)
			{	wxString path=fn.Left(pos);
				if (!wxSetWorkingDirectory(path))
					error(FILE_ERROR, new List(
						new String("couldn't change dir to '"),
						new String(path),	_ListEnd));
				fn.Remove(0,pos+1);
			}
			myfn=new String(fn);
		}
	}
	unmark(0);
	if (!throwing)
	{	console->SetFocus();
		if (!_Startup->var() || _Startup->var()==Unbound)
		{	File::currentLine=myfn;
			PairP exec_list = File::currentLine->parser(true);
			Evaluator::val_status = no_value_allowed;
			if (exec_list) exec_list->eval_driver();
		}
		if throwing
		{	     if (throw_node==_System  ) return;
			else if (throw_node==_Error   ) error_print();
			else if (throw_node!=_Toplevel)
			{	error(NO_CATCH_TAG, throw_node);	error_print();
			}
			stoppingFlag=RUN;
		}
		if (stopping || outputting)
		{	c_out << "; You must be in a procedure to use OUTPUT or STOP." << endL;
			stoppingFlag=RUN;
		}
		argStack.clean();
	}
	graph->updateGraph();
	console->charmodeOff();

	console->readLine='\n';
	console->lineAvailable=true;
}

void EdFrame::togglePause(Edit *edit, int linenr, StringP fn, PairP parent)
{
	PairP list=parent->first(), last=0;
//		cout << parent << endL;
	while (list)
	{	if (list->first()  &&  list->first()->is.list)
			togglePause(edit, linenr, fn, list);
		else
		{	if (LineP(list)->number == linenr+1  &&  list->first())
			{	edit->togglemark(linenr);

				if (list->first() == NodeP(_Pause))
				{	if (list->tail())
					{	list->setfirst(list->tail()->first());
						list->settail(list->tail()->tail());
					}
					break;
				}
				Line::nr=linenr+1;
				Line::fn=fn;
				list->settype_LINE();
				LineP line=new Line(_Pause, list);
				if (last)
					last->settail(line);
				else
					parent->setfirst(line);
				break;
			}
		}
		last=list;
		next(list);
	}
}

void EdFrame::togglePauseTree(Edit *edit, int linenr, StringP fn, PairP parent)
{
	PairP list=TreeP(parent)->tree(), last=0;
	while (list)
	{	if (LineP(list)->number == linenr+1)
		{	
			if (list->first()  &&  list->first()->first() == NodeP(_Pause))
			{	if (list->tail())
				{	list->setfirst(list->tail()->first());
					list->settail(list->tail()->tail());
				}
				break;
			}
			Line::nr=linenr+1;
			Line::fn=fn;
			LineP line=new Line(new Line(_Pause, 0), list);
			if (last)
				last->settail(line);
			else
			{	TreeP(parent)->settree(line);
				TreeP(parent->first())->settype_TREE();
				TreeP(parent->first())->settree(line);
			}
			break;
		}
		togglePauseTree2(edit, linenr, fn, list);
		last=list;
		next(list);
	}
}

void EdFrame::togglePauseTree2(Edit *edit, int linenr, StringP fn, PairP parent)
{
	PairP list=parent->first(), last=0;
	while (list)
	{	if (LineP(list)->number == linenr+1)
		{	
			if (list->first()->is.list)
				togglePauseTree2(edit, linenr, fn, list);
			else
			{
				if (list->first()  &&  list->first()->first() == NodeP(_Pause))
				{	if (list->tail())
					{	list->setfirst(list->tail()->first());
						list->settail(list->tail()->tail());
					}
					break;
				}
				Line::nr=linenr+1;
				Line::fn=fn;
				LineP line=new Line(new Quote(new Line(_Pause, list->first()->first())), 0);
				if (last)
					last->settail(line);
				else
					parent->setfirst(line);
				break;
			}
		}
		last=list;
		next(list);
	}
}

void EdFrame::OnBreakpoint(wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	int linenr=edit->GetCurrentLine(), 
		pos=edit->PositionFromLine(linenr);

	if (edit->GetLine(linenr).Left(6)=="pause ")
	{	int oldpos=edit->GetCurrentPos();
		edit->SetSelection(pos, pos+6);
		edit->DeleteBack();
		if (oldpos-pos < 6)
			edit->GotoPos(pos);
		else
			edit->GotoPos(oldpos-6);
	}else
		edit->InsertText(pos, "pause ");

	StringP filename=new String(edit->GetFilename());
	StringP fn=0;
	for (PairP fnlist=Line::filenames;  fnlist;  next(fnlist))
		if (fnlist->first() && fnlist->first()->equal(filename))
			fn=StringP(fnlist->first());

	if (!fn)
	{	wxMessageBox("Setting breakpoint not possible: "
			"File not found in the Logo root table", 
			"aUCBLogo Editor Error", wxOK, this);
		return;
	}
	ProcedureP proc=root->findProc(fn, linenr+1); // in Logo the lines start from 1!!!

	if (proc)
	{	togglePause(edit, linenr, fn, proc->bodylist());
//		cout << proc->bodylist() << endL;
		togglePauseTree(edit, linenr, fn, proc->bodylist());
//		cout << proc->bodylist() << endL;
		TreeP(proc->bodylist())->untreeify();
		proc->bodylist()->make_tree_from_body();
//		cout << proc->bodylist() << endL;
	}	
}

void EdFrame::OnSetHelpFileName(wxCommandEvent &WXUNUSED(event))
{
	wxFileDialog dlg(this, _T("Set help file name"), _T(""),
		theApp->helpFileName, _T("HTML Help Package (*.hhp)|*.hhp"),
					  wxOPEN | wxFILE_MUST_EXIST);
	if (dlg.ShowModal() != wxID_OK) return;
	theApp->helpFileName = dlg.GetPath ();
}

void EdFrame::OnSetRunDir(wxCommandEvent &WXUNUSED(event))
{
	wxDirDialog dlg(this, _T("Set run dir"), theApp->runDirectory);
	if (dlg.ShowModal() != wxID_OK) return;
	theApp->runDirectory = dlg.GetPath();
}

// properties event handlers
void EdFrame::OnProperties (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	if (!edit) return;

	EditProperties dlg(edit, 0);
}

// print event handlers
void EdFrame::OnPrintSetup (wxCommandEvent &WXUNUSED(event))
{
	(*g_pageSetupData) = *g_printData;
	wxPageSetupDialog pageSetupDialog(this, g_pageSetupData);
	pageSetupDialog.ShowModal();
	(*g_printData) = pageSetupDialog.GetPageSetupData().GetPrintData();
	(*g_pageSetupData) = pageSetupDialog.GetPageSetupData();
}

void EdFrame::OnPrintPreview (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	wxPrintDialogData printDialogData( *g_printData);
	wxPrintPreview *preview =
		new wxPrintPreview (new EditPrint (edit),
							new EditPrint (edit),
							&printDialogData);
	if (!preview->Ok())
	 {   delete preview;
		wxMessageBox (_("There was a problem with previewing.\n\
						 Perhaps your current printer is not correctly?"),
					  _("Previewing"), wxOK);
		return;
	}
	wxRect rect = DeterminePrintSize();
	wxPreviewFrame *frame = new wxPreviewFrame (preview, this, _("Print Preview"));
	frame->SetSize (rect);
	frame->Centre(wxBOTH);
	frame->Initialize();
	frame->Show(true);
}

void EdFrame::OnPrint (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	wxPrintDialogData printDialogData(*g_printData);
	wxPrinter printer (&printDialogData);
	EditPrint printout (edit);
	if (!printer.Print (this, &printout, true))
	 {   if (wxPrinter::GetLastError() == wxPRINTER_ERROR)
		  {
				wxMessageBox (_("There was a problem with printing.\n\
						 Perhaps your current printer is not correctly?"),
					  _("Previewing"), wxOK);
			return;
		}
	}
	(*g_printData) = printer.GetPrintDialogData().GetPrintData();
}

// edit events
void EdFrame::OnEdit (wxCommandEvent &event)
{
	int page=notebook->GetSelection();
	if (page < 0)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (edit) edit->ProcessEvent (event);
}

void EdFrame::OnSelChange(wxNotebookEvent &event)
{
	SetTitle(((Edit*)notebook->GetPage(event.GetSelection()))->GetFilename());
}

void EdFrame::OnPageNext(wxCommandEvent &WXUNUSED(event))
{
	notebook->AdvanceSelection();
}

void EdFrame::OnPagePrev(wxCommandEvent &WXUNUSED(event))
{
	notebook->AdvanceSelection(false);
}

void EdFrame::OnMakeBackupFiles(wxCommandEvent &WXUNUSED(event))
{
	theApp->backup=!theApp->backup;
}

// private functions
void EdFrame::CreateMenu ()
{
	// File menu
	wxMenu *menuFile = new wxMenu;
	menuFile->Append (wxID_NEW, _("&New ..\tCtrl+N"));
	menuFile->Append (wxID_OPEN, _("&Open ..\tCtrl+O"));
	menuFile->Append (wxID_SAVE, _("&Save\tCtrl+S"));
	menuFile->Append (wxID_SAVEAS, _("Save &as ..\tCtrl+Shift+S"));
	menuFile->Append (wxID_CLOSE, _("&Close\tCtrl+F4"));
	menuFile->AppendSeparator();
	menuFile->Append (myID_PROPERTIES, _("Proper&ties ..\tCtrl+I"));
	menuFile->AppendSeparator();
	menuFile->Append (wxID_PRINT_SETUP, _("Print Set&up .."));
	menuFile->Append (wxID_PREVIEW, _("Print Pre&view\tCtrl+Shift+P"));
	menuFile->Append (wxID_PRINT, _("&Print ..\tCtrl+P"));
	menuFile->AppendSeparator();
	menuFile->Append (wxID_EXIT, _("&Exit\tAlt+X"));



	// Edit menu
	wxMenu *menuEdit = new wxMenu;
	menuEdit->Append (wxID_UNDO, _("&Undo\tCtrl+Z"));
	menuEdit->Append (wxID_REDO, _("&Redo\tCtrl+Shift+Z"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_CUT, _("Cu&t\tCtrl+X"));
	menuEdit->Append (wxID_COPY, _("&Copy\tCtrl+C"));
	menuEdit->Append (wxID_PASTE, _("&Paste\tCtrl+V"));
	menuEdit->Append (wxID_CLEAR, _("&Delete\tDel"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_FIND, _("&Find\tCtrl+F"));
	menuEdit->Append (myID_FINDNEXT, _("Find &next\tF3"));
	menuEdit->Append (myID_FINDPREVIOUS, _("Find pr&evious\tShift+F3"));
	menuEdit->Append (myID_REPLACE, _("&Replace\tCtrl+H"));
	menuEdit->Append (myID_REPLACENEXT, _("Replace &again\tCtrl+F3"));
	menuEdit->Append (myID_REPLACEALL, _("Rep&lace all"));
	menuEdit->AppendCheckItem (myID_FINDWHOLEWORD, _("Find &whole words"));
	menuEdit->Check(myID_FINDWHOLEWORD, true);
	menuEdit->AppendCheckItem (myID_SKIP_COMMENTS, _("&Skip comments"));
	menuEdit->Check(myID_SKIP_COMMENTS, false);
	menuEdit->AppendSeparator();
	menuEdit->Append (myID_BRACEMATCH, _("&Match brace\tCtrl+M"));
	menuEdit->Append (myID_GOTO, _("&Goto\tCtrl+G"));
	menuEdit->Append (myID_GOTODEF, _("Goto definition\tF4"));
	menuEdit->Append (myID_GOTOPREV, _("Goto previous position\tShift+F4"));
	menuEdit->AppendSeparator();
	menuEdit->Append (myID_INDENTINC, _("&Indent increase\tTab"));
	menuEdit->Append (myID_INDENTRED, _("I&ndent reduce\tShift+Tab"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_SELECTALL, _("Select &all\tCtrl+A"));
	menuEdit->Append (myID_SELECTLINE, _("Select &line\tCtrl+L"));

	// hilight submenu
	wxMenu *menuHilight = new wxMenu;
	int Nr;
	for (Nr = 0; Nr < g_LanguagePrefsSize; Nr++)
	 {   menuHilight->Append (myID_HILIGHTFIRST + Nr,
							 g_LanguagePrefs [Nr].name);
	}

	// charset submenu
	wxMenu *menuCharset = new wxMenu;
	menuCharset->Append (myID_CHARSETANSI, _("&ANSI (Windows)"));
	menuCharset->Append (myID_CHARSETMAC, _("&MAC (Macintosh)"));

	// View menu
	wxMenu *menuView = new wxMenu;
	menuView->Append (myID_HILIGHTLANG, _("&Hilight language .."), menuHilight);
	menuView->AppendSeparator();
	menuView->AppendCheckItem (myID_FOLDTOGGLE, _("&Toggle current fold\tCtrl+T"));
	menuView->AppendCheckItem (myID_FOLDTOGGLEALL, _("&Toggle all folds\tCtrl+Alt+T"));
	menuView->AppendCheckItem (myID_OVERTYPE, _("&Overwrite mode\tIns"));
	menuView->AppendCheckItem (myID_WRAPMODEON, _("&Wrap mode\tCtrl+U"));
	menuView->AppendSeparator();
	menuView->AppendCheckItem (myID_DISPLAYEOL, _("Show line &endings"));
	menuView->AppendCheckItem (myID_INDENTGUIDE, _("Show &indent guides"));
	menuView->AppendCheckItem (myID_LINENUMBER, _("Show line &numbers"));
	menuView->AppendCheckItem (myID_LONGLINEON, _("Show &long line marker"));
	menuView->AppendCheckItem (myID_WHITESPACE, _("Show white&space"));
	menuView->AppendSeparator();
	menuView->Append (myID_USECHARSET, _("Use &code page of .."), menuCharset);

	// change case submenu
	wxMenu *menuChangeCase = new wxMenu;
	menuChangeCase->Append (myID_CHANGEUPLOW, _("&Correct case"));
	menuChangeCase->Append (myID_CHANGEUPPER, _("&Upper case"));
	menuChangeCase->Append (myID_CHANGELOWER, _("&Lower case"));

	// convert EOL submenu
	wxMenu *menuConvertEOL = new wxMenu;
	menuConvertEOL->Append (myID_CONVERTCR, _("CR (&Linux)"));
	menuConvertEOL->Append (myID_CONVERTCRLF, _("CR+LF (&Windows)"));
	menuConvertEOL->Append (myID_CONVERTLF, _("LF (&Macintosh)"));

	// Run menu
	wxMenu *menuRun = new wxMenu;
	menuRun->Append (myID_RUN, _("&Run\tF9"));
	menuRun->Append (myID_STEP, _("&Step\tF7"));
	menuRun->Append (myID_BREAKPOINT, _("&Set breakpoint\tCtrl-F8"));
	menuRun->Append (myID_SETRUNDIR, _("&Set run directory"));

	// Extra menu
	wxMenu *menuExtra = new wxMenu;
	menuExtra->AppendCheckItem (myID_BACKUP, _("Make &backup files"));
	menuExtra->AppendCheckItem (myID_READONLY, _("&Readonly mode"));
	menuExtra->Check(myID_BACKUP, theApp->backup);
	menuExtra->AppendSeparator();
	menuExtra->Append (myID_CHANGECASE, _("Change &case to .."), menuChangeCase);
	menuExtra->AppendSeparator();
	menuExtra->Append (myID_CONVERTEOL, _("Convert line &endings to .."), menuConvertEOL);
	menuExtra->Append (myID_MAKETOASSIGN, _("Convert &make to ="));

	// Window menu
	wxMenu *menuWindow = new wxMenu;
	menuWindow->Append (myID_PAGEPREV, _("&Previous\tCtrl+Shift+Tab"));
	menuWindow->Append (myID_PAGENEXT, _("&Next\tCtrl+Tab"));

	// Help menu
	wxMenu *menuHelp = new wxMenu;
	menuHelp->Append (wxID_HELP, _("&Keyword lookup\tF1"));
	menuHelp->Append (myID_SETHELPFILENAME, _("&Set help file name"));
	menuHelp->Append (wxID_ABOUT, _("&About ..\tShift+F1"));

	// construct menu
	m_menuBar->Append (menuFile  , _("&File"));
	m_menuBar->Append (menuEdit  , _("&Edit"));
	m_menuBar->Append (menuView  , _("&View"));
	m_menuBar->Append (menuRun   , _("&Run"));
	m_menuBar->Append (menuExtra , _("Ex&tra"));
	m_menuBar->Append (menuWindow, _("&Window"));
	m_menuBar->Append (menuHelp  , _("&Help"));
	SetMenuBar (m_menuBar);
}

void EdFrame::FileOpen (wxString fname)
{
	wxFileName w(fname); w.Normalize(); fname = w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page!=-1)
	{	notebook->SetSelection(page);
		notebook->GetPage(page)->SetFocus();
		return;
	}
	Edit *edit = new Edit(notebook);
	wxString s=w.GetFullName();
	notebook->AddPage(edit, s, true);
	edit->LoadFile(fname);
	SetTitle(edit->GetFilename());
	edit->SetFocus();
}

void EdFrame::FileClose(charP filename)
{
	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
		return;

	notebook->SetSelection(page);
	wxCommandEvent cev;
	OnFileClose(cev);
}

void EdFrame::FileSave(charP filename)
{
	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
		return;

	notebook->SetSelection(page);
	((Edit*)notebook->GetPage(page))->SaveFile ();
}

void EdFrame::mark(charP filename, int linenr, int markernr)
{
	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
	{	FileOpen(fname);
		for (unsigned int i=0; i<notebook->GetPageCount(); i++)
			if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
				page=i;
	}
	if (page==-1)
		return;

	notebook->SetSelection(page);
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->mark(linenr-1, markernr);
	console->SetFocus();
}

void EdFrame::unmark(int markernr)
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->unmark(markernr);

}

wxRect EdFrame::DeterminePrintSize ()
{
	wxSize scr = wxGetDisplaySize();

	// determine position and size (shifting 16 left and down)
	wxRect rect = GetRect();
	rect.x += 16;
	rect.y += 16;
	rect.width = wxMin (rect.width, (scr.x - rect.x));
	rect.height = wxMin (rect.height, (scr.x - rect.y));

	return rect;
}
