#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/listctrl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Tree.h"
#include "Object.h"
#include "List.h"
#include "Int.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

#include "Vars.h"
#include "Calls.h"
#include "LogoApp.h"

#include "EdDefsExt.h"
#include "Edit.h"
#include "EdFrame.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif


NodeP throw_node=0;
PairP err_mesg=0;

ErrorType erract_errtype;

bool inpause=false,
	instep=false,
	stepArmed=false;


void error_print(void) 
{
	ControlType save_flag=stoppingFlag;
	
	if (!err_mesg) return;
	
	stoppingFlag=RUN;
	File::backslashes=true;
	
	File::print(stdout, err_mesg->second(), 5, 80);

	if (err_mesg->tail()->tail()  &&  err_mesg->third()) 
	{	c_out << "  in " << err_mesg->third() 
			<< " " << err_mesg->fourth()
			<< "(" << err_mesg->fifth() << ")" << endL;
		
		charP fn=StringP(err_mesg->fourth())->ptr();
		int linenr=IntP(err_mesg->fifth())->int_;
		theApp->edFrame->mark(fn, linenr, 0);
	}
	err_mesg=0;
	c_out << endL;
	
	File::backslashes=false;
	stoppingFlag=save_flag;
}

PairP error(ErrorType error_type, NodeP error_desc, int nr)
{
	if throwing return PairP(Unbound);

	lupdateCalls(0);
	lupdateVars(0);

	bool recoverable=false, warning=false, uplevel=false;
	PairP err_act=0, val = (PairP)Unbound;
	
	switch(error_type) 
	{
	case FATAL:
			c_out << "Logo: Fatal Internal Error." << endL;
			exit(1);

	case OUT_OF_MEM_UNREC:
			c_out << "Logo: Out of Memory." << endL;
			exit(1);

	case OUT_OF_MEM:
			GC_::reserve_tank_use();
			if (theApp && theApp->heapReserve)
			{	delete theApp->heapReserve;
				theApp->heapReserve=0;
			}
			err_mesg = new Pair("; out of space. Delete something or clearScreen!");
			break;

	case STACK_OVERFLOW:
			err_mesg = new Pair("; stack overflow");		break;

	case TURTLE_OUT_OF_BOUNDS:
			err_mesg = new Pair("; turtle out of bounds");	break;

	case BAD_GRAPH_INIT:
			err_mesg = new Pair("; couldn't initialize graphics"); break;
			break;

	case BAD_DATA_UNREC:
			err_mesg = new List
				(	new String("; "),
					fun,
					new String("doesn\'t like"), 
					error_desc,
					new String("as input"), 
					_ListEnd
				);
			break;

	case DIDNT_OUTPUT:
			if (didnt_output_name) 
				Evaluator::last_call=didnt_output_name;
			
			if (!error_desc && Evaluator::didnt_get_output) 
			{	error_desc = Evaluator::didnt_get_output->fun;
					      ufun = Evaluator::didnt_get_output->ufun;
				    this_line = Evaluator::didnt_get_output->this_line;
			}
			err_mesg = new List
				(	new String("; "),
					Evaluator::last_call,
					new String("didn\'t output to"),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case NOT_ENOUGH:
			if (!error_desc)
				err_mesg = new List
					(	new String("; not enough inputs to"),
						fun, 
						_ListEnd
					);
			else
				err_mesg = new List
					(	new String("; not enough inputs to"),
						error_desc, 
						_ListEnd
					);
			break;

	case BAD_DATA:
			if (nr==1)
				err_mesg = new List
					(	new String("; "),
						fun,
						new String("doesn\'t like"), 
						error_desc,
						new String("as input"), 
						_ListEnd
					);
			else
				err_mesg = new List
					(	new String("; "),
						fun,
						new String("doesn\'t like"), 
						error_desc,
						new List(
							new String("as input "), 



							new Int(nr), _ListEnd), 
						_ListEnd
					);
			recoverable=true;
			break;

	case APPLY_BAD_DATA:
			err_mesg = new List
				(	new String("; APPLY doesn\'t like"),
					error_desc,
					new String("as input"), 
					_ListEnd
				);
			recoverable=true;
			break;

	case TOO_MUCH:
			err_mesg = new Pair("; too much inside ()\'s");	break;

	case DK_WHAT_UP:
			uplevel=true;

	case DK_WHAT:
			err_mesg = new List
				(	error_desc, 
					new String(" ;-)"),	// "You don`t say what to do with x" was too annoying
					_ListEnd
				);
			break;

	case PAREN_MISMATCH:
			err_mesg = new Pair("; too many (\'s");	break;

	case NO_VALUE:
			err_mesg = new List
				(	new String("; "),
					error_desc,
					new String("has no value"),
					_ListEnd
				);
			recoverable=true;
			break;

	case UNEXPECTED_PAREN:
			err_mesg = new Pair("; unexpected \')\'");	break;

	case UNEXPECTED_BRACKET:
			err_mesg = new Pair("; unexpected \']\'");	break;

	case UNEXPECTED_BRACE:
			err_mesg = new Pair("; unexpected \'}\'");	break;

	case DK_HOW:
			recoverable=true;

	case DK_HOW_UNREC:
			err_mesg = new List
				(	new String("; I don\'t know how  to"),
					error_desc, 
					_ListEnd
				);
			break;

	case NO_CATCH_TAG:
			err_mesg = new List
				(	new String("; Can't find catch tag for"),
					error_desc, 
					_ListEnd
				);
			break;

	case ALREADY_DEFINED:
			err_mesg = new List
				(	new String("; "),
					error_desc,
					new String("is already defined"),
					_ListEnd
				);
			break;

	case STOP_ERROR:
			err_mesg = new Pair("; Stopping...");		break;

	case ALREADY_DRIBBLING:
			err_mesg = new Pair("; Already dribbling");	break;

	case FILE_ERROR:
			err_mesg = new List
				(	new String("; File system error:"),
					error_desc, 
					_ListEnd
				);
			break;

	case IF_WARNING:
			err_mesg = new Pair("; Assuming you mean IFELSE, not IF");
			warning=true;
			break;

	case SHADOW_WARN:
			err_mesg = new List
				(	new String("; "),
					error_desc,
					new String("shadowed by local in procedure call"),
					_ListEnd
				);
			warning=true;
			break;

	case USER_ERR:
			if (error_desc==Unbound)
				err_mesg = new Pair("; Throw \"Error");
			else 
			{	uplevel=true;
				if (error_desc->is.list)
					err_mesg = (PairP)error_desc;
				else
					err_mesg = new Pair(error_desc, 0);
			}
			break;

	case IS_PRIM:
			err_mesg = new List
				(	new String("; "),
					error_desc,
					new String("is a primitive"),

					_ListEnd
				);
			break;

	case NOT_INSIDE:
			err_mesg = new Pair("; Can't use TO inside a procedure");	break;

	case AT_TOPLEVEL:
			err_mesg = new List
				(	new String("; Can only use"),
					error_desc,
					new String("inside a procedure"),
					_ListEnd
				);
			break;

	case NO_TEST:
			err_mesg = new List
				(	new String("; "),
					fun, 
					new String("without TEST"),
					_ListEnd
				);
			break;

	case ERR_MACRO:
			err_mesg = new List
				(	new String("; Macro returned"),
					error_desc,
					new String("instead of a list"),
					_ListEnd
				);
			break;

	case DEEPEND:
			if (!error_desc) 
				err_mesg = new Pair("; END inside multi-line instruction.\n");
			else 
			{	err_mesg = new List
					(	new String("; END inside multi-line instruction in"),
						error_desc,
						_ListEnd
					);
			}
			break;

	case NOT_YET_SUPPORTED:
			err_mesg = new List
				(	new String("; "),
					fun,
					new String("is not yet supported"), 
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	default:
			c_out << "Unknown error condition - internal error." << endL;
			exit(1);
	}

	didnt_output_name=0;

	if (ufun  &&  uplevel) 
	{	ufun = Evaluator::last_ufun;
		this_line = Evaluator::last_line;
	}
	if (ufun && this_line && this_line->is.line)
		err_mesg = new List
			(	err_mesg, 
				ufun,
				new String(LineP(this_line)->filename),
				new Int(LineP(this_line)->number),
				_ListEnd
			);
	else

		err_mesg = new List
			(	err_mesg, 
				0, 
				0,
				0,
				_ListEnd
			);

	err_mesg = new Pair
		(	new Int(error_type), 
			err_mesg
		);
	if (warning) 
	{	error_print();
		return (PairP)Unbound;
	}
	err_act = (PairP)_ErrorAction->var();

	if (err_act  &&  err_act != Unbound) 
	{	if (error_type != erract_errtype) 
		{	ValStatus save_val_status = Evaluator::val_status;
			
			erract_errtype = error_type;
			_ErrorAction->setvar(0);
			Evaluator::val_status = value_maybe_ok_in_macro;

			val = PairP(err_act->err_eval_driver());

			Evaluator::val_status = save_val_status;
			_ErrorAction->setvar( err_act );

			if (recoverable==true  &&  val != Unbound) 
			{	return val;
			}else 
				if (recoverable==false  &&  val != Unbound) 
				{	c_out << "; You don't say what to do with " << val << endL;
					val = (PairP)Unbound;
					throw_node = _Toplevel;
				} else 
				{	return (PairP)Unbound;
				}
		}else 
		{	c_out << "; _Erract loop" << endL;
			throw_node=_Toplevel;
		}
	}else 
	{	// no erract
		throw_node=_Error;
	}
	stoppingFlag=THROWING;
	output_node = (PairP)Unbound;
	return val;
}

NodeP lerror(PairPP) 
{
	NodeP val=err_mesg;
	err_mesg=0;
	return val;
}

NodeP lpause(PairPP) 
{
	bool oldinpause=inpause;
	inpause=true;
	stepArmed=false;

	lupdateCalls(0);
	lupdateVars(0);

	NodeP val=Unbound;
	if (err_mesg) error_print();
	ObjectP uname=ufun;

	c_out << "continue ; Pausing " << uname;
	if (this_line && this_line->is.line)
		c_out << " in line " << this_line->number;
	c_out << "..." << endL;
	if (this_line  
	&&  this_line->is.line  
	&&  this_line->filename
	&&  this_line->filename->is.string)
		theApp->edFrame->mark(this_line->filename->ptr(), this_line->number, 1);

	jmp_buf sav_iblk;
	memcpy(&sav_iblk, &File::inputBlockingBuffer, sizeof(jmp_buf));

	int saveInputBlocking=File::inputBlocking;
	File::inputBlocking=0;

	ValStatus save_val_status = Evaluator::val_status;

	while (running) 
	{	StringP str=0;	str=c_in.read("");	// "? ");
		PairP elist=0;	if (!throwing) elist = str->parser(true);

		Evaluator::val_status=value_maybe_ok_in_macro;
		if (elist) 
			elist->eval_driver();
		if (throwing) 
		{	if (throw_node->equal(_Pause)) 
			{	val=output_node;
				output_node = (PairP)Unbound;
				stoppingFlag=RUN;

				memcpy(&File::inputBlockingBuffer, &sav_iblk, sizeof(jmp_buf));

				File::inputBlocking = saveInputBlocking;
				Evaluator::val_status = save_val_status;
				if (uname) 
				{	ufun = uname;
				}

				theApp->edFrame->unmark(1);
				inpause=oldinpause;
				return val;
			}
			else if (throw_node->equal(_Error)) 
			{	error_print();
				stoppingFlag=RUN;
			}
		}
	}
	memcpy(&File::inputBlockingBuffer, &sav_iblk, sizeof(jmp_buf));

	File::inputBlocking = saveInputBlocking;
	File::unblockInput();
	Evaluator::val_status = save_val_status;
	if (uname) 
		ufun=uname;
	
	inpause=oldinpause;
	return val;
}

NodeP lcontinue(PairPP arg) 
{
	stoppingFlag=THROWING;
	throw_node=_Pause;
	if (arg!=argStack)
		output_node = arg[1];
	else
		output_node = (PairP)Unbound;
	return Unbound;
}
