#include <wx/dir.h>
#include <wx/filename.h>
#include <wx/file.h>
#include <wx/datstrm.h>
#include <wx/textctrl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
	#include <crtdbg.h>
	#include <io.h>
#endif
	
#include "Main.h"

#include "IntToPtr.h"
#include "PtrToInt.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "List.h"
#include "Line.h"
#include "Object.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Struct.h"
#include "TextControl.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

extern Function lContents, lprintOut;

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

bool File::interactive = true,
	 File::backslashes = false;

char File::digits=0, File::precision=6;

int File::stringlen=0, File::rstringlen=0,  File::indent=0,  
	File::marginX=0, File::marginY=0,
	File::toPending=0,
	File::inputBlocking=0;

const FILEP File::textcontrol=FILEP(1234567);

jmp_buf File::inputBlockingBuffer;

StringP File::currentLine=0,  File::deependProcName=0;

FileP File::list=0, File::Null=0, File::Stdin=0, File::Stdout=0,
		File::CIn=0, File::COut=0, File::Reader=0, File::Writer=0, 
		File::Loader=0, File::Dribbler=0, File::Stringbuffer=0,
		File::TextCtrl=0;

charP File::stringptr=0, File::stringstart=0, File::stringend=0,
		File::rstringptr=0, File::rstringstart=0, File::rstringend=0;

TextControlNodeP File::textControl=0;

void File::init()
{
	  Null  = new File(   0    , StringP(_Null_Word)          );
	 Stdin  = new File( stdin  , StringP(new String("stdin" )));
	 Stdout = new File( stdout , StringP(new String("stdout")));
	TextCtrl= new File(textcontrol, StringP(new String("textctrl" )));

	CIn   =Stdin;
	COut  =Stdout;
	Reader=Stdin;
	Loader=Stdin;
	Writer=Stdout;
	Dribbler    =Null;
	Stringbuffer=Null;
}

void File::inc_gc() {fifo << name() << mode() << tail();}

int  File::youngs() 
{	return gmark(name())+gmark(mode())+gmark(tail());
}
int  File::youngsPtr() 
{	return gmarkPtr(name())+gmarkPtr(mode())+gmarkPtr(tail());
}
void File::check_oldyoungs() 
{	check_oldyoung(name()); 
	check_oldyoung(mode()); 
	check_oldyoung(tail());
}

void File::setPtrs()
{	settail(FileP((  File*)GC_::intToPtr->find(int(_tail))));
	setname(StringP((String*)GC_::intToPtr->find(int(_name))));
	setmode(StringP((String*)GC_::intToPtr->find(int(_mode))));
	if (mode())
	{	setfile(fopen(name()->ptr(), mode()->ptr()));
		if (_pos)
			fseek(file(), _pos, 0);
	}
}
wxDataInputStream &File::read(wxDataInputStream &istr) 
{	istr >> (int&)_tail >> (int&)_name >> (int&)_mode >> _pos;
	int f;
	istr >> f;
	if (f==1)
		setfile(stdin);
	else if (f==2)
		setfile(stdout);
	else if (f==3)
		setfile(textcontrol);
	return istr;
}
wxDataOutputStream &File::write(wxDataOutputStream &ostr) 
{	if (file() && file()!=textcontrol)
		_pos=ftell(file());
	else
		_pos=0;
	ostr << type 
		<< GC_::ptrToInt->find(tail()) 
		<< GC_::ptrToInt->find(name()) 
		<< GC_::ptrToInt->find(mode())
		<< _pos;
	if (file()==stdin)
		ostr << int(1);
	else if (file()==stdout)
		ostr << int(2);
	else if (file()==textcontrol)
		ostr << int(3);
	else
		ostr << int(0);
	return ostr;
}

int File::eof() 
{	if (file())
		return feof(file());
	return rstringlen > 0 ? 0: 1;
}
int File::getc_() 
{	if (file())
		return getc(file());
	else
		if (rstringlen > 0)
			return *rstringptr++;
		else
			error(FILE_ERROR, new String
				("can't read after end of memory block"));
	return 0;
}
int File::ungetc(int c) {return ::ungetc(c, file());}
int File::putc_(int c) {return putc(c, file());}
int File::close() {int result=fclose(file()); setfile(0); return result;}
int File::pos() 
{	return file() ?  
		ftell(file()) : (stringend-stringstart)-stringlen;
}
int File::seek(int pos) 
{	if (file())
		return fseek(file(), pos, 0);
	else
		if (pos <= stringend-stringstart)
		{	stringptr=stringstart+pos;
			stringlen=stringend-stringptr;
			return 0;
		}
		else
			return -1;
}

size_t File::read(void *buffer, size_t size, size_t count) 
{	return fread(buffer, size, count, file());
}

void File::unblockInput() 
{	if (inputBlocking) 
	{	inputBlocking=0;
		longjmp( inputBlockingBuffer, 1);
	}
}

bool File::exists(const char *filename)
{
	FILE *fp = fopen(filename, "r");
	if (fp==0)
	{	// That fopen failed does _not_ definitely mean the file isn't there 
		// but for now this is ok
		return false;
	}
	fclose(fp);
	return true;
}

NodeP lisFile(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	return Node::boolnode(wxFileExists(name->ptr()));
}

NodeP lisDirectory(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	return Node::boolnode(wxDirExists(name->ptr()));
}

NodeP lgetWorkingDirectory(PairPP arg)
{
	return new String(wxGetCwd());
}

NodeP lmakeDirectory(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	if (!wxMkdir(name->ptr()))
		error(FILE_ERROR, new String(
			StringP(new String("I can't make the directory ")), name));
	return Unbound;
}

NodeP lFiles(PairPP arg)
{
	wxString filespec=wxEmptyString;
	if (exist(1) && arg[1])
	{	StringP fs=arg[1]->a_string(1);
		if throwing return Unbound;
		filespec=fs->ptr();
	}
	int flags=wxDIR_DEFAULT;
	if (exist(2))
	{	flags=arg[2]->a_int(2);
		if throwing return Unbound;
	}
	wxDir dir(wxGetCwd());

	if (!dir.IsOpened())
	{	error(FILE_ERROR, new String(
			wxString("I can't create the directory ")+filespec));
		return Unbound;
	}
	wxString filename;

	bool cont = dir.GetFirst(&filename, filespec, flags);
	PairP result=0, last=0;
	while (cont)
	{	StringP str=new String(filename);
		PairP l=new Pair(str, 0);
		if (last)
			last->settail(l);
		else
			result=l;
		last=l;
		cont = dir.GetNext(&filename);
	}
	return result;
}

FILEP String::open_file(charP access) 
{
	if (this==Unbound) return 0;

	charP fnstr = charP( GC_::malloc(length()+1) );
	if (!fnstr) return 0;
	
	copy_noparity(fnstr, ptr(), length());

	FILEP tstrm=fopen(fnstr, access);
	::free(fnstr);
	return tstrm;
}

NodeP ldribble(PairPP arg) 
{
	if (dribbler)
		error(ALREADY_DRIBBLING, 0);
	else
	{	StringP name = arg[1]->a_string(1);
		if (!throwing)
		{	FILEP file = name->open_file("w+");
			if (!file) 
			{	if (!throwing)
					error(FILE_ERROR, 0);
			}else 
				File::Dribbler = new File(file, 
											StringP(new String(name)), 
											StringP(new String("w+")), FileP(0));
		}
	}
	return Unbound;
}

NodeP lnoDribble(PairPP) 
{
	if (dribbler) 
	{	dribbler.close();
		File::Dribbler=File::Null;
	}
	return Unbound;
}

FileP Node::findInFilelist(bool remove) 
{
	return string()->findInFilelist(remove);
}
FileP File::findInFilelist(bool remove)
{
	FileP fp=0;

	if (remove)
	{	FileP prev=0;

		for (fp=File::list;  fp;  next(fp))
		{
			if (FileP(this)==fp)
			{	if (remove) 
				{	if (!prev)
						File::list = tail();
					else
						prev->settail( tail() );
				}
				break;
			}
			prev=fp;
		}
	}
	return fp;
}
FileP String::findInFilelist(bool remove)
{
	FileP fp=0, prev=0;
	
	for (fp=File::list;  fp;  next(fp))
	{
		if (equal(fp->name())) 
		{	if (remove) 
			{	if (!prev)
					File::list = fp->tail();
				else
					prev->settail( fp->tail() );
			}
			break;
		}
		prev=fp;
	}
	return fp;
}

NodeP Pair::open_file_helper(charP mode) 
{
	StringP name=a_string(1);

	if (name->findInFilelist(false))
		error(FILE_ERROR, new String("File already open"));
	else
	{	FILEP file = name->open_file(mode);
		if (file) 
			File::list = new File(file, StringP(new String(name)), 
											StringP(new String(mode)), File::list);
		else
			error(FILE_ERROR, new String(
				StringP(new String("I can't open file ")), name));
	}
	return Unbound;
}

NodeP lopenRead     (PairPP arg) {return arg[1]->open_file_helper("r");}
NodeP lopenReadBin  (PairPP arg) {return arg[1]->open_file_helper("rb");}
NodeP lopenWrite    (PairPP arg) {return arg[1]->open_file_helper("w");}
NodeP lopenWriteBin (PairPP arg) {return arg[1]->open_file_helper("wb");}
NodeP lopenAppend   (PairPP arg) {return arg[1]->open_file_helper("a");}
NodeP lopenAppendBin(PairPP arg) {return arg[1]->open_file_helper("ab");}
NodeP lopenUpdate   (PairPP arg) {return arg[1]->open_file_helper("a+");}
NodeP lopenUpdateBin(PairPP arg) {return arg[1]->open_file_helper("ab+");}

NodeP lallOpen(PairPP){return File::list;}

NodeP lclose(PairPP arg) 
{
	FileP file = arg[1]->findInFilelist(true);
	if (!file)
		error(FILE_ERROR, new String("File not open"));
	else
		file->close();

	return Unbound;
}

NodeP lsetWriter(PairPP arg) 
{
	if (noarg || !arg[1] || NodeP(arg[1])==File::Stdout) 
		 File::Writer=File::Stdout;
	else if (type_of(arg[1])==INTARRAY)
	{	IntArrayP ia=IntArrayP(arg[1]);
		File::stringptr=File::stringstart=charP(ia->data());
		File::stringlen=ia->length()*sizeof(int);
		File::stringend=File::stringstart+File::stringlen;
		File::Writer=File::Null;
	}
	else if (type_of(arg[1])==INT16ARRAY)
	{	Int16ArrayP ia=Int16ArrayP(arg[1]);
		File::stringptr=File::stringstart=charP(ia->data());
		File::stringlen=ia->length()*sizeof(short);
		File::stringend=File::stringstart+File::stringlen;
		File::Writer=File::Null;
	}
	else if (type_of(arg[1])==TEXTCONTROL)
	{	File::Writer=File::TextCtrl;
		File::textControl=TextControlNodeP(arg[1]);
	}
	else 
	{	FileP file = arg[1]->findInFilelist(false);
		if (file) 
			File::Writer=file;
		else
			error(FILE_ERROR, new String("File not open"));
	}
	return Unbound;
}

NodeP lsetReader(PairPP arg) 
{
	if (noarg || !arg[1] || NodeP(arg[1])==File::Stdin) 
	{	 File::Reader=File::Stdin;
	}
	else if (type_of(arg[1])==INTARRAY)
	{	IntArrayP ia=IntArrayP(arg[1]);
		File::rstringptr=File::rstringstart=charP(ia->data());
		File::rstringlen=ia->length()*sizeof(int);
		File::rstringend=File::rstringstart+File::rstringlen;
		File::Reader=File::Null;
	}
	else if (type_of(arg[1])==INT16ARRAY)
	{	Int16ArrayP ia=Int16ArrayP(arg[1]);
		File::rstringptr=File::rstringstart=charP(ia->data());
		File::rstringlen=ia->length()*sizeof(short);
		File::rstringend=File::rstringstart+File::rstringlen;
		File::Reader=File::Null;
	}
	else 
	{	FileP file = arg[1]->findInFilelist(false);
		if (file) 
			File::Reader=file;
		else
			error(FILE_ERROR, new String("File not open"));
	}
	return Unbound;
}

NodeP lReader  (PairPP){return File::Reader;}
NodeP lWriter  (PairPP){return File::Writer;}
NodeP lDribbler(PairPP){return File::Dribbler;}

NodeP leraseFile(PairPP arg) 
{
	StringP filename = arg[1]->string();
	if (filename==Unbound) return Unbound;

	charP fnstr = charP( GC_::malloc(filename->length()+1) );
	if (!fnstr) return Unbound;

	String::copy(fnstr, filename->ptr(), filename->length());
	remove(fnstr);
	free(fnstr);
	return Unbound;
}

NodeP lsave(PairPP arg) 
{
	FILEP oldfile = writer.file();
	StringP str=arg[1]->a_string(1);
	writer.setfile( str->open_file("w+") );

	if (writer) 
	{	arg[1] = PairP(lContents(argStack));
		lprintOut(arg);
		writer.close();
	}else
		error( FILE_ERROR, new List(new String("save: Could not open file"), str, _ListEnd) );

	writer.setfile(oldfile);
	return Unbound;
}

charP String::addSeparator(const char *path) 
{
	static char result[200];
	strcpy(result, path);
	if (result[0]) 
		strcat(result, "/");
	return result;
}

void Node::silent_load(const char *prefix) 
{
	// This procedure is called three ways:
	//	  silent_load(0, *argv)         loads *argv
	//	  silent_load(proc, logolib)    loads logolib/proc
	//	  silent_load(proc, 0)          loads proc.lg
	// The "/" or ".lg" is supplied by this procedure as needed.
	// In the case that this procedure is called to load a procedure from the
	// logo library, it must first truncate the name of the procedure to
	// eight characters, to find the filename (so as to be compatible with MS-DOS)

	if (throwing || (!prefix && !this)) return;

	char load_path[256];
	strcpy
	(	load_path, 
		!prefix ?  ""  
		:  (!this ?  prefix  
		:  String::addSeparator(prefix))
	);

	ValStatus save_val_status = Evaluator::val_status;
	StringP str=0;

	if (this) 
	{	str = string();
		if (str==Unbound) 
			return;
		if ( !strncmp(str->ptr(), ".", str->length()) ) 
			return;
		String::copy_noparitylow(
			load_path+strlen(load_path),  str->ptr(),  str->length());
//		if (prefix==0)
		strcat(load_path, ".lg");
	}
	wxFileName fullLoadPath(load_path);
	fullLoadPath.Normalize();
	wxString fullPath=fullLoadPath.GetFullPath();
	strcpy(load_path, fullPath);

	FILEP old_file = loader.file();
	StringP old_line=File::currentLine;
	loader.setfile( fopen(load_path, "r") );

	if (loader) 
	{
		_Startup->setvar(0);
		Line::readnr=1;
		Line::nr=1;
		Line::fn=Line::findFile(StringP(new String(load_path)));

//		cout << "silent_load Line::nr=1;" << endL;

		while (!loader.eof() && !throwing) 
		{
			File::currentLine = loader.read("");
			PairP exec_list = File::currentLine->parser(true);

			Evaluator::val_status=no_value_allowed;
			if (exec_list) 
				exec_list->eval_driver();
		}
		loader.close();
		loader.setfile(old_file);

		PairP st = PairP(_Startup->var());
		st->runstartup();
		Evaluator::val_status=save_val_status;
	}else 
		if (!str) 
			c_out << "File not found: " << prefix << endL;

	loader.setfile(old_file);
	File::currentLine=old_line;
}

NodeP lload(PairPP arg) 
{
	ValStatus save_val_status = Evaluator::val_status;
	StringP str = arg[1]->a_string(1),
	              old_line = File::currentLine;
	if throwing return Unbound;

	FILEP old_file = loader.file();

	wxFileName fullLoadPath(str->ptr());
	fullLoadPath.Normalize();
	wxString fullPath=fullLoadPath.GetFullPath();
	str=new String(fullPath);

	loader.setfile( str->open_file("r") );

	if (loader) 
	{
		_Startup->setvar(0);
		int oldreadnr=Line::readnr;
		int oldnr=Line::nr;
		StringP oldfn=Line::fn;
		Line::readnr=1;
		Line::nr=1;
		Line::fn=Line::findFile(str);
//		cout << "_load Line::nr=1;" << endL;
		
		while (!loader.eof() && !throwing) 
		{
			File::currentLine = loader.read("");
			PairP exec_list = File::currentLine->parser(true);
			
			Evaluator::val_status = no_value_allowed;
			if (exec_list) 
				exec_list->eval_driver();
		}
		loader.close();
		loader.setfile(old_file);
		Line::readnr=oldreadnr;
		Line::nr=oldnr;
		Line::fn=oldfn;
	
		PairP st = PairP(_Startup->var());
		st->runstartup();
		Evaluator::val_status=save_val_status;
	}else
		error( FILE_ERROR, new List(new String("load: Could not open file"), str, _ListEnd) );

	loader.setfile(old_file);
	File::currentLine=old_line;
	return Unbound;
}

NodeP lreadList(PairPP) 
{
	NodeP val=0;
	val = reader.read("")->parser(false);
	if (reader.eof()) 
		return _Null_Word;
	return val;
}

NodeP lreadWord(PairPP) 
{
	NodeP val=0;
	val = reader.read("RW"); // fake prompt flags no auto-continue
//	if (reader.eof())
//		return 0;
	return val;
}

NodeP lreadRawLine(PairPP) 
{
	NodeP val=0;
	val = reader.read("RAW"); // fake prompt flags no specials
//	if (reader.eof())
//		return 0;
	return val;
}

NodeP lreadChar(PairPP) 
{
	char c;
	File::charmodeOn();
	File::inputBlocking++;

	if (!setjmp(File::inputBlockingBuffer))
	{	
		if (File::interactive && reader.file()==stdin)
		{
			c=File::getch();

			if (c=='q'-'a'+1)		// control-q
			{	File::toPending=0;
				error(STOP_ERROR,0);
			}
			if (c=='p'-'a'+1)		// control-p
			{	logo_pause(0);
				return lreadChar(argStack);
			}
		}else
			c=reader.getc_();
	}
	File::inputBlocking=0;
	if (reader.eof()) return 0;

	return String::make
	(	&c, 0, 1,
		backslash.is(c) ?  STRING  
						:  BACKSLASHED, 
		String::copy
	);
}

NodeP lreadCharExt(PairPP) 
{
	int c;
	File::charmodeOn();
	File::inputBlocking++;

	if (!setjmp(File::inputBlockingBuffer))

	{	if (File::interactive && reader.file()==stdin)
			c=File::getchExt();
		else
			c=reader.getc_();
	}
	File::inputBlocking=0;
	if (reader.eof()) return 0;

	return new Int(c);
}

NodeP lreadChars(PairPP arg) 
{
	unsigned int c, i;
	NodeType type=STRING;
	String::HeadP strhead=0;
	charP strptr=0;
	
	c = arg[1]->a_pos_int(1)->int_;

	if (throwing) return Unbound;

	File::charmodeOn();
	File::inputBlocking++;

	if (!setjmp(File::inputBlockingBuffer))
	{	strhead = String::HeadP( GC_::malloc(c+sizeof(int)+1) );
		if (!strhead) return Unbound;

		strptr=strhead->str;
		reader.read(strptr, 1, c);
		strhead->refCount=0;
	}
	File::inputBlocking=0;

	if (throwing) return Unbound;
	if (reader.eof()) 
	{	free(strhead);
		return 0;
	}
	for (i=0; i < c; i++)
		if (backslash.is(strptr[i])) 
			type=BACKSLASHED;

	return String::make(strptr, strhead, (int)c, type, String::copy);
}

NodeP Node::readbin(FileP file) {error(BAD_DATA, this); return Unbound;}
NodeP Int8::readbin(FileP file)
{
	char num;
	File::charmodeOn();
	File::inputBlocking++;

	if (!setjmp(File::inputBlockingBuffer))
	{	
		if (File::interactive && file->file()==stdin)
			num=File::getch();
		else
			if (file->file())
			{	fread(&num, sizeof(char), 1, file->file());
				if (file->eof()) return 0;
			}else
				if (File::rstringlen > 0)
				{	File::rstringlen-=sizeof(char);
					if (File::rstringlen >= 0)
					{	num = *charP(File::rstringptr);	// reading from string
						File::rstringptr+=sizeof(char);
					}else
						error(FILE_ERROR, new String
							("memory block too small to read char"));
				}
	}
	File::inputBlocking=0;
	if (file->eof()) return 0;
	int_=num;
	return this;
}
NodeP UInt8::readbin(FileP file)
{
	unsigned char num;
	File::charmodeOn();
	File::inputBlocking++;

	if (!setjmp(File::inputBlockingBuffer))
	{	
		if (File::interactive && file->file()==stdin)
			num=File::getch();
		else
			if (file->file())
			{	fread(&num, sizeof(unsigned char), 1, file->file());
				if (file->eof()) return 0;
			}else
				if (File::rstringlen > 0)
				{	File::rstringlen-=sizeof(unsigned char);
					if (File::rstringlen >= 0)
					{	num = *charP(File::rstringptr);	// reading from string
						File::rstringptr+=sizeof(unsigned char);
					}else
						error(FILE_ERROR, new String
							("memory block too small to read UInt8"));
				}
	}
	File::inputBlocking=0;
	if (file->eof()) return 0;

	int_=num;
	return this;
}
NodeP Int16::readbin(FileP file)
{
	short num;
	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read int16 binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(short), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(short);
			if (File::rstringlen >= 0)
			{	num = *shortP(File::rstringptr);	// reading from string
				File::rstringptr+=sizeof(short);
			}else
				error(FILE_ERROR, new String
					("memory block too small to read short"));
		}
	int_=num;
	return this;
}
NodeP Int::readbin(FileP file)
{
	int num;
	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read int binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(int), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(int);
			if (File::rstringlen >= 0)
			{	num = *intP(File::rstringptr);	// reading from string
				File::rstringptr+=sizeof(int);
			}else
				error(FILE_ERROR, new String
					("memory block too small to read int"));
		}
	int_=num;
	return this;
}
NodeP Float::readbin(FileP file)
{
	double num;
	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read double binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(double), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(double);
			if (File::rstringlen >= 0)
			{	num = *doubleP(File::rstringptr);	// reading from string
				File::rstringptr+=sizeof(double);
			}else
				error(FILE_ERROR, new String
					("memory block too small to read float"));
		}
	float_=num;
	return this;
}
NodeP Complexd::readbin(FileP file)
{
	complexd num;
	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read complex binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(&num, sizeof(complexd), 1, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(complexd);
			if (File::rstringlen >= 0)
			{	num = *complexdP(File::rstringptr);	// reading from string
				File::rstringptr+=sizeof(complexd);
			}else
				error(FILE_ERROR, new String
					("memory block too small to read complex"));
		}
	complex_=num;
	return this;
}
NodeP String::readbin(FileP file)
{
	charP cp=ptr();

	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read word binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(cp, sizeof(char), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(char)*_length;
			if (File::rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*cp++ = *charP(File::rstringptr);	// reading from string
					File::rstringptr+=sizeof(char);
				}
			else
				error(FILE_ERROR, new String
					("memory block too small to read word"));
		}
	return this;
}
NodeP IntArray::readbin(FileP file)
{
	intP ip=data();

	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read intarray binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(int), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(int)*_length;
			if (File::rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *intP(File::rstringptr);	// reading from string
					File::rstringptr+=sizeof(int);
				}
			else
				error(FILE_ERROR, new String
					("memory block too small to read intarray"));
		}
	return this;
}
NodeP Int16Array::readbin(FileP file)
{
	shortP ip=data();

	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read int16array binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(short), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(short)*_length;
			if (File::rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *shortP(File::rstringptr);	// reading from string
					File::rstringptr+=sizeof(short);
				}
			else
				error(FILE_ERROR, new String
					("memory block too small to read int16array"));
		}
	return this;
}
NodeP FloatArray::readbin(FileP file)
{
	doubleP ip=data();

	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read floatarray binary from keyboard"));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(double), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (File::rstringlen > 0)
		{	File::rstringlen-=sizeof(double)*_length;
			if (File::rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *doubleP(File::rstringptr);	// reading from string
					File::rstringptr+=sizeof(double);
				}
			else
				error(FILE_ERROR, new String
					("memory block too small to read floatarray"));
		}
	return this;
}
NodeP Struct::readbin(FileP file)
{
	if (File::interactive && file->file()==stdin)
	{	error(FILE_ERROR, new String
			("Can't read struct binary from keyboard"));
		return Unbound;
	}
	for (ObjectP obj=first();  obj;  obj=ObjectP(obj->proc()))
	{	NodeP nd=obj->var()->readbin(file);
		obj->setvar(nd);
	}
	if (file->eof()) return 0;
	return this;
}

NodeP lreadInt8Bin(PairPP) 
{
	NodeP n=new Int8();
	return n->readbin(File::Reader);
}

NodeP lreadUInt8Bin(PairPP) 
{
	NodeP n=new UInt8();
	return n->readbin(File::Reader);
}

NodeP lreadInt16Bin(PairPP) 
{
	NodeP n=new Int16();
	return n->readbin(File::Reader);
}

NodeP lreadIntBin(PairPP) 
{
	NodeP n=new Int();
	return n->readbin(File::Reader);
}

NodeP lreadFloatBin(PairPP) 

{
	NodeP n=new Float();
	return n->readbin(File::Reader);
}

NodeP lreadComplexBin(PairPP) 
{
	NodeP n=new Complexd();
	return n->readbin(File::Reader);
}

NodeP lreadWordBin(PairPP arg) 
{
	int length=arg[1]->a_int(1);
	if throwing return Unbound;

	StringP s=new String(length);
	if (throwing || !s) return Unbound;
	return s->readbin(File::Reader);
}

NodeP lreadIntArrayBin(PairPP arg) 
{
	int size=arg[1]->a_int(1);
	if throwing return Unbound;

	IntArrayP a=new IntArray(size);
	if (throwing || !a) return Unbound;
	return a->readbin(File::Reader);
}

NodeP lreadInt16ArrayBin(PairPP arg) 
{
	int size=arg[1]->a_int(1);
	if throwing return Unbound;

	Int16ArrayP a=new Int16Array(size);
	if (throwing || !a) return Unbound;
	return a->readbin(File::Reader);
}

NodeP lreadFloatArrayBin(PairPP arg) 
{
	int size=arg[1]->a_int(1);
	if throwing return Unbound;

	FloatArrayP a=new FloatArray(size);
	if (throwing || !a) return Unbound;
	return a->readbin(File::Reader);
}

NodeP lreadStructBin(PairPP arg) 
{
	PairP li=arg[1]->a_list(1);

	if (!li) 
		error(BAD_DATA, li);
	if throwing return Unbound;

	StructP s=new Struct(li);
	if (throwing || !s) return Unbound;
	return s->readbin(File::Reader);
}

NodeP lisEof(PairPP) 
{
	int c=reader.getc_();
	if (c==EOF) return _True;
	reader.ungetc(c);
	return _False;
}

NodeP lisKey(PairPP) 
{
	if (reader.file()==stdin && File::interactive) 
	{	File::charmodeOn();
		fflush(stdout);
		return Node::boolnode(File::kbhit() != 0);
	}else
	{	int c=reader.getc_();
		if (reader.eof())
			return _False;
		else 
		{	reader.ungetc(c);
			return _True;
		}
	}
}

NodeP lReaderPos(PairPP) {return new Int(reader.pos());}
NodeP lWriterPos(PairPP) {return new Int(writer.pos());}

NodeP lsetReaderPos(PairPP arg) 
{
	IntP pos = arg[1]->a_pos_int(1);
	if (!throwing) 
		reader.seek(pos->int_);
	return Unbound;
}

NodeP lsetWriterPos(PairPP arg) 
{
	IntP pos = arg[1]->a_pos_int(1);

	if (!throwing) 
		writer.seek(pos->int_);
	return Unbound;
}

NodeP lFileSize(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	FileP file = filename->findInFilelist(false);
	if (file) 
#ifdef WIN32
		return new Int(_filelength(file->file()->_file));
#else
	{	wxFile wxfile(file->name()->ptr());
		return new Int(wxfile.Length());
	}	
//		error(NOT_YET_SUPPORTED, new String("FileSize"));
//		return Unbound;
#endif
	else
		error(FILE_ERROR, new String("File not open"));
	return Unbound;
}

NodeP lFileTime(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	time_t fmt=wxFileModificationTime(filename->ptr());
	wxDateTime t(fmt);
	PairP result=new List(
		new Int(t.GetHour()), new Int(t.GetMinute()), new Int(t.GetSecond()), 
		new Int(t.GetMillisecond()), 
		new Int(t.GetWeekDay()), 
		new Int(t.GetDay()), new Int(t.GetMonth()+1), new Int(t.GetYear()), _ListEnd);
	// There's a bug in GetMonth(), so we need the +1!
	return result;
}

NodeP lLogoComspec(PairPP)
{
	return new String(environment.comspec);

}


NodeP lsetLogoComspec(PairPP arg)
{
	StringP dir=arg[1]->a_string(1);
	if throwing return Unbound;
	StringP d=new String(dir);
	environment.comspec = d->ptr();
	return Unbound;
}

NodeP lLogoEditor(PairPP)
{
	return new String(environment.editor);
}

NodeP lsetLogoEditor(PairPP arg)
{
	StringP dir=arg[1]->a_string(1);
	if throwing return Unbound;
	StringP d=new String(dir);
	environment.editor = d->ptr();
	environment.editorname = environment.editor.AfterLast('/');
	return Unbound;
}

NodeP lLogoHelpDir(PairPP)
{
	return new String(environment.helpfiles);
}

NodeP lsetLogoHelpDir(PairPP arg)
{
	StringP dir=arg[1]->a_string(1);
	if throwing return Unbound;
	StringP d=new String(dir);
	environment.helpfiles = d->ptr();
	return Unbound;
}

NodeP lLogoLibDir(PairPP)
{
	return new String(environment.logolib);
}

NodeP lsetLogoLibDir(PairPP arg)
{
	StringP dir=arg[1]->a_string(1);
	if throwing return Unbound;
	StringP d=new String(dir);
	environment.logolib = d->ptr();
	return Unbound;
}

NodeP lLogoTempDir(PairPP)
{
	return new String(environment.tempdir);
}

NodeP lsetLogoTempDir(PairPP arg)
{
	StringP dir=arg[1]->a_string(1);
	if throwing return Unbound;
	StringP d=new String(dir);
	environment.tempdir = d->ptr();
	return Unbound;
}
