//___________________________________________________________________
//
//                           Class  Frame
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/textctrl.h>
#include <wx/sizer.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Float.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "BoxSizer.h"
#include "Point.h"
#include "Graph.h"
#include "Frame.h"
#include "UInt8.h"
#include "Color.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"

#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Frame, wxFrame)
	EVT_CHAR		(Frame::OnChar		)
	EVT_KEY_DOWN(Frame::OnKeyDown	)
	EVT_KEY_UP	(Frame::OnKeyUp	)
	EVT_CLOSE	(Frame::OnClose	)
END_EVENT_TABLE()

Frame::Frame(FrameNodeP frameNode_):
onChar(0), onKeyUp(0), onKeyDown(0), shape(0), boxSizer(0),
frameNode(frameNode_)
{
	wxWindow *parent=0;
	if (frameNode->parentFrame)
		parent=frameNode->parentFrame->frame;
	else if (frameNode->parentGraph)
		parent=frameNode->parentGraph->graph;
	Create(parent, -1, "Frame",
			wxDefaultPosition, wxDefaultSize, 
			wxDEFAULT_FRAME_STYLE, "Frame");
}

Frame::Frame(FrameNodeP frameNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString& title_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxString& name):
wxFrame(parent, id, title_, pos, size, style, name),
onChar(0), onKeyUp(0), onKeyDown(0), shape(0), boxSizer(0),
frameNode(frameNode_),
title(title_)
{
}

Frame::~Frame()
{
	if (frameNode)
		frameNode->frame=0;
	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::FrameChar:
		case Event::FrameKeyDown:
		case Event::FrameKeyUp:
			if (((EventFrame*)ev)->frame==this)
				((EventFrame*)ev)->frame=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void Frame::OnClose(wxCloseEvent &event)
{
	if (frameNode)
		frameNode->frame=0;
	event.Skip();
}

#define FRAME_ON(NAME)															\
void Frame::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventFrame##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

FRAME_ON(Char)
FRAME_ON(KeyDown)
FRAME_ON(KeyUp)

bool Frame::setShape(PairP list)
{
	if (!list)
		return true;
	int n=list->count();
	wxPoint *points=new wxPoint[n];
	int i=0;
	while (list)
	{	PairP p=list->first()->a_vector(2);
		if throwing 
		{	delete points;
			return false;
		}
		int x=int(p->first ()->getnumber()),
			 y=int(p->second()->getnumber());
		points[i]=wxPoint(x,y);
		i++;
		next(list);
	}
	SetShape(wxRegion(n, points));
	delete points;
	return true;
}

wxString Frame::wxstring()
{
	wxString str;
	str << "Frame [" 
		<< GetTitle() << "]["
		<< GetPosition().x << ' ' << GetPosition().y << "][" 
		<< GetSize().x << ' ' << GetSize().y << "] "
		<< GetWindowStyle() << " [" << GetName() << ']';
	return str;
}

wxDataInputStream &Frame::read(wxDataInputStream &istr) 
{	
	wxString name, fontName;

	wxPoint pos;
	wxSize size;
	int style, fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> title >> pos.x >> pos.y >> size.x >> size.y >> style >> name
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> (int&)shape
		>> (int&)boxSizer;

	SetTitle(title); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetName(name); 
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont font=GetFont();
	font.SetPointSize(fontSize);
	font.SetFaceName(fontName);
	font.SetStyle(fontStyle);
	font.SetWeight(fontWeight);
	SetFont(font);
	return istr;
}
wxDataOutputStream &Frame::write(wxDataOutputStream &ostr) 
{
	wxColour f=GetForegroundColour(), b=GetBackgroundColour();
	wxFont font=GetFont();
	ostr
		<< GetTitle() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle()) << GetName()
		<< f.Red() << f.Green() << f.Blue() 
		<< b.Red() << b.Green() << b.Blue() 
		<< font.GetPointSize() << font.GetFaceName() 
		<< font.GetStyle() << font.GetWeight()
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GC_::ptrToInt->find(shape)
		<< GC_::ptrToInt->find(boxSizer);
	return ostr;
}

#define NODE_ON(x)						\
PairP FrameNode::on##x()				\
{												\
	if (frame)								\
		return frame->on##x;				\
	return 0;								\
}												\
												\
void FrameNode::seton##x(NodeP nd)	\
{												\
	if (frame)								\
	{	frame->on##x = PairP(nd);		\
		check_oldyoung(nd);				\
	}											\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)

PairP FrameNode::shape()
{
	if (frame)
		return frame->shape;
	return 0;
}

void FrameNode::realsetshape(PairP shape_)
{
	if (frame && frame->setShape(shape_))
		setshape(shape_);
}

void FrameNode::setshape(PairP shape_)
{
	if (frame)
	{	frame->shape=shape_;
		check_oldyoung(shape_);
	}
}

BoxSizerNodeP FrameNode::boxSizer()
{
	if (frame)
		return frame->boxSizer;
	return 0;
}

void FrameNode::setSizer(BoxSizerNodeP boxSizer_)
{
	if ((frame!=0) && boxSizer_)
	{	frame->boxSizer=boxSizer_;
		check_oldyoung(boxSizer_);
		frame->SetSizer(boxSizer_->boxSizer);
		frame->Layout();
	}
}

void FrameNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void FrameNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

FrameNode::FrameNode(): 
	Node(FRAME), frame(0), 	
	parentFrame(0),
	parentGraph(0),
	children(0) 
{}

FrameNode::FrameNode(FrameNodeP parentFrame_,
							GraphNodeP parentGraph_): 
	Node(FRAME), frame(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_),
	children(0)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void FrameNode::free() 
{	
	if (frame) 
	{	frame->frameNode=0;
		WindowPairP c=children;
		while (c)
		{	if (c->first())
				c->first()->setParentFrame(0);
			c=WindowPairP(c->tail());
		}				
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		frame->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::FrameChar:
			case Event::FrameKeyDown:
			case Event::FrameKeyUp:
				if (((EventFrame*)ev)->frame==frame)
					((EventFrame*)ev)->frame=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		frame=0;
	}
}
void FrameNode::inc_gc() 
{	fifo << children << onChar() << onKeyDown() << onKeyUp() << shape();
}
int FrameNode::youngs() 
{	return gmark(children)
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp())
		+gmark(shape());
}
int FrameNode::youngsPtr()
{	return gmarkPtr(children)
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp())
		+gmarkPtr(shape());
}
void FrameNode::check_oldyoungs() 
{	check_oldyoung(children);
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
	check_oldyoung(shape());
}

wxDataInputStream &FrameNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr 
			>> (int&)parentFrameNode 
			>> (int&)parentGraphNode
			>> (int&)children;
		frame=new Frame(FrameNodeP(this));
		frame->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &FrameNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (frame)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame)
			<< GC_::ptrToInt->find(parentGraph)
			<< GC_::ptrToInt->find(children);
		frame->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void FrameNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	children = (WindowPair*)GC_::intToPtr->find(int(children));
	check_oldyoung(children);
	if (frame)
	{	setonChar   ((Pair*)GC_::intToPtr->find(int(frame->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(frame->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(frame->onKeyUp  )));
		setshape		(PairP((Pair*)GC_::intToPtr->find(int(frame->shape	   ))));
		frame->boxSizer = (BoxSizerNode*)GC_::intToPtr->find(int(frame->boxSizer));
		frame->Show();
	}
}

void FrameNode::aftersetPtrs()
{
	if (parentFrame || parentGraph)
		return;

	recreate();
}

void FrameNode::recreate()
{
	FrameP f=frame, f2=0;
	if (!frame)
		return;
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	f2=new Frame(FrameNodeP(this),
			p, -1, f->GetTitle(), f->GetPosition(), 
			f->GetSize(), f->GetWindowStyle(), f->GetName());
	f2->SetForegroundColour(f->GetForegroundColour());
	f2->SetBackgroundColour(f->GetBackgroundColour());
	wxFont font=f->GetFont();
	f2->SetFont(font);
	frame=f2;
	setonChar   (f->onChar   );
	setonKeyDown(f->onKeyDown);
	setonKeyUp  (f->onKeyUp  );
	realsetshape(f->shape    );
	frame->boxSizer=f->boxSizer;
	delete f;

	frame=f2;
	frame->Show();

	PairP c=children;
	while (c)
	{	if (c->first())
			c->first()->recreate();
		next(c);
	}
}

void FrameNode::aftersetPtrs3()
{
	if (frame)
		setSizer(frame->boxSizer);
}

StringP FrameNode::string()
{
	if (frame)
		return new String(frame->wxstring());
	return _Null_Word->name();
}

wxString FrameNode::wxstring()
{
	if (frame)
		return frame->wxstring();
	return wxEmptyString;
}

NodeP lFrame(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=0;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=0;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString title=_T("Frame");
	if (exist(2))
	{	StringP t=arg[2]->a_string(2);
		if throwing return Unbound;
		title=t->ptr();
	}
	long style=wxDEFAULT_FRAME_STYLE;
	if (exist(3) && arg[3])
	{	int s=arg[3]->a_int(3);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4) && arg[4])
	{	PairP p=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(5) && arg[5])
	{	PairP s=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=title;
	if (exist(6))
	{	StringP n=arg[6]->a_string(6);
		if throwing return Unbound;
		name=n->ptr();
	}
	FrameNodeP frameNode=new FrameNode(
		parentFrameNode,
		parentGraphNode);
	FrameP frame=new Frame(frameNode,
			parent, id, title, pos, size, style, name);
	frame->Show();
	frameNode->frame=frame;
	return frameNode;
}

NodeP lFrameDestroy(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;

	WindowPairP c=frame->children;
	while (c)
	{	if (c->first())
			c->first()->setParentFrame(0);
		c=WindowPairP(c->tail());
	}				
	{	FrameNodeP p=frame->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(frame));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=frame->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(frame));
			p->check_oldyoung(p->children);
		}
	}
	frame->frame->Destroy();
	frame->frame=0;
	return Unbound;
}

NodeP lFrameSetFocus(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	frame->frame->SetFocus();
	return Unbound;
}

NodeP lFrameEnable(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	frame->frame->Enable(state!=0);
	return Unbound;
}

NodeP lFrameMaximize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	frame->frame->Maximize(state!=0);
	return Unbound;
}

NodeP lFrameIconize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	frame->frame->Iconize(state!=0);
	return Unbound;
}

NodeP lFrameFullScreen(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	int style=arg[3]->a_int(3);
	if throwing return Unbound;
	frame->frame->ShowFullScreen(state!=0, style);
	return Unbound;
}

NodeP lFrameSetClientSize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int w=arg[2]->a_int(2);
	if throwing return Unbound;
	int h=arg[3]->a_int(3);
	if throwing return Unbound;
	frame->frame->SetClientSize(w,h);
	return Unbound;
}

NodeP lFrameSetBackgroundColor(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	frame->frame->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lFrameSetColor(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	frame->frame->SetForegroundColour(wc);
	return Unbound;
}

NodeP lFrameSetFontSize(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=frame->frame->GetFont();
	f.SetPointSize(s);
	frame->frame->SetFont(f);
	return Unbound;
}

NodeP lFrameSetFontName(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=frame->frame->GetFont();
	f.SetFaceName(s->ptr());
	frame->frame->SetFont(f);
	return Unbound;
}

NodeP lFrameSetFontStyle(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=frame->frame->GetFont();
	f.SetStyle(s);
	frame->frame->SetFont(f);
	return Unbound;
}

NodeP lFrameSetFontWeight(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=frame->frame->GetFont();
	f.SetWeight(s);
	frame->frame->SetFont(f);
	return Unbound;
}

NodeP lFrameSetShape(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	frame->realsetshape(list);
	return Unbound;
}

NodeP lFrameSetSizer(PairPP arg)
{
	FrameNodeP frame=arg[1]->a_frame(1);
	if throwing return Unbound;
	BoxSizerNodeP boxSizer=arg[2]->a_boxSizer(2);
	if throwing return Unbound;
	frame->setSizer(boxSizer);
	return Unbound;
}

#define _FRAME_ON(NAME)								\
NodeP lFrameOn##NAME(PairPP arg)					\
{															\
	FrameNodeP frame=arg[1]->a_frame(1);		\
	if throwing return Unbound;					\
	PairP commandlist=arg[2]->a_runnable(2);	\
	if throwing return Unbound;					\
	frame->seton##NAME(commandlist);				\
	return Unbound;									\
}

_FRAME_ON(Char)
_FRAME_ON(KeyDown)
_FRAME_ON(KeyUp)
