//___________________________________________________________________
//
//                          Class Graph
//
//                          Definitions
//___________________________________________________________________

#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
	#define DOSWIN32
	#include <windows.h>
	#include <vfw.h>
#endif

#include <wx/glcanvas.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/dcmemory.h>
#include <wx/html/helpctrl.h>
#include <wx/log.h>

#include <wx/image.h>
#include <wx/datstrm.h>
#include <wx/txtstrm.h>
#include <wx/metafile.h>

#include <GL/gl.h>
#include <GL/glu.h>

#ifndef WIN32
	#include <setjmp.h>
	#define FT_FREETYPE_H </usr/include/freetype2/freetype/freetype.h>
	#define FT_ERRORS_H </usr/include/freetype2/freetype/fterrors.h>
	#define FT_CONFIG_CONFIG_H </usr/include/freetype2/freetype/config/ftconfig.h>
	#define FT_CONFIG_OPTIONS_H </usr/include/freetype2/freetype/config/ftoption.h>
	#define FT_CONFIG_STANDARD_LIBRARY_H </usr/include/freetype2/freetype/config/ftstdlib.h>
	#define FT_MODULE_ERRORS_H </usr/include/freetype2/freetype/ftmoderr.h>
	#define FT_BEGIN_HEADER
	#define FT_END_HEADER
	#define FT_ERROR_DEFINITIONS_H </usr/include/freetype2/freetype/fterrdef.h>
	#define FT_TYPES_H </usr/include/freetype2/freetype/fttypes.h>
	#define FT_SYSTEM_H </usr/include/freetype2/freetype/ftsystem.h>
	#define FT_IMAGE_H </usr/include/freetype2/freetype/ftimage.h>
	#define FT_GLYPH_H </usr/include/freetype2/freetype/ftglyph.h>
#endif
#include <FTGL/FTGLPolygonFont.h>

#include <gl2ps.h>

#include "Main.h"
#include "Error.h"
#include "File.h"
#include "Int.h"
#include "UInt8.h"
#include "Pair.h"
#include "List.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Frame.h"
#include "Init.h"
#include "Eval.h"
#include "Stack.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Mouse.h"
#include "FTGLFontManager.h"
#include "Font.h"
#include "Turtle.h"
#include "TurtleNode.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Recorder.h"

#ifdef WIN32
	#include "AVIGen.h"
#endif

#include "IntToPtr.h"
#include "PtrToInt.h"

#include "Event.h"
#include "Splitter.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Shadow.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

BEGIN_EVENT_TABLE(Graph, wxGLCanvas)
	EVT_SIZE					(Graph::OnSize					)
	EVT_ERASE_BACKGROUND	(Graph::OnErase				) 
	EVT_CHAR					(Graph::OnChar					)
	EVT_KEY_DOWN			(Graph::OnKeyDown				)
	EVT_KEY_UP				(Graph::OnKeyUp				)
	EVT_LEFT_DOWN			(Graph::OnMouseLeftDown		)
	EVT_RIGHT_DOWN			(Graph::OnMouseRightDown	)
	EVT_MIDDLE_DOWN		(Graph::OnMouseMiddleDown	)
	EVT_LEFT_UP				(Graph::OnMouseLeftUp		)
	EVT_RIGHT_UP			(Graph::OnMouseRightUp		)
	EVT_MIDDLE_UP			(Graph::OnMouseMiddleUp		)
	EVT_LEFT_DCLICK		(Graph::OnMouseLeftDClick	)	
	EVT_RIGHT_DCLICK		(Graph::OnMouseRightDClick	)
	EVT_MIDDLE_DCLICK		(Graph::OnMouseMiddleDClick)
	EVT_MOTION				(Graph::OnMouseMotion		)
	EVT_MOUSE_EVENTS		(Graph::OnMouse				)
END_EVENT_TABLE()

void mySwapBuffers()
{
	graph->swapBuffers();
}

int Graph::attribs[]=
#ifndef WIN32
	{};
#else
	{	WX_GL_RGBA,
		WX_GL_DOUBLEBUFFER,
		WX_GL_DEPTH_SIZE,
		32,
		WX_GL_STENCIL_SIZE,
		1,
		0
	};

#endif

Graph::Graph(GraphNodeP graphNode_):
	wxGLCanvas((wxWindow*)theApp->mainFrame->splitter, wxID_ANY,
		wxDefaultPosition, wxDefaultSize, 0,
		wxGLCanvasName, attribs, wxNullPalette),
turtle(0),
onChar(0), onKeyUp(0), onKeyDown(0), 
onMouseLeftDown(0)  , onMouseRightDown(0)  , onMouseMiddleDown(0)  ,
onMouseLeftUp(0)    , onMouseRightUp(0)    , onMouseMiddleUp(0)    ,
onMouseLeftDClick(0), onMouseRightDClick(0), onMouseMiddleDClick(0),
onMouseMotion(0),
graphNode(graphNode_),
turtleNode(0)
{
	wxWindow *parent=0;
	if (graphNode && graphNode->parentFrame)
		parent=graphNode->parentFrame->frame;
	else if (graphNode && graphNode->parentGraph)
		parent=graphNode->parentGraph->graph;
	else
		parent=theApp->mainFrame->graph;

	create(parent, -1, wxDefaultPosition, wxDefaultSize, 
		wxFULL_REPAINT_ON_RESIZE, "Graph", attribs,
		wxNullPalette, false);
}

Graph::Graph(GraphNodeP graphNode_, 
				 wxWindow* parent, wxWindowID id, 
				 const wxPoint& pos, const wxSize& size, 
				 long style, const wxString& name, 
				 int* attribList, const wxPalette& palette,
				 bool pictureAutosave):
wxGLCanvas(parent, id, pos, size, style, name, attribs),
onChar(0), onKeyUp(0), onKeyDown(0),
onMouseLeftDown(0)  , onMouseRightDown(0)  , onMouseMiddleDown(0)  ,
onMouseLeftUp(0)    , onMouseRightUp(0)    , onMouseMiddleUp(0)    ,
onMouseLeftDClick(0), onMouseRightDClick(0), onMouseMiddleDClick(0),
onMouseMotion(0),
graphNode(graphNode_),
turtleNode(0)
{
	create(parent, id, pos, size, 
		style | wxFULL_REPAINT_ON_RESIZE, name, attribs,
		palette, pictureAutosave);

	turtle=new Turtle;
	recorder=new Recorder(this);
	recorderSave=recorder;
	turtle->create(recorder, screen);
}

void Graph::create(wxWindow* parent, wxWindowID id, 
				 const wxPoint& pos, const wxSize& size, 
				 long style, const wxString& name, 
				 int* attribList, const wxPalette& palette,
				 bool pictureAutosave)
{
	inread=false;
	was1=printing=saving=false;
	oldPalette = NULL;
	doubleBuffer=theApp->doubleBuffer;
	if (!doubleBuffer)
	{	glReadBuffer(GL_FRONT);
		glDrawBuffer(GL_FRONT);
	}
	screen=new Screen_();
	mouse=new Mouse();

	wxSize s=GetClientSize();
	screen->create(s.GetWidth(), s.GetHeight());

	saveSize=wxSize(800,600);

	GraphP og=graph;
	graph=this;
	SetCurrent();
	setOpenGLState();
	graph=og;

	aviGen=0;
	autosave=pictureAutosave;
	pictureFileName=theApp->pictureFileName;
	shadowObject=new ShadowObject;
}

Graph::~Graph()
{
	reset();
	if (graphNode)
		graphNode->graph=0;
	if (recorder)
	{	recorder->refCount--;
		if (recorder->refCount==0)
			delete recorder;
	}
	if (screen)
		screen->destroy();
	if (mouse)
		delete mouse;
	if (screen)
		delete screen;
	if (shadowObject)
		delete shadowObject;

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::GraphChar:
		case Event::GraphKeyDown:
		case Event::GraphKeyUp:
			if (((EventGraph*)ev)->graph==this)
				((EventGraph*)ev)->graph=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void Graph::reset()
{
	if (::graph==this)
	{	if (::graph && ::graph->m_glContext 
		&&  ::graph != theApp->mainFrame->graph
		&&  theApp->mainFrame->initialized)
		{	::graph=theApp->mainFrame->graph;
#ifdef WIN32		
			wglMakeCurrent(
				(HDC)::graph->m_glContext->m_hDC,
				::graph->m_glContext->m_glContext);
#else
			glXMakeCurrent((Display*) wxGetDisplay(),
				Window(::graph->m_glContext->m_window->GetClientAreaWindow()),
				::graph->m_glContext->m_glContext);
#endif
		}
	}
}

#define Graph_ON(NAME)															\
void Graph::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventGraph##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

Graph_ON(Char)
Graph_ON(KeyDown)
Graph_ON(KeyUp)

wxString Graph::wxstring()
{
	wxString str;
	str << "Graph [" 
		<< GetName() << "]["
		<< GetPosition().x << ' ' << GetPosition().y << "] [" 
		<< GetSize().x << ' ' << GetSize().y << "] " 
		<< GetWindowStyle() << " [" << GetName() << ']';
	return str;
}

wxDataInputStream &Graph::read(wxDataInputStream &istr) 
{	
	inread=true;
	wxString title, name;
	wxPoint pos;
	int style;
	
	recorder=new Recorder(this);
	recorderSave=recorder;

	istr >> title >> pos.x >> pos.y >> size.x >> size.y >> style >> name
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> (int&)onMouseLeftDown
		>> (int&)onMouseRightDown
		>> (int&)onMouseMiddleDown
		>> (int&)onMouseLeftUp
		>> (int&)onMouseRightUp
		>> (int&)onMouseMiddleUp
		>> (int&)onMouseLeftDClick
		>> (int&)onMouseRightDClick
		>> (int&)onMouseMiddleDClick
		>> (int&)onMouseMotion
		>> (int&)turtleNode
		>> active
		>> screen->eye >> screen->center >> screen->up
		>> screen->Min.x >> screen->Min.y >> screen->Min.z
		>> screen->Max.x >> screen->Max.y >> screen->Max.z



		>> *recorder;
//	turtle->read(istr);
//	turtle->refCount+=2;

	SetTitle(title); 
	SetPosition(pos);
	SetWindowStyle(style);
	SetName(name); 
	return istr;
}

wxDataOutputStream &Graph::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetTitle() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle()) << GetName();

#define WRITE_ON(NAME)	ostr << GC_::ptrToInt->find(on##NAME);
	WRITE_ON(Char)
	WRITE_ON(KeyDown)
	WRITE_ON(KeyUp)
	WRITE_ON(MouseLeftDown)
	WRITE_ON(MouseRightDown)
	WRITE_ON(MouseMiddleDown)

	WRITE_ON(MouseLeftUp)
	WRITE_ON(MouseRightUp)
	WRITE_ON(MouseMiddleUp)

	WRITE_ON(MouseLeftDClick)
	WRITE_ON(MouseRightDClick)
	WRITE_ON(MouseMiddleDClick)

	WRITE_ON(MouseMotion)
	ostr << GC_::ptrToInt->find(turtleNode);

	if (this==::graph)
		ostr << 1;
	else
		ostr << 0;
	ostr << screen->eye << screen->center << screen->up
		<< screen->Min.x << screen->Min.y << screen->Min.z
		<< screen->Max.x << screen->Max.y << screen->Max.z;
	ostr << *recorder;
//	turtle->write(ostr);
	return ostr;
}

void Graph::setPtrs()
{	
	turtleNode=(TurtleNode*)GC_::intToPtr->find(int(turtleNode));
	turtle=turtleNode->turtle;
	recorder->turtle=turtle;
	turtle->create(recorder, screen);

	CommandListIterator i(recorder);
	for (i=recorder->begin();  i;  i++)
		i->setPtrs();

	if (active)
	{
#ifdef WIN32
		wglMakeCurrent(
			(HDC)m_glContext->m_hDC, m_glContext->m_glContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(m_glContext->m_window->GetClientAreaWindow()),
			m_glContext->m_glContext);
#endif
		::graph=this;
	}
	inread=false;
	SetSize(size);
	wxSizeEvent sev(size);
	OnSize(sev);
}

void Graph::setOpenGLState()
{
	GLfloat fMaxObjSize, fAspect;

	glClearDepth(1.0f);
	glDepthFunc(GL_LEQUAL);
//	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	GLfloat light_ambient[] = {0.1f,0.1f,0.1f,1.0f};
	GLfloat light_diffuse[] = {1,1,1,1};
	GLfloat light_specular[] = {1,1,1,1};
	/*  light_position is NOT default value */
	GLfloat light_position[] = {1,1,1,0};

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

//	glEnable (GL_LIGHTING);
//	glEnable (GL_LIGHT0);

//	glEnable(GL_COLOR_MATERIAL);
//	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	if (oldRect.GetHeight())
		fAspect = (GLfloat)oldRect.GetWidth()/oldRect.GetHeight();
	else    // don't divide by zero, not that we should ever run into that...
		fAspect = 1.0f;
	nearPlane = 3.0f;
	farPlane = 10000;//7.0f;
	fMaxObjSize = (screen->Max.x-screen->Min.x)/2;//3.0f;
	fRadius = nearPlane + fMaxObjSize/2;

	setFrustum(
		screen->Max.x - screen->Min.x, 
		screen->Max.y - screen->Min.y); 
	screen->getMatrices();
	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

//	glPushMatrix();
//	glTranslatef(0, 0, -fRadius);
}

void Graph::OnPaint(wxPaintEvent& event)
{
	if (recorder->inReplay) 
		return;

	if (!(printing || saving))
		swapBuffers();
}

void Graph::swapBuffers()
{
	if (m_glContext)
	{
#ifdef WIN32
		HDC oldDC=wglGetCurrentDC();
		HGLRC oldContext=wglGetCurrentContext();
		wglMakeCurrent((HDC) m_glContext->m_hDC, m_glContext->m_glContext);
#else
		GLXDrawable oldDrawable=glXGetCurrentDrawable();
		GLXContext oldContext=glXGetCurrentContext();
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(m_glContext->m_window->GetClientAreaWindow()),
			m_glContext->m_glContext);
#endif
		if (turtle)
			if (!turtle->polyMode && !turtle->tessMode && !turtle->surfaceMode)
				glFinish();
		if (doubleBuffer)
#ifdef WIN32
			::SwapBuffers((HDC) m_glContext->m_hDC);    //blits the backbuffer into DC
#else
			m_glContext->SwapBuffers();
#endif
		else
		{	glReadBuffer(GL_FRONT);
			glDrawBuffer(GL_FRONT);
		}
#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
	}
}

void Graph::OnErase(wxEraseEvent &event)
{
	wxPaintEvent pe;
	OnPaint(pe);
}

void Graph::setMouse(wxMouseEvent &event, Mouse *mouse) 
{
	int x=event.GetX(), y=event.GetY();

	if (event.LeftUp())
		mouse->buttons&=~1;
	if (event.RightUp())
		mouse->buttons&=~2;
	if (event.MiddleUp())
		mouse->buttons&=~4;

	if (event.LeftDown())
		mouse->buttons|=1;
	if (event.RightDown())
		mouse->buttons|=2;
	if (event.MiddleDown())
		mouse->buttons|=4;

	if (event.ShiftDown())
		mouse->buttons|=8;
	else
		mouse->buttons&=~8;

	if (event.ControlDown())
		mouse->buttons|=16;
	else
		mouse->buttons&=~16;

	if (event.AltDown())
		mouse->buttons|=32;
	else
		mouse->buttons&=~32;

	if (event.MetaDown())
		mouse->buttons|=64;
	else
		mouse->buttons&=~64;

	wxSize s=GetClientSize();
	y=s.GetHeight()-y;
	gluUnProject(x, y, 0.9965, 
		screen->modelview_matrix, 
		screen->projection_matrix, 
		screen->viewport, 
		&mouse->x, &mouse->y, &mouse->z);
	event.Skip();
}

void Graph::OnMouse(wxMouseEvent &event) 
{
	setMouse(event, mouse);
}

#define EVENT_ON_MOUSE(NAME)								\
void Graph::OnMouse##NAME(wxMouseEvent &event)		\
{																	\
	setMouse(event, mouse);									\
	if (this==theApp->mainFrame->graph)					\
		Event::fifo << new EventMouse##NAME(mouse);	\
	Event::fifo << new EventGraphMouse##NAME(this, mouse);\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)

EVENT_ON_MOUSE(Motion)

bool Graph::OnMouseWheel(int nFlags, short zDelta, wxPoint pt) 
{
	mouse->wheel=pt.x;
	mouse->buttons=nFlags;
	return TRUE;
}

void Graph::setFrustum(double w, double h)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if (screen->perspective)
	{	glFrustum(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);
		if (h!=0)
			gluPerspective(45.0f, (GLdouble)w/h, nearPlane, farPlane);
		glTranslated(0,0,-fRadius);
	}
	else
	{	if (h==0 || w==0)
		{	glOrtho(-1,-1,1,1,-1,1);
			winorigin=Point(0,0,0);
		}else
		{	double y=screen->Min.y, x=screen->Min.x,
					 Y=screen->Max.y, X=screen->Max.x;
			if (double(w)/h >= (X-x)/(Y-y))
			{	//glOrtho(-y*w/h, y*w/h,-y, y, -1, 1);
				glOrtho((y*w)/h, (Y*w)/h, y, Y, -1, 1);
			}else
			{	//glOrtho(-x, x, -x*h/w, x*h/w, -1, 1);
				glOrtho(x, X, (x*h)/w, (X*h)/w, -1, 1);
			}
			screen->getMatrices();
			double sx, sy, dummyz;
			gluProject(x, y, 0, 
				screen->modelview_matrix, 
				screen->projection_matrix, 
				screen->viewport, 
				&winorigin.x, &winorigin.y, &dummyz);
			winorigin.x=floor(winorigin.x+0.5);
			winorigin.y=floor(winorigin.y+0.5);
			winorigin.z=0;
			gluProject(X, Y, 0, 
				screen->modelview_matrix, 
				screen->projection_matrix, 
				screen->viewport, 
				&sx, &sy, &dummyz);
			screen->winrangex = int(sx-winorigin.x);
			screen->winrangey = int(sy-winorigin.y);
		}
	}
	//glOrtho(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);

	//glFrustum(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);
	//gluPerspective(45.0f, (GLdouble)cx/cy, nearPlane, farPlane);
	//glTranslated(0,0,-fRadius);
	glMatrixMode(GL_MODELVIEW);
}
void Graph::OnSize(wxSizeEvent &event)
{
	if (m_glContext && !inread)
	{

#ifdef WIN32
		HDC oldDC=wglGetCurrentDC();
		HGLRC oldContext=wglGetCurrentContext();
		wglMakeCurrent((HDC) m_glContext->m_hDC, m_glContext->m_glContext);
#else
		GLXDrawable oldDrawable=glXGetCurrentDrawable();
		GLXContext oldContext=glXGetCurrentContext();
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(m_glContext->m_window->GetClientAreaWindow()),
			m_glContext->m_glContext);
#endif

		if (turtle->polyMode && turtle->polyPointNr > 1)
			glEnd();

		wxSize s=event.GetSize();
		int x=s.GetWidth(), y=s.GetHeight();
		
		if(y > 0)
		{	glViewport(0, 0, x, y);
			oldRect.SetWidth(x);
			oldRect.SetHeight(y);
			setFrustum(x, y);
			screen->getMatrices();
		}
		recorder->stop();
		screen->destroy();
		screen->create(x,y);
		recorder->playAgain();
//		swapBuffers();

#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
	}
}

#ifdef WIN32
bool Graph::setDCPixelFormat(HDC hDC, DWORD dwFlags)
{
	PIXELFORMATDESCRIPTOR 
		pixelDesc = 
		{	sizeof(PIXELFORMATDESCRIPTOR), //nSize
			1,	//nVersion
			dwFlags,	//dwFlags
			PFD_TYPE_RGBA,	//iPixelType
			24, //cColorBits
			8, //cRedBits
			0, //cRedShift
			8, //cGreenBits
			8, //cGreenShift
			8, //cBlueBits
			16, //cBlueShift
			8, //cAlphaBits
			24, //cAlphaShift
			0, //cAccumBits
			0, //cAccumRedBits
			0, //cAccumGreenBits
			0, //cAccumBlueBits
			0, //cAccumAlphaBits
			32, //cDepthBits
			0, //cStencilBits

			0, //cAuxBuffers
			PFD_MAIN_PLANE, //iLayerType
			0, //bReserved
			0, //dwLayerMask
			0, //dwVisibleMask
			0, //dwDamageMask
		};
	
	int nPixelIndex = ChoosePixelFormat(hDC, &pixelDesc);
	if (nPixelIndex == 0) // Choose default
	{	nPixelIndex = 1;
		if (DescribePixelFormat(hDC, nPixelIndex, 
			sizeof(PIXELFORMATDESCRIPTOR), &pixelDesc) == 0)
			return false;
	}
	if (!SetPixelFormat(hDC, nPixelIndex, &pixelDesc))
		return false;

	return true;
}
#endif

void Graph::print(wxDC *dc)
{
	printing=true;
#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif

	int w, h;
	dc->GetSize(&w, &h);

#ifdef WIN32
	BITMAPINFO bmi;
	memset(&bmi, 0, sizeof(BITMAPINFO));

	BITMAPINFOHEADER bmiHeader=
		{	sizeof(BITMAPINFOHEADER), //biSize
			w, //biWidth
			h, //biHeight
			1, //biPlanes
			24, //biBitCount
			BI_RGB, //biCompression
			w*h*4, //biSizeImage
		};
	bmi.bmiHeader=bmiHeader;

	HDC hDC = ::GetDC((HWND)m_hWnd);
	LPVOID bitmapBits;
	HBITMAP dib = ::CreateDIBSection(
		hDC, &bmi, DIB_RGB_COLORS, &bitmapBits, NULL, (DWORD)0);
	::ReleaseDC((HWND)m_hWnd, hDC);

	HDC mdc=::CreateCompatibleDC(0);
	HBITMAP oldbm = (HBITMAP)SelectObject(mdc, dib);

	setDCPixelFormat(mdc, 
		PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL);

	HGLRC rc=wglCreateContext(mdc), 
		oldrc=wglGetCurrentContext();
	wglMakeCurrent(mdc, rc);
#else
	Display     *dis = (Display*)wxGetDisplay();
	int       scrnum = DefaultScreen(dis);
	GLXContext  fbrc = glXGetCurrentContext();
	GLXDrawable fbdc = glXGetCurrentDrawable();

	if (fbdc == 0)
	{	error(FILE_ERROR, new String("unable to get device context in Graph::print"));
		return;
	}
	if (!fbrc)
	{	error(FILE_ERROR, new String("unable to get render context in Graph::print"));
		return;
	}
	int attrib[] =
	{	GLX_DOUBLEBUFFER,  False,
		GLX_RED_SIZE,      8,
		GLX_GREEN_SIZE,    8,
		GLX_BLUE_SIZE,     8,
		GLX_ALPHA_SIZE,    8,
//		GLX_STENCIL_SIZE,  8,
		GLX_DEPTH_SIZE,    24,
		GLX_RENDER_TYPE,   GLX_RGBA_BIT,
		GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT | GLX_WINDOW_BIT,
		None
	};

	int PBattrib[] =
	{	GLX_PBUFFER_WIDTH,   w,
		GLX_PBUFFER_HEIGHT,  h,
		GLX_LARGEST_PBUFFER, False,
		None
	};

// choose a pixel format that meets our minimum requirements
	int count = 0;
	GLXFBConfig *config = glXChooseFBConfig(dis, scrnum, attrib, &count);

	if(config == NULL || count == 0)
	{	error(FILE_ERROR, new String("no fitting pbuffer pixel format found in Graph::print"));
		return;
	}

	GLXDrawable pbdc = glXCreatePbuffer(dis, config[0], PBattrib);
	GLXContext  pbrc = glXCreateNewContext(dis, config[0], GLX_RGBA_TYPE, fbrc, true);
	XFree(config);
	glXMakeCurrent(dis, pbdc, pbrc);
#endif
	glViewport(0,0, w,h);
	setFrustum(w,h);
	setOpenGLState();
	
	screen->getMatrices();

	wxSize saveWinsize(screen->winsizex, screen->winsizey);
	screen->winsizex=w;
	screen->winsizey=h;

	ftglFontManager->clear();
	PairP oldt=Texture::clone_all();

	saving=true;
	recorder->playAgain();	// here the drawing takes place
	saving=false;

	int dummy;
	glReadPixels(0,0, 0,0, GL_RGB, GL_UNSIGNED_BYTE, &dummy);
#ifdef WIN32
	BitBlt((HDC)dc->GetHDC(), 0,0, w,h, mdc, 0,0, SRCCOPY);	

	wglMakeCurrent(oldDC, oldContext);
	wglDeleteContext(rc);
	DeleteObject(dib);
	DeleteDC(mdc);
#else
	if (w & 3)
		w=(w & ~3)+4;
	ColorP pixels=(ColorP)malloc(w*h*sizeof(ColorP));
	if (pixels)
	{	glReadPixels(0,0, w,h, GL_RGB, GL_UNSIGNED_BYTE, pixels);
		wxImage im(w,h, (unsigned char *)pixels);
		im=im.Mirror(false);
		wxBitmap bm=wxBitmap(im, -1);
		dc->DrawBitmap(bm, 0,0);
	}
	glXMakeCurrent(dis, oldDrawable, oldContext);
	glXDestroyPbuffer(dis, pbdc);
	glXDestroyContext(dis, pbrc);
#endif
	screen->winsizex=saveWinsize.GetWidth();
	screen->winsizey=saveWinsize.GetHeight();
	
	ftglFontManager->clear();
	Texture::unclone_all(oldt);

	printing=false;
}

NodeP Graph::saveScreen(charP filename)
{
	int 
		x=saveSize.x,
		y=saveSize.y;
	if (x & 3)
		x=(x & ~3)+4;

#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif
#ifdef WIN32
	HDC olddc=wglGetCurrentDC();

	BITMAPINFO bmi;
	memset(&bmi, 0, sizeof(BITMAPINFO));

	BITMAPINFOHEADER bmiHeader=
		{	sizeof(BITMAPINFOHEADER), //biSize
			x, //biWidth
			y, //biHeight
			1, //biPlanes
			24, //biBitCount
			BI_RGB, //biCompression
			x*y*3, //biSizeImage
		};
	bmi.bmiHeader=bmiHeader;

	HDC hDC = ::GetDC((HWND)m_hWnd);
	void *bitmapbits;
	HBITMAP dib = ::CreateDIBSection(
		hDC, &bmi, DIB_RGB_COLORS, &bitmapbits, NULL, (DWORD)0);
	::ReleaseDC((HWND)m_hWnd, hDC);

	HDC mdc=::CreateCompatibleDC(0);
	HBITMAP oldbm = (HBITMAP)SelectObject(mdc, dib);

	setDCPixelFormat(mdc, 
		PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL);

	HGLRC rc=wglCreateContext(mdc), 
		oldrc=wglGetCurrentContext();
	wglMakeCurrent(mdc, rc);
#else
	Display     *dis = (Display*)wxGetDisplay();
	int       scrnum = DefaultScreen(dis);
	GLXContext  fbrc = glXGetCurrentContext();
	GLXDrawable fbdc = glXGetCurrentDrawable();

	if (fbdc == 0)
	{	error(FILE_ERROR, new String("unable to get device context in Graph::print"));
		return Unbound;
	}
	if (!fbrc)
	{	error(FILE_ERROR, new String("unable to get render context in Graph::print"));
		return Unbound;
	}
	int attrib[] =
	{	GLX_DOUBLEBUFFER,  False,
		GLX_RED_SIZE,      8,
		GLX_GREEN_SIZE,    8,
		GLX_BLUE_SIZE,     8,
		GLX_ALPHA_SIZE,    8,
//		GLX_STENCIL_SIZE,  8,
		GLX_DEPTH_SIZE,    24,
		GLX_RENDER_TYPE,   GLX_RGBA_BIT,
		GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT | GLX_WINDOW_BIT,
		None
	};

	int PBattrib[] =
	{	GLX_PBUFFER_WIDTH,   x,
		GLX_PBUFFER_HEIGHT,  y,
		GLX_LARGEST_PBUFFER, False,
		None
	};

// choose a pixel format that meets our minimum requirements
	int count = 0;
	GLXFBConfig *config = glXChooseFBConfig(dis, scrnum, attrib, &count);

	if(config == NULL || count == 0)
	{	error(FILE_ERROR, new String("no fitting pbuffer pixel format found in Graph::print"));
		return Unbound;
	}

	GLXDrawable pbdc = glXCreatePbuffer(dis, config[0], PBattrib);
	GLXContext  pbrc = glXCreateNewContext(dis, config[0], GLX_RGBA_TYPE, fbrc, true);
	XFree(config);
	glXMakeCurrent(dis, pbdc, pbrc);
#endif
	glViewport(0, 0, x, y);
	setFrustum(x, y);
	setOpenGLState();
#ifndef WIN32
	glColor4ubv((unsigned char*)&turtle->screen->color);
	glBegin(GL_QUADS);
	{	double x=screen->Max.x, y=screen->Max.y;
		glVertex2d(-x,-y);
		glVertex2d( x,-y);
		glVertex2d( x, y);
		glVertex2d(-x, y);
	}
	glEnd();
	glColor4ubv((unsigned char*)&turtle->pen.color);
#endif
	screen->getMatrices();

	wxSize saveWinsize(screen->winsizex, screen->winsizey);
	screen->winsizex=x;
	screen->winsizey=y;

	ftglFontManager->clear();
	PairP oldt=Texture::clone_all();

	saving=true;
	recorder->playAgain();	// here the drawing takes place
	saving=false;

	unsigned char *pixels=(unsigned char*)GC_::malloc(x*y*3);
	if (pixels)
	{	glReadPixels(0,0, x,y, GL_RGB, GL_UNSIGNED_BYTE, (void*)pixels);
		wxImage im(x,y, pixels);
		im=im.Mirror(false);
		if (!im.SaveFile(filename))
			error(FILE_ERROR, new List(
				new String("saveScreen to file"), 
				new String(filename),
				new String("failed"), _ListEnd));
	}
#ifdef WIN32
	wglMakeCurrent(olddc, oldrc);
	wglDeleteContext(rc);
	DeleteObject(dib);
	DeleteDC(mdc);	
#else
	glXMakeCurrent(dis, oldDrawable, oldContext);
	glXDestroyPbuffer(dis, pbdc);
	glXDestroyContext(dis, pbrc);
#endif
	screen->winsizex=saveWinsize.GetWidth();
	screen->winsizey=saveWinsize.GetHeight();
	ftglFontManager->clear();
	Texture::unclone_all(oldt);
	return Unbound;
}

void Graph::saveVector(wxString filename,
							  int sort, int options, int colornr)
{
	int state = GL2PS_OVERFLOW, buffsize = 0;

	int x=saveSize.x,
		 y=saveSize.y;
	GLint viewport[4]={0,0, x,y};

	wxString ext=filename.AfterLast('.').Upper();
	int format=GL2PS_PS;
	if (ext=="EPS") format=GL2PS_EPS;
	if (ext=="PDF") format=GL2PS_PDF;
	if (ext=="TEX") format=GL2PS_TEX;
	if (ext=="SVG") format=GL2PS_SVG;
	if (ext=="PGF") format=GL2PS_PGF;

	FILE *fp=fopen(filename, "wb");
	if(!fp)
	{	error(FILE_ERROR, new String(
			wxString("Unable to open file "
			+filename+" for writing"+endL)));
		return;
	}
	while(state == GL2PS_OVERFLOW)
	{	buffsize += 1024*1024;
		gl2psBeginPage(filename, "aUCBLogo", viewport, 
							format, sort, options,
							GL_RGBA, 0, NULL, 
							colornr, colornr, colornr, 
							buffsize, fp, filename);
		glViewport(0,0, x,y);
		setFrustum(x,y);
		setOpenGLState();
		screen->getMatrices();

		saving=true;
		recorder->playAgain();	// here the drawing takes place
		saving=false;

		state = gl2psEndPage();
	}
	fclose(fp);
	wxSizeEvent sev(GetSize());
	OnSize(sev);
}

void Graph::videoStart(charP filename, int rate)
{
#ifdef WIN32
	if (aviGen)
	{	error(FILE_ERROR, new String("Video output stream is already open (forgotten VideoEnd?)"));
		return;
	}
	aviGen=new AVIGen(filename, this, rate);
	int err=aviGen->InitEngine();
	if (err)
	{	error(FILE_ERROR, new String(aviGen->aviError));
		aviGen->ReleaseEngine();
		delete aviGen;
		aviGen=0;
	}
#else
	error(NOT_YET_SUPPORTED, new String("AVI generation not yet supported!"));
#endif
}

void Graph::videoFrame()
{
#ifdef WIN32
	if (!aviGen)
	{	error(FILE_ERROR, new String("Video output stream is not open (use VideoStart!)"));
		return;
	}
	int x=(screen->winsizex/4)*4, y=screen->winsizey;

	charP pixels=charP(GC_::malloc(x*y*3));
	if (pixels)
	{	glReadPixels(0,0, x,y, GL_BGR, GL_UNSIGNED_BYTE, (void*)pixels);
		aviGen->AddFrame(pixels);
		free(pixels);
	}
#else
	error(NOT_YET_SUPPORTED, new String("AVI generation not yet supported!"));
#endif
}

void Graph::videoEnd()
{
#ifdef WIN32
	if (!aviGen)
	{	error(FILE_ERROR, new String("Video output stream is not open (use VideoStart!)"));
		return;
	}
	aviGen->ReleaseEngine();
	delete aviGen;
	aviGen=0;
#else
	error(NOT_YET_SUPPORTED, new String("AVI generation not yet supported!"));
#endif
}

void Graph::screen_reset(int centerP)
{
	if (centerP) 
	{	turtle->x=turtle->y=turtle->z=0;
		turtle->outOfBounds=false;
		bool shown=turtle->shown;

		turtle->shown=false;
		turtle->setOrientation(Point(0,0,0));
		turtle->moveTo();
		turtle->shown=shown;
		turtle->reset3d();
	}
	turtle->draw();
	recorder->clear();

	if (recorder->recording)
	{	*recorder << new R_SetScreenColor(screen->color);
		if (turtle->x!=0.0 || turtle->y!=0.0) 
			*recorder << new R_MoveTo(*turtle, *recorder);
		if (!turtle->pen.down) 
			*recorder << new R_SetPenDown(*turtle);
		if (!turtle->pen.roundLineEnds) 
			*recorder << new R_SetPenRLE(*turtle);
		*recorder << new R_SetCyLines(*turtle);
		if (turtle->pen.internalMode!=Pen::Paint) 
			*recorder << new R_SetPenMode(*turtle);
		*recorder << new R_SetPenColor(*turtle);
		*recorder << new R_SetPenSize(*turtle);
		if (screen->perspective)
			*recorder << new R_Perspective();
	}
}

void Graph::screen_clear(int centerp) 
{
	turtle->prepareToDraw();
	turtle->clearScreen();
	screen_reset(centerp);
}

void Graph::update()
{
//	Invalidate();
	wxPaintEvent e;
	OnPaint(e);
}

void Graph::updateGraph()
{
	if (turtle->direct)
		swapBuffers();
}

void Graph::autoload()
{
	theApp->Yield();
	if (autosave)
		if (wxFileExists(pictureFileName))
			recorder->load(pictureFileName);
}

void Graph::save()
{
	if (autosave)
		recorder->save(pictureFileName);
}
//___________________________________________________________________

Bitmap::Bitmap(int width_,int height_, 
					int winx_ ,int winy_  , Color *pixels_): 
	Node(BITMAP_) 
{
	width=width_;	height=height_; 
	winx=winx_; winy=winy_; 
	pixels=pixels_;
}

void Bitmap::free()
{
	if (pixels)
		::free(pixels);
}


wxTextInputStream &Bitmap::read(wxTextInputStream &istr) 
{	
	istr >> width >> height >> winx >> winy;
	int size=winx*winy;
	pixels=(ColorP)GC_::malloc(size*sizeof(Color));
	if (!pixels)
	{	width=height=winx=winy=0;
		return istr;
	}
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	istr >> *itemptr;
		itemptr++;
	}
	return istr;
}
wxDataInputStream &Bitmap::read(wxDataInputStream &istr) 
{	
	istr >> width >> height >> winx >> winy;
	int size=winx*winy;
	pixels=(ColorP)GC_::malloc(size*sizeof(Color));
	if (!pixels)
	{	width=height=winx=winy=0;
		return istr;
	}
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	istr >> *itemptr;
		itemptr++;
	}
	return istr;
}

wxTextOutputStream &Bitmap::write(wxTextOutputStream &ostr) 
{
	ostr << type << ' ' << width << ' ' << height << ' ' 
		<< winx << ' ' << winy << ' ';
	int size=winx*winy;
	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	ostr << *itemptr;
		itemptr++;
	}
	ostr << ' ';
	return ostr;
}
wxDataOutputStream &Bitmap::write(wxDataOutputStream &ostr) 
{
	ostr << type << width << height << winx << winy;
	int size=winx*winy;

	ColorP itemptr=pixels;
	for (int i=size;  i;  i--)
	{	ostr << *itemptr;
		itemptr++;
	}
	return ostr;

}

StringP Bitmap::string()
{
	char s[32];
	sprintf(s, "Bitmap [%i %i]", width, height);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void Bitmap::makeTransparent(Color c)
{
	for (unsigned int i=0; i < winx*winy; i++)
		if (pixels[i]==c)
			pixels[i].a=0;
}

void Bitmap::setPixel(int x, int y, Color c)
{
	if (x >= 0 && x < int(winx) && y >= 0 && y < int(winy))
		pixels[x+y*winx]=c;
}

Color Bitmap::Pixel(int x, int y)
{
	if (x >= 0 && x < int(winx) && y >= 0 && y < int(winy))
		return pixels[x+y*winx];
	return 123456789;
}
//___________________________________________________________________

#define NODE_ON(x)						\
PairP GraphNode::on##x()				\
{												\
	if (graph)								\
		return graph->on##x;				\
	return 0;								\
}												\
												\
void GraphNode::seton##x(NodeP nd)	\
{												\
	if (graph)								\
	{	graph->on##x = PairP(nd);		\
		check_oldyoung(nd);				\
	}											\
}

TurtleNodeP GraphNode::turtleNode()
{
	if (graph)
		return graph->turtleNode;
	return 0;
}

void GraphNode::setturtleNode(TurtleNodeP tn)
{
	if (graph)
	{	graph->turtleNode = tn;
		check_oldyoung(tn);
	}
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(MouseLeftDown)
NODE_ON(MouseRightDown)
NODE_ON(MouseMiddleDown)

NODE_ON(MouseLeftUp)
NODE_ON(MouseRightUp)
NODE_ON(MouseMiddleUp)

NODE_ON(MouseLeftDClick)

NODE_ON(MouseRightDClick)
NODE_ON(MouseMiddleDClick)

NODE_ON(MouseMotion)

void GraphNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void GraphNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

GraphNode::GraphNode(): 

	Node(GRAPH), graph(0), 
	parentFrame(0), parentGraph(0), children(0), textures(0)
{}

GraphNode::GraphNode(FrameNodeP parentFrame_, 
							GraphNodeP parentGraph_): 
	Node(GRAPH), graph(0), 
	parentFrame(parentFrame_), 
	parentGraph(parentGraph_),
	children(0), textures(0)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void GraphNode::free() 
{	
	if (graph) 
	{	graph->reset();
		graph->graphNode=0;
#ifdef WIN32		
		HDC oldDC=wglGetCurrentDC();
		HGLRC oldContext=wglGetCurrentContext();
#else
		GLXDrawable oldDrawable=glXGetCurrentDrawable();
		GLXContext oldContext=glXGetCurrentContext();
#endif
		delete graph;	//graph->Destroy();
#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif

		WindowPairP c=children;
		while (c)
		{	if (c->first())
				c->first()->setParentFrame(0);
			c=WindowPairP(c->tail());
		}				
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		PairP tl=textures;
		while (tl)
		{	TextureP(tl->first())->_graph=0;
			next(tl);
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::GraphChar:
			case Event::GraphKeyDown:
			case Event::GraphKeyUp:
			case Event::GraphMouseLeftDown:
			case Event::GraphMouseRightDown:
			case Event::GraphMouseMiddleDown:
			case Event::GraphMouseLeftUp:
			case Event::GraphMouseRightUp:
			case Event::GraphMouseMiddleUp:
			case Event::GraphMouseLeftDClick:
			case Event::GraphMouseRightDClick:
			case Event::GraphMouseMiddleDClick:
				if (((EventGraph*)ev)->graph==graph)
					((EventGraph*)ev)->graph=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		graph=0;
	}
}
void GraphNode::inc_gc() 
{
	fifo << children 
		<< onChar() << onKeyDown() << onKeyUp()
		<< onMouseLeftDown()  << onMouseRightDown()  << onMouseMiddleDown()  
		<<	onMouseLeftUp()    << onMouseRightUp()    << onMouseMiddleUp()    
		<<	onMouseLeftDClick()<< onMouseRightDClick()<< onMouseMiddleDClick()
		<<	onMouseMotion()
		<< turtleNode()
		<< textures;
}
int GraphNode::youngs() 
{	
	return gmark(children)+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp())
		+gmark(onMouseLeftDown())  +gmark(onMouseRightDown())  +gmark(onMouseMiddleDown())
		+gmark(onMouseLeftUp())    +gmark(onMouseRightUp())    +gmark(onMouseMiddleUp()) 
		+gmark(onMouseLeftDClick())+gmark(onMouseRightDClick())+gmark(onMouseMiddleDClick())
		+gmark(onMouseMotion())
		+gmark(turtleNode())
		+gmark(textures);
}
int GraphNode::youngsPtr()
{	return gmarkPtr(children)
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp())
		+gmarkPtr(onMouseLeftDown())  +gmarkPtr(onMouseRightDown())  +gmarkPtr(onMouseMiddleDown())
		+gmarkPtr(onMouseLeftUp())    +gmarkPtr(onMouseRightUp())    +gmarkPtr(onMouseMiddleUp()) 
		+gmarkPtr(onMouseLeftDClick())+gmarkPtr(onMouseRightDClick())+gmarkPtr(onMouseMiddleDClick())
		+gmarkPtr(onMouseMotion())
		+gmarkPtr(turtleNode())
		+gmarkPtr(textures);
}
void GraphNode::check_oldyoungs() 
{	check_oldyoung(children);
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
	check_oldyoung(onMouseLeftDown());
	check_oldyoung(onMouseRightDown());
	check_oldyoung(onMouseMiddleDown());
	check_oldyoung(onMouseLeftUp());
	check_oldyoung(onMouseRightUp());
	check_oldyoung(onMouseMiddleUp());
	check_oldyoung(onMouseLeftDClick());
	check_oldyoung(onMouseRightDClick());
	check_oldyoung(onMouseMiddleDClick());
	check_oldyoung(onMouseMotion());
	check_oldyoung(turtleNode());
	check_oldyoung(textures);
}

wxDataInputStream &GraphNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	parentGraph=theApp->mainFrame->graphNode;
		FrameNodeP parentFrameNode=0;
		GraphNodeP parentGraphNode=0;
		istr 
			>> (int&)parentFrameNode 
			>> (int&)parentGraphNode
			>> (int&)children
			>> (int&)textures;
		graph=new Graph(GraphNodeP(this));
		graph->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &GraphNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (graph)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame)
			<< GC_::ptrToInt->find(parentGraph)
			<< GC_::ptrToInt->find(children)
			<< GC_::ptrToInt->find(textures);
		graph->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void GraphNode::setPtrs()
{
	parentFrame= ( FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph= ( GraphNode*)GC_::intToPtr->find(int(parentGraph));
	children	  = (WindowPair*)GC_::intToPtr->find(int(children));
	textures	  = (      Pair*)GC_::intToPtr->find(int(textures));
	check_oldyoung(textures);
	if (graph)
	{
		if (!parentFrame && !parentGraph)
		{	theApp->mainFrame->splitter->ReplaceWindow(
				theApp->mainFrame->graph, graph);

#ifdef WIN32
			HDC oldDC=wglGetCurrentDC();
			HGLRC oldContext=wglGetCurrentContext();
#else
			GLXDrawable oldDrawable=glXGetCurrentDrawable();
			GLXContext oldContext=glXGetCurrentContext();
#endif
			delete theApp->mainFrame->graph;
#ifdef WIN32
			wglMakeCurrent(oldDC, oldContext);
#else
			glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
			theApp->mainFrame->graph=graph;
			theApp->mainFrame->graphNode=this;

			wxSizeEvent sev;
			graph->OnSize(sev);
		}

#define SETON(NAME)	\
	seton##NAME((Pair*)GC_::intToPtr->find(int(graph->on##NAME)));

		SETON(Char)
		SETON(KeyDown)
		SETON(KeyUp)
		SETON(MouseLeftDown)
		SETON(MouseRightDown)
		SETON(MouseMiddleDown)


		SETON(MouseLeftUp)
		SETON(MouseRightUp)
		SETON(MouseMiddleUp)

		SETON(MouseLeftDClick)
		SETON(MouseRightDClick)
		SETON(MouseMiddleDClick)

		SETON(MouseMotion)

		graph->setPtrs();
	}
}


void GraphNode::aftersetPtrs() 
{
	if (parentFrame || parentGraph)
		return;

	recreate();
}

void GraphNode::recreate()
{
	wxWindow *p=(wxWindow*)theApp->mainFrame->splitter;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	GraphP g=graph, g2=graph;
	if (parentGraph || parentFrame)
	{	g2=new Graph(GraphNodeP(this),
			p, -1, g->GetPosition(), g->GetSize(), 
			g->GetWindowStyle(), g->GetName(), 0, wxNullPalette, false);

		delete g2->screen;
		delete g2->mouse;
		delete g2->recorder;
		delete g2->turtle;

		g2->active=g->active;
		g2->screen=g->screen;	g->screen=0;
		g2->mouse =g->mouse;		g->mouse=0;
		g2->turtle=g->turtle;	g->turtle=0;
		g2->turtleNode=g->turtleNode;	g->turtleNode=0;
		g2->recorder=g->recorder; g->recorder=0;
		g2->recorder->graph=g2;
		g2->recorderSave=g2->recorder;

		graph=g2;

	#define SET_ON(NAME) seton##NAME(graph->on##NAME);	
		SET_ON(Char)
		SET_ON(KeyDown)
		SET_ON(KeyUp)

		SET_ON(MouseLeftDown)
		SET_ON(MouseRightDown)
		SET_ON(MouseMiddleDown)

		SET_ON(MouseLeftUp)
		SET_ON(MouseRightUp)
		SET_ON(MouseMiddleUp)

		SET_ON(MouseLeftDClick)
		SET_ON(MouseRightDClick)
		SET_ON(MouseMiddleDClick)

		SET_ON(MouseMotion)

		graph=g;
		delete g;
		graph=g2;
		if (graph->active)
		{	::graph=graph;
#ifdef WIN32
			wglMakeCurrent(
				(HDC)::graph->GetContext()->m_hDC,
				::graph->GetContext()->m_glContext);
#else
			glXMakeCurrent((Display*) wxGetDisplay(),

				Window(::graph->m_glContext->m_window->GetClientAreaWindow()),
				::graph->m_glContext->m_glContext);
#endif
		}
	}
	PairP c=children;
	while (c)
	{	c->first()->recreate();
		next(c);
	}
#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif
	wxGLContext *glContext=0;
	if (graph)
		glContext=graph->glContext();
	if (glContext)
#ifdef WIN32
		wglMakeCurrent((HDC)glContext->m_hDC, glContext->m_glContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(glContext->m_window->GetClientAreaWindow()),
			glContext->m_glContext);
#endif

	if (textures)
	{	textures = WindowPairP(textures->destructiveRemove(0));
		check_oldyoung(textures);
	}
	PairP t=textures;
	while (t)
	{	TextureP(t->first())->create();
		next(t);
	}
	if (parentGraph || parentFrame)
	{
		graph->setOpenGLState();
		wxSizeEvent sev(graph->GetSize());
		graph->OnSize(sev);
		graph->OnSize(sev);
	}
	if (glContext)
#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
}

StringP GraphNode::string()
{
	if (graph)
		return new String(graph->wxstring());
	return _Null_Word->name();
}

wxString GraphNode::wxstring()
{
	if (graph)
		return graph->wxstring();
	return wxEmptyString;
}

NodeP lGraph(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=0;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=0;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	else
	{	error(BAD_DATA, arg[1], 1);
		return Unbound;
	}
	long style=wxFULL_REPAINT_ON_RESIZE;
	if (exist(2) && arg[2])
	{	int s=arg[2]->a_int(2);
		if throwing return Unbound;
		style=s;	
	}

	wxPoint pos=wxDefaultPosition;
	if (exist(3) && arg[3])
	{	PairP p=arg[3]->a_int_vector(3);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(4) && arg[4])
	{	PairP s=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=_T("Graph");
	if (exist(5))
	{	StringP n=arg[5]->a_string(5);
		if throwing return Unbound;
		name=n->ptr();
	}
	GraphNodeP graphNode=new GraphNode(parentFrameNode, parentGraphNode);
	GraphP graph=new Graph(graphNode,
			parent, id, pos, size, style, name, 0, wxNullPalette, false);
	graphNode->graph=graph;
	::graph=graph;
	if (graph && graph->GetContext())
#ifdef WIN32
		wglMakeCurrent(
			(HDC)graph->GetContext()->m_hDC,
			     graph->GetContext()->m_glContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(graph->m_glContext->m_window->GetClientAreaWindow()),
			graph->m_glContext->m_glContext);
#endif
	graph->turtleNode=new TurtleNode(graph->turtle, graphNode);
	wxSizeEvent sev;
	graph->OnSize(sev);
	return graphNode;
}

NodeP lGraphDestroy(PairPP arg)
{
	GraphNodeP graphNode=arg[1]->a_graph(1);
	if throwing return Unbound;
	if (graphNode != theApp->mainFrame->graphNode)
	{	WindowPairP c=graphNode->children;
		while (c)
		{	if (c->first())
				c->first()->setParentFrame(0);
			c=WindowPairP(c->tail());
		}				
		FrameNodeP p=graphNode->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(graphNode));
			p->check_oldyoung(p->children);
		}
		GraphNodeP pg=graphNode->parentGraph;
		if (pg)
		{	pg->children = WindowPairP(pg->children->destructiveRemove(graphNode));
			pg->check_oldyoung(pg->children);
		}
		graphNode->graph->Destroy();
		graphNode->graph=0;
	}
	return Unbound;
}

NodeP lGraphCurrent(PairPP arg)
{
	return graph->graphNode;
}

NodeP lGraphSetCurrent(PairPP arg)
{
	if (!arg[1])
		graph=theApp->mainFrame->graph;
	else
	{	GraphNodeP graphNode=arg[1]->a_graph(1);
		if throwing return Unbound;
		graph=graphNode->graph;
	}
	if (graph && graph->GetContext())
#ifdef WIN32
		wglMakeCurrent(
			(HDC)graph->GetContext()->m_hDC,
			     graph->GetContext()->m_glContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(),
			Window(graph->m_glContext->m_window->GetClientAreaWindow()),
			graph->m_glContext->m_glContext);
#endif
	Float::reset_FPU();
	return Unbound;
}

#define _GRAPH_ON(NAME)								\
NodeP lGraphOn##NAME(PairPP arg)					\
{															\
	GraphNodeP graph=arg[1]->a_graph(1);		\
	if throwing return Unbound;					\
	PairP commandlist=arg[2]->a_runnable(2);	\
	if throwing return Unbound;					\
	graph->seton##NAME(commandlist);				\
	return Unbound;									\
}

_GRAPH_ON(Char)
_GRAPH_ON(KeyDown)
_GRAPH_ON(KeyUp)
_GRAPH_ON(MouseLeftDown)
_GRAPH_ON(MouseRightDown)
_GRAPH_ON(MouseMiddleDown)

_GRAPH_ON(MouseLeftUp)
_GRAPH_ON(MouseRightUp)
_GRAPH_ON(MouseMiddleUp)

_GRAPH_ON(MouseLeftDClick)
_GRAPH_ON(MouseRightDClick)
_GRAPH_ON(MouseMiddleDClick)

_GRAPH_ON(MouseMotion)
