//___________________________________________________________________
//
//                        Class Graph
//___________________________________________________________________

extern Function lGraph, lGraphDestroy, 
	lGraphCurrent, lGraphSetCurrent,
	lGraphOnChar, lGraphOnKeyDown, lGraphOnKeyUp,
	lGraphOnMouseLeftDown  , lGraphOnMouseRightDown  , lGraphOnMouseMiddleDown  ,
	lGraphOnMouseLeftUp    , lGraphOnMouseRightUp    , lGraphOnMouseMiddleUp    ,
	lGraphOnMouseLeftDClick, lGraphOnMouseRightDClick, lGraphOnMouseMiddleDClick,
	lGraphOnMouseMotion;

class Mouse;
class Screen_;
class Turtle;
class Recorder;

class Console;
class AVIGen;
class ShadowObject;

class Graph: public wxGLCanvas
{
	public:
		bool printing, saving, doubleBuffer, was1, autosave, inread;
		int active;
		float fRadius;
		wxSize saveSize, size;
		Point winorigin;
		wxString pictureFileName;
		Screen_ *screen;
		Turtle *turtle;
		Recorder *recorder, *recorderSave;
		Mouse *mouse;
		AVIGen *aviGen;
		ShadowObject *shadowObject;

		PairP 
			onChar, onKeyUp, onKeyDown, 
			onMouseLeftDown  , onMouseRightDown  , onMouseMiddleDown  ,
			onMouseLeftUp    , onMouseRightUp    , onMouseMiddleUp    ,
			onMouseLeftDClick, onMouseRightDClick, onMouseMiddleDClick,
			onMouseMotion;
		GraphNodeP graphNode;
		TurtleNodeP turtleNode;

		static int attribs[];

		Graph(GraphNodeP graphNode_);
		Graph(GraphNodeP graphNode_,
			wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, const wxSize& size, 
			long style=0, const wxString& name="GLCanvas", 
			int* attribList = 0, const wxPalette& palette = wxNullPalette,
			bool pictureAutosave = true);

		~Graph();

		void create(wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, const wxSize& size, 
			long style=0, const wxString& name="GLCanvas", 
			int* attribList = 0, const wxPalette& palette = wxNullPalette,
			bool pictureAutosave = true);
		void reset();

		wxString wxstring();
		void setPtrs();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		void 
			OnSize(wxSizeEvent &event),
			OnChar(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp(wxKeyEvent &event),
			OnPaint(wxPaintEvent &event),
			OnMove (wxMoveEvent  &event),
			OnErase(wxEraseEvent &event),
			swapBuffers(),
			setFrustum(double w, double h),
			setOpenGLState(),
			screen_reset(int centerP),
			screen_clear(int centerp),
			updateGraph(),
			autoload(), save(), 
			saveVector(wxString filename,
						  int sort, int options, int colornr),
			update(), 
			print(wxDC* dc),
			videoStart(charP filename, int rate),
			videoFrame(),
			videoEnd();

		NodeP saveScreen(charP filename);
	
		wxGLContext *glContext() {return m_glContext;}

	protected:
		float nearPlane, farPlane;
		wxRect oldRect;
#ifdef WIN32
		wxSize m_szPage, saveWinsize;
		BITMAPINFO m_bmi;
		HBITMAP m_hDib, m_hOldDib;
		void *m_pBitmapBits;
		HDC m_hMemDC, m_hOldDC;
		HGLRC m_hMemRC, m_hOldRC;
#endif
		wxPalette palette, *oldPalette;

		void 
			OnDestroy(),
			OnSetFocus(wxFocusEvent &event),
			setMouse					(wxMouseEvent &event, Mouse *mouse),
			OnMouse					(wxMouseEvent& event),
			OnMouseLeftDown		(wxMouseEvent &event),
			OnMouseRightDown		(wxMouseEvent &event),
			OnMouseMiddleDown		(wxMouseEvent &event),
			OnMouseLeftUp  		(wxMouseEvent &event),
			OnMouseRightUp  		(wxMouseEvent &event),
			OnMouseMiddleUp  		(wxMouseEvent &event),
			OnMouseLeftDClick		(wxMouseEvent &event),
			OnMouseRightDClick	(wxMouseEvent &event),
			OnMouseMiddleDClick	(wxMouseEvent &event),
			OnMouseMotion			(wxMouseEvent &event),

			beginPrinting(wxDC* pDC),
			endPrinting(wxDC* pDC),
			beginSaving(),
			endSaving();

		bool 
			OnMouseWheel(int nFlags, short zDelta, wxPoint pt),
			OnPreparePrinting();

#ifdef WIN32
		bool setDCPixelFormat(HDC hDC, DWORD dwFlags);
#endif

//		virtual void OnPrint(CDC* pDC, CPrintInfo* pInfo);
//		virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);

		DECLARE_EVENT_TABLE()
};
typedef Graph *GraphP;

extern Graph *graph;

#define TURTLE		(graph->turtle)
#define SCREEN		(graph->screen)
#define MEMSCREEN	(graph->memscreen)
#define RECORDER	(*(graph->recorder))

class GraphNode: public Node
{
	public:
		GraphP graph;
		FrameNodeP parentFrame;
		GraphNodeP parentGraph;
		WindowPairP children;
		PairP textures;

		static const GraphNode aGraphNode;

		operator_new;
		PairP onChar(), onKeyDown(), onKeyUp(),
			onMouseLeftDown()  , onMouseRightDown()  , onMouseMiddleDown()  ,
			onMouseLeftUp()    , onMouseRightUp()    , onMouseMiddleUp()    ,
			onMouseLeftDClick(), onMouseRightDClick(), onMouseMiddleDClick(),
			onMouseMotion();
		TurtleNodeP turtleNode();
		void 
			setonChar				(NodeP function), 
			setonKeyDown			(NodeP function),
			setonKeyUp				(NodeP function),
			setonMouseLeftDown	(NodeP function), 
			setonMouseRightDown	(NodeP function), 
			setonMouseMiddleDown	(NodeP function),
			setonMouseLeftUp		(NodeP function), 
			setonMouseRightUp		(NodeP function), 
			setonMouseMiddleUp	(NodeP function),
			setonMouseLeftDClick	(NodeP function), 
			setonMouseRightDClick(NodeP function), 
			setonMouseMiddleDClick(NodeP function),
			setonMouseMotion		(NodeP function),
			setturtleNode(TurtleNodeP tn);

		GraphNode();
		GraphNode(
			FrameNodeP parentFrame_, 
			GraphNodeP parentGraph_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();
		
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs();
		void aftersetPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		StringP _typeof();

		void setParentFrame(FrameNodeP frame);
		void setParentGraph(GraphNodeP graph);
};
