#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/textctrl.h>
#include <wx/listctrl.h>
#include <wx/glcanvas.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/dialog.h>
#include <wx/html/helpctrl.h>
#include <wx/image.h>
#include <wx/datstrm.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <math.h>
#include <stdio.h>

#include <gl2ps.h>

#include "Main.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "File.h"
#include "Pair.h"
#include "Object.h"
#include "List.h"
#include "Int.h"
#include "Int16.h"
#include "UInt8.h"
#include "Array.h"
#include "IntArray.h"
#include "FloatArray.h"
#include "String.h"
#include "TurtleNode.h"
#include "GraphicNode.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Mouse.h"
#include "Font.h"
#include "Turtle.h"

#include "Recorder.h"

#include "Console.h"
#include "Graph.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "AboutDlg.h"
#include "LogoApp.h"
#include "Shadow.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________

TurtleNode::TurtleNode(Turtle *turtle_, GraphNodeP graphNode_): 
	Node(TURTLE_), turtle(turtle_), graphNode(graphNode_)
{
//	turtle->refCount++;
//	graphNode=theApp->mainFrame->graphNode;
}

void TurtleNode::free()
{
	if (!turtle)
		return;

/*	if (turtle->refCount > 0)
		--turtle->refCount;

	if (turtle == TURTLE)
		return;

	if (turtle->refCount==0)
	{
*/		turtle->destroy();
		if (turtle)
		{	delete turtle;
			turtle=0;
		}
//	}
}

wxDataInputStream &TurtleNode::read(wxDataInputStream &istr) 
{
	turtle=new Turtle;
	turtle->read(istr);
	istr >> (int&)graphNode;
	return istr;
}

wxDataOutputStream &TurtleNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	turtle->write(ostr);
	ostr << GC_::ptrToInt->find(graphNode);
	return ostr;
}

void TurtleNode::setPtrs()
{
	graphNode=(GraphNode*)GC_::intToPtr->find(int(graphNode));
	if (graphNode)
	{	turtle->screen=graphNode->graph->screen;
		turtle->recorder=graphNode->graph->recorder;
	}else
	{	turtle->screen=graph->screen;
		turtle->recorder=graph->recorder;
	}
}

StringP TurtleNode::string()
{
	char format[32], s[256];
	sprintf(format, "Turtle {%%-.%ig %%-.%ig %%-.%ig}", 
		File::precision, File::precision, File::precision);

	sprintf(s, format, turtle->x, turtle->y, turtle->z);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

GraphicNode::GraphicNode(Recorder *recorder_): Node(GRAPHIC)
{
	recorder=recorder_;
}

void GraphicNode::free()
{
	if (recorder && !--recorder->refCount)
	{	delete recorder;
		recorder=0;
	}
}

void GraphicNode::setPtrs()
{	
	CommandListIterator i(recorder);
	for (i=recorder->begin();  i;  i++)
		i->setPtrs();
}
wxDataInputStream &GraphicNode::read(wxDataInputStream &istr) 
{
	int nr;
	istr >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(graph);
		Recorder::intToPtr->insert(recorder);
		istr >> *recorder;
	}
	else
		recorder->refCount++;
	return istr;
}
wxDataOutputStream &GraphicNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
		ostr << int(0) << *recorder;
	}
	else
		ostr << nr;
	return ostr;
}

StringP GraphicNode::string()
{
	return new String("Graphic");
}

//___________________________________________________________________

NodeP lMatrix(PairPP)
{//	return SCREEN->matrix();
	double *m;
	m=SCREEN->modelview_matrix;
	return new List
	(	new List(new Float(m[ 0]),new Float(m[ 1]),new Float(m[ 2]), _ListEnd),
		new List(new Float(m[ 4]),new Float(m[ 5]),new Float(m[ 6]), _ListEnd),
		new List(new Float(m[ 8]),new Float(m[ 9]),new Float(m[10]), _ListEnd),
		new List(new Float(m[12]),new Float(m[13]),new Float(m[14]), _ListEnd), _ListEnd);
}

NodeP lIdentityMatrix(PairPP)
{
	FloatP O=new Float(0), l=new Float(1);
	return new List
	(	new List( l,O,O, _ListEnd),
		new List( O,l,O, _ListEnd), 
		new List( O,O,l, _ListEnd), 
		new List( O,O,O, _ListEnd), _ListEnd);
}

NodeP lsetIdentityMatrix(PairPP)
{
//	TURTLE->setMatrix(
//		Point(1,0,0), Point(0,1,0), Point(0,0,1), Point(0,0,0));
	glLoadIdentity();
	glGetDoublev(GL_MODELVIEW_MATRIX, SCREEN->modelview_matrix);
	return Unbound;
}

NodeP lTurtleMatrix(PairPP) {return TURTLE->matrix();}
NodeP lsetTurtleMatrix(PairPP) 
{
	TURTLE->draw();
	TURTLE->setTurtleMatrix(); 
	TURTLE->draw();
	return Unbound;
}

NodeP lsetMatrix(PairPP arg)
{
	PairP m=arg[1]->a_list(1);
	if (!throwing)
	{	if (m && m->first() 
		&& m->tail() && m->second() 
		&& m->tail()->tail() && m->third() 
		&& m->tail()->tail()->tail() && m->fourth()) 
		{	PairP 
				 basex=m-> first()->a_3dvector(1),
				 basey=m->second()->a_3dvector(1),
				 basez=m-> third()->a_3dvector(1),
				center=m->fourth()->a_3dvector(1);

			if (!throwing)
				TURTLE->setMatrix(basex,basey,basez,center);
		}
		else error(BAD_DATA, m);
	}
	return Unbound;
}

NodeP lpushMatrix(PairPP) 
{
	TURTLE->draw();
	glPushMatrix(); 
	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_PushMatrix();
	return Unbound;
}

NodeP lpopMatrix(PairPP) 
{
	TURTLE->draw();
	glPopMatrix(); 
	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_PopMatrix();
	return Unbound;
}

NodeP lsetScreenRange(PairPP arg)
{
	int minx=arg[1]->a_int(1);
	if throwing return Unbound;
	int miny=arg[2]->a_int(2);
	if throwing return Unbound;
	int maxx=arg[3]->a_int(3);
	if throwing return Unbound;
	int maxy=arg[4]->a_int(4);
	if throwing return Unbound;
	graph->screen->Min.x=minx;
	graph->screen->Min.y=miny;
	graph->screen->Max.x=maxx;
	graph->screen->Max.y=maxy;

	int w,h;
	graph->GetClientSize(&w, &h);
	TURTLE->draw();
	glViewport(0, 0, w, h);
	graph->setFrustum(w, h);
	graph->screen->getMatrices();
	TURTLE->draw();
	return Unbound;
}

NodeP lScreenRange(PairPP)
{
	return new List(
		new Float(graph->screen->Min.x), 
		new Float(graph->screen->Min.y), 
		new Float(graph->screen->Max.x), 
		new Float(graph->screen->Max.y), _ListEnd);
}

void Node::setPixel(NodeP color) {error(BAD_DATA, this);}
void Pair::setPixel(NodeP color)
{	if (first()->islist())
	{	PairP a=this;
		if (color->islist())
		{	PairP clist=PairP(color);
			while (a && clist)
			{	a->first()->setPixel( clist->first() );
				next(a);  
				next(clist);
			}
		}else
		{	while (a)
			{	a->first()->setPixel(color);
				next(a);
			}
		}
		if (!throwing) return;
	}else
	{	PairP vec=this;//a_vector();
		IntP C=PairP(color)->a_integer(2);
		if (!throwing)
		{	NodeP X=vec->first(), Y=vec->second();
			double
				x= X->getnumber(),
				y= Y->getnumber(),
				z=0;
			if (vec->tail()->tail())
				z=vec->third()->getnumber();
			int c= Pen::realcolor(C->int_);
			TURTLE->setPixel(Point(x,y,z), c);
			return;
		}
	}
	error(BAD_DATA, this);
}
void Array::setPixel(NodeP color)
{	NodePP d=data();
	int i=length();
	if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
	{	NodePP c=ArrayP(color)->data();
		while (i--)
			(*d++)->setPixel(*c++);
	}else 
		while (i--)
			(*d++)->setPixel(color);
}
void FloatArray::setPixel(NodeP color)
{	int i=length();
	if (i!=3)
	{	error(BAD_DATA, this);
		return;
	}
	doubleP d=data();
	Color c=Pen::realcolor(color->a_color(2));
	if throwing return;
	TURTLE->setPixel(Point(d[0],d[1],d[2]), c);
}
NodeP lsetPixel(PairPP arg)
{
	if (arg[1]) 
	{	TURTLE->draw();
		arg[1]->setPixel(arg[2]);
		TURTLE->draw();
	}
	return Unbound;
}

void Node::setPixelXY(NodeP yin, NodeP color) {error(BAD_DATA, this);}
void Int::setPixelXY(NodeP yin, NodeP color)
{
	IntP C=color->a_integer(2);
	double
		x=int_,
		y=yin->getnumber(),
		z=0;
	int c= Pen::realcolor(C->int_);
	TURTLE->setPixel(Point(x,y,z), c);
}
void Float::setPixelXY(NodeP yin, NodeP color)
{
	IntP C=color->a_integer(2);
	double
		x=float_,
		y=yin->getnumber(),
		z=0;
	int c= Pen::realcolor(C->int_);
	TURTLE->setPixel(Point(x,y,z), c);
}
void Pair::setPixelXY(NodeP yin, NodeP color)
{
	if (!yin || !yin->is.list)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	PairP x=this, y=PairP(yin);
	if (color->islist())
	{	PairP c=PairP(color);
		while (x && y && c)
		{	x->first()->setPixelXY( y->first(), c->first() );
			next(x);
			next(y);
			next(c);
		}
	}else
	{	while (x && y)
		{	x->first()->setPixelXY( y->first(), color);
			next(x);
			next(y);
		}
	}
}
void Array::setPixelXY(NodeP yin, NodeP color)
{	if (type_of(yin) == ARRAY)
	{	NodePP d=data();
		int i=length();
		ArrayP y=ArrayP(yin);
		if (y->length() != i)
		{	error(BAD_DATA, yin, 2);
			return;
		}
		NodePP dy=y->data();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
				(*d++)->setPixelXY(*dy++, *c++);
		}else 
			while (i--)
				(*d++)->setPixelXY(*dy++, color);
	}
	else if (type_of(yin) == INT_ || type_of(yin) == FLOAT_)
	{	NodePP d=data();
		int i=length();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
				(*d++)->setPixelXY(yin, *c++);
		}else 
			while (i--)
				(*d++)->setPixelXY(yin, color);
	}
	else
	{	error(BAD_DATA, yin, 2);
		return;
	}
}
void FloatArray::setPixelXY(NodeP yin, NodeP color)
{	if (type_of(yin) != FLOATARRAY)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP x=data();
	int i=length();
	FloatArrayP ya=FloatArrayP(yin);
	if (ya->length() != i)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP y=ya->data();
	if (color->type==INTARRAY  &&  IntArrayP(color)->length()==i)
	{	ColorP c=ColorP(IntArrayP(color)->data());
		while (i--)
		{	int ic= Pen::realcolor(*c++);
			TURTLE->setPixel(Point(*x++,*y++,0), ic);
		}
	}else if (color->type==INT_)
	{	int c= Pen::realcolor(IntP(color)->int_);
		while (i--)
		{	TURTLE->setPixel(Point(*x++,*y++,0), c);
		}
	}
	else
		error(BAD_DATA, this);
}
NodeP lsetPixelXY(PairPP arg)
{
	if (arg[1]) 
	{	TURTLE->draw();
		arg[1]->setPixelXY(arg[2], arg[3]);
		TURTLE->draw();
	}
	return Unbound;
}

void Node::setPixelXYZ(NodeP yin, NodeP zin, NodeP color) {error(BAD_DATA, this);}
void Int::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	IntP C=color->a_integer(2);
	double
		x=int_,
		y=yin->getnumber(),
		z=zin->getnumber();
	int c= Pen::realcolor(C->int_);
	TURTLE->setPixel(Point(x,y,z), c);
}
void Float::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	IntP C=color->a_integer(2);
	double
		x=float_,
		y=yin->getnumber(),
		z=zin->getnumber();
	int c= Pen::realcolor(C->int_);
	TURTLE->setPixel(Point(x,y,z), c);
}
void Pair::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	if (!yin || !yin->is.list)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	if (!zin || !zin->is.list)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	PairP x=this, y=PairP(yin), z=PairP(zin);
	if (color->islist())
	{	PairP c=PairP(color);
		while (x && y && z && c)
		{	x->first()->setPixelXYZ( y->first(), z->first(), c->first() );
			next(x);
			next(y);
			next(z);
			next(c);
		}
	}else
	{	while (x && y && z)
		{	x->first()->setPixelXYZ( y->first(), z->first(), color);
			next(x);
			next(y);
			next(z);
		}
	}
}
void Array::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{	if (type_of(yin) == ARRAY)
	{	NodePP d=data();
		int i=length();
		ArrayP y=ArrayP(yin);
		if (y->length() != i)
		{	error(BAD_DATA, yin, 2); return;
		}
		if (type_of(zin) != ARRAY)
		{	error(BAD_DATA, zin, 3); return;
		}
		ArrayP z=ArrayP(zin);
		if (z->length() != i)
		{	error(BAD_DATA, zin, 3); return;
		}
		NodePP 
			dy=y->data(), 
			dz=z->data();
		if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
		{	NodePP c=ArrayP(color)->data();
			while (i--)
				(*d++)->setPixelXYZ(*dy++, *dz++, *c++);
		}else 
			while (i--)
				(*d++)->setPixelXYZ(*dy++, *dz++, color);
		return;
	}
	else if (type_of(yin) == INT_ || type_of(yin) == FLOAT_)
	{	if (type_of(zin) == INT_ || type_of(zin) == FLOAT_)
		{	NodePP d=data();
			int i=length();
			if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
			{	NodePP c=ArrayP(color)->data();
				while (i--)
					(*d++)->setPixelXYZ(yin, zin, *c++);
			}else 
				while (i--)
					(*d++)->setPixelXYZ(yin, zin, color);
			return;
		}
		else
		{	error(BAD_DATA, zin, 3); return;
		}
	}
	error(BAD_DATA, yin, 2); return;
}
void FloatArray::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{	if (type_of(yin) != FLOATARRAY)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP x=data();
	int i=length();
	FloatArrayP ya=FloatArrayP(yin);
	if (ya->length() != i)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	if (type_of(zin) != FLOATARRAY)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	FloatArrayP za=FloatArrayP(zin);
	if (za->length() != i)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	doubleP 
		y=ya->data(),
		z=za->data();
	if (color->type==INTARRAY  &&  IntArrayP(color)->length()==i)
	{	ColorP c=ColorP(IntArrayP(color)->data());
		while (i--)
		{	int ic= Pen::realcolor(*c++);
			TURTLE->setPixel(Point(*x++, *y++, *z++), ic);
		}
	}else if (color->type==INT_)
	{	int c= Pen::realcolor(IntP(color)->int_);
		while (i--)
		{	TURTLE->setPixel(Point(*x++, *y++, *z++), c);
		}
	}
	else
		error(BAD_DATA, this);
}
NodeP lsetPixelXYZ(PairPP arg)
{
	if (arg[1]) 
	{	TURTLE->draw();
		arg[1]->setPixelXYZ(arg[2], arg[3], arg[4]);
		TURTLE->draw();
	}
	return Unbound;
}

NodeP lPixel(PairPP)
{
	TURTLE->draw();
	NodeP p=new Int(TURTLE->Pixel());
	TURTLE->draw();
	return p;
}

void line_draw(NodeP a, NodeP color)
{	PairP vec= PairP(a)->a_vector(1);
	IntP C=PairP(color)->a_integer(2);
	if (!throwing)
	{	NodeP X=vec->first(), Y=vec->second();
		double
			x=X->getnumber(),
			y=Y->getnumber();
		int c= C->int_;
		if (c ==-1)
			TURTLE->moveTo(x,y);
		else
		{	if (TURTLE->pen.rawcolor !=c)
				TURTLE->setColor(c);
			if (RECORDER.recording)
				RECORDER << new R_SetPenColor(*TURTLE);
			TURTLE->lineTo(x,y);
			*TURTLE=Point(x,y,0);
		}

	}
}
void move_draw(NodeP a, NodeP color)
{	PairP vec= PairP(a)->a_vector(1);
	IntP C=PairP(color)->a_integer(2);
	if (!throwing)
	{	NodeP X=vec->first(), Y=vec->second();
		double
			x=X->getnumber(),
			y=Y->getnumber();
		int c= C->int_;
		if (c ==-1)
			TURTLE->moveTo(x,y);
		else
		{	if (TURTLE->pen.rawcolor !=c)
				TURTLE->setColor(c);
			if (RECORDER.recording)
				RECORDER << new R_SetPenColor(*TURTLE);
			TURTLE->moveTo(x,y);
		}
	}
}
void Node::line(NodeP color) {error(BAD_DATA, this);}
void Pair::line(NodeP color)
{	if (first()->islist())
	{	PairP a=this;
		if (a && a->first() && a->first()->first()->islist())
			if (color->islist())
			{	PairP clist=PairP(color);
				while (a && a->first() && clist) 
				{	a->first()->line(clist->first());
					next(a);
					next(clist);
				}
				if (!a && !clist && !throwing) return;
			}else
			{	while (a && a->first()) 
				{	a->first()->line(color);
					next(a);
				}
				if (!a && !throwing) return;
			}
		else
			if (color->islist())
			{	PairP clist=PairP(color);
				if (a && clist) 
					move_draw(a->first(), clist->first());
				while (a && clist)
				{	line_draw(a->first(), clist->first());
					next(a);  
					next(clist);
				}
				if (!a && !clist && !throwing) return;
			}else
			{	if (a) move_draw(a->first(), color);
				while (a)
				{	line_draw(a->first(), color);
					next(a);
				}
				if (!a && !throwing) return;
			}
	}else if (first()->type==ARRAY)
	{	PairP a=this;
		while (a)
		{	if (a->first() && a->first()->type==ARRAY)
				a->first()->line(color);
			else break;
			next(a);
		}
		if (!a && !throwing) return;
	}else
	{	move_draw(this, color);
		if (!throwing) return;
	}
	error(BAD_DATA, this);
}
void Array::line(NodeP color)
{	NodePP d=data();
	int i=length();
	if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
	{	NodePP c=ArrayP(color)->data();
		if (*d) move_draw(*d, *c);
		while (i-- && *d)
			line_draw(*d++, *c++);
		if (i==-1 && !throwing) return; 
	}else 
	{	if (*d) move_draw(*d, color);
		while (i-- && *d)
			line_draw(*d++, color);
		if (i==-1 && !throwing) return;
	}
	error(BAD_DATA, this);
}
void FloatArray::line(NodeP color)
{	error(BAD_DATA, this);
}
NodeP lLine(PairPP arg)
{
	if (arg[1]) 
	{	TURTLE->draw();	
		Point t=*TURTLE;
		arg[1]->line(arg[2]);
		*TURTLE=t;
		TURTLE->draw();
	}
	return Unbound;
}

NodeP lright(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->right(a);
	}
	return Unbound;
}

NodeP lleft(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->right(-a);
	}
	return Unbound;
}

NodeP lupPitch(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->upPitch(a);
	}
	return Unbound;
}

NodeP ldownPitch(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->upPitch(-a);
	}
	return Unbound;
}

NodeP lrightRoll(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->rightRoll(a);
	}
	return Unbound;
}

NodeP lleftRoll(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->rightRoll(-a);
	}
	return Unbound;
}

NodeP lspinX(PairPP arg)
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->spinX(a);
	}
	return Unbound;
}

NodeP lspinY(PairPP arg)
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->spinY(a);
	}
	return Unbound;
}

NodeP lspinZ(PairPP arg)
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double a=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->spinZ(a);
	}
	return Unbound;
}

NodeP lforward(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double d=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->forward(d);
	}
	return Unbound;
}

NodeP lback(PairPP arg) 
{
	NodeP val = arg[1]->a_number(1);
	if (!throwing) 
	{	double d=val->getnumber();
		TURTLE->prepareToDraw();
		TURTLE->forward(-d);
	}
	return Unbound;
}

NodeP lshowTurtle(PairPP){TURTLE->show(); return Unbound;}
NodeP lhideTurtle(PairPP){TURTLE->hide();  return Unbound;}
NodeP lisshown(PairPP){return (TURTLE->shown ?  _True  :  _False);}

NodeP lHeading(PairPP) {return new Float(TURTLE->getHeading());}
NodeP lpitch(PairPP) {return new Float(TURTLE->pitch());}
NodeP lroll(PairPP) {return new Float(TURTLE->roll());}

NodeP lOrientation(PairPP) 
{
	return 
		new Pair(new Float(TURTLE->roll()), 
		new Pair(new Float(TURTLE->pitch()), 
		new Pair(new Float(TURTLE->getHeading()), 0)));
}

NodeP lsetHeading(PairPP arg) 
{
	double heading = arg[1]->a_float(1);
	if (!throwing) 
		TURTLE->setHeading(heading);
	return Unbound;
}

NodeP lsetRoll(PairPP arg) 
{
	double roll = arg[1]->a_float(1);
	if (!throwing) 
		TURTLE->setRoll(roll);
	return Unbound;
}

NodeP lsetPitch(PairPP arg) 
{
	double pitch = arg[1]->a_float(1);
	if (!throwing) 
		TURTLE->setPitch(pitch);
	return Unbound;
}

NodeP lsetOrientation(PairPP arg)
{
	PairP o=arg[1]->a_3dvector(1);
	if (!throwing)
	{	TURTLE->setOrientation(Point(
			o-> first()->getnumber(),
			o->second()->getnumber(),
			o-> third()->getnumber()));
	}
	return Unbound;
}

NodeP ltowards(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
	{	NodeP 
			nodeX = v->first(),
			nodeY = v->second();

		Point target(
				nodeX->type == FLOAT_ ?   FloatP(nodeX)->float_  :  IntP(nodeX)->int_,
				nodeY->type == FLOAT_ ?   FloatP(nodeY)->float_   :  IntP(nodeY)->int_,
				0);
		return new Float(TURTLE->towards_helper(target));
	}
	return Unbound;
}

NodeP ltowardsXYZ(PairPP arg) 
{
	PairP v = arg[1]->a_3dvector(1);
	if (!throwing) 
	{	double
			x=v-> first()->getnumber(),
			y=v->second()->getnumber(),
			z=v-> third()->getnumber(),
			rx,ry,rz;

		Point target(x,y,z), diff=target-*TURTLE;
		diff=diff/diff.Norm();

      // Compute angle between Y Axis and Vector [DOT] (Rotation about Z)
      rz = acos(diff.y)/Float::deg2rad;

      // Project point onto X-Z plane and renormalize to measure roll
      diff.y = 0.0;
		double n=diff.Norm();
      if (n!=0)
			diff=diff/n;

      // Compute angle between X Axis and Projection [DOT] (Rotation about Y)
      ry = acos(diff.x)/Float::deg2rad;

      // Heading was limited to 0 to 180 but we must be able to roll -180 to 180
      // if point is in positive Z hemisphere then "roll" left to it.
      if (diff.z > 0.0) ry = -ry;

      // No Pitch needed
      rx = 0.0;

		return 
			new Pair(new Float(ry), 
			new Pair(new Float(rx), 
			new Pair(new Float(rz), 0)));
	}
	return Unbound;
}

NodeP lDistance(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
	{	NodeP 
			nodeX = v->first(),
			nodeY = v->second();

		Point target(
				nodeX->type == FLOAT_? FloatP(nodeX)->float_:IntP(nodeX)->int_,
				nodeY->type == FLOAT_? FloatP(nodeY)->float_:IntP(nodeY)->int_,
				0);
		return new Float(TURTLE->distance(target));

	}
	return Unbound;
}

NodeP lDistanceXYZ(PairPP arg) 
{
	PairP v = arg[1]->a_3dvector(1);
	if (!throwing) 
	{	NodeP 
			nodeX=v->first(),
			nodeY=v->second(),
			nodeZ=v->third();

		Point target(
				nodeX->type == FLOAT_? FloatP(nodeX)->float_:IntP(nodeX)->int_,
				nodeY->type == FLOAT_? FloatP(nodeY)->float_:IntP(nodeY)->int_,
				nodeZ->type == FLOAT_? FloatP(nodeZ)->float_:IntP(nodeZ)->int_);
		return new Float(TURTLE->distance(target));
	}
	return Unbound;
}

NodeP lxCor(PairPP) {return new Float(TURTLE->x);}
NodeP lyCor(PairPP) {return new Float(TURTLE->y);}
NodeP lzCor(PairPP) {return new Float(TURTLE->z);}

NodeP lPos(PairPP) 
{
	return 
		new Pair(new Float(TURTLE->x), 
		new Pair(new Float(TURTLE->y), 0));
}

NodeP lPosXYZ(PairPP) 
{
	return 
		new Pair(new Float(TURTLE->x), 
		new Pair(new Float(TURTLE->y), 
		new Pair(new Float(TURTLE->z), 0)));
}

NodeP lScrunch(PairPP) 
{
	return new List
	(	new Float(TURTLE->screen->scale.x), 
		new Float(TURTLE->screen->scale.y), _ListEnd
	);
}

NodeP lHome(PairPP) 
{
	TURTLE->home();
	return Unbound;
}

NodeP lclearScreen(PairPP arg) 
{	graph->screen_clear(true); 
	return Unbound;
}

NodeP lclean(PairPP arg) 
{	graph->screen_clear(false); 
	return Unbound;
}

void Node::setXYZ(NodeP ynode, NodeP znode)
{	error(BAD_DATA, this, 1);
}
void String::setXYZ(NodeP ynode, NodeP znode)
{	if (numberP())
		number()->setXYZ(ynode, znode);
	else 
		error(BAD_DATA, this, 1);
}
void Int::setXYZ(NodeP ynode, NodeP znode)
{	if (ynode->numberP())
	{	if (znode)
		{	if (znode->numberP())
				TURTLE->setPos(this, ynode, znode);
			else 
			{	error(BAD_DATA, ynode, 2);	return;
			}
		}else
			TURTLE->setPos(this, ynode, 0);
	}else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}
void Float::setXYZ(NodeP ynode, NodeP znode)
{	if (ynode->numberP())
	{	if (znode)
		{	if (znode->numberP())
				TURTLE->setPos(this, ynode, znode);
			else 
			{	error(BAD_DATA, ynode, 2);	return;
			}
		}else
			TURTLE->setPos(this, ynode, 0);
	}else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}
void Pair::setXYZ(NodeP ynode, NodeP znode)
{	PairP x=this;
	if (ynode->islist())
	{	PairP y=PairP(ynode);
		if (znode)
			if (znode->islist())
			{	PairP z=PairP(znode);
				while (x && y && z)
				{	if (x->first() && y->first() && z->first())
						x->first()->setXYZ(y->first(), z->first());
					else
					{	error(BAD_DATA, this, 2);	return;
					}
					next(x);
					next(y);
					next(z);
				}
			}else
			{	error(BAD_DATA, this, 2);	return;
			}
		else
		{	while (x && y)
			{	if (x->first() && y->first())
					x->first()->setXYZ(y->first(), 0);
				else
				{	error(BAD_DATA, this, 2);	return;
				}
				next(x);
				next(y);
			}
		}
	}
	else
	{	error(BAD_DATA, this, 2);	return;
	}
}
void Array::setXYZ(NodeP ynode, NodeP znode)
{	int l=length();
	if (ynode->type==ARRAY && ArrayP(ynode)->length()==l)
	{	if (znode)
			if (znode->type==ARRAY && ArrayP(znode)->length()==l)
			{	NodePP x=data(), 
					y=ArrayP(ynode)->data(),
					z=ArrayP(znode)->data();
				while (l--)
				{	if (*x)
						(*x++)->setXYZ(*y++, *z++);
				}
			}
			else 
			{	error(BAD_DATA, znode, 3);	return;
			}
		else
		{	NodePP x=data(), 
				y=ArrayP(ynode)->data();
			while (l--)
			{	if (*x)
					(*x++)->setXYZ(*y++, 0);
			}
		}
	}
	else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}
void FloatArray::setXYZ(NodeP ynode, NodeP znode)
{	int l=length();
	if (ynode->type==FLOATARRAY && FloatArrayP(ynode)->length()==l)
	{	if (znode)
		{	if (znode->type==FLOATARRAY && FloatArrayP(znode)->length()==l)
			{	doubleP x=data(), 
					y=FloatArrayP(ynode)->data(),
					z=FloatArrayP(znode)->data();
				while (l--)
				{	TURTLE->setPos(*x++, *y++, *z++);
				}
			}
			else 
			{	error(BAD_DATA, znode, 3);	return;
			}
		}
		else
		{	doubleP x=data(), 
				y=FloatArrayP(ynode)->data();
			while (l--)
			{	TURTLE->setPos(*x++, *y++, 0);
			}
		}
	}
	else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}

NodeP lsetPos(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
		TURTLE->setPos(v->first(), v->second(), NIL);
	return Unbound;
}

NodeP l_setPos(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
		TURTLE->_setPos(v->first(), v->second(), NIL);
	return Unbound;
}

NodeP lsetPosXYZ(PairPP arg) 
{
	PairP v = arg[1]->a_3dvector(1);
	if (!throwing) 
		TURTLE->setPos(v->first(), v->second(), v->third());
	return Unbound;
}

NodeP l_setPosXYZ(PairPP arg) 
{
	PairP v = arg[1]->a_3dvector(1);
	if (!throwing) 
		TURTLE->_setPos(v->first(), v->second(), v->third());
	return Unbound;
}

NodeP lsetXYZ(PairPP arg) 
{
 	NodeP
		x = arg[1],
		y = arg[2],
		z = arg[3];
	if (!x)
	{	error(BAD_DATA, x, 1);
		return Unbound;
	}
	if (!y)
	{	error(BAD_DATA, y, 2);
		return Unbound;
	}
	if (!z)
	{	error(BAD_DATA, z, 3);
		return Unbound;
	}
	x->setXYZ(y, z);
	return Unbound;
}

NodeP lsetXY(PairPP arg) 
{
	NodeP 
		x = arg[1],
		y = arg[2];
	if (!x)
	{	error(BAD_DATA, x, 1);
		return Unbound;
	}
	if (!y)
	{	error(BAD_DATA, y, 2);
		return Unbound;
	}
	x->setXYZ(y, 0);
	return Unbound;
}

NodeP lsetX(PairPP arg) 
{
	NodeP xnode = arg[1]->a_number(1);
	if (!throwing) 
		TURTLE->setPos(xnode, NIL, NIL);
	return Unbound;
}

NodeP lsetY(PairPP arg) 
{
	NodeP ynode = arg[1]->a_number(1);
	if (!throwing) 
		TURTLE->setPos(NIL, ynode, NIL);
	return Unbound;

}

NodeP lsetZ(PairPP arg) 
{
	NodeP znode = arg[1]->a_number(1);
	if (!throwing) 
		TURTLE->setPos(NIL, NIL, znode);
	return Unbound;

}

NodeP lsetSpherePos(PairPP arg) 
{
	double
		    r = arg[1]->a_float(1),
		  phi = arg[2]->a_float(2),
		theta = arg[3]->a_float(3);
	if (!throwing) 
		TURTLE->setSpherePos(r, phi, theta);
	return Unbound;
}

NodeP lsetCylinderPos(PairPP arg) 
{
	double
		    r = arg[1]->a_float(1),
		  phi = arg[2]->a_float(2),
		    y = arg[3]->a_float(3);
	if (!throwing) 
		TURTLE->setCylinderPos(r, phi, y);
	return Unbound;
}

NodeP lwrap(PairPP)
{
	TURTLE->wrap(); 
	if (RECORDER.recording)
		RECORDER << new R_wrap();
	return Unbound;
}

NodeP lFence(PairPP arg) 
{
	lwrap(arg);	    // get turtle inside the fence
	TURTLE->draw();
	TURTLE->mode=Turtle::fencemode;
	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_mode(*TURTLE);
	return Unbound;
}

NodeP lWindow(PairPP) 
{
	TURTLE->draw();
	TURTLE->mode=Turtle::windowmode;
	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_mode(*TURTLE);
	return Unbound;
}

NodeP lfill(PairPP arg) 
{
	bool surface;
	if (exist(1))
	{	int a=arg[1]->a_bool(1);
		if throwing return Unbound;
		surface=a? true: false;
	}else 
		surface=false;
	TURTLE->fill(surface);
	if (RECORDER.recording)
		RECORDER << new R_Fill(surface);
	return Unbound;
}

NodeP lfillRect(PairPP arg)
{
	PairP v1=arg[1]->a_vector(1),
			v2=arg[2]->a_vector(2);
	if (!throwing)
	{	NodeP X1=v1->first(), Y1=v1->second(),
				X2=v2->first(), Y2=v2->second();
		double x1=X1->getnumber(), y1=Y1->getnumber(),
				 x2=X2->getnumber(), y2=Y2->getnumber();
		Point p1=Point(x1,y1,0), p2=Point(x2,y2,0);

		TURTLE->fillRect(p1, p2);
		if (RECORDER.recording)
			RECORDER << new R_FillRect(*TURTLE, p1,p2);
	}
	return Unbound;
}

NodeP lfillEllipse(PairPP arg)
{
	NodeP 
		radiusx = arg[1]->a_number(1),
		radiusy = arg[2]->a_number(2);
	
	if (!throwing) 
	{	double 
			rx=radiusx->getnumber(), 
			ry=radiusy->getnumber();
		Point r=Point(rx,ry,0);
		TURTLE->prepareToDraw();
		TURTLE->fillEllipse(r);
		if (RECORDER.recording)
			RECORDER << new R_FillEllipse(*TURTLE, r);
	}
	return Unbound;

}

NodeP lfillCircle(PairPP arg)
{
	NodeP radius = arg[1]->a_number(1);
	
	if (!throwing) 
	{	double rx=radius->getnumber();
		Point r=Point(rx,rx,0);
		TURTLE->prepareToDraw();
		TURTLE->fillEllipse(r);
		if (RECORDER.recording)
			RECORDER << new R_FillEllipse(*TURTLE, r);
	}
	return Unbound;
}

NodeP lSphere(PairPP arg)
{
	NodeP radius = arg[1]->a_number(1);
	if throwing return Unbound;
	double r=radius->getnumber();
	int slices=0;
	if (exist(2))
	{	slices=arg[2]->a_int(2);
		if throwing return Unbound;
	}
	int stacks=0;
	if (exist(3))
	{	stacks=arg[3]->a_int(3);
		if throwing return Unbound;
	}
	TURTLE->prepareToDraw();
	if (TURTLE->shadows)
		TURTLE->partialEllipsoidSh(r, 0, 360, 40, 0, 180, 40);
	else
		TURTLE->sphere(r, slices, stacks);
	if (RECORDER.recording)
		RECORDER << new R_Sphere(*TURTLE, r, slices, stacks);
	return Unbound;
}

NodeP lEllipsoid(PairPP arg)
{
	NodeP 
		rx = arg[1]->a_number(1),
		ry = arg[2]->a_number(2),
		rz = arg[3]->a_number(3);
	if throwing return Unbound;
	int slices=0;
	if (exist(4))
	{	slices=arg[4]->a_int(4);
		if throwing return Unbound;
	}
	int stacks=0;
	if (exist(5))
	{	stacks=arg[5]->a_int(5);
		if throwing return Unbound;
	}
	Point r=Point(rx->getnumber(),ry->getnumber(),rz->getnumber());
	TURTLE->prepareToDraw();
	TURTLE->ellipsoid(r);
	if (RECORDER.recording)
		RECORDER << new R_Ellipsoid(*TURTLE, r, slices, stacks);
	return Unbound;
}

NodeP lpartialEllipsoid(PairPP arg)
{
	NodeP 
		rx = arg[1]->a_number(1),
		ry = arg[2]->a_number(2),
		rz = arg[3]->a_number(3),
		slicesstart=arg[4]->a_number(4),
		slicesend=arg[5]->a_number(5),
		stacksstart=arg[7]->a_number(7),
		stacksend=arg[8]->a_number(8);
	int slices=arg[6]->a_int(6);
	int stacks=arg[9]->a_int(9);
	if throwing return Unbound;
	Point r=Point(rx->getnumber(),ry->getnumber(),rz->getnumber());
	double 
		sls=slicesstart->getnumber(),
		sle=slicesend->getnumber(),
		sts=stacksstart->getnumber(),
		ste=stacksend->getnumber();
	TURTLE->prepareToDraw();
	if (TURTLE->shadows)
		TURTLE->partialEllipsoidSh(r, sls, sle, slices, sts, ste, stacks);
	else
		TURTLE->partialEllipsoid(r, sls, sle, slices, sts, ste, stacks);
	if (RECORDER.recording)
		RECORDER << new R_partialEllipsoid(*TURTLE, 
			r, sls, sle, slices, sts, ste, stacks);
	return Unbound;
}

NodeP lfillPie(PairPP arg)
{
	NodeP 
		angle   = arg[1]->a_number(1),
		radiusx = arg[2]->a_number(2);
	
	if (!throwing) 
	{	double 
			ang=angle->getnumber(),
			rx=radiusx->getnumber();
		Point r=Point(rx,rx,0);
		TURTLE->prepareToDraw();
		TURTLE->fillPie(ang,r);
		if (RECORDER.recording)
			RECORDER << new R_FillPie(*TURTLE, ang,r);
	}
	return Unbound;
}

NodeP lLabel(PairPP arg) 
{
	char textbuf[300];
	File::stringptr=textbuf;
	File::stringlen=300-1;
	stringbuffer << arg[1];
	*File::stringptr='\0';

	if (!throwing) 
	{	TURTLE->prepareToDraw();
		TURTLE->draw();
		TURTLE->Label(textbuf);
		if (RECORDER.recording)
			RECORDER << new R_Label(*TURTLE, textbuf);
		TURTLE->draw();
	}
	return Unbound;
}

NodeP lLabelSize(PairPP arg) 
{
	if (noarg)
	{	Font_ &f=TURTLE->font;
		return 
			new Pair(new Float(f.sizex),
			new Pair(new Float(f.sizey), 0));
	}
	char textbuf[300];
	File::stringptr=textbuf;
	File::stringlen=300-1;
	stringbuffer << arg[1];
	*File::stringptr='\0';

	if (!throwing) 
	{	Point s=TURTLE->LabelSize(textbuf);
		return 
			new Pair(new Float(s.x), 
			new Pair(new Float(s.y), 
			new Pair(new Float(s.z), 0)));
	}
	return Unbound;
}

NodeP lsetLabelSize(PairPP arg)
{
	PairP l=arg[1]->a_vector(1);
	if (!throwing)
	{	NodeP xn=l->first(), yn=l->second();
		double x=xn->getnumber(), y=yn->getnumber();
		TURTLE->setLabelSize(x,y);
	}
	return Unbound;
}

NodeP lLabelFont(PairPP)
{
	StringP fontname=new String(TURTLE->getLabelFont());
	return fontname;
}

NodeP lsetLabelFont(PairPP arg)
{
	StringP font=arg[1]->a_string(1);
	if (!throwing)
	{	TURTLE->setLabelFont(font->ptr());
	}
	return Unbound;
}

NodeP lLabelWeight(PairPP arg)
{
	return new Float(TURTLE->font.weight);
}

NodeP lsetLabelWeight(PairPP arg)
{
	int weight=arg[1]->a_int(1);
	if (!throwing)
	{	TURTLE->setLabelWeight(weight);
	}
	return Unbound;
}

NodeP lLabelAlign(PairPP arg)
{
	Font_ &f=TURTLE->font;
	return 
		new Pair(new Int(f.alignx),
		new Pair(new Int(f.aligny), 0));
}

NodeP lsetLabelAlign(PairPP arg)
{
	int	alignx=arg[1]->a_int(1),
			aligny=arg[2]->a_int(2);
	if (!throwing)
		TURTLE->setLabelAlign(alignx,aligny);
	return Unbound;
}

NodeP lisPenDown  (PairPP){return TURTLE->pen.down ?  _True  :  _False;}
NodeP lPenMode    (PairPP){return TURTLE->getNodePenMode();}
NodeP lPenColor   (PairPP){return new Int(TURTLE->pen.color);}
NodeP lScreenColor(PairPP){return new Int(SCREEN->color);}

NodeP lPenSize(PairPP) 
{
	return new List
	(	new Float(TURTLE->pen.size.x),
		new Float(TURTLE->pen.size.y), _ListEnd
	);
}

NodeP lPenPattern(PairPP) {return 0;} // no Patterns yet!

NodeP lPenDown(PairPP) 
{
	TURTLE->pen.down=true; 
	if (RECORDER.recording)
		RECORDER << new R_SetPenDown(*TURTLE); 
	return Unbound;
}

NodeP lPenUp(PairPP) 
{
	TURTLE->pen.down=false; 
	if (RECORDER.recording)
		RECORDER << new R_SetPenDown(*TURTLE); 
	return Unbound;
}

NodeP lPenPaint(PairPP) 
{
	TURTLE->pen.internalMode=Pen::Paint;
	TURTLE->penDown();
	if (RECORDER.recording)
		RECORDER << new R_SetPenMode(*TURTLE);
	return lPenDown(argStack);
}

NodeP lPenErase(PairPP) 
{
	TURTLE->pen.internalMode=Pen::Erase;
	TURTLE->penErase();
	if (RECORDER.recording)
		RECORDER << new R_SetPenMode(*TURTLE);
	return lPenDown(argStack);
}

NodeP lPenReverse(PairPP) 
{
	TURTLE->pen.internalMode=Pen::Reverse;
	TURTLE->penReverse();
	if (RECORDER.recording)
		RECORDER << new R_SetPenMode(*TURTLE);
	return lPenDown(argStack);
}

NodeP lRGB(PairPP arg)	// RGB=Red, Green, Blue
{
	NodeP r,g,b;
	if (!exist(2) && arg[1] && arg[1]->is.list)
	{	PairP l=arg[1]->a_rgb(1);
		r=l->first();
		g=l->second();
		b=l->third();
	}else
	{	r=arg[1]->a_number(1), 
		g=arg[2]->a_number(2), 
		b=arg[3]->a_number(3);
	}
	if throwing return Unbound;
	Color c=Pen::rgb(r->getnumber(), g->getnumber(), b->getnumber());
	return new Int(c);
}

NodeP lRGBA(PairPP arg)	// RGBA=Red, Green, Blue, Alpha
{
	NodeP r,g,b,a;
	if (!exist(2) && arg[1] && arg[1]->is.list)
	{	PairP l=arg[1]->a_list(1);
		r=l->first();
		g=l->second();
		b=l->third();
		a=l->fourth();
	}else
	{	r=arg[1]->a_number(1), 
		g=arg[2]->a_number(2), 
		b=arg[3]->a_number(3);
		a=arg[4]->a_number(4);
	}
	if throwing return Unbound;
	Color c=Pen::rgba(r->getnumber(), g->getnumber(), b->getnumber(), a->getnumber());
	return new Int(c);
}

NodeP lreRGB(PairPP arg)
{
	Color c=arg[1]->a_int(1);
	if throwing return Unbound;
	return new List
	(	new Float(double(c.r)/0xff), 
		new Float(double(c.g)/0xff), 
		new Float(double(c.b)/0xff), _ListEnd
	);
}

NodeP lreRGBA(PairPP arg)
{
	Color c=arg[1]->a_int(1);
	if throwing return Unbound;
	return new List
	(	new Float(double(c.r)/0xff), 
		new Float(double(c.g)/0xff), 
		new Float(double(c.b)/0xff), 
		new Float(double(c.a)/0xff), _ListEnd
	);
}

NodeP lHSB(PairPP arg)	// HSB=Hue Saturation Brightness
{
	NodeP 
		h=arg[1]->a_number(1), 
		s=arg[2]->a_number(2), 

		b=arg[3]->a_number(3);
	if throwing return Unbound;
	Color c=Pen::hsb(h->getnumber(), s->getnumber(), b->getnumber());
	return new Int(c);
}

NodeP lHSBA(PairPP arg)	// HSB=Hue Saturation Brightness Alpha
{
	NodeP 
		h=arg[1]->a_number(1), 
		s=arg[2]->a_number(2), 
		b=arg[3]->a_number(3),
		a=arg[4]->a_number(4);
	if throwing return Unbound;
	Color c=Pen::hsba(
		h->getnumber(), 
		s->getnumber(), 
		b->getnumber(), 
		a->getnumber());
	return new Int(c);
}

NodeP lreHSB(PairPP arg)	// color to HSB
{
	Color	c=arg[1]->a_color(1);
	if throwing return Unbound;
	return Pen::rehsb(
		double(c.r)/0xff, 
		double(c.g)/0xff, 
		double(c.b)/0xff);
}

NodeP lreHSBA(PairPP arg)	// color to HSBA
{
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	return Pen::rehsba(
		double(c.r)/0xff, 
		double(c.g)/0xff, 
		double(c.b)/0xff, 
		double(c.a)/0xff);
}

NodeP laddColors(PairPP arg)
{
	Color
		c1=arg[1]->a_int(1),
		c2=arg[2]->a_int(2),
		c12;
	if (!throwing)
	{	int 
			r=c1.r+c2.r,
			g=c1.g+c2.g,
			b=c1.b+c2.b;
		if (r>255) r=255;
		if (g>255) g=255;
		if (b>255) b=255;
		c12.r=r;
		c12.g=g;
		c12.b=b;
		return new Int(c12);
	}
	return Unbound;
}

NodeP laddColorsMod(PairPP arg)
{
	Color
		c1=arg[1]->a_int(1),
		c2=arg[2]->a_int(2),
		c12;
	if (!throwing)
	{	int 
			r=(c1.r+c2.r) & 0xff,
			g=(c1.g+c2.g) & 0xff,
			b=(c1.b+c2.b) & 0xff,
			a=(c1.a+c2.a)/2;
		c12.r=r;
		c12.g=g;
		c12.b=b;
		c12.a=a;
		return new Int(c12);
	}
	return Unbound;
}

NodeP lgetColorDatabase(PairPP)
{
	wxStringToColourHashMap *cols=wxTheColourDatabase->GetColourHashMap();
	wxStringToColourHashMap::iterator it;
	PairP result=0;
	for(it=cols->begin();  it!=cols->end();  ++it)
	{	wxString name = it->first;
		wxColour *c = it->second;
		push(result, new List(
			new String(name), 
			new Float(c->Red()/255.0),
			new Float(c->Green()/255.0),
			new Float(c->Blue()/255.0), _ListEnd));
	}
	result=result->mergeSort();
	return result;
}

NodeP lsetPenColor(PairPP arg) 
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	TURTLE->setColor(c);
		if (!TURTLE->surfaceMode && RECORDER.recording)
			RECORDER << new R_SetPenColor(*TURTLE);
	}
	return Unbound;
}

NodeP lsetFloodColor(PairPP arg) 
{
	Color c=arg[1]->a_color(1);
	if (!throwing) 
	{	TURTLE->pen.floodcolor=TURTLE->pen.realcolor(c);
		if (RECORDER.recording)
			RECORDER << new R_SetFloodColor(*TURTLE);
	}
	return Unbound;
}

NodeP lFloodColor(PairPP arg) 
{
	return new Int(TURTLE->pen.floodcolor);
}

NodeP lsetShadowColor(PairPP arg) 
{
	Color c=arg[1]->a_color(1);
	if (!throwing) 
	{	TURTLE->shadowcolor=TURTLE->pen.realcolor(c);
		if (RECORDER.recording)
			RECORDER << new R_SetShadowColor(*TURTLE);
	}
	return Unbound;
}

NodeP lShadowColor(PairPP arg) 
{
	return new Int(TURTLE->shadowcolor);
}

NodeP lsetScreenColor(PairPP arg) 
{
	Color c=arg[1]->a_color(1);
	if (!throwing) 
	{	TURTLE->prepareToDraw();
		TURTLE->setScreenColor(c);
		if (RECORDER.recording)
			RECORDER << new R_SetScreenColor(SCREEN->color);
		graph->screen_reset(true);
	}
	return Unbound;
}


NodeP lsetPenSize(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
	{	TURTLE->prepareToDraw();
		TURTLE->setPenSize( 
			v-> first()->getnumber(),
			v->second()->getnumber());
		if (RECORDER.recording)
			RECORDER << new R_SetPenSize(*TURTLE);
	}
	return Unbound;
}

NodeP lsetPenPattern(PairPP arg) 
{
	while (arg[1]->isnolist() && !throwing)
		arg[1]=error(BAD_DATA, arg[1]);

	if (!throwing) 
	{	TURTLE->prepareToDraw();
		TURTLE->pen.setListPattern(arg[1]);
		if (RECORDER.recording)
			RECORDER << new R_SetPenPattern(*TURTLE);
	}
	return Unbound;
}

NodeP lsetScrunch(PairPP arg) 
{
	NodeP
		nodeX = arg[1]->a_number(1),
		nodeY = arg[2]->a_number(2);

	if (!throwing) 
	{	TURTLE->prepareToDraw();
		TURTLE->draw();
		TURTLE->screen->scale.x = nodeX->getnumber();
		TURTLE->screen->scale.y = nodeY->getnumber();
		TURTLE->draw();
	}
	return Unbound;
}

NodeP lMousePos(PairPP) 
{
	console->dispatchMessages();
	return 
		new Pair(new Float(graph->mouse->getX()), 
		new Pair(new Float(graph->mouse->getY()), 
		new Pair(new Float(graph->mouse->getZ()), 0)));
}

NodeP lMouseX(PairPP){return new Float(graph->mouse->getX());}
NodeP lMouseY(PairPP){return new Float(graph->mouse->getY());}
NodeP lMouseZ(PairPP){return new Float(graph->mouse->getZ());}

NodeP lTextMousePos(PairPP) 
{
	return 
		new Pair(new Int(graph->mouse->textx), 
		new Pair(new Int(graph->mouse->texty), 0));
}

NodeP lTextMouseX(PairPP){return new Int(graph->mouse->textx);}
NodeP lTextMouseY(PairPP){return new Int(graph->mouse->texty);}

NodeP lMouseButtons(PairPP)
{
	return new Int(graph->mouse->button());
}

NodeP lArc(PairPP arg)
{
	NodeP
		 angleNd = arg[1]->a_number(1),
		radiusNd = arg[2]->a_number(2);
	
	if (!throwing) 
	{	double 
			 angle= angleNd->getnumber(), 
			radius=radiusNd->getnumber();

		TURTLE->arc(angle, radius);
	}
	return Unbound;
}

NodeP lCircle(PairPP arg)
{
	NodeP
		radiusNd = arg[1]->a_number(1);
	
	if (!throwing) 
	{	double radius=radiusNd->getnumber();
		TURTLE->arc(360, radius);
	}
	return Unbound;
}

NodeP lEllipseArc(PairPP arg)
{
	NodeP
		      angleNd = arg[1]->a_number(1),
		minorRadiusNd = arg[2]->a_number(2),
		majorRadiusNd = arg[3]->a_number(3),
		 startAngleNd = arg[4]->a_number(4);
	
	if (!throwing) 
	{	double 
			      angle=      angleNd->getnumber(), 
			minorRadius=minorRadiusNd->getnumber(),
			majorRadius=majorRadiusNd->getnumber(),
			 startAngle= startAngleNd->getnumber();

		TURTLE->EllipseArc(angle, minorRadius, majorRadius, startAngle);
	}
	return Unbound;
}

NodeP lEllipse(PairPP arg)
{
	NodeP
		minorRadiusNd = arg[1]->a_number(1),
		majorRadiusNd = arg[2]->a_number(2);
	
	if (!throwing) 
	{	double 
			minorRadius=minorRadiusNd->getnumber(),
			majorRadius=majorRadiusNd->getnumber();

		TURTLE->EllipseArc(360, minorRadius, majorRadius, 0);
	}
	return Unbound;
}

NodeP lrefresh(PairPP) {RECORDER.on (); return Unbound;}
NodeP lnoRefresh(PairPP) {RECORDER.off(); return Unbound;}
NodeP lisRefresh(PairPP) {return Node::boolnode(RECORDER.recording);}

NodeP lredraw(PairPP) {RECORDER.playAgain(); return Unbound;}

NodeP lsavePicture    (PairPP arg) 
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	RECORDER.save    (name->ptr()); return Unbound;
}
NodeP lloadPicture    (PairPP arg) 
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	RECORDER.load    (name->ptr()); return Unbound;
}
NodeP lsavePictureText(PairPP arg) 
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	RECORDER.saveText(name->ptr()); return Unbound;
}
NodeP lloadPictureText(PairPP arg) 
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	RECORDER.loadText(name->ptr()); return Unbound;
}
NodeP lsavePostScript (PairPP arg) 
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	RECORDER.writePS (name->ptr()); return Unbound;
}

NodeP lsetSaveSize(PairPP arg)
{
	PairP l=arg[1]->a_vector(1);
	if (!throwing)
	{	graph->saveSize.SetWidth ( int(l-> first()->getnumber()) );	
		graph->saveSize.SetHeight( int(l->second()->getnumber()) );
	}
	return Unbound;
}

NodeP lSaveSize(PairPP arg)
{
	return new List(
		new Int(graph->saveSize.GetWidth()),
		new Int(graph->saveSize.GetHeight()), _ListEnd);
}

NodeP lsaveScreen(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	return graph->saveScreen(filename->ptr());
}

NodeP lsaveScreenVector(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	int sort=GL2PS_SIMPLE_SORT;
	if (exist(2) && arg[2])
	{	sort=arg[2]->a_int(2);
		if throwing return Unbound;
	}
	int options=GL2PS_DRAW_BACKGROUND;
	if (exist(3) && arg[3])
	{	options=arg[3]->a_int(3);
		if throwing return Unbound;
	}
	int colornr=0;
	if (exist(4))
	{	colornr=arg[4]->a_int(3);
		if throwing return Unbound;
	}
	graph->saveVector(filename->ptr(), sort, options, colornr);
	return Unbound;
}

NodeP lupdateGraph(PairPP)
{
	graph->update();
	return Unbound;
}

NodeP lsetUpdateGraph(PairPP arg)
{
	int b=arg[1]->a_bool(1);
	if (!throwing)
	{	TURTLE->direct= b==1 ? true : false;
	}
	return Unbound;
}

NodeP lloadImage(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if (throwing)
		return Unbound;

	wxImage img;
	if (!img.LoadFile(filename->ptr()))
	{	error(BAD_DATA, filename);
		return Unbound;
	}

	Color *rgba=(Color*)GC_::malloc(img.GetWidth()*img.GetHeight()*4),
		*rgbap=rgba;
	if (!rgba)
		return Unbound;
	if (img.HasAlpha())
	{	for (int y=img.GetHeight()-1;  y >= 0;  y--)
		{	for (int x=0;  x< img.GetWidth();  x++)
			{	rgbap->r = img.GetRed(x,y);
				rgbap->g = img.GetGreen(x,y);
				rgbap->b = img.GetBlue(x,y);
				rgbap->a = img.GetAlpha(x,y);
				rgbap++;
			}
		}
	}
	else
	{	for (int y=img.GetHeight()-1;  y >= 0;  y--)
		{	for (int x=0;  x< img.GetWidth();  x++)
			{	rgbap->r = img.GetRed(x,y);
				rgbap->g = img.GetGreen(x,y);
				rgbap->b = img.GetBlue(x,y);
				rgbap->a = 0xff;
				rgbap++;
			}
		}
	}

	return new Bitmap(
		img.GetWidth(),img.GetHeight(), 
		img.GetWidth(),img.GetHeight(), rgba);
}

NodeP lsaveImage(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	StringP filename=arg[2]->a_string(2);
	if (throwing)
		return Unbound;

	short mx=b->winx, my=b->winy;

	wxImage img(mx, my, false);
	img.InitAlpha();

	Color *rgba=b->pixels,
		*rgbap=rgba;

	for (int y=my-1; y>=0; y--)
	{	for (int x=0; x<mx; x++)
		{	img.SetRGB(x,y, rgbap->r, rgbap->g, rgbap->b);
			img.SetAlpha(x,y, rgbap->a);
			rgbap++;
		}
	}
	if (!img.SaveFile(filename->ptr()))
		error(BAD_DATA, filename, 2);	
	return Unbound;
}

NodeP lBitCopy(PairPP arg)
{
	double 
		w=arg[1]->a_float(1),
		h=arg[2]->a_float(2);
	if (!throwing)
		return NodeP(TURTLE->bitCopy(w,h));
	return Unbound;
}

NodeP lBitPaste(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	if (!throwing)
		TURTLE->bitPaste(b);
	return Unbound;
}

NodeP lBitMakeTransparent(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	int c=arg[2]->a_color(2);
	if (!throwing)
		b->makeTransparent(c);
	return Unbound;
}

NodeP lBitSetPixel(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	if (arg[2]->type==INT_ ||  arg[2]->type==FLOAT_)
	{	int x=arg[2]->getnumber();
		if (arg[3]->type==INT_ ||  arg[3]->type==FLOAT_)
		{	int y=arg[3]->getnumber();
			Color c=arg[4]->a_color(4);
			b->setPixel(x, y, c);
		}
	}
	else if (arg[2]->type==INTARRAY)
	{	IntArrayP xa=IntArrayP(arg[2]);
		intP x=xa->data();
		int l=xa->length();
		if (arg[3]->type==INT_)
		{	int y=IntP(arg[3])->int_;
			if (arg[4]->type==INTARRAY)
			{	IntArrayP cia=IntArrayP(arg[4]);
				if (l != cia->length())
				{	error(BAD_DATA, cia, 4);
					return Unbound;
				}
				intP c=cia->data();
				while (l--)
					b->setPixel(*x++, y, *c++);
			}
			else if (arg[4]->type==INT_)
			{	int c=IntP(arg[4])->int_;
				while (l--)
					b->setPixel(*x++, y, c);
			}
			else
				error(BAD_DATA, arg[4], 4);
		}
		else
			error(BAD_DATA, arg[3], 3);
	}
	else
		error(BAD_DATA, arg[2], 2);
	return Unbound;
}

NodeP lBitPixel(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	int x=arg[2]->a_float(2),
		y=arg[3]->a_float(3);
	if (!throwing)
		return new Int(b->Pixel(x, y));
	return Unbound;
}

NodeP lBitMaxX(PairPP arg)
{

	BitmapP b=arg[1]->a_bitmap(1);
	if throwing return Unbound;
	return new Int(b->winx-1);
}

NodeP lBitMaxY(PairPP arg)
{
	BitmapP b=arg[1]->a_bitmap(1);
	if throwing return Unbound;
	return new Int(b->winy-1);
}

NodeP lscroll(PairPP arg)
{
	PairP 
		size=arg[1]->a_vector(1),
		displacement=arg[2]->a_vector(2);

	if (!throwing)
		TURTLE->scroll(
			int(size-> first()->getnumber()),
			int(size->second()->getnumber()),
			displacement-> first()->getnumber(),
			displacement->second()->getnumber());
	return Unbound;
}

NodeP lscrollCalibrate(PairPP arg)
{
	PairP 
		displacement=arg[1]->a_vector(1);

	if (!throwing)
	{	Point s=TURTLE->scrollCalibrate(			
			displacement-> first()->getnumber(),
			displacement->second()->getnumber());
		return new List(
			new Float(s.x), new Float(s.y), _ListEnd);
	}
	return Unbound;
}

NodeP lTurtle(PairPP)
{
	return graph->turtleNode;
}

NodeP lnewTurtle(PairPP)
{
	Turtle *t=new Turtle;
	t->create(graph->recorder, SCREEN);
	t->draw();
	return new TurtleNode(t, graph->graphNode);//theApp->mainFrame->graphNode);
}

NodeP lsetTurtle(PairPP arg)
{
	TurtleNodeP tn=arg[1]->a_turtleNode(1);
	if (!throwing)
	{	graph->turtleNode=tn;
		Turtle *t=tn->turtle;
		t->recorder=graph->recorder;
		t->screen=graph->screen;
		TURTLE=t;
		RECORDER.turtle=t;
		t->setColor(t->pen.color);
		if (RECORDER.recording)
			RECORDER << new R_SetPenColor(*TURTLE);
		Point p=*t;
		t->getInWindow();
		t->moveTo();
		*t=p;
	}
	return Unbound;
}

NodeP lperspective(PairPP)
{
	TURTLE->draw();
	SCREEN->perspective=true;
	graph->setFrustum(graph->GetSize().x,graph->GetSize().y);

	SCREEN->eye=Point(400,400,600);
	SCREEN->center=Point(0,0,0);
	SCREEN->up=Point(0,1,0);
	SCREEN->setEye();
	SCREEN->getMatrices();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

	glFogf(GL_FOG_START, 300);
	glFogf(GL_FOG_END, 1000);
	glFogf(GL_FOG_DENSITY, 0.04f);
	float c[4]={0,0,0,0};
	glFogfv(GL_FOG_COLOR, (float*)&c);
	glFogi(GL_FOG_MODE, GL_LINEAR);

	if (RECORDER.recording)
		RECORDER << new R_Perspective();

	TURTLE->reset3d();
	TURTLE->mode=Turtle::windowmode;
	TURTLE->draw();
	return Unbound;
}

NodeP lunPerspective(PairPP)
{
	TURTLE->draw();
	SCREEN->perspective=false;
	graph->setFrustum(SCREEN->winsizex, SCREEN->winsizey);
	glLoadIdentity();
	SCREEN->getMatrices();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);

	TURTLE->reset3d();
	TURTLE->draw();
	return Unbound;
}


NodeP lsetEye(PairPP arg)
{
	Point 
		   eye=arg[1]->a_3darray(1),
		center=arg[2]->a_3darray(2),
		    up=arg[3]->a_3darray(3);

	if (!throwing)
	{	TURTLE->draw();
		SCREEN->eye=eye;
		SCREEN->center=center;
		SCREEN->up=up;
		SCREEN->setEye();
		SCREEN->getMatrices();
		TURTLE->draw();
	}
	return Unbound;
}

NodeP lsetLightPos(PairPP arg)
{
	Point pos=arg[1]->a_3darray(1);
	if (!throwing)
	{	TURTLE->lightPos=pos;
		float p[4]={pos.x,pos.y,pos.z,0};
		glLightfv(GL_LIGHT0, GL_POSITION, p);
		if (RECORDER.recording)
			RECORDER << new R_SetLightPos(pos);
	}
	return Unbound;
}

NodeP lsetLightAmbient(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glLightfv(GL_LIGHT0, GL_AMBIENT, cf);
		if (RECORDER.recording)
			RECORDER << new R_Light(GL_AMBIENT, c);
	}
	return Unbound;
}

NodeP lsetLightDiffuse(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,

			float(c.a)/255
		};
		glLightfv(GL_LIGHT0, GL_DIFFUSE, cf);
		if (RECORDER.recording)
			RECORDER << new R_Light(GL_DIFFUSE, c);
	}
	return Unbound;
}

NodeP lsetLightSpecular(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glLightfv(GL_LIGHT0, GL_SPECULAR, cf);
		if (RECORDER.recording)
			RECORDER << new R_Light(GL_SPECULAR, c);
	}
	return Unbound;
}

NodeP lsetMaterialAmbient(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, cf);
		if (RECORDER.recording)
			RECORDER << new R_Material(GL_AMBIENT, c);
	}
	return Unbound;
}

NodeP lsetMaterialDiffuse(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, cf);
		if (RECORDER.recording)
			RECORDER << new R_Material(GL_DIFFUSE, c);
	}
	return Unbound;
}

NodeP lsetMaterialSpecular(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, cf);
		if (RECORDER.recording)
			RECORDER << new R_Material(GL_SPECULAR, c);
	}
	return Unbound;
}

NodeP lsetMaterialEmission(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if (!throwing)
	{	float cf[4]=
		{	float(c.r)/255,
			float(c.g)/255,
			float(c.b)/255,
			float(c.a)/255
		};
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, cf);
		if (RECORDER.recording)
			RECORDER << new R_Material(GL_EMISSION, c);
	}
	return Unbound;
}

NodeP lsetMaterialShininess(PairPP arg)
{
	float s=arg[1]->a_float(1);
	if throwing return Unbound;
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, s);
	if (RECORDER.recording)
		RECORDER << new R_Shininess(s);
	return Unbound;
}

NodeP lsetFogDensity(PairPP arg)
{
	float density=float(arg[1]->a_float(1));
	if throwing return Unbound;
	glFogf(GL_FOG_DENSITY, density);
	return Unbound;
}

NodeP lsetFogRange(PairPP arg)
{
	float start=float(arg[1]->a_float(1));
	if throwing return Unbound;
	float end=float(arg[2]->a_float(2));
	if throwing return Unbound;
	glFogf(GL_FOG_START, start);
	glFogf(GL_FOG_END, end);
	return Unbound;
}

NodeP lsetFogColor(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	float col[4]={c.r/255,c.g/255,c.b/255,c.a/255};
	glFogfv(GL_FOG_COLOR, (float*)&col);
	return Unbound;
}

NodeP lsetFogMode(PairPP arg)
{
	int mode=arg[1]->a_int(1);
	if throwing return Unbound;
	glFogi(GL_FOG_MODE, mode);
	return Unbound;
}

NodeP lPolyStart(PairPP)
{
	TURTLE->polyStart();
	return Unbound;
}

NodeP lPolyEnd(PairPP)
{
	TURTLE->polyEnd();
	return Unbound;
}

NodeP lTessStart(PairPP)
{
	TURTLE->tessStart();
	return Unbound;
}

NodeP lTessContour(PairPP)
{
	TURTLE->tessContour();
	return Unbound;
}

NodeP lTessEnd(PairPP)
{
	TURTLE->tessEnd();
	return Unbound;
}

NodeP lsetTessWindingRule(PairPP arg)
{
	int rule=arg[1]->a_int(1);
	if (!throwing)
	{	if (rule < 0 || rule > 4)
		{	error(BAD_DATA, arg[1]);
			return Unbound;
		}
		TURTLE->tessWindingRule=GLU_TESS_WINDING_ODD+rule;
		if (RECORDER.recording)
			RECORDER << new R_TessWindingRule(rule);
	}
	return Unbound;
}

NodeP lSurfaceStart(PairPP)
{
	TURTLE->surfaceStart();
	return Unbound;
}

NodeP lSurfaceColumn(PairPP)
{
	TURTLE->surfaceColumn();
	return Unbound;
}

NodeP lSurfaceEnd(PairPP)
{
	TURTLE->surfaceEnd();
	return Unbound;
}

NodeP lenablePolySmooth(PairPP)
{	glEnable(GL_POLYGON_SMOOTH);
	if (RECORDER.recording)
		RECORDER << new R_SetPolySmooth(true);
	return Unbound;
}
NodeP ldisablePolySmooth(PairPP)
{	glDisable(GL_POLYGON_SMOOTH); 
	if (RECORDER.recording)
		RECORDER << new R_SetPolySmooth(false); 
	return Unbound;
}

NodeP lenableLineSmooth(PairPP)
{	glEnable(GL_LINE_SMOOTH);
	if (RECORDER.recording)
		RECORDER << new R_SetLineSmooth(true);
	return Unbound;
}
NodeP ldisableLineSmooth(PairPP)
{	glDisable(GL_LINE_SMOOTH);
	if (RECORDER.recording)
		RECORDER << new R_SetLineSmooth(false);
	return Unbound;
}

NodeP lenablePointSmooth(PairPP)
{	glEnable(GL_POINT_SMOOTH);
	if (RECORDER.recording)
		RECORDER << new R_SetPointSmooth(true);
	return Unbound;
}
NodeP ldisablePointSmooth(PairPP)
{	glDisable(GL_POINT_SMOOTH);
	if (RECORDER.recording)
		RECORDER << new R_SetPointSmooth(false);
	return Unbound;
}

NodeP lenableDepthTest(PairPP)
{	glEnable(GL_DEPTH_TEST);
	if (RECORDER.recording)
		RECORDER << new R_SetDepthTest(true);
	return Unbound;
}
NodeP ldisableDepthTest(PairPP)
{	glDisable(GL_DEPTH_TEST); 
	if (RECORDER.recording)
		RECORDER << new R_SetDepthTest(false);
	return Unbound;
}

NodeP lenableLighting(PairPP)
{	glEnable(GL_LIGHTING);
	if (RECORDER.recording)
		RECORDER << new R_SetLighting(true);
	return Unbound;
}
NodeP ldisableLighting(PairPP)
{	glDisable(GL_LIGHTING); 
	if (RECORDER.recording)
		RECORDER << new R_SetLighting(false);
	return Unbound;
}

NodeP lenableDither(PairPP)
{	glEnable(GL_DITHER);
	if (RECORDER.recording)
		RECORDER << new R_SetDither(true);
	return Unbound;
}
NodeP ldisableDither(PairPP)
{	glDisable(GL_DITHER); 
	if (RECORDER.recording)
		RECORDER << new R_SetDither(false);
	return Unbound;
}

NodeP lenableTexture(PairPP)
{	glEnable(GL_TEXTURE_2D); SCREEN->textured=true; 
	if (RECORDER.recording)
		RECORDER << new R_SetTextured(true); 
	return Unbound;
}

NodeP ldisableTexture(PairPP)
{	glDisable(GL_TEXTURE_2D); SCREEN->textured=false; 
	if (RECORDER.recording)
		RECORDER << new R_SetTextured(false); 
	return Unbound;
}

NodeP lenableRoundLineEnds(PairPP)
{	TURTLE->pen.roundLineEnds=true; 
	if (RECORDER.recording)
		RECORDER << new R_SetPenRLE(*TURTLE); 
	return Unbound;
}
NodeP ldisableRoundLineEnds(PairPP)
{	TURTLE->pen.roundLineEnds=false; 
	if (RECORDER.recording)
		RECORDER << new R_SetPenRLE(*TURTLE); 
	return Unbound;
}

NodeP lenableCylinderLines(PairPP)
{	TURTLE->cylinderLines=true; 
	if (RECORDER.recording)
		RECORDER << new R_SetCyLines(*TURTLE); 
	return Unbound;
}
NodeP ldisableCylinderLines(PairPP)
{	TURTLE->cylinderLines=false; 
	if (RECORDER.recording)
		RECORDER << new R_SetCyLines(*TURTLE); 
	return Unbound;
}

NodeP lenableShadows(PairPP)
{	TURTLE->shadows=true; 
	if (RECORDER.recording)
		RECORDER << new R_SetShadows(*TURTLE); 
	return Unbound;
}
NodeP ldisableShadows(PairPP)
{	TURTLE->shadows=false; 
	if (RECORDER.recording)
		RECORDER << new R_SetShadows(*TURTLE); 
	return Unbound;
}

NodeP lenableFog(PairPP)
{	glEnable(GL_FOG); 
	if (RECORDER.recording)
		RECORDER << new R_SetFogEnabled(true); 
	return Unbound;
}
NodeP ldisableFog(PairPP)
{	glDisable(GL_FOG); 
	if (RECORDER.recording)
		RECORDER << new R_SetFogEnabled(false); 
	return Unbound;
}

NodeP lsetTexWrapping(PairPP arg)
{
	int wrapping=arg[1]->a_bool(1);
	if throwing return Unbound;
	if (wrapping)
	{	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}
	else
	{	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
	return Unbound;
}

NodeP ldoubleBuffer(PairPP)
{	
	glReadBuffer(GL_BACK);
	glDrawBuffer(GL_BACK);
	graph->doubleBuffer=true; 
	theApp->doubleBuffer=true; 
	return Unbound;
}
NodeP lsingleBuffer(PairPP)
{
	graph->doubleBuffer=false; 
	theApp->doubleBuffer=false; 
	return Unbound;
}

NodeP lsetDepthFunc(PairPP arg)
{
	int f=arg[1]->a_int(1);
	if (f < 0 || f > 7)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	glDepthFunc(GL_NEVER+f);
	if (RECORDER.recording)
		RECORDER << new R_SetDepthFunc(f); 
	return Unbound;
}

NodeP lTexture(PairPP arg)
{
	if (!arg[1])
	{	ldisableTexture(0);
	}
	else if (arg[1]->type==BITMAP_)
	{	BitmapP b=BitmapP(arg[1]);
		TextureP tex=new Texture(b);
		SCREEN->textured=true; 
		if (RECORDER.recording)
			RECORDER << new R_Texture(tex); 
		return tex;
	}
	else if (arg[1]->type==TEXTURE)
	{	TextureP tex=TextureP(arg[1]);
		glBindTexture(GL_TEXTURE_2D, tex->_texnr);
		SCREEN->textured=true; 
		if (RECORDER.recording)
			RECORDER << new R_Texture(tex); 
	}
	else
		error(BAD_DATA, arg[1]);
	return Unbound;
}

NodeP lsetTexXY(PairPP arg) 
{
	double 
		x = arg[1]->a_float(1),
		y = arg[2]->a_float(2);
	if (!throwing) 
		TURTLE->texpos=Point(x, y, 0);
	return Unbound;
}

NodeP lsetTexPos(PairPP arg) 
{
	PairP v = arg[1]->a_vector(1);
	if (!throwing) 
		TURTLE->texpos=Point(v->first()->getnumber(), v->second()->getnumber(), 0);
	return Unbound;
}

NodeP ldeleteTextures(PairPP)
{
	if (R_Texture::maxnr)
	{	unsigned *textures=new unsigned[R_Texture::maxnr];
		for (int i=1; i<=R_Texture::maxnr; i++)
			textures[i-1]=i;
		glDeleteTextures(R_Texture::maxnr, textures);
		R_Texture::maxnr=0;
		delete textures;
	}
	R_Texture::textures=0;
	return Unbound;
}

NodeP lVideoStart(PairPP arg)
{
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	int rate;
	if (exist(2))
		rate=arg[2]->a_int(2);
	else
		rate=30;
	if throwing return Unbound;

	graph->videoStart(filename->ptr(), rate);
	return Unbound;
}

NodeP lVideoFrame(PairPP arg)
{
	graph->videoFrame();
	return Unbound;
}

NodeP lVideoEnd(PairPP arg)
{
	graph->videoEnd();
	return Unbound;
}


NodeP lGraphicStart(PairPP)
{
	if (graph->recorder != graph->recorderSave)
	{	error(BAD_DATA, new String("You must call GraphicEnd before a new GraphicStart!"));
		return Unbound;
	}
	graph->recorderSave=graph->recorder;
	graph->recorder=new Recorder(graph);
	graph->turtle->recorder=graph->recorder;
	return Unbound;
}

NodeP lGraphicEnd(PairPP)
{
	if (graph->recorder == graph->recorderSave)
	{	error(BAD_DATA, new String("You must call GraphicStart before GraphicEnd!"));
		return Unbound;
	}
	Recorder *r=graph->recorder;
	graph->recorder=graph->recorderSave;
	graph->turtle->recorder=graph->recorder;

	if (RECORDER.recording)
		RECORDER << new R_Graphic(*TURTLE, r, 1);

	return new GraphicNode(r);
}

NodeP ldrawGraphic(PairPP arg)
{
	GraphicNodeP g=arg[1]->a_graphicNode(1);
	double scale=1;
	bool isscaled=false;
	if (exist(2))
	{	scale=arg[2]->a_float(2);
		isscaled=true;
	}
	if throwing return Unbound;

	bool rr=RECORDER.recording;
	RECORDER.recording=false;

	TURTLE->draw();
	static Turtle save(false);
	save=*TURTLE;
	glPushMatrix();
	TURTLE->setTurtleMatrix();
	GLboolean normalized;
	if (isscaled)
	{	glScaled(scale, scale, scale);
		normalized=glIsEnabled(GL_NORMALIZE);
		glEnable(GL_NORMALIZE);
	}
	*TURTLE=Point(0,0,0);
	TURTLE->reset3d();
	TURTLE->heading=0;

	Turtle *saveRecT=g->recorder->turtle;
	g->recorder->turtle=TURTLE;
	g->recorder->screen=graph->screen;
	CommandListIterator i(g->recorder);
	i=g->recorder->begin();
	while (i)
	{	i->play(*g->recorder);
		++i;
	}
	g->recorder->turtle=saveRecT;
	if (!normalized)
		glDisable(GL_NORMALIZE);
	glPopMatrix();
	*TURTLE=save;
	TURTLE->draw();

	RECORDER.recording=rr;
	if (RECORDER.recording)
		RECORDER << new R_Graphic(*TURTLE, g->recorder, scale);
	return Unbound;
}

NodeP lcastShadows(PairPP)
{
	ShadowObject *so=graph->shadowObject;
	for (int i=0;  i<so->nPlanes;  i++)	// Loop Through All Object Planes
		so->calcPlane(&(so->planes[i]));	// Compute Plane Equations For All Faces

//	so->setConnectivity();	// O(n^2), so don't use it

	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	so->castShadows(TURTLE->lightPos, 1);

	glDisable(GL_CULL_FACE);
	return Unbound;
}

NodeP lclearShadows(PairPP)
{
	graph->shadowObject->clear();
	return Unbound;
}

