#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/frame.h>
#include <wx/glcanvas.h>
#include <wx/listctrl.h>
#include <wx/splitter.h>
#include <wx/html/helpctrl.h>
#include <wx/slider.h>
#include <wx/gauge.h>
#include <wx/button.h>
#include <wx/tglbtn.h>
#include <wx/radiobut.h>
#include <wx/dirdlg.h>
#include <wx/filedlg.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/listbox.h>
#include <wx/sizer.h>
#include <wx/dir.h>
#include <wx/wxMidi.h>

#include <GL/glu.h>

#include <math.h>

#include <gl2ps.h>
#include <m_apm.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Primitiv.h"
#include "Object.h"
#include "Table.h"
#include "Line.h"
#include "Tree.h"
#include "Int.h"
#include "BigFloat.h"
#include "UInt8.h"
#include "DL.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Frame.h"
#include "BoxSizer.h"

#include "Coms.h"
#include "Error.h"
#include "Eval.h"
#include "Files.h"
#include "Graphics.h"
#include "Init.h"
#include "Lists.h"
#include "Math_.h"
#include "Mem.h"
#include "Parse.h"
#include "Print.h"
#include "Term.h"
#include "Workspce.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Mouse.h"
#include "Font.h"
#include "Turtle.h"
#include "TurtleNode.h"

#include "Event.h"
#include "io.h"
#include "Dialogs.h"
#include "Console.h"
#include "Graph.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Midi.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#pragma optimize("",off)
#endif

#define wxCONSTANTS(x)											\
	x(SIMPLE_BORDER) x(DOUBLE_BORDER)						\
	x(SUNKEN_BORDER) x(RAISED_BORDER)						\
	x(STATIC_BORDER) x(NO_BORDER)								\
	x(TRANSPARENT_WINDOW) x(TAB_TRAVERSAL)					\
	x(WANTS_CHARS) x(NO_FULL_REPAINT_ON_RESIZE)			\
	x(VSCROLL) x(HSCROLL) x(ALWAYS_SHOW_SB)				\
	x(CLIP_CHILDREN) x(FULL_REPAINT_ON_RESIZE)			\
																		\
	x(WS_EX_VALIDATE_RECURSIVELY)								\
	x(WS_EX_BLOCK_EVENTS) x(WS_EX_TRANSIENT)				\
	x(WS_EX_PROCESS_IDLE) x(WS_EX_PROCESS_UI_UPDATES)	\
																		\
	x(DEFAULT_FRAME_STYLE)										\
	x(ICONIZE) x(CAPTION) x(MINIMIZE) x(MINIMIZE_BOX)	\
	x(MAXIMIZE) x(MAXIMIZE_BOX) x(CLOSE_BOX)				\
	x(STAY_ON_TOP) x(SYSTEM_MENU) x(RESIZE_BORDER)		\
	x(FRAME_TOOL_WINDOW) x(FRAME_NO_TASKBAR)				\
	x(FRAME_FLOAT_ON_PARENT)									\
	x(FRAME_EX_CONTEXTHELP)										\
	x(FRAME_SHAPED) x(FRAME_EX_METAL)						\
																		\
	x(DEFAULT_DIALOG_STYLE) x(THICK_FRAME)	x(NO_3D)		\
	x(DIALOG_NO_PARENT) x(DIALOG_EX_CONTEXTHELP)			\
	x(DIALOG_EX_METAL)											\
																		\
	x(CHANGE_DIR)													\
	x(DD_DEFAULT_STYLE)											\
	x(DD_NEW_DIR_BUTTON)											\
																		\
	x(OPEN) x(SAVE) x(OVERWRITE_PROMPT)						\
	x(FILE_MUST_EXIST) x(MULTIPLE)							\
																		\
	x(CENTRE) x(YES) x(OK) x(NO) x(YES_NO)					\
	x(CANCEL) x(NO_DEFAULT)										\
	x(ICON_EXCLAMATION) x(ICON_HAND) x(ICON_WARNING)	\
	x(ICON_ERROR) x(ICON_QUESTION)							\
	x(ICON_INFORMATION) x(ICON_STOP)							\
	x(ICON_ASTERISK) x(ICON_MASK)								\
	x(FORWARD) x(BACKWARD) x(RESET) x(HELP) x(MORE)		\
	x(SETUP)															\
																		\
	x(SL_HORIZONTAL) x(SL_VERTICAL)							\
	x(SL_TICKS) x(SL_LABELS) x(SL_LEFT) x(SL_TOP)		\
	x(SL_RIGHT) x(SL_BOTTOM) x(SL_BOTH)						\
	x(SL_SELRANGE) x(SL_INVERSE)								\
																		\
	x(BU_LEFT) x(BU_TOP) x(BU_RIGHT) x(BU_BOTTOM)		\
	x(BU_EXACTFIT)													\
																		\
	x(ID_ADD) x(ID_APPLY) x(ID_BOLD) x(ID_CANCEL)		\
   x(ID_CLEAR) x(ID_CLOSE) x(ID_COPY) x(ID_CUT)			\
	x(ID_DELETE) x(ID_FIND) x(ID_REPLACE)					\
	x(ID_BACKWARD) x(ID_DOWN) x(ID_FORWARD) x(ID_UP)	\
	x(ID_HELP) x(ID_HOME) x(ID_INDENT) x(ID_INDEX)		\
	x(ID_ITALIC) x(ID_JUSTIFY_CENTER)						\
	x(ID_JUSTIFY_FILL) x(ID_JUSTIFY_LEFT)					\
	x(ID_JUSTIFY_RIGHT) x(ID_NEW) x(ID_NO) x(ID_OK)		\
	x(ID_OPEN) x(ID_PASTE) x(ID_PREFERENCES)				\
	x(ID_PRINT) x(ID_PREVIEW) x(ID_PROPERTIES)			\
	x(ID_EXIT) x(ID_REDO) x(ID_REFRESH) x(ID_REMOVE)	\
	x(ID_REVERT_TO_SAVED) x(ID_SAVE) x(ID_SAVEAS)		\
	x(ID_STOP) x(ID_UNDELETE) x(ID_UNDERLINE)				\
	x(ID_UNDO) x(ID_UNINDENT) x(ID_YES) x(ID_ZOOM_100)	\
	x(ID_ZOOM_FIT) x(ID_ZOOM_IN) x(ID_ZOOM_OUT)			\
																		\
	x(CHK_2STATE) x(CHK_3STATE)								\
	x(CHK_ALLOW_3RD_STATE_FOR_USER) x(ALIGN_RIGHT)		\
																		\
	x(RB_GROUP) x(RB_SINGLE) x(RB_USE_CHECKBOX)			\
																		\
	x(GA_HORIZONTAL) x(GA_VERTICAL) x(GA_SMOOTH)			\
																		\
	x(SP_ARROW_KEYS) x(SP_WRAP)								\
																		\
	x(TE_PROCESS_ENTER) x(TE_PROCESS_TAB)					\
	x(TE_MULTILINE) x(TE_PASSWORD) x(TE_READONLY)		\
	x(TE_RICH) x(TE_RICH2) x(TE_AUTO_URL)					\
	x(TE_NOHIDESEL) x(TE_LEFT) x(TE_CENTRE)				\
	x(TE_RIGHT) x(TE_DONTWRAP) x(TE_CHARWRAP)				\
	x(TE_WORDWRAP) x(TE_BESTWRAP) x(TE_CAPITALIZE)		\
																		\
	x(ALIGN_LEFT) x(ALIGN_CENTRE) x(ST_NO_AUTORESIZE)	\
	x(FONTSTYLE_NORMAL) x(FONTSTYLE_SLANT)					\
	x(FONTSTYLE_ITALIC)											\
	x(FONTWEIGHT_NORMAL) x(FONTWEIGHT_LIGHT)				\
	x(FONTWEIGHT_BOLD)											\
																		\
	x(TOP) x(BOTTOM) x(LEFT) x(RIGHT) x(ALL) x(EXPAND)	\
	x(SHAPED) x(FIXED_MINSIZE)									\
	x(ALIGN_CENTER) x(ALIGN_TOP) x(ALIGN_BOTTOM)			\
	x(ALIGN_CENTER_VERTICAL) x(ALIGN_CENTER_HORIZONTAL)\
																		\
	x(HORIZONTAL) x(VERTICAL)									\
																		\
	x(LB_SINGLE) x(LB_MULTIPLE) x(LB_EXTENDED)			\
	x(LB_HSCROLL) x(LB_ALWAYS_SB) x(LB_NEEDED_SB)		\
	x(LB_SORT)														\
																		\
	x(CB_SIMPLE) x(CB_DROPDOWN) x(CB_READONLY)			\
	x(CB_SORT)	x(NOT_FOUND)									\
																		\
	x(LC_LIST) x(LC_REPORT) x(LC_VIRTUAL) x(LC_ICON)	\
	x(LC_SMALL_ICON) x(LC_ALIGN_TOP) x(LC_ALIGN_LEFT)	\
	x(LC_AUTOARRANGE) x(LC_EDIT_LABELS) x(LC_NO_HEADER)\
	x(LC_SINGLE_SEL)												\
	x(LC_SORT_ASCENDING) x(LC_SORT_DESCENDING)			\
	x(LC_HRULES) x(LC_VRULES)									\
	x(LIST_FORMAT_LEFT) x(LIST_FORMAT_RIGHT)				\
	x(LIST_FORMAT_CENTRE)										\
																		\
	x(DIR_FILES) x(DIR_DIRS) x(DIR_HIDDEN)					\
	x(DIR_DOTDOT) x(DIR_DEFAULT)								\
																		\
	x(FULLSCREEN_NOMENUBAR)	x(FULLSCREEN_NOTOOLBAR)		\
	x(FULLSCREEN_NOSTATUSBAR) x(FULLSCREEN_NOBORDER)	\
	x(FULLSCREEN_NOCAPTION) x(FULLSCREEN_ALL)				\


#define Keycodes(x) \
	x(BACK		) \
	x(TAB			) \
	x(RETURN		) \
	x(ESCAPE		) \
	x(SPACE		) \
	x(DELETE		) \
	x(START		) \
	x(LBUTTON	) \
	x(RBUTTON	) \
	x(CANCEL		) \
	x(MBUTTON	) \
	x(CLEAR		) \
	x(SHIFT		) \
	x(ALT			) \
	x(CONTROL	) \
	x(MENU		) \
	x(PAUSE		) \
	x(CAPITAL	) \
	x(PRIOR		) \
	x(NEXT		) \
	x(END			) \
	x(HOME		) \
	x(LEFT		) \
	x(UP			) \
	x(RIGHT		) \
	x(DOWN		) \
	x(SELECT		) \
	x(PRINT		) \
	x(EXECUTE	) \
	x(SNAPSHOT	) \
	x(INSERT		) \
	x(HELP		) \
	x(NUMPAD0	) \
	x(NUMPAD1	) \
	x(NUMPAD2	) \
	x(NUMPAD3	) \
	x(NUMPAD4	) \
	x(NUMPAD5	) \
	x(NUMPAD6	) \
	x(NUMPAD7	) \
	x(NUMPAD8	) \
	x(NUMPAD9	) \
	x(MULTIPLY	) \
	x(ADD			) \
	x(SEPARATOR	) \
	x(SUBTRACT	) \
	x(DECIMAL	) \
	x(DIVIDE		) \
	x(F1			) \
	x(F2			) \
	x(F3			) \
	x(F4			) \
	x(F5			) \
	x(F6			) \
	x(F7			) \
	x(F8			) \
	x(F9			) \
	x(F10			) \
	x(F11			) \
	x(F12			) \
	x(F13			) \
	x(F14			) \
	x(F15			) \
	x(F16			) \
	x(F17			) \
	x(F18			) \
	x(F19			) \
	x(F20			) \
	x(F21			) \
	x(F22			) \
	x(F23			) \
	x(F24			) \
	x(NUMLOCK	) \
	x(SCROLL		) \
	x(PAGEUP		) \
	x(PAGEDOWN	) \
					  \
	x(NUMPAD_SPACE		) \
	x(NUMPAD_TAB		) \
	x(NUMPAD_ENTER		) \
	x(NUMPAD_F1			) \
	x(NUMPAD_F2			) \
	x(NUMPAD_F3			) \
	x(NUMPAD_F4			) \
	x(NUMPAD_HOME		) \
	x(NUMPAD_LEFT		) \
	x(NUMPAD_UP			) \
	x(NUMPAD_RIGHT		) \
	x(NUMPAD_DOWN		) \
	x(NUMPAD_PRIOR		) \
	x(NUMPAD_PAGEUP	) \
	x(NUMPAD_NEXT		) \
	x(NUMPAD_PAGEDOWN	) \
	x(NUMPAD_END		) \
	x(NUMPAD_BEGIN		) \
	x(NUMPAD_INSERT	) \
	x(NUMPAD_DELETE	) \
	x(NUMPAD_EQUAL		) \
	x(NUMPAD_MULTIPLY	) \
	x(NUMPAD_ADD		) \
	x(NUMPAD_SEPARATOR) \
	x(NUMPAD_SUBTRACT	) \
	x(NUMPAD_DECIMAL	) \
	x(NUMPAD_DIVIDE	) \
							  \
	x(WINDOWS_LEFT		) \
	x(WINDOWS_RIGHT	) \
	x(WINDOWS_MENU		) \
	x(COMMAND			) \

#define GL_CONSTANTS(x) \
	x(GL_LINEAR) x(GL_EXP) x(GL_EXP2)

#define GL2PS_CONSTANTS(x)	\
	x(NO_SORT) x(SIMPLE_SORT) x(BSP_SORT)					\
	x(NONE) x(DRAW_BACKGROUND) x(SIMPLE_LINE_OFFSET)	\
	x(SILENT) x(BEST_ROOT) x(NO_TEXT) x(NO_PIXMAP)		\
	x(LANDSCAPE) x(NO_PS3_SHADING) x(NO_BLENDING)		\
	x(OCCLUSION_CULL) x(USE_CURRENT_VIEWPORT)				\
	x(TIGHT_BOUNDING_BOX) x(COMPRESS)						\

#define define(x)	_##x=0,
#define defineType(x) type_##x=0,
#define definewx(x)	_wx##x=0,
#define defineWXK_(x)	_WXK_##x=0,
#define defineGL2PS_(x)	_GL2PS_##x=0,

ObjectP 
	Permanent_Objects(define) 
	Permanent_Objects2(define)
	Node_Type_Strings(defineType, UNUSED)
	wxCONSTANTS(definewx) 
	Keycodes(defineWXK_)
	GL2PS_CONSTANTS(defineGL2PS_)
	GL_CONSTANTS(define)
	_=0;

PairP _Not_Enough_node=0;

//extern Function lcallCode;

#define Primitives_prefix__no_arg(x) \
	x( allFullScreen			) \
	x( allOpen 					) \
	x( beginBusyCursor		) \
	x( boldTextMode			) \
	x( break						) \
	x( buried 					) \
	x( bye						) \
	x( Calls						) \
	x( castShadows				) \
	x( clean						) \
	x( clearScreen				) \
	x( clearShadows			) \
	x( clearText	 			) \
/*	x( callCode					)*/ \
	x( ConsoleSetFocus		) \
	x( Contents					) \
	x( continueLoop			) \
	x( Cursor 					) \
	x( deleteTextures			) \
	x(	disableCylinderLines	) \
	x( disableDepthTest		) \
	x( disableDither			) \
	x( disableFog				) \
	x( disableLighting		) \
	x( disableLineSmooth		) \
	x( disablePointSmooth	) \
	x( disablePolySmooth		) \
	x( disableRoundLineEnds	) \
	x( disableShadows			) \
	x(	disableTextMouseEvents) \
	x( disableTexture			) \
	x( dispatchMessages		) \
	x( doubleBuffer			) \
	x( Dribbler					) \
	x(	enableCylinderLines	) \
	x( enableDepthTest		) \
	x( enableDither			) \
	x( enableFog				) \
	x( enableLighting			) \
	x( enableLineSmooth		) \
	x( enablePointSmooth		) \
	x( enablePolySmooth		) \
	x( enableRoundLineEnds	) \
	x( enableShadows			) \
	x(	enableTextMouseEvents) \
	x( enableTexture			) \
	x( endBusyCursor			) \
	x( eraseAll					) \
	x( eraseNames				) \
	x( eraseProcedures		) \
	x( erasePropertyLists	) \
	x( error 					) \
	x( Fence 					) \
	x( FloodColor	 			) \
	x( fullScreen	 			) \
	x( genSym		 			) \
	x( getColorDatabase		) \
	x( CharUnderCursor		) \
	x( getWorkingDirectory	) \
	x( GraphCurrent			) \
	x( GraphicEnd				) \
	x( GraphicStart			) \
	x( Heading 					) \
	x( hideTurtle 				) \
	x( Home 						) \
	x( IdentityMatrix			) \
	x( insertMode				) \
	x( italicTextMode			) \
	x( KeyboardValue			) \
	x( LabelAlign				) \
	x( LabelFont				) \
	x( LabelWeight				) \
	x( ListControlColumn		) \
	x( ListControlRow			) \
	x( ListControlText		) \
	x( LogoComspec				) \
	x( LogoConstants			) \
	x( LogoEditor				) \
	x( LogoHelpDir				) \
	x( LogoInstance			) \
	x( LogoLibDir				) \
	x( LogoTempDir				) \
	x( LogoVersion				) \
	x( Matrix					) \
	x( MidiAllSoundsOff		) \
	x( MidiClose				) \
	x( MidiCountDevices		) \
	x( MidiOutStreamsFinished) \
	x( MidiOutStreamsStart	) \
	x( MidiOutStreamsStop	) \
	x( MIPS						) \
	x( MousePos					) \
	x( MouseButtons			) \
	x( MouseX					) \
	x( MouseY					) \
	x( MouseZ					) \
	x( Names 					) \
	x( newTurtle				) \
	x( Nodes 					) \
	x( noDribble	 			) \
	x( noRefresh	 			) \
	x( notFullScreen			) \
	x( Orientation				) \
	x( OSScreenSize			) \
	x( overwriteMode			) \
	x( pause 					) \
	x( PenColor					) \
	x( PenDown 					) \
	x( PenErase					) \
	x( PenMode 					) \
	x( PenPaint					) \
	x( PenPattern	 			) \
	x( PenReverse	 			) \
	x( PenSize 					) \
	x( PenUp 					) \
	x( perspective				) \
	x( pitch						) \
	x( Pixel						) \
	x( plainTextMode			) \
	x( PolyStart				) \
	x( PolyEnd					) \
	x( popMatrix				) \
	x( Pos						) \
	x( PosXYZ					) \
	x( Primitives				) \
	x( Procedures				) \
	x( PropertyLists			) \
	x( pushMatrix				) \
	x( readChar					) \
	x( readCharExt				) \
	x( readComplexBin			) \
	x( Reader					) \
	x( ReaderPos	 			) \
	x( readIntBin				) \
	x( readInt16Bin			) \
	x( readInt8Bin				) \
	x( readFloatBin			) \
	x( readList					) \
	x( readUInt8Bin			) \
	x( readRawLine				) \
	x( readWord					) \
	x( redraw 					) \
	x( refresh 					) \
	x( repeatCount				) \
	x( rnd						) \
	x( roll						) \
	x( SaveSize					) \
	x( ScreenColor				) \
	x( ScreenRange				) \
	x( Scrunch 					) \
	x( setIdentityMatrix		) \
	x( setTurtleMatrix		) \
	x( ShadowColor				) \
	x( showTurtle 				) \
	x( shrinkStacks			) \
	x( singleBuffer			) \
	x( SoundsStop				) \
	x( SoundsFinished			) \
	x( stepped 					) \
	x( SurfaceColumn			) \
	x( SurfaceEnd				) \
	x( SurfaceStart			) \
	x( TessContour				) \
	x( TessEnd					) \
	x( TessStart				) \
	x( TextMousePos			) \
	x( TextMouseX				) \
	x( TextMouseY				) \
	x( TextScreen 				) \
	x( Time						) \
	x( TimeFine 				) \
	x( TimeMilli 				) \
	x( TimerFreq 				) \
	x( Timeu 					) \
	x( TimeuRes 				) \
	x( TonesStop				) \
	x( TonesFinished			) \
	x( traced 					) \
	x( Turtle					) \
	x( TurtleMatrix			) \
	x( underlineTextMode		) \
	x( unPerspective			) \
	x( updateGraph				) \
	x( updateVars				) \
	x( updateCalls				) \
	x( VideoEnd					) \
	x( VideoFrame				) \
	x( Window 					) \
	x( WordUnderCursor		) \
	x( wrap						) \
	x( Writer					) \
	x( WriterPos				) \
	x( xCor						) \
	x( yCor						) \
	x( zCor						) \

#define Primitives_prefix__no_arg_alias(x) \
	x( cS			, clearScreen				) \
	x( cT			, clearText					) \
	x(	disCL		, disableCylinderLines	) \
	x(	disDT		, disableDepthTest		) \
	x( disLS		, disableLineSmooth		) \
	x( disPS		, disablePolySmooth		) \
	x( disRLE	, disableRoundLineEnds	) \
	x( disTex	, disableTexture			) \
	x(	enCL		, enableCylinderLines	) \
	x(	enDT		, enableDepthTest			) \
	x( enLS		, enableLineSmooth		) \
	x( enPS		, enablePolySmooth		) \
	x( enRLE		, enableRoundLineEnds	) \
	x( enTex		, enableTexture			) \
	x( erAll		, eraseAll					) \
	x( erNs		, eraseNames				) \
	x( erPLs		, erasePropertyLists		) \
	x( erPs		, eraseProcedures			) \
	x( fS			, fullScreen				) \
	x( hT			, hideTurtle				) \
	x( IDM		, IdentityMatrix			) \
	x( PC			, PenColor					) \
	x( PD			, PenDown					) \
	x( PE			, PenErase					) \
	x( PLists	, PropertyLists			) \
	x( PPt		, PenPaint					) \
	x( PU			, PenUp						) \
	x( PX			, PenReverse				) \
	x( rC			, readChar					) \
	x( rCE		, readCharExt				) \
	x( repCount	, repeatCount				) \
	x( rL			, readList					) \
	x( rW			, readWord					) \
	x( SC			, ScreenColor				) \
	x( setIDM	, setIdentityMatrix		) \
	x( setTM		, setTurtleMatrix			) \
	x( sT			, showTurtle				) \
	x( TM			, TurtleMatrix				) \
	x( TS			, TextScreen				) \
	

#define Primitives_prefix_one_arg(x) \
	x( _setPos				) \
	x( _setPosXYZ			) \
	x(	ArcSin				) \
	x(	ArcCos				) \
	x(	Arity					) \
	x( ASCII					) \
	x( back					) \
	x( BackslashEncode	) \
	x( BigFloatSetPrecision) \
	x( BitMaxX				) \
	x( BitMaxY				) \
	x( BitNot				) \
	x( BitPaste				) \
	x( BoxSizer				) \
	x( bury					) \
	x( butFirst				) \
	x( butFirsts			) \
	x( butLast				) \
	x( ButtonDestroy 		) \
	x( Char					) \
	x( changeDir			) \
	x( CheckBoxDestroy	) \
	x( CheckBoxValue		) \
	x( ChoiceBoxCount		) \
	x( ChoiceBoxDestroy	) \
	x( Circle				) \
	x( close					) \
	x( ComboBoxCount		) \
	x( ComboBoxDestroy	) \
	x( conjugate			) \
	x( Cos					) \
	x( count					) \
	x( deepCopy				) \
	x( Distance				) \
	x( DistanceXYZ			) \
	x( dribble				) \
	x( downPitch			) \
	x( DynamicLibrary		) \
	x( editFile				) \
	x( erase					) \
	x( eraseFile			) \
	x( exp					) \
	x( factorize			) \
	x( Faculty				) \
	x( FileSize				) \
	x( FileTime				) \
	x( fillCircle			) \
	x( first					) \
	x( firsts				) \
	x( forward				) \
	x( FrameDestroy 		) \
	x( FrameSetFocus		) \
	x( fullText				) \
	x( GaugeDestroy 		) \
	x( GaugeValue	 		) \
	x( GraphDestroy 		) \
	x( GraphSetCurrent	) \
	x( ignore				) \
	x( imag					) \
	x( IntControlDestroy	) \
	x( invertMatrix		) \
	x( Label					) \
	x( last					) \
	x( left					) \
	x( leftRoll				) \
	x( ListBoxCount		) \
	x( ListBoxDestroy		) \
	x( ListControlColumnCount	) \
	x( ListControlDestroy		) \
	x( ListControlGet				) \
	x( ListControlItemCount		) \
	x( LN						) \
	x( load					) \
	x( loadImage			) \
	x( loadNodes			) \
	x( loadPicture			) \
	x( loadPictureText	) \
	x( Log10					) \
	x( lowerCase			) \
	x( makeDirectory		) \
	x( mergePairs			) \
	x( mergeSort			) \
	x( MidiDeviceInfo		) \
	x( MidiMessage			) \
	x( not					) \
	x( OnChar				) \
	x( OnKeyDown			) \
	x( OnKeyUp				) \
	x(	OnMouseLeftDown		) \
	x(	OnMouseRightDown		) \
	x(	OnMouseMiddleDown		) \
	x(	OnMouseLeftUp			) \
	x(	OnMouseRightUp			) \
	x(	OnMouseMiddleUp		) \
	x(	OnMouseLeftDClick		) \
	x( OnMouseRightDClick	) \
	x( OnMouseMiddleDClick	) \
	x( OnMouseMotion			) \
	x(	OnTextMouseLeftDown		) \
	x(	OnTextMouseRightDown		) \
	x(	OnTextMouseMiddleDown	) \
	x(	OnTextMouseLeftUp			) \
	x(	OnTextMouseRightUp		) \
	x(	OnTextMouseMiddleUp		) \
	x(	OnTextMouseLeftDClick	) \
	x( OnTextMouseRightDClick	) \
	x( OnTextMouseMiddleDClick	) \
	x( OnTextMouseMotion			) \
	x( openAppend			) \
	x( openAppendBin		) \
	x( openRead				) \
	x( openReadBin			) \
	x( openUpdate			) \
	x( openUpdateBin		) \
	x( openWrite			) \
	x( openWriteBin		) \
	x( parse					) \
	x( pick					) \
	x( playWaveFast		) \
	x( pop					) \
	x( PortIn				) \
	x( printOut				) \
	x( printOutTitles		) \
	x( PropertyList		) \
	x( radCos				) \
	x( RadioButtonDestroy) \
	x( RadioButtonValue	) \
	x( radSin				) \
	x( radTan				) \
	x(	radArcSin			) \
	x(	radArcCos			) \
	x( random				) \
	x( rawASCII				) \
	x( readChars			) \
	x( readFloatArrayBin	) \
	x( readIntArrayBin	) \
	x( readInt16ArrayBin	) \
	x( readStructBin		) \
	x( readWordBin			) \
	x(	real					) \
	x( reHSB					) \
	x( reHSBA				) \
	x(	remdup				) \
	x( reRGB					) \
	x( reRGBA				) \
	x( reverse				) \
	x( right					) \
	x( rightRoll			) \
	x( round					) \
	x( runParse				) \
	x( save					) \
	x( saveNodes			) \
	x( savePicture			) \
	x( savePictureText	) \
	x( savePostScript		) \
	x( saveScreen			) \
	x(	scrollCalibrate	) \
	x( setCallsSplitter	) \
	x( setCaseIgnored		) \
	x( setCursor			) \
	x( setDepthFunc		) \
	x( setFloodColor		) \
	x( setFogColor			) \
	x( setFogDensity		) \
	x( setFogMode			) \
	x( setHeading			) \
	x( setLabelFont		) \
	x( setLabelSize		) \
	x( setLabelWeight		) \
	x( setLightAmbient	) \
	x( setLightDiffuse	) \
	x( setLightPos			) \
	x( setLightSpecular	) \
	x( setLogoComspec		) \
	x( setLogoEditor		) \
	x( setLogoHelpDir		) \
	x( setLogoLibDir		) \
	x( setLogoTempDir		) \
	x( setMargins			) \
	x( setMaterialAmbient) \
	x( setMaterialDiffuse) \
	x( setMaterialEmission) \
	x( setMaterialShininess) \
	x( setMaterialSpecular) \
	x( setMatrix			) \
	x( setOrientation		) \
	x( setPenColor			) \
	x( setPenPattern		) \
	x( setPenSize			) \
	x( setPitch				) \
	x( setPos				) \
	x( setPosXYZ			) \
	x( setPrintPrecision	) \
	x( setReader			) \
	x( setReaderPos		) \
	x( setRoll				) \
	x( setSaveSize			) \
	x( setScreenColor		) \
	x( setShadowColor		) \
	x( setStackNoisy		) \
	x( setTessWindingRule) \
	x( setTexPos			) \
	x( setTexWrapping		) \
	x( setTextFont			) \
	x( setTextSize			) \
	x( setTurtle			) \
	x( setUpdateGraph		) \
	x( setVarsSplitter	) \
	x( setWriter			) \
	x( setWriterPos		) \
	x( setX					) \
	x( setY					) \
	x( setZ					) \
	x( shuffle				) \
	x( Signum				) \
	x( Sin					) \
	x( SizeOf				) \
	x( SliderDestroy		) \
	x( SliderValue			) \
	x( spinX					) \
	x( spinY					) \
	x( spinZ					) \
	x( Sqr					) \
	x( Sqrt					) \
	x( standout				) \
	x( StaticTextDestroy	) \
	x( StaticTextLabel	) \
	x( StaticTextSetLabel) \
	x( step					) \
	x( StringBuffer		) \
	x( StringBufferToWord) \
	x( Struct				) \
	x( Table					) \
	x( tag					) \
	x( Tan					) \
	x( test					) \
	x( Text					) \
	x( TextControlCursor	) \
	x( TextControlDestroy) \
	x( TextControlInsertMode) \
	x( TextControlOverwriteMode) \
	x( TextControlSetInsertionPointEnd) \
	x( Texture				) \
	x( Thing					) \
	x( ToggleButtonDestroy	) \
	x( ToggleButtonValue	) \
	x( toList				) \
	x( toListList			) \
	x( towards				) \
	x( towardsXYZ			) \
	x( trace					) \
	x( truncate				) \
	x( TypeOf				) \
	x( unBury				) \
	x( unStep				) \
	x( unTrace				) \
	x( updateVarsOnStep	) \
	x( upperCase			) \
	x( upPitch				) \
	x( wait					) \
	x( waitMS				) \
	x( waituS				) \

#define Primitives_prefix_one_arg_alias(x) \
	x( bk			, back				) \
	x( bF			, butFirst			) \
	x( bFs		, butFirsts			) \
	x( bL			, butLast			) \
	x( cd			, changeDir			) \
	x( dequeue	, pop					) \
	x( down		, downPitch			) \
	x( er			, erase				) \
	x( erF		, eraseFile			) \
	x( fd			, forward			) \
	x( Float		, _float				) \
	x( BigFloat	, _bigfloat			) \
	x( Int		, _int				) \
	x( Int16		, toint16			) \
	x( Int8		, toint8				) \
	x( UInt8		, touint8			) \
	x( loadPic	, loadPicture		) \
	x( lt			, left				) \
	x( lR			, leftRoll			) \
	x( minus		, sub					) \
	x( PList		, PropertyList		) \
	x( pO			, printOut			) \
	x( pOT		, printOutTitles	) \
	x( rt			, right				) \
	x( rr			, rightRoll			) \
	x( rCs		, readChars			) \
	x( savePic	, savePicture		) \
	x( savePS	, savePostScript	) \
	x(	scrollCal, scrollCalibrate	) \
	x( setFC		, setFloodColor	) \
	x( setH		, setHeading		) \
	x( setM		, setMatrix			) \
	x( setPC		, setPenColor		) \
	x( setPS		, setPenSize		) \
	x( setSC		, setScreenColor	) \
	x(	trunc		, truncate			) \
	x(	uP			, upPitch			) \
	

#define Primitives_prefix_two_arg(x) \
	x( addColors			) \
	x( addColorsMod		) \
	x( Arc					) \
	x( ArcTan				) \
	x( aShift				) \
	x( BitCopy				) \
	x( BitItem				) \
	x( BitMakeTransparent) \
	x( ButtonEnable		) \
	x( ButtonOnClick		) \
	x( CheckBoxEnable		) \
	x( CheckBoxOnClick	) \
	x( CheckBoxSet			) \
	x( ChoiceBoxAppend		) \
	x( ChoiceBoxEnable		) \
	x( ChoiceBoxOnChar		) \
	x( ChoiceBoxOnKeyDown	) \
	x( ChoiceBoxOnKeyUp		) \
	x( ChoiceBoxOnSelect		) \
	x( ChoiceBoxRemoveItem	) \
	x( ChoiceBoxSetBackgroundColor) \
	x( ChoiceBoxSetColor		) \
	x( ChoiceBoxSetChoices	) \
	x( ChoiceBoxSetFontName	) \
	x( ChoiceBoxSetFontSize	) \
	x( ChoiceBoxSetFontStyle) \
	x( ChoiceBoxSetFontWeight) \
	x( ChoiceBoxSetSelection) \
	x( clearPortBit			) \
	x( ComboBoxAppend			) \
	x( ComboBoxEnable			) \
	x( ComboBoxOnChange		) \
	x( ComboBoxOnChar			) \
	x( ComboBoxOnEnter		) \
	x( ComboBoxOnSelect		) \
	x( ComboBoxOnKeyDown		) \
	x( ComboBoxOnKeyUp		) \
	x( ComboBoxOnSelect		) \
	x( ComboBoxRemoveItem	) \
	x( ComboBoxSetBackgroundColor) \
	x( ComboBoxSetColor		) \
	x( ComboBoxSetChoices	) \
	x( ComboBoxSetFontName	) \
	x( ComboBoxSetFontSize	) \
	x( ComboBoxSetFontStyle	) \
	x( ComboBoxSetFontWeight) \
	x( ComboBoxSetSelection	) \
	x( ComboBoxSetValue		) \
	x( copyDef				) \
	x( Cross					) \
	x( define				) \
	x( destructiveRemove	) \
	x( Ellipse				) \
	x( fillEllipse			) \
	x( fillPie				) \
	x( fillRect				) \
	x( FloatControlEnable	) \
	x( FloatControlSetValue	) \
	x( FloatControlOnChange	) \
	x( fPut					) \
	x( FrameEnable			) \
	x( FrameIconize		) \
	x( FrameMaximize		) \
	x( FrameOnChar			) \
	x( FrameOnKeyDown		) \
	x( FrameOnKeyUp		) \
	x( FrameSetBackgroundColor) \
	x( FrameSetColor		) \
	x( FrameSetFontName	) \
	x( FrameSetFontSize	) \
	x( FrameSetFontStyle	) \
	x( FrameSetFontWeight) \
	x( FrameSetShape		) \
	x( FrameSetSizer		) \
	x( GaugeSetBackgroundColor) \
	x( GaugeSetColor		) \
	x( GaugeSetRange 		) \
	x( GaugeSetValue 		) \
	x( GraphOnChar			) \
	x( GraphOnKeyDown		) \
	x( GraphOnKeyUp		) \
	x(	GraphOnMouseLeftDown		) \
	x(	GraphOnMouseRightDown	) \
	x(	GraphOnMouseMiddleDown	) \
	x(	GraphOnMouseLeftUp		) \
	x(	GraphOnMouseRightUp		) \
	x(	GraphOnMouseMiddleUp		) \
	x(	GraphOnMouseLeftDClick	) \
	x( GraphOnMouseRightDClick	) \
	x( GraphOnMouseMiddleDClick) \
	x( GraphOnMouseMotion		) \
	x( gcd					) \
	x( getPortBit			) \
	x( getProperty			) \
	x( GraphOnChar			) \
	x( GraphOnKeyDown		) \
	x( GraphOnKeyUp		) \
	x( IntControlEnable	) \
	x( IntControlSetValue) \
	x( IntControlOnChange) \
	x( iSeq					) \
	x( leftPortShift		) \
	x( Line					) \
	x( ListBoxAppend		) \
	x( ListBoxEnable		) \
	x( ListBoxOnChar		) \
	x( ListBoxOnDClick	) \
	x( ListBoxOnKeyDown	) \
	x( ListBoxOnKeyUp		) \
	x( ListBoxOnSelect	) \
	x( ListBoxRemoveItem) \
	x( ListBoxSetBackgroundColor	) \
	x( ListBoxSetColor				) \
	x( ListBoxSetChoices				) \
	x( ListBoxSetFontName			) \
	x( ListBoxSetFontSize			) \
	x( ListBoxSetFontStyle			) \
	x( ListBoxSetFontWeight			) \
	x( ListBoxSetSelections			) \
	x( ListControlEnable				) \
	x( ListControlInsertItem		) \
	x( ListControlDeleteItem		) \
	x( ListControlGetColumn			) \
	x( ListControlGetRow				) \
	x( ListControlOnChar				) \
	x( ListControlOnColClick		) \
	x( ListControlOnItemSelected	) \
	x( ListControlOnItemActivated	) \
	x( ListControlOnKeyDown			) \
	x( ListControlOnKeyUp			) \
	x( ListControlSet					) \
	x( ListControlSetBackgroundColor) \
	x( ListControlSetColor			) \
	x( ListControlSetFontName		) \
	x( ListControlSetFontSize		) \
	x( ListControlSetFontStyle		) \
	x( ListControlSetFontWeight	) \
	x(	ListControlSort	) \
	x( lowPassFilter		) \
	x( lPut					) \
	x( lShift				) \
	x( make					) \
	x( mdItem				) \
	x( Member				) \
	x( merge					) \
	x( MidiProgramChange	) \
	x( mod					) \
	x( notPortBit			) \
	x( playWave				) \
	x( PortOut				) \
	x( Power					) \
	x( push					) \
	x( queue					) \
	x( radArcTan			) \
	x( RadioButtonEnable	) \
	x( RadioButtonOnClick) \
	x( RadioButtonSet		) \
	x( Remainder			) \
	x( remove				) \
	x( removeItem			) \
	x( removeProperty		) \
	x( resize				) \
	x( rightPortShift		) \
	x( rotate				) \
	x(	saturateAbove		) \
	x(	saturateBelow		) \
	x(	saveImage			) \
	x(	scroll				) \
	x( setFogRange			) \
	x( setLabelAlign		) \
	x( setPixel				) \
	x( setPortBit			) \
	x( setScrunch			) \
	x( setTextColor		) \
	x( setTextSelection	) \
	x( setTexXY				) \
	x( setXY					) \
	x( SliderEnable		) \
	x( SliderOnScroll		) \
	x( SliderSetValue		) \
	x( Sound					) \
	x( StaticTextSetLabel) \
	x( StaticTextSetBackgroundColor) \
	x( StaticTextSetColor		) \
	x( StaticTextSetFontName	) \
	x( StaticTextSetFontSize	) \
	x( StaticTextSetFontStyle	) \
	x( StaticTextSetFontWeight	) \
	x( TextControlAppend			) \
	x( TextControlEnable			) \
	x( TextControlOnChange		) \
	x( TextControlOnChar			) \
	x( TextControlOnEnter		) \
	x( TextControlOnKeyDown		) \
	x( TextControlOnKeyUp		) \
	x( TextControlSetBackgroundColor) \
	x( TextControlSetColor		) \
	x( TextControlSetCursor		) \
	x( TextControlSetFontName	) \
	x( TextControlSetFontSize	) \
	x( TextControlSetFontStyle	) \
	x( TextControlSetFontWeight	) \
	x( TextControlSetValue		) \
	x( TextControlWrite			) \
	x( ToggleButtonEnable		) \
	x( ToggleButtonOnClick		) \
	x( ToggleButtonSetValue		) \
	x( Tone					) \
	x( xAdd					) \
	x( xCopy					) \
	x( xDiv					) \
	x( xMod					) \
	x( xMul					) \
	x( xSub					) \

#define Primitives_prefix_two_arg_alias(x) \
	x( BitTrans, BitMakeTransparent) \
	x( Difference, sub			) \
	x( gProp, getProperty		) \
	x( Modulo, mod					) \
	x( remProp, removeProperty	) \
	x( setTC , setTextColor		) \


#define Primitives_prefix_three_arg(x) \
	x( DynamicLibraryCall) \
	x( BitPixel				) \
	x( ChoiceBoxSetItem	) \
	x( ComboBoxSetItem	) \
	x( FloatControlSetRange) \
	x( Form					) \
	x( FrameFullScreen	) \
	x( FrameSetClientSize) \
	x( HSB					) \
	x( IntControlSetRange) \
	x( intForm				) \
	x( Items					) \
	x( ListBoxSetItem		) \
	x( ListControlGetItem) \
	x( ListControlSetColumn		) \
	x( ListControlSetRow	) \
	x( mandelIterate		) \
	x( mdSetItem			) \
	x( MidiNoteOn			) \
	x( MidiNoteOff			) \
	x( putProperty			) \
	x( replace				) \
	x( rSeq					) \
	x( rSeqFloatArray		) \
	x( setItem				) \
	x( setItems				) \
	x( setCylinderPos		) \
	x( setEye				) \
	x( setPixelXY			) \
	x( setSpherePos		) \
	x( setXYZ				) \
	x( SliderSetRange		) \

#define Primitives_prefix_three_arg_alias(x) \
	x( DLCall, DynamicLibraryCall	) \
	x( pProp	, putProperty			) \
	x( rSeqFA, rSeqFloatArray		) \

#define Primitives_prefix_four_arg(x) \
	x( BitSetPixel			) \
	x( EllipseArc			) \
	x( HSBA					) \
	x( ListControlSetItem) \
	x( setScreenRange		) \
	x( setPixelXYZ			) \


#define Primitives_General_Binary(x) \
	x( and			) \
	x( BitAnd		) \
	x( BitOr			) \
	x( BitXOr		) \
	x( List			) \
	x( or				) \
	x( Sentence		) \
	x( Word			) \

#define Primitives_General_Prefix(x) \
	x( Array						,1,1,2) \
	x( BoxSizerAdd				,5,5,5) \
	x( Button					,3,3,8) \
	x( CheckBox					,2,3,8) \
	x( ChoiceBox				,3,4,7) \
	x( ChoiceBoxSelection	,0,0,1) \
	x( ComboBox					,4,7,10) \
	x( ComboBoxSelection		,0,0,1) \
	x( ComboBoxValue			,0,0,1) \
	x( DirSelector				,0,0,3) \
	x( drawGraphic				,1,1,2) \
	x( Ellipsoid				,3,3,5) \
	x( Files						,0,0,2) \
	x( FileSelector			,1,1,7) \
	x( fill						,0,0,1) \
	x( FloatArray				,1,1,2) \
	x( FloatControl			,6,8,12) \
	x( FloatControlValue		,0,0,1) \
	x( Frame						,3,5,6) \
	x( Gauge						,3,3,7) \
	x( GC							,0,0,1) \
	x( getColorFromUser		,0,1,1) \
	x( getFontFromUser		,0,0,1) \
	x(	getMultipleChoices	,3,3,6) \
	x(	getNumberFromUser		,6,6,7) \
	x(	getPasswordFromUser	,3,3,5) \
	x(	getSingleChoice		,3,3,6) \
	x(	getSingleChoiceIndex	,3,3,6) \
	x(	getTextFromUser		,3,3,5) \
	x( Graph						,1,1,5) \
	x( IntArray					,1,1,2) \
	x( Int16Array				,1,1,2) \
	x( IntControl				,5,6,9) \
	x( IntControlValue		,0,0,1) \
	x( LabelSize				,0,1,1) \
	x( ListBox					,3,5,8) \
	x( ListBoxSelections		,0,0,1) \
	x( ListControl				,2,5,8) \
	x( ListControlInsertColumn,3,3,5) \
	x( MessageBox				,1,1,4) \
	x( MidiOpen					,0,1,2) \
	x( MidiOutStream			,2,2,4) \
	x( partialEllipsoid		,9,9,9) \
	x( reRandom					,0,0,1) \
	x( RadioButton				,2,3,8) \
	x( RGB						,1,3,3) \
	x( RGBA						,1,4,4) \
	x( saveScreenVector		,1,1,4) \
	x( Shell						,0,1,2) \
	x( ShellSpawn				,0,1,2) \
	x( Slider					,5,6,9) \
	x( SliderValue				,0,0,1) \
	x( Sounds					,1,1,2) \
	x( Sphere					,1,1,3) \
	x( splitScreen				,0,0,1) \
	x( StaticText				,2,2,6) \
	x( TextControl				,2,3,8) \
	x( TextControlValue		,0,0,1) \
	x( throw						,1,1,2) \
	x( ToggleButton			,3,3,7) \
	x( Tones						,1,1,2) \
	x( VideoStart				,1,1,2) \
\
	x( combine		,1,2,-1) \
	x( print			,0,1,-1) \
	x( show			,0,1,-1) \
	x( type			,0,1,-1) \
	x( typeBin		,0,1,-1) \
\
	x( abs			,1,1,-1) \
	x( local			,1,1,-1) \
	x( maxNorm		,1,1,-1) \
	x( Norm			,1,1,-1) \


#define Primitives_General_Prefix_alias(x) \
	x( Item		, Item			, 2, 2, 3) \
	x( pr			, print			, 0, 1,-1) \
	x( Product	, mul				, 0, 2,-1) \
	x( Quotient	, div				, 1, 2, 2) \
	x( Se			, Sentence		, 0, 2,-1) \
	x( sS			, splitScreen	, 0, 0, 1) \
	x( sum		, add				, 0, 2,-1) \
	x( ?			, questionmark	, 0, 0, 1) \


#define Predicates__no_arg(x) \
	x( CaseIgnored	) \
	x( Eof			) \
	x( BusyCursor	) \
	x( Key			) \
	x( PenDown		) \
	x( Refresh		) \
	x( shown			) \

#define Predicates_one_arg(x) \
	x( AlNum			) \
	x( Alpha			) \
	x( Array			) \
	x( ASCII			) \
	x( backslashed	) \
	x( BigFloat		) \
	x( Char			) \
	x( circular		) \
	x( Cntrl			) \
	x( Complex_		) \
	x( CSym			) \
	x( defined		) \
	x( Digit			) \
	x( Directory	) \
	x( empty			) \
	x( File			) \
	x( Float			) \
	x( FloatArray	) \
	x( Graph			) \
	x( Int			) \
	x( IntArray		) \
	x( Int16			) \
	x( Int16Array	) \
	x( Int8			) \
	x( List			) \
	x( Lower			) \
	x( Macro			) \
	x( Name			) \
	x( Number		) \
	x( Prime			) \
	x( Primitive	) \
	x( Print			) \
	x( Procedure	) \
	x( Punct			) \
	x( Space			) \
	x( Table			) \
	x( UInt8			) \
	x( Upper			) \
	x( Word			) \
	x( xDigit		) \

#define Predicates_two_arg(x) \
	x( before		) \
	x( equal			) \
	x( greater		) \
	x( greaterEqual) \
	x( less			) \
	x( lessEqual	) \
	x( Member		) \
	x( NameInTable	) \
	x( Substring	) \


#define Primitives_prefix_ok(x) \
	x( continue		) \
	x( edit			) \
	x( help			) \

#define Primitives_prefix_ok_alias(x) \
	x( co, continue) \
	x( ed, edit		) \


#define Primitives_prefix_macro_arg(x) \
	x( to				) \

#define Primitives_macro_one_arg(x) \
	x( forever		) \
	x( goTo			) \
	x( ifFalse		) \
	x( ifTrue		) \
	x( run			) \
	x( runResult	) \

#define Primitives_macro_one_arg_alias(x) \
	x( ifF, ifFalse) \
	x( ifT, ifTrue	) \

#define Primitives_macro_two_arg(x) \
	x( apply			) \
	x( case			) \
	x( catch			) \
	x( do_until		) \
	x( do_while		) \
	x( for			) \
	x( repeat		) \
	x( until			) \
	x( while			) \

#define Primitives_macro_three_arg(x) \
	x( ifElse		) \

#define make_prefix(x,n)			{#x	, n,n,n, prefix, l##x	},
#define make_prefix_alias(x,y,n)	{#x	, n,n,n, prefix, l##y	},
#define make_general_binary(x)	{#x	, 0,2,-1,prefix, l##x	},
#define make_macro(x,n)				{#x	, n,n,n, priority_MACRO	, l##x	},
#define make_macro_alias(x,y,n)	{#x	, n,n,n, priority_MACRO	, l##y	},

#define make_prefix_ok(x) \
	{#x, Primitive::ok_if_no_arg,1,1, prefix, l##x	},

#define make_prefix_ok_alias(x,y) \
	{#x, Primitive::ok_if_no_arg,1,1, prefix, l##y	},

#define make_general_prefix(x,min,std,max) \
	{#x	, min,std,max, prefix, l##x	},

#define make_general_prefix_alias(x,y,min,std,max) \
	{#x	, min,std,max, prefix, l##y	},

#define make_predicate(x,n) \
	{#x"P", n,n,n, prefix, lis##x	}, \
	{#x"?", n,n,n, prefix, lis##x	},

#define make_prefix__no_arg(x)	make_prefix(x,0)
#define make_prefix_one_arg(x)	make_prefix(x,1)
#define make_prefix_two_arg(x)	make_prefix(x,2)
#define make_prefix_three_arg(x)	make_prefix(x,3)
#define make_prefix_four_arg(x)	make_prefix(x,4)
#define make_prefix_macro_arg(x)	make_prefix(x,-1)

#define make_prefix__no_arg_alias(x,y)	make_prefix_alias(x,y,0)
#define make_prefix_one_arg_alias(x,y)	make_prefix_alias(x,y,1)
#define make_prefix_two_arg_alias(x,y)	make_prefix_alias(x,y,2)
#define make_prefix_three_arg_alias(x,y)	make_prefix_alias(x,y,3)

#define make_macro_one_arg(x)		make_macro(x,1)
#define make_macro_two_arg(x)		make_macro(x,2)
#define make_macro_three_arg(x)	make_macro(x,3)

#define make_macro_one_arg_alias(x,y)	make_macro_alias(x,y,1)

#define make_predicate__no_arg(x)	make_predicate(x,0)
#define make_predicate_one_arg(x)	make_predicate(x,1)
#define make_predicate_two_arg(x)	make_predicate(x,2)

PrimType Primitive::primitives[]=
{
	Primitives_prefix__no_arg		(make_prefix__no_arg)
	Primitives_prefix_one_arg		(make_prefix_one_arg)
	Primitives_prefix_two_arg		(make_prefix_two_arg)
	Primitives_prefix_three_arg	(make_prefix_three_arg)

	Primitives_prefix_four_arg		(make_prefix_four_arg)

	Primitives_prefix__no_arg_alias(make_prefix__no_arg_alias)
	Primitives_prefix_one_arg_alias(make_prefix_one_arg_alias)
	Primitives_prefix_two_arg_alias(make_prefix_two_arg_alias)
	Primitives_prefix_three_arg_alias(make_prefix_three_arg_alias)

	{"max",0,2,-1,prefix, ::lmax},
	{"min",0,2,-1,prefix, ::lmin},
	Primitives_General_Binary			(make_general_binary)
	Primitives_General_Prefix			(make_general_prefix)
	Primitives_General_Prefix_alias	(make_general_prefix_alias)

	Predicates__no_arg			(make_predicate__no_arg)
	Predicates_one_arg			(make_predicate_one_arg)
	Predicates_two_arg			(make_predicate_two_arg)

	Primitives_prefix_ok			(make_prefix_ok)
	Primitives_prefix_ok_alias	(make_prefix_ok_alias)

	{"--",1,1,1,priority_tight_minus	, lsub		},
	{"+" ,1,1,1,priority_addition		, ladd		},
	{"-" ,1,1,1,priority_addition		, lsub		},
	{"*" ,1,1,1,priority_product		, lmul		},
	{"/" ,1,1,1,priority_product		, ldiv		},
//	{"%" ,1,1,1,priority_product		, lmod		},
	{"^" ,1,1,1,priority_power			, lPower		},

//	{"!" ,1,1,1,priority_faculty		, lFaculty	},
	{"." ,1,1,1,priority_dot			, ldot		},
	{">" ,2,2,2,priority_comparsion	, lisgreater},
	{"<" ,2,2,2,priority_comparsion	, lisless	},
	{">=",2,2,2,priority_comparsion	, lisgreaterEqual	},
	{"<=",2,2,2,priority_comparsion	, lislessEqual		},
	{"==",2,2,2,priority_comparsion	, lisequal	},
	{"!=",2,2,2,priority_comparsion	, lisunequal},
	{ "=",2,2,2,priority_assignment	, lmake		},
	{"+=",2,2,2,priority_assignment	, lxAdd		},
	{"-=",2,2,2,priority_assignment	, lxSub		},
	{"*=",2,2,2,priority_assignment	, lxMul		},
	{"/=",2,2,2,priority_assignment	, lxDiv		},
//	{"%=",2,2,2,priority_assignment	, lxMod		},
	{"or2",0,2,-1,priority_logical	, lor			},
	{"and2",0,2,-1,priority_logical	, land		},

	Primitives_prefix_macro_arg(make_prefix_macro_arg)

	Primitives_macro_one_arg	(make_macro_one_arg)
	Primitives_macro_two_arg	(make_macro_two_arg)
	Primitives_macro_three_arg	(make_macro_three_arg)

	Primitives_macro_one_arg_alias(make_macro_one_arg_alias)

	{"_defMacro"	,2,2,2, prefix, l_defMacro		},
	{"_eq"			,2,2,2, prefix, l_eq				},
	{"_Macro"	,-1,-1,-1, prefix, l_Macro			},
	{"_setBF"		,2,2,2, prefix, l_setButFirst	},
	{"_setButFirst",2,2,2, prefix, l_setButFirst	},
	{"_setFirst"	,2,2,2, prefix, l_setFirst		},
	{"_setItem"		,3,3,3, prefix, l_setItem		},
	{"?Rest"			,0,0,0, prefix, lquestionmarkRest},

	{"if"				,2,2,3, priority_MACRO , lif		},
	{"foreach"		,2,2,-1,priority_MACRO , lforeach},

	{"_maybeOutput",1,1,1, priority_MAYBE , loutput	},
	{"output"		,1,1,1, priority_OUTPUT, loutput	},
	{"op"				,1,1,1, priority_OUTPUT, loutput	},
	{"stop"			,0,0,0, priority_STOP  , lstop	},

	{0,0,0,0,prefix,0}
};

ObjectP permanent(charP name) 
{
	StringP str = new String(name);
	ObjectP result = str->intern();
	result->bit.permanent=true;
	return result;
}

void Primitive::init(bool caseIgnored)
{
	String::caseIgnored=caseIgnored;

	for (int i=0;  primitives[i].name;  i++)
	{
		PrimType &prim=primitives[i];
		PrimitiveP proc=new Primitive(prim);

		StringP name;
		if (caseIgnored)
			name=new String(prim.name, 0, 
				strlen(prim.name), STRING, String::copy_noparitylow);
		else
			name=new String(prim.name, 0, 
				strlen(prim.name), STRING, String::copy_noparity);

		ObjectP obj=root->find(name);
		if (!obj)	// don't insert duplicate primitive names
			obj=root->insert(name);
//		else cout << name << " is double!" << endL;
		
		obj->setprim(proc);

		if (proc->type == MACRO)
			obj->bit.proc_macro=true;

		obj->bit.permanent=true;
	}
}

void init(void) 
{
	   GC_::init();

	_Null_Word = permanent("\0");

#define           CREATE(x)	_##x = permanent(#x);
#define         CREATEwx(x)	_wx##x = permanent("wx"#x);
#define       CREATEWXK_(x)	_WXK_##x = permanent("WXK_"#x);
#define     CREATEGL2PS_(x)	_GL2PS_##x = permanent("GL2PS_"#x);
#define             BURY(x)	_##x->bit.val_buried=true;
#define           BURYwx(x)	_wx##x->bit.val_buried=true;
#define         BURYWXK_(x)	_WXK_##x->bit.val_buried=true;
#define       BURYGL2PS_(x)	_GL2PS_##x->bit.val_buried=true;
#define SET_FILE_POINTER(x)	x=FileP(_##x->var());

	Permanent_Objects(CREATE)

	 File::init();
	 Line::init();

	Primitive::init(true);
	Primitive::init(false);
	String::caseIgnored=true;

	_True        ->setvar(_True  );
	_False       ->setvar(_False);

	     _Zero = permanent( "0");
	      _One = permanent( "1");
	_Minus_One = permanent("-1"); 

	  Int::init();
	Float::init();
	BigFloat::init();

	    Special_Vars(BURY)
	Number_Constants(BURY)

	  _Plus_Sign = permanent("+");
	 _Minus_Sign = permanent("-");
	_Minus_Tight = permanent("--");
	   _Mul_Sign = permanent("*");
	   _Div_Sign = permanent("/");
//	   _Mod_Sign = permanent("%");

	 _Left_Paren = permanent("(");
	_Right_Paren = permanent(")");
	      _Query = permanent("?");
			  _Dot = permanent(".");
	_SingleQuote = permanent("'");
	       _Else = permanent("else");
	      _Word_ = permanent("word");
	       _Void = permanent("void");
	  _IntArray_ = permanent("intarray");
	_Int16Array_ = permanent("int16array");
	_FloatArray_ = permanent("floatarray");

	_Not_Enough_node = new Pair(0,0);
	Tree::the_generation = new Int(0);

#define Make_permanent(x) type_##x = permanent
#define Paren(y) (y);

	Node_Type_Strings(Make_permanent, Paren)
	GC_::initialized=true;

	      wxCONSTANTS(CREATEwx)
			   Keycodes(CREATEWXK_)
		  GL_CONSTANTS(CREATE)
	  GL2PS_CONSTANTS(CREATEGL2PS_)

		  wxCONSTANTS(BURYwx)
			  Keycodes(BURYWXK_)
		 GL_CONSTANTS(BURY)
	 GL2PS_CONSTANTS(BURYGL2PS_)

#define SET_wxCONSTANT(x) _wx##x->setvar(new Int(wx##x));
	wxCONSTANTS(SET_wxCONSTANT)
#define SET_WXK_CONSTANT(x) _WXK_##x->setvar(new Int(WXK_##x));
	Keycodes(SET_WXK_CONSTANT)
#define SET_GL_CONSTANT(x) _##x->setvar(new Int(##x));
	GL_CONSTANTS(SET_GL_CONSTANT)
#define SET_GL2PS_CONSTANT(x) _GL2PS_##x->setvar(new Int(GL2PS_##x));
	GL2PS_CONSTANTS(SET_GL2PS_CONSTANT)

	LoadIODLL();

	theApp->mainFrame->graphNode=new GraphNode();
	theApp->mainFrame->graphNode->graph=graph;
	theApp->mainFrame->graphNode->graph->graphNode=
		theApp->mainFrame->graphNode;
	theApp->mainFrame->graph->turtleNode=new TurtleNode(
		theApp->mainFrame->graph->turtle,
		theApp->mainFrame->graphNode);

	midiInit();
}
