//___________________________________________________________________
//
//                          Class IntControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "spinctli.h"

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "IntControl.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(IntControl, wxSpinCtrlInt)
	EVT_SPINCTRL	(-1, IntControl::OnChange)
END_EVENT_TABLE()

IntControl::IntControl(IntControlNodeP intControlNode_, int style):
	intControlNode(intControlNode_)
{
	wxWindow *parent=console;
	if (intControlNode->parentFrame)
		parent=intControlNode->parentFrame->frame;
	else if (intControlNode->parentGraph)
		parent=intControlNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			1,1,1,1,
			"IntControl");
}

IntControl::IntControl(IntControlNodeP intControlNode_, 
			wxWindow* parent, wxWindowID id, 
			int value, int minValue, int maxValue, 
			PairP onChange_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxSpinCtrlInt(parent, id, 
			wxEmptyString,
			pos, size, style, 
			minValue, maxValue, value,
			name),
onChange(onChange_),
intControlNode(intControlNode_)
{
	SetValue(value);
	intControlNode->intControl=this;
	intControlNode->style=style;
}

IntControl::~IntControl()
{
	if (intControlNode)
		intControlNode->intControl=0;
}

void IntControl::OnChange(wxSpinEvent &event)
{
	Event::fifo << new EventIntControl(this);
	event.Skip();
}

wxString IntControl::wxstring()
{
	wxString str;
	str << "IntControl " << GetMin() << ' ' << GetValue() << ' ' << GetMax() << ' ' 
		<< '[' << GetName() << "] ";
	if (onChange)
		str << onChange->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &IntControl::read(wxDataInputStream &istr) 
{	
	int value, min, max;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> name >> (int&)onChange
		>> pos.x >> pos.y >> size.x >> size.y;

	SetRange(min, max);
	SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	return istr;
}
wxDataOutputStream &IntControl::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetMin() << GetValue() << GetMax() << GetName() 
		<< GC_::ptrToInt->find(onChange)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP IntControlNode::onChange()
{
	if (intControl)
		return intControl->onChange;
	return 0;
}

void IntControlNode::setonChange(NodeP nd) 
{
	if (intControl)
	{	intControl->onChange = PairP(nd);  check_oldyoung(nd);
	}
}

void IntControlNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void IntControlNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

IntControlNode::IntControlNode(): 
	Node(INTCONTROL), intControl(0), parentFrame(0), parentGraph(0) {}

IntControlNode::IntControlNode(FrameNodeP parentFrame_, 
										 GraphNodeP parentGraph_): 
	Node(INTCONTROL), intControl(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void IntControlNode::free() 
{	
	if (intControl) 
	{	intControl->intControlNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		intControl->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tIntControl:
				if (((EventIntControl*)ev)->intControl==intControl)
					((EventIntControl*)ev)->intControl=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		intControl=0;
	}
}
void IntControlNode::inc_gc() {fifo << onChange();}
int IntControlNode::youngs() {return gmark(onChange());}
int IntControlNode::youngsPtr() {return gmarkPtr(onChange());}
void IntControlNode::check_oldyoungs() {check_oldyoung(onChange());}

wxDataInputStream &IntControlNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		intControl=new IntControl(IntControlNodeP(this), style);
		intControl->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &IntControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (intControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph) 
			<< style;
		intControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void IntControlNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (intControl)
		setonChange((Pair*)GC_::intToPtr->find(int(intControl->onChange)));
}

void IntControlNode::recreate()
{	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	IntControlP i=intControl, i2=0;
	i2=new IntControl(IntControlNodeP(this),
			p, -1, i->GetValue(), i->GetMin(), i->GetMax(),
			i->onChange, i->GetPosition(), i->GetSize(), style, 
			wxDefaultValidator, i->GetName());
	intControl=i2;
	setonChange(i->onChange);
	delete i;
	intControl=i2;


}

StringP IntControlNode::string()
{
	if (intControl)
		return new String(intControl->wxstring());
	return _Null_Word->name();
}

wxString IntControlNode::wxstring()
{
	if (intControl)
		return intControl->wxstring();
	return wxEmptyString;
}

NodeP lIntControl(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("IntControl");
	if (arg[2])
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int minValue;
	{	int v=arg[3]->a_int(3);
		if throwing return Unbound;
		minValue=v;	
	}
	int value;
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		value=v;	
	}
	int maxValue;
	{	int v=arg[5]->a_int(5);
		if throwing return Unbound;
		maxValue=v;	
	}
	PairP onChange=0;
	if (exist(6))
	{	PairP os=arg[6]->a_runnable(6);
		if throwing return Unbound;
		onChange=os;
	}
	long style = wxSP_ARROW_KEYS;
	if (exist(7))
	{	int s=arg[7]->a_int(7);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(8))
	{	PairP p=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(9))
	{	PairP s=arg[9]->a_int_vector(9);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	IntControlNodeP intControlNode=new IntControlNode(
		parentFrameNode, 
		parentGraphNode);
	new IntControl(intControlNode,
			parent, id, value, minValue, maxValue, 
			onChange, pos, size, style, wxDefaultValidator, name);
	return intControlNode;
}

NodeP lIntControlValue(PairPP arg)
{
	if (exist(1))
	{	IntControlNodeP intControl=arg[1]->a_intControl(1);
		if throwing return Unbound;
		return new Int(intControl->intControl->GetValue());
	}
	else
		return new Int(Event::intControlValue);
}

NodeP lIntControlSetValue(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
	intControl->intControl->SetValue(value);
	return Unbound;
}

NodeP lIntControlSetRange(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int min=arg[2]->a_int(2);
	if throwing return Unbound;
	int max=arg[3]->a_int(3);
	if throwing return Unbound;
	intControl->intControl->SetRange(min, max);
	return Unbound;
}

NodeP lIntControlOnChange(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	intControl->setonChange(function);
	return Unbound;
}

NodeP lIntControlEnable(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	intControl->intControl->Enable(state!=0);
	return Unbound;
}

NodeP lIntControlDestroy(PairPP arg)
{
	IntControlNodeP intControl=arg[1]->a_intControl(1);
	if throwing return Unbound;
	{	FrameNodeP p=intControl->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(intControl));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=intControl->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(intControl));
			p->check_oldyoung(p->children);
		}
	}
	intControl->intControl->Destroy();
	return Unbound;
}
