//___________________________________________________________________
//
//                         Class ListControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/listctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "File.h"

#include "Event.h"
#include "ListControl.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyListControl: public wxListCtrl
{
public:
	MyListControl(wxWindow* parent, wxWindowID id, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = wxLC_REPORT, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = "ListControl");

	// MSW sends extra kill focus event
	virtual ~MyListControl()
	{
		if (parent) 
			parent->listControl = 0;
		parent=0;
	}

	ListControl *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to ListControl
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);
	void OnItemSelected(wxListEvent &event);
	void OnItemActivated(wxListEvent &event);
	void OnColClick(wxListEvent &event);

	void setRow(int row, PairP rowdata);
	void setColumn(int col, PairP coldata);
	void set(PairP data);

	PairP getRow(int row);
	PairP getColumn(int col);
	PairP get();

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyListControl,wxListCtrl)
	EVT_CHAR		(MyListControl::OnChar		)
	EVT_KEY_DOWN(MyListControl::OnKeyDown	)
	EVT_KEY_UP	(MyListControl::OnKeyUp		)
	EVT_LIST_ITEM_SELECTED	(-1,	MyListControl::OnItemSelected	)
	EVT_LIST_ITEM_ACTIVATED	(-1,	MyListControl::OnItemActivated)
	EVT_LIST_COL_CLICK		(-1,	MyListControl::OnColClick		)
END_EVENT_TABLE()

MyListControl::MyListControl(wxWindow* parent_, wxWindowID id, 
		const wxPoint& pos, 
		const wxSize& size, 
		long style, 
		const wxValidator& validator, 
		const wxString& name):
wxListCtrl(parent_, id, pos, size, style, validator, name)
{
	parent = (ListControl*)parent_;
}

void MyListControl::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyListControl::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyListControl::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyListControl::OnItemSelected(wxListEvent &event)
{
	if (parent) 
		parent->OnItemSelected(event);
}

void MyListControl::OnItemActivated(wxListEvent &event)
{
	if (parent) 
		parent->OnItemActivated(event);
}

void MyListControl::OnColClick(wxListEvent &event)
{
	if (parent) 
		parent->OnColClick(event);
}

void MyListControl::setRow(int row, PairP rowdata)
{
	if (GetItemCount() <= row)
		InsertItem(row, wxEmptyString);

	int col=0;
	while (rowdata)
	{	NodeP nd=rowdata->first();
		if (nd)
		{	StringP str=nd->string();
			str=new String(str);
			SetItem(
				row, col, str->ptr());
		}
		col++;
		next(rowdata);
	}
}

void MyListControl::setColumn(int col, PairP coldata)
{
	if (GetColumnCount() <= col)
		InsertColumn(col, wxEmptyString);

	int row=0;
	while (coldata)
	{	NodeP nd=coldata->first();
		if (nd)
		{	StringP str=nd->string();
			str=new String(str);
			SetItem(
				row, col, str->ptr());
		}
		row++;
		next(coldata);
	}
}

void MyListControl::set(PairP data)
{
	int row=0;
	while (data)
	{	PairP rowdata=data->first();
		setRow(row, rowdata);
		row++;
		next(data);
	}
}

PairP MyListControl::getRow(int row)
{
	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT);
	PairP rowdata=0;
	li.SetId(row);
	for (int col=GetColumnCount()-1; col>=0; col--)
	{	li.SetColumn(col);
		GetItem(li);
		StringP str=new String(li.GetText());
		push(rowdata, str);
	}
	return rowdata;
}

PairP MyListControl::getColumn(int col)
{
	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT);
	PairP coldata=0;
	li.SetColumn(col);
	for (int row=GetItemCount()-1; row>=0; row--)
	{	li.SetId(row);
		GetItem(li);
		StringP str=new String(li.GetText());
		push(coldata, str);
	}
	return coldata;
}

PairP MyListControl::get()
{
	PairP data=0;
	for (int row=GetItemCount()-1; row>=0; row--)
	{	PairP rowdata=getRow(row);
		push(data, rowdata);
	}
	return data;
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ListControl, wxControl)
	EVT_CHAR				(		ListControl::OnChar		)
	EVT_KEY_DOWN		(		ListControl::OnKeyDown	)
	EVT_KEY_UP			(		ListControl::OnKeyUp		)
	EVT_LIST_ITEM_SELECTED	(-1,	ListControl::OnItemSelected	)
	EVT_LIST_ITEM_ACTIVATED	(-1,	ListControl::OnItemActivated	)
	EVT_LIST_COL_CLICK		(-1,	ListControl::OnColClick			)
END_EVENT_TABLE()

ListControl::ListControl(ListControlNodeP listControlNode_, int style):
label(0), listControl(0), bestSize(-999, -999),
onChar(0), onKeyUp(0), onKeyDown(0),
onItemSelected(0), onItemActivated(0), onColClick(0),
listControlNode(listControlNode_)
{
	wxWindow *parent=console;
	if (listControlNode->parentFrame)
		parent=listControlNode->parentFrame->frame;
	else if (listControlNode->parentGraph)
		parent=listControlNode->parentGraph->graph;
	create(parent, -1, 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, "ListControl");
}

ListControl::ListControl(ListControlNodeP listControlNode_, 
			wxWindow* parent, wxWindowID id, 
			PairP onItemSelected_,
			PairP onItemActivated_,
			PairP onColClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), listControl(0), bestSize(-999, -999), 
onChar(0), onKeyUp(0), onKeyDown(0),
onItemSelected(onItemSelected_), 
onItemActivated(onItemActivated_),
onColClick(onColClick_),
listControlNode(listControlNode_)
{
	bestSize = wxSize(-999, -999);
	listControlNode->listControl=this;
	listControlNode->style=style;
	create(parent, id, pos, size, style, validator, name);
}

bool ListControl::create(wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		listControl = new MyListControl(this, id, 
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		listControl = new MyListControl(this, id, 
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void ListControl::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (listControl)   
		listControl->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ListControl::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxListCtrl *t=new wxListCtrl((wxWindow*)this, wxID_ANY, 
			wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ListControl::~ListControl()
{
	if (listControlNode)
		listControlNode->listControl=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (listControl)
	{	delete listControl;
		listControl=0;
	}

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::ListControlItemSelected:
		case Event::ListControlItemActivated:
		case Event::ListControlColClick:
		case Event::ListControlChar:
		case Event::ListControlKeyDown:
		case Event::ListControlKeyUp:
			if (((EventListControl*)ev)->listControl==this)
				((EventListControl*)ev)->listControl=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

#define LB_ON(NAME)																\
void ListControl::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventListControl##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ListControl::OnItemSelected(wxListEvent &event)
{
	Event::fifo << new EventListControlItemSelected(this, event);
}

void ListControl::OnItemActivated(wxListEvent &event)
{
	Event::fifo << new EventListControlItemActivated(this, event);
}

void ListControl::OnColClick(wxListEvent &event)
{
	Event::fifo << new EventListControlColClick(this,event);
}

wxString ListControl::wxstring()
{
	wxString str;
	str << "ListControl " 
		<< '[' << GetName() << "] ";
	if (onItemSelected)
		str << onItemSelected->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onItemActivated)
		str << onItemActivated->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onColClick)
		str << onColClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &ListControl::read(wxDataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name 
		>> (int&)onItemSelected
		>> (int&)onItemActivated
		>> (int&)onColClick
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int cols, rows;
	istr >> cols >> rows;

	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_WIDTH);
	wxString str;
	int width, state;

	for (int col=0; col<cols; col++)
	{	istr >> str >> width;
		li.SetText(str);
		li.SetWidth(width);
		li.SetColumn(col);
		listControl->InsertColumn(col, li);
	}
	for (int row=0; row<rows; row++)
	{	li.SetId(row);
		li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_STATE);
		listControl->InsertItem(row, wxEmptyString);
		for (int col=0; col<cols; col++)
		{	istr >> str >> state;
			li.SetText(str);
			li.SetState(state);
			li.SetColumn(col);
			li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_STATE);
			li.SetStateMask(-1);
			listControl->SetItem(li);
		}
	}

	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ListControl::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetName() 
		<< GC_::ptrToInt->find(onItemSelected)
		<< GC_::ptrToInt->find(onItemActivated)
		<< GC_::ptrToInt->find(onColClick)
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=listControl->GetForegroundColour(), 
			b=listControl->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=listControl->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int cols=listControl->GetColumnCount(),
		rows=listControl->GetItemCount();
	ostr << cols << rows;

	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_WIDTH);

	for (int col=0; col<cols; col++)
	{	li.SetColumn(col);
		listControl->GetColumn(col, li);
		ostr << li.GetText() << li.GetWidth();
	}
	for (int row=0; row<rows; row++)
	{	li.SetId(row);
		for (int col=0; col<cols; col++)
		{	li.SetColumn(col);
			li.SetMask(wxLIST_MASK_TEXT);
			listControl->GetItem(li);
			ostr << li.GetText();

			li.SetMask(wxLIST_MASK_STATE);
			int res=listControl->GetItemState(li, -1);
			ostr << res;
		}
	}
	return ostr;
}

bool ListControl::SetBackgroundColour(const wxColour &c)
{
	if (listControl)
		return listControl->SetBackgroundColour(c);
	return false;
}

bool ListControl::SetForegroundColour(const wxColour &c)
{
	if (listControl)
		return listControl->SetForegroundColour(c);
	return false;
}

bool ListControl::SetFont(const wxFont &f)
{
	if (listControl)
		return listControl->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)							\
PairP ListControlNode::on##x()					\
{													\
	if (listControl)								\
		return listControl->on##x;				\
	return 0;									\
}													\
													\
void ListControlNode::seton##x(NodeP nd)	\
{													\
	if (listControl)								\
	{	listControl->on##x = PairP(nd);		\
		check_oldyoung(nd);					\
	}												\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(ItemSelected)
NODE_ON(ItemActivated)
NODE_ON(ColClick)

void ListControlNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ListControlNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ListControlNode::ListControlNode(): 

	Node(LISTCONTROL), listControl(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

ListControlNode::ListControlNode(FrameNodeP parentFrame_, 
											GraphNodeP parentGraph_): 
	Node(LISTCONTROL), listControl(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ListControlNode::free() 
{	
	if (listControl) 
	{	listControl->listControlNode=0;
		listControl->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}

		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::ListControlItemSelected:
			case Event::ListControlItemActivated:
			case Event::ListControlColClick:
			case Event::ListControlChar:
			case Event::ListControlKeyDown:
			case Event::ListControlKeyUp:
				if (((EventListControl*)ev)->listControl==listControl)
					((EventListControl*)ev)->listControl=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		listControl=0;
	}
}
void ListControlNode::inc_gc() 
{	fifo << onItemSelected() << onItemActivated() << onColClick()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ListControlNode::youngs() 
{	return gmark(onItemSelected())+gmark(onItemActivated())+gmark(onColClick())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ListControlNode::youngsPtr() 
{	return gmarkPtr(onItemSelected())+gmarkPtr(onItemActivated())
		+gmarkPtr(onColClick())
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp());
}
void ListControlNode::check_oldyoungs() 
{	check_oldyoung(onItemSelected());
	check_oldyoung(onItemActivated());
	check_oldyoung(onColClick());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

wxDataInputStream &ListControlNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		listControl=new ListControl(ListControlNodeP(this), style);
		listControl->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &ListControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (listControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph)
			<< style;
		listControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ListControlNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (listControl)
	{	ListControlP t=listControl;
		setonItemSelected ((Pair*)GC_::intToPtr->find(int(t->onItemSelected )));
		setonItemActivated((Pair*)GC_::intToPtr->find(int(t->onItemActivated)));
		setonColClick     ((Pair*)GC_::intToPtr->find(int(t->onColClick     )));
		setonChar   ((Pair*)GC_::intToPtr->find(int(t->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(t->onKeyUp  )));
	}
}

void ListControlNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ListControlP t=listControl, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ListControl(ListControlNodeP(this),
			p, -1, 
			t->onItemSelected, t->onItemActivated, t->onColClick, 
			pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	listControl=t2;
	setonItemSelected (t->onItemSelected );
	setonItemActivated(t->onItemActivated);
	setonColClick     (t->onColClick     );
	setonChar   (t->onChar   );
	setonKeyDown(t->onKeyDown);
	setonKeyUp  (t->onKeyUp  );

	MyListControl 
		*lc=t->listControl, 
		*lc2=t2->listControl;

	t2->listControl->SetForegroundColour(lc->GetForegroundColour());
	t2->listControl->SetBackgroundColour(lc->GetBackgroundColour());
	wxFont f=lc->GetFont();
	t2->listControl->SetFont(f);

	int cols=lc->GetColumnCount(), 
		 rows=lc->GetItemCount();

	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_WIDTH);
	wxString str;

	for (int col=0; col<cols; col++)
	{	li.SetColumn(col);
		lc->GetColumn(col, li);
		lc2->InsertColumn(col, li);
	}
	for (int row=0; row<rows; row++)
	{	li.SetId(row);
		lc2->InsertItem(row, wxEmptyString);
		for (int col=0; col<cols; col++)
		{	li.SetColumn(col);
			if (!col)
			{	li.SetMask(wxLIST_MASK_TEXT | wxLIST_MASK_STATE);
				li.SetStateMask(-1);
			}else
				li.SetMask(wxLIST_MASK_TEXT);
			lc->GetItem(li);
			lc2->SetItem(li);
		}
	}
	
	delete t;
	listControl=t2;
}

StringP ListControlNode::string()
{
	if (listControl)
		return new String(listControl->wxstring());
	return _Null_Word->name();
}

wxString ListControlNode::wxstring()
{
	if (listControl)
		return listControl->wxstring();
	return wxEmptyString;
}

NodeP lListControl(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("ListControl");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	PairP onItemSelected=0;
	if (exist(3))
	{	PairP os=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onItemSelected=os;
	}
	PairP onItemActivated=0;
	if (exist(4))
	{	PairP oa=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onItemActivated=oa;
	}
	PairP onColClick=0;
	if (exist(5))
	{	PairP oc=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onColClick=oc;
	}
	long style = wxLC_REPORT;
	if (exist(6))
	{	int s=arg[6]->a_int(6);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(8))
	{	PairP s=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	ListControlNodeP listControlNode=new ListControlNode(
		parentFrameNode, 
		parentGraphNode);
	new ListControl(listControlNode,
			parent, id, 
			onItemSelected, onItemActivated, onColClick, 
			pos, size, style, wxDefaultValidator, name);
	return listControlNode;
}

NodeP lListControlInsertColumn(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int col=arg[2]->a_int(2);
	if throwing return Unbound;
	StringP label=arg[3]->a_string(3);
	if throwing return Unbound;
	int width=-1;
	if (exist(4))
	{	width=arg[4]->a_int(4);
		if throwing return Unbound;
	}
	int format=wxLIST_FORMAT_LEFT;
	if (exist(5))
	{	format=arg[5]->a_int(5);
		if throwing return Unbound;
	}
	listControl->listControl->listControl->InsertColumn(
		col, label->ptr(), format, width);
	return Unbound;
}

NodeP lListControlInsertItem(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	listControl->listControl->listControl->InsertItem(row, 0);
	return Unbound;
}

NodeP lListControlSetItem(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	int col=arg[3]->a_int(3);
	if throwing return Unbound;
	StringP str=arg[4]->a_string(4);
	if throwing return Unbound;
	StringP strreal=new String(str);
	listControl->listControl->listControl->SetItem(row, col, strreal->ptr());
	return Unbound;
}

NodeP lListControlGetItem(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	int col=arg[3]->a_int(3);
	if throwing return Unbound;
	wxListItem li;
	li.SetId(row);
	li.SetColumn(col);
	li.SetMask(wxLIST_MASK_TEXT);
	listControl->listControl->listControl->GetItem(li);
	StringP str=new String(li.GetText());
	return str;
}

NodeP lListControlDeleteItem(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);

	if throwing return Unbound;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	if (row < 0 || row >= listControl->listControl->listControl->GetItemCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	listControl->listControl->listControl->DeleteItem(row);
	return Unbound;
}

NodeP lListControlSetRow(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	PairP rowdata=arg[3]->a_list(3);
	if throwing return Unbound;
	listControl->listControl->listControl->setRow(row, rowdata);
	return Unbound;
}

NodeP lListControlSetColumn(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int col=arg[2]->a_int(2);
	if throwing return Unbound;
	PairP coldata=arg[3]->a_list(3);
	if throwing return Unbound;
	listControl->listControl->listControl->setColumn(col, coldata);
	return Unbound;
}

NodeP lListControlSet(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	PairP data=arg[2]->a_list(2);
	if throwing return Unbound;
	listControl->listControl->listControl->set(data);
	return Unbound;
}

NodeP lListControlGetRow(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	MyListControl *lc=listControl->listControl->listControl;
	int row=arg[2]->a_int(2);
	if throwing return Unbound;
	if (lc->GetItemCount() <= row)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	PairP rowdata=lc->getRow(row);
	return rowdata;
}

NodeP lListControlGetColumn(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	MyListControl *lc=listControl->listControl->listControl;
	int col=arg[2]->a_int(2);
	if throwing return Unbound;
	if (lc->GetColumnCount() <= col)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	PairP coldata=lc->getColumn(col);
	return coldata;
}

NodeP lListControlGet(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	MyListControl *lc=listControl->listControl->listControl;
	PairP data=lc->get();
	return data;
}

NodeP lListControlItemCount(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int count=listControl->listControl->listControl->GetItemCount();
	return new Int(count);
}

NodeP lListControlColumnCount(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int count=listControl->listControl->listControl->GetColumnCount();
	return new Int(count);
}

NodeP lListControlRow(PairPP arg)
{
	int row=Event::listControlRow;
	return new Int(row);
}

NodeP lListControlColumn(PairPP arg)
{
	int col=Event::listControlCol;
	return new Int(col);
}

NodeP lListControlText(PairPP arg)
{
	StringP text=new String(Event::listControlText);
	return text;
}

NodeP lListControlSort(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	MyListControl *lc=listControl->listControl->listControl;
	int sortcol=arg[2]->a_int(2);
	if throwing return Unbound;
	if (sortcol < 0 
	||  sortcol >= lc->GetColumnCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	int cc=lc->GetColumnCount(),
		rc=lc->GetItemCount();

	wxListItem li;
	li.SetMask(wxLIST_MASK_TEXT);
	PairP data=0;
	int row;
	for (row=rc-1; row>=0; row--)
	{	PairP rd=0;
		NodeP key=0;
		li.SetId(row);
		for (int col=cc-1; col>=0; col--)
		{	li.SetColumn(col);
			lc->GetItem(li);
			NodeP nd=new String(li.GetText());
			if (nd->numberP())
				nd=nd->number();
			push(rd, nd);
			if (col==sortcol)
				key=nd;
		}
		if (sortcol>0)
			push(rd, key);
		push(data, rd);
	}
	data=data->mergeSort();

	row=0;
	while (data)
	{	PairP rd=data->first();
		if (sortcol>0)
			next(rd);
		lc->setRow(row, rd);
		row++;
		next(data);
	}
	return Unbound;
}

#define _TC_ON(NAME)									\
NodeP lListControlOn##NAME(PairPP arg)			\
{															\
	ListControlNodeP listControl=arg[1]->a_listControl(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	listControl->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(ItemSelected)
_TC_ON(ItemActivated)
_TC_ON(ColClick)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lListControlSetBackgroundColor(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	listControl->listControl->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lListControlSetColor(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	listControl->listControl->SetForegroundColour(wc);
	return Unbound;
}

NodeP lListControlSetFontSize(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listControl->listControl->GetFont();
	f.SetPointSize(s);
	listControl->listControl->SetFont(f);
	return Unbound;
}

NodeP lListControlSetFontName(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=listControl->listControl->GetFont();
	f.SetFaceName(s->ptr());
	listControl->listControl->SetFont(f);
	return Unbound;
}

NodeP lListControlSetFontStyle(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listControl->listControl->GetFont();
	f.SetStyle(s);
	listControl->listControl->SetFont(f);
	return Unbound;
}

NodeP lListControlSetFontWeight(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listControl->listControl->GetFont();
	f.SetWeight(s);
	listControl->listControl->SetFont(f);
	return Unbound;
}

NodeP lListControlEnable(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	listControl->listControl->Enable(state!=0);
	return Unbound;

}

NodeP lListControlDestroy(PairPP arg)
{
	ListControlNodeP listControl=arg[1]->a_listControl(1);
	if throwing return Unbound;
	{	FrameNodeP p=listControl->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(listControl));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=listControl->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(listControl));
			p->check_oldyoung(p->children);
		}
	}
	listControl->listControl->Destroy();
	return Unbound;
}
