//___________________________________________________________________
//
//                         ListControl Header
//___________________________________________________________________

extern Function 
	lListControl, lListControlDestroy,
	lListControlInsertColumn,
	lListControlInsertItem, 
	lListControlSetItem, lListControlGetItem, lListControlDeleteItem, 
	lListControlSetRow, lListControlSetColumn, lListControlSet, 
	lListControlGetRow, lListControlGetColumn, lListControlGet, 
	lListControlItemCount, lListControlColumnCount,
	lListControlColumn, lListControlRow, lListControlText,
	lListControlSort,
	lListControlSetBackgroundColor, lListControlSetColor,
	lListControlSetFontSize, lListControlSetFontName,
	lListControlSetFontStyle, lListControlSetFontWeight,
	lListControlOnChar, lListControlOnKeyDown, lListControlOnKeyUp,
	lListControlOnItemSelected, lListControlOnItemActivated,
	lListControlOnColClick,
	lListControlEnable;

class wxStaticText;
class wxListEvent;
class MyListControl;

class ListControl: public wxControl
{
	public:
		wxStaticText *label;
		MyListControl *listControl;
		wxSize bestSize;

		PairP onChar, onKeyUp, onKeyDown,
			onItemSelected, onItemActivated, 
			onColClick, onColEndDrag;
		ListControlNodeP listControlNode;

		ListControl(ListControlNodeP listControlNode_, int style);
		ListControl(ListControlNodeP listControlNode_,
			wxWindow* parent, wxWindowID id, 
			PairP onItemSelect_=PairP(0),
			PairP onItemActivated_=PairP(0),
			PairP onColClick_=PairP(0),
			const wxPoint& pos=wxDefaultPosition, 
			const wxSize& size=wxDefaultSize, 
			long style = 0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = "ListControl");

		bool create(wxWindow* parent, wxWindowID id, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name);

		~ListControl();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		void 
			OnClose	(wxCloseEvent &event),
			OnItemSelected	(wxListEvent &event),
			OnItemActivated(wxListEvent &event),
			OnColClick		(wxListEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef ListControl *ListControlP;

class ListControlNode: public Node
{
	public:
		ListControlP listControl;
		FrameNodeP parentFrame;
		GraphNodeP parentGraph;
		int style;

		static const ListControlNode aListControlNode;

		operator_new;
		PairP onItemSelected(), onItemActivated(), onColClick(),
			onChar(), onKeyDown(), onKeyUp();
		void
			setonItemSelected	(NodeP function),
			setonItemActivated(NodeP function),
			setonColClick		(NodeP function),
			setonChar			(NodeP function), 
			setonKeyDown		(NodeP function),
			setonKeyUp			(NodeP function);

		ListControlNode();
		ListControlNode(FrameNodeP parentFrame_, GraphNodeP parentGraph_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();
		
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		StringP _typeof();

		void setParentFrame(FrameNodeP frame);
		void setParentGraph(GraphNodeP graph);
};
