#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/string.h>

#include <m_apm.h>

#include "Main.h"
#include "String.h"
#include "Object.h"
#include "List.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "BigFloat.h"
#include "Complex.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "File.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Point.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

const Float::x87_ControlWord 
		Float::x87_overflow_mask = {1,0,1,1, 0,0,0,0, 0,0,0};	// 0xD was: 0x1F

void Float::reset_FPU()
{
	// for FPU=80x87:
	x87_ControlWord cw = {1,1,1,1, 1,1,0,1, 3, 0, 1};	// should be 0x1320
	cw.precision_control = _64bit;
	cw.rounding_control = nearest;
#ifdef WIN32
	_asm finit
	_asm fldcw cw;
#else
	asm("finit;  fldcw %0": :"m" (cw));
#endif
}

void Float::init() 
{
	reset_FPU();
	_FloatMax   ->setvar(new Float(max_));
	_Pi         ->setvar(new Float(pi ));
	_Determinant->setvar(new Float( 0 ));
}

void BigFloat::init() 
{
	m_apm_cpp_precision(30);
	deg2rad=MAPM(MM_PI)/180;
	_BigFloatPi->setvar(new BigFloat(MAPM(MM_PI)));
}

void Int  ::init() 
{
	seedMT(time(0));
		  _Zero->setvar(new Int( 0)); 
	      _One->setvar(new Int( 1));
	_Minus_One->setvar(new Int(-1));
	_IntMax   ->setvar(new Int(max));
	_Int16Max ->setvar(new Int16(Int16::max));
	_Int8Max  ->setvar(new Int8 (Int8 ::max));
	_UInt8Max ->setvar(new UInt8(UInt8::max));
}

NodeP lBigFloatSetPrecision(PairPP arg)
{
	int precision=arg[1]->a_int(1);
	m_apm_cpp_precision(precision);
	BigFloat::deg2rad=MAPM(MM_PI)/180;
	return Unbound;
}

double Float::abs(double x) {return x<0 ? -x : x;}
double Complexd::abs(complexd x) 
{	return ::sqrt(Float::sqr(x.real())+Float::sqr(x.imag()));
}
double Float::sqr(double x) {return x*x;}
complexd Complexd::sqr(complexd x) {return x*x;}

double Float::cutError(double x) 
{
	if ((x>0 && x<epsilon)||(x<0 && x>-epsilon)) 
		return 0; 
	return x;
}

int    Int  ::Remainder(   int x,    int d) 
{	if (!d) 
		return Int::max+1;
	return x%d;
}
double Float::Remainder(double x, double d) 
{	if (d==0.0) 
		return Float::NaN;
	return fmod(x,d);
}
MAPM BigFloat::Remainder(MAPM x, MAPM d) 
{	if (d==0.0) 
		return Float::NaN;
	return x%d;
}
complexd Complexd::Remainder(complexd x, double d) 
{	return complexd(fmod(x.real(),d), fmod(x.imag(),d));
}
complexd Complexd::Remainder(double x, complexd d) 
{	return Float::NaN;
}
complexd Complexd::Remainder(complexd x, complexd d) 
{	return Float::NaN;
}

int Int::mod(int x, int d) 
{
	if (!d) return Int::max+1;
	int m=x%d;
	if ((m < 0) != (d < 0)) return m+d;
	return m;


}
double Float::mod(double x, double d) 
{
	double m=fmod(x,d);
	if ((m < 0) != (d < 0)) return m+d;
	return m;
}
MAPM BigFloat::mod(MAPM x, MAPM d) 
{
	MAPM m=x % d;
	if ((m < 0) != (d < 0)) return m+d;
	return m;
}
complexd Complexd::mod(complexd x, double d) 
{
	double r=fmod(x.real(),d), i=fmod(x.imag(),d);

	if ((r < 0) != (d < 0)) 
		if ((i < 0) != (d < 0)) 
			return complexd(r+d,i+d);
		else
			return complexd(r+d,i);

	if ((i < 0) != (d < 0)) 
		return complexd(r,i+d);
	return complexd(r,i);
}
complexd Complexd::mod(double x, complexd d) 
{	return Float::NaN;
}
complexd Complexd::mod(complexd x, complexd d) 
{	return Float::NaN;
}

int Int::clip(double d)
{
	if (d > Int::max) return  Int::max;
	if (d <-Int::max) return -Int::max-1;
	return int(d);
}

short Int16::clip(double d)
{
	if (d > Int16::max) return  Int16::max;
	if (d <-Int16::max) return -Int16::max-1;
	return short(d);
}

short Int16::clip(int d)
{
	if (d > Int16::max) return  Int16::max;

	if (d <-Int16::max) return -Int16::max-1;
	return short(d);
}

char Int8::clip(double d)
{
	if (d > Int8::max) return  Int8::max;
	if (d <-Int8::max) return -Int8::max-1;
	return char(d);
}

char Int8::clip(int d)
{
	if (d > Int8::max) return  Int8::max;
	if (d <-Int8::max) return -Int8::max-1;
	return char(d);
}

unsigned char UInt8::clip(double d)
{
	if (d > UInt8::max) return UInt8::max;
	if (d < 0         ) return 0;
	return (unsigned char)(d);
}

unsigned char UInt8::clip(int d)
{
	if (d > UInt8::max) return UInt8::max;
	if (d < 0         ) return 0;
	return (unsigned char)(d);
}

NodeP Node::boolnode(bool tf) {return tf ?  _True : _False;}

void Node::checkboolerror(int tf, NodeP a, NodeP b) 
{
	if (tf==Node::boolerror)
		error(BAD_DATA, new List(a, b, _ListEnd));
}

int Node::compare(NodeP nd) 
{
	if (!nd) return boolerror;

	if (NodeP(this)==nd) return 0;
	if (!is.word  ||  !nd->is.word)
		return boolerror;

	error(FATAL, this);	// should never be reached
	return 0;
}
int Object::compare(NodeP nd) 
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;
	else return name()->compare(nd);
}
int Int::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	int i=int_- IntP(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	double f = int_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 

	{	double f = int_- ComplexdP(nd)->Norm()->getnumber();

		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	int i=int_- Int16P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	int i=int_- Int8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	int i=int_- UInt8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = int_- BigFloatP(nd)->float_;

		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b=0;
	if (nd->type==OBJECT)
		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}
int Int16::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	int i=int_- IntP(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	double f = int_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 
	{	double f = int_- ComplexdP(nd)->Norm()->getnumber();
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	int i=int_- Int16P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	int i=int_- Int8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	int i=int_- UInt8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = int_- BigFloatP(nd)->float_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b=0;
	if (nd->type==OBJECT)
		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}

int Int8::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	int i=int_- IntP(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}


	else if (nd->type==FLOAT_) 
	{	double f = int_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 
	{	double f = int_- ComplexdP(nd)->Norm()->getnumber();
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	int i=int_- Int16P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	int i=int_- Int8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	int i=int_- UInt8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = int_- BigFloatP(nd)->float_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b=0;
	if (nd->type==OBJECT)
		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}
int UInt8::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	int i=int_- IntP(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	double f = int_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 
	{	double f = int_- ComplexdP(nd)->Norm()->getnumber();
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	int i=int_- Int16P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);

	}
	else if (nd->type==INT8) 
	{	int i=int_- Int8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	int i=int_- UInt8P(nd)->int_;
		return i==0 ?  0  :  (i > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = int_- BigFloatP(nd)->float_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b=0;
	if (nd->type==OBJECT)
		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}
int Float::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	double f = float_- IntP(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	double f = float_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 
	{	double f = float_- ComplexdP(nd)->Norm()->getnumber();
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	double f = float_- Int16P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	double f = float_- Int8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	double f = float_- UInt8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = float_- BigFloatP(nd)->float_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b;
	if (nd->type==OBJECT)

		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}
int BigFloat::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	MAPM f = float_- IntP(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	MAPM f = float_- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==COMPLEX) 
		return Node::boolerror;
	else if (nd->type==INT16) 
	{	MAPM f = float_- Int16P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	MAPM f = float_- Int8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	MAPM f = float_- UInt8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	MAPM f = float_- BigFloatP(nd)->float_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	NodeP b;
	if (nd->type==OBJECT)

		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return string()->compare(nd);
}
int Complexd::compare(NodeP nd)
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;

	if (nd->type==INT_) 
	{	double f = Norm()->getnumber()- IntP(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==FLOAT_) 
	{	double f = Norm()->getnumber()- FloatP(nd)->float_;
		return f==0.0 ?  0  :  (f > 0.0 ?  1 : -1);
	}
	else if (nd->type==INT16) 
	{	double f = Norm()->getnumber()- Int16P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==INT8) 
	{	double f = Norm()->getnumber()- Int8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==UINT8) 
	{	double f = Norm()->getnumber()- UInt8P(nd)->int_;
		return f==0 ?  0  :  (f > 0 ?  1 : -1);
	}
	else if (nd->type==BIGFLOAT) 
	{	return Node::boolerror;
	}
	NodeP b;
	if (nd->type==OBJECT)
		b = ObjectP(nd)->name()->number();
	else
		b = nd->number();

	if (b != Unbound) 
		return compare(b);
	else
		return Node::boolerror;
}
int String::compare(NodeP nd)
{
	if (!nd || nd->islist() || nd->isarray()) 
		return Node::boolerror;

	String &b = *nd->string();

	if (type==STRING  &&  b.type==STRING) 
	{	if (length()==b.length()  &&  ptr()==b.ptr())
			return 0;
		else 
		{	int cmp_len = length() < b.length() ?
							  length() : b.length();
// case-dependent comparsion should be done by changing the strcomp function pointer
			int icmp = caseIgnored ? 
					comp_low (ptr(), b.ptr(), cmp_len) :
					comp_case(ptr(), b.ptr(), cmp_len);

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);

			if (length() != b.length()  &&  icmp==0)
				icmp = length() - b.length();

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);
			return icmp;
		}
	}else if (is.backslashed || b.is.backslashed) 
	{	if (length()==b.length()  &&  ptr()==b.ptr())
			return 0;
		else 
		{	int cmp_len = length() < b.length() ?
							  length() : b.length();
			int icmp = caseIgnored ?
					comp_noparitylow(ptr(), b.ptr(), cmp_len) :
					comp_noparity   (ptr(), b.ptr(), cmp_len);

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);

			if ((length() != b.length()) && icmp==0)
				icmp = length() - b.length();

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);
			return icmp;
		}
	}else 
		return Node::boolerror;
}
int Pair::compare(NodeP nd)
{
	if (nd && nd->isnolist())
		return Node::boolerror;


	PairP l=this, b=PairP(nd);
	while (l && b)
	{	if (!l->first())
		{	if (b->first())
				return -1;
		}else
		{	int c=l->first()->compare(b->first());
			if (c!=0)
				return c;
		}
		next(l);
		next(b);
	}
	if (l)
		return 1;
	if (b)
		return -1;

	return 0;
}
int Array::compare(NodeP nd)
{	if (nd->type != ARRAY) 
		return Node::boolerror;

	if (this==nd)
		return 0;


	ArrayP b=ArrayP(nd);
	NodePP dp=data(), dpb=b->data();

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{
		int c=(*dp++)->compare(*dpb++);
		if (c)
			return c;
	}
	if (!i & !j)
		return 0;

	if (i)

		return 1;

	return -1;
}
int IntArray::compare(NodeP nd)
{	if (nd->type != INTARRAY) 
		return Node::boolerror;

	if (this==nd)
		return 0;

	IntArrayP b=IntArrayP(nd);
	intP dp=_data, dpb=b->_data;

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{
		int c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;

	if (i)

		return 1;

	return -1;
}
int Int16Array::compare(NodeP nd)
{	if (nd->type != INT16ARRAY) 
		return Node::boolerror;

	if (this==nd)
		return 0;

	Int16ArrayP b=Int16ArrayP(nd);
	shortP dp=_data, dpb=b->_data;

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{
		int c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;

	if (i)

		return 1;

	return -1;
}
int FloatArray::compare(NodeP nd)
{	if (nd->type != FLOATARRAY) 
		return Node::boolerror;

	if (this==nd)
		return 0;

	FloatArrayP b=FloatArrayP(nd);
	doubleP dp=_data, dpb=b->_data;

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{
		double c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;

	if (i)

		return 1;

	return -1;
}

NodeP lisbefore(PairPP arg) 
{
	StringP 
		a = arg[1]->a_string(1),
		b = arg[2]->a_string(2);
	return Node::boolnode( a->compare(b) < 0 );
}

NodeP lisless(PairPP arg) 
{
	NodeP 
		a=arg[1],
		b=arg[2];
	if (b->is.number)
	{	int tf=b->compare(a);
		Node::checkboolerror(tf,a,b);
		if (!throwing)
			return Node::boolnode( tf > 0 );
		return Unbound;
	}
	int tf=a->compare(b);
	Node::checkboolerror(tf,a,b);
	if (!throwing)
		return Node::boolnode( tf < 0 );
	return Unbound;
}

NodeP lisgreater(PairPP arg) 
{
	NodeP 

		a=arg[1],
		b=arg[2];
	if (b->is.number)
	{	int tf=b->compare(a);
		Node::checkboolerror(tf,a,b);
		if (!throwing)
			return Node::boolnode( tf < 0 );
		return Unbound;
	}
	int tf=a->compare(b);
	Node::checkboolerror(tf,a,b);
	if (!throwing)
		return Node::boolnode( tf > 0 );
	return Unbound;
}

NodeP lislessEqual(PairPP arg) 
{
	NodeP 
		a=arg[1],
		b=arg[2];
	if (b->is.number)
	{	int tf=b->compare(a);
		Node::checkboolerror(tf,a,b);
		if (!throwing)
			return Node::boolnode( tf >= 0 );
		return Unbound;
	}
	int tf=a->compare(b);
	Node::checkboolerror(tf,a,b);
	if (!throwing)
		return Node::boolnode( tf <= 0 );
	return Unbound;
}

NodeP lisgreaterEqual(PairPP arg) 
{
	NodeP 
		a=arg[1],
		b=arg[2];
	if (b->is.number)
	{	int tf=b->compare(a);
		Node::checkboolerror(tf,a,b);
		if (!throwing)
			return Node::boolnode( tf <= 0 );
		return Unbound;
	}
	int tf=a->compare(b);
	Node::checkboolerror(tf,a,b);
	if (!throwing)
		return Node::boolnode( tf >= 0 );
	return Unbound;
}

bool Node::equalP(NodeP b) {return equal(b);}
bool Pair::equalP(NodeP b)
{	if (b->isnolist()) 
		return false;

	PairP alist=PairP(this), blist=PairP(b);

	while (alist && blist) 
	{	if (alist->first())
		{	if (! alist->first()->equalP(blist->first()))
				return false;
		}
		else if (blist->first())
			return false;
		next(alist);
		next(blist);
		if (checkThrowing()) break;
	}
	return !alist && !blist;
}
bool Array::equalP(NodeP b)
{	if (b->type != ARRAY) 
		return false;

	if (this==b)
		return true;

	ArrayP ba=ArrayP(b);
	NodePP dp=data(), dpb=ba->data();

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
	{	if (!(*dp))
		{	if (*dpb)
				return false;
			dp++;
			dpb++;
		}else if (!(*dp++)->equalP(*dpb++))
			return false;
	}
	if (!i & !j)
		return true;

	return false;
}
bool IntArray::equalP(NodeP b)
{	if (b->type != INTARRAY) 
		return false;

	if (this==b)
		return true;

	IntArrayP ba=IntArrayP(b);
	intP dp=_data, dpb=ba->_data;

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	
	if (!i & !j)
		return true;

	return false;
}
bool Int16Array::equalP(NodeP b)
{	if (b->type != INT16ARRAY) 


		return false;

	if (this==b)
		return true;

	Int16ArrayP ba=Int16ArrayP(b);
	shortP dp=_data, dpb=ba->_data;

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	
	if (!i & !j)
		return true;

	return false;
}
bool FloatArray::equalP(NodeP b)
{	if (b->type != FLOATARRAY) 
		return false;

	if (this==b)
		return true;

	FloatArrayP ba=FloatArrayP(b);
	doubleP dp=_data, dpb=ba->_data;

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	
	if (!i & !j)
		return true;

	return false;
}
NodeP l_eq    (PairPP arg) {return Node::boolnode(arg[1]==arg[2]);}
NodeP lisequal(PairPP arg) 
{	return 
		arg[1]? Node::boolnode( arg[1]->equalP(arg[2])) 
			:	(arg[2]?  _False 
					:	_True);
}
NodeP lisunequal(PairPP arg) 
{	return 
		arg[1] ?  Node::boolnode(!arg[1]->equalP(arg[2]))
			:	(arg[2] ?  _True 
					:	_False);
}


NodeP lnot(PairPP arg) 
{
	int b = arg[1]->a_bool(1);
	if (!throwing) 
	{	if (b) return _False;
		else   return _True;
	}
	return Unbound;
}

NodeP land(PairPP arg) 
{
	int nr=1;
	while (arg!=argStack && !throwing) 
	{	if ((*++arg)->a_bool(nr++)==0) 
			return _False;
	}
	if throwing return Unbound;
	return _True;
}

NodeP lor(PairPP arg) 
{
	int nr=1;
	while (arg!=argStack && !throwing) 
	{	if ((*++arg)->a_bool(nr++)==1) 
			return _True;
	}
	if throwing return Unbound;
	return _False;
}

NodeP lFaculty(PairPP arg)
{
	IntP num = arg[1]->a_pos_int(1);
	int n=num->int_;
	if (n < 13)
	{	int f=1;
		for (int i=n;  i>0;  i--)
			f*=i;
		return new Int(f);
	}
	else if (n <21)
	{	__int64 f=1;
		for (int i=n;  i>0;  i--)
			f*=i;
		return new Float(double(f));	// rounding may occur here
	}
	else
	{	MAPM f=1;
		for (int i=n;  i>0;  i--)
			f*=i;
		return new BigFloat(f);
	}
	return Unbound;
}

NodeP lisPrime(PairPP arg)
{
	NodeP num = arg[1]->a_number(1);
	if (num->type==BIGFLOAT)
	{	MAPM n=BigFloatP(num)->float_;
		MAPM maxi=sqrt(n);

		if (n % 2 == 0)
			return _False;

		for (MAPM i=3;  i <= maxi;  i+=2)
		{	if (n % i == 0)
				return _False;
			if (checkThrowing())
				return _False;
		}
		return _True;
	}
	double d=num->getnumber();

	if (d < 0) d=-d;

	if (d < 4) return _True;
	
	if (d < Int::max)
	{	unsigned n=unsigned(d);
		unsigned maxi=unsigned(sqrt(n));

		if (n % 2 == 0)
			return _False;

		for (unsigned i=3;  i <= maxi;  i+=2)
			if (n % i == 0)
				return _False;
	}
	else if (d < Int::i64max)
	{	__int64 n=(__int64)(num->getnumber());	// don`t wanna loose precision
		unsigned maxi=unsigned(sqrt(double(n)));

		if (n % 2 == 0)
			return _False;

		for (unsigned i=3;  i <= maxi;  i+=2)
			if (n % i == 0)
				return _False;
	}else
		error(BAD_DATA, num); // number too big for exact representation
	
	return _True;
}

int Int::gcd(int a, int b)
{
	if (!b) return a;
	return gcd(b, a % b);
}

NodeP lgcd(PairPP arg)
{
	int 
		a=arg[1]->a_int(1),
		b=arg[2]->a_int(2);

	if (!throwing)
	{	if (!b) return new Int(a);
		return new Int(Int::gcd(b, a % b));
	}
	return Unbound;
}

NodeP lfactorize(PairPP arg)
{
	NodeP num = arg[1]->a_number(1);
	PairP list=0;
	if (num->type==BIGFLOAT)
	{	MAPM d=BigFloatP(num)->float_;
		if (d<0) 
		{	push(list, new Int(-1));	
			if (d==-1) return list;
			d=-d;
		}
		if (d < 4)
		{	push(list, new Int(d.to_int()));
		}
		else
		{
	factorizeBig:
			MAPM maxi=ceil(sqrt(d));

			if (d % 2 == 0)
			{	push(list, new Int(2));
				d/=2;  goto factorizeBig;
			}
			for (MAPM i=3;  i <= maxi;  i+=2)
			{	if (checkThrowing())
					return Unbound;
				if (d % i == 0)
				{	push(list, new BigFloat(i));
					d/=i;  goto factorizeBig;
				}	
			}
			if (d!=1)
				push(list, new BigFloat(d));
			return list;
		}
	}
	double d = num->getnumber();

	if (d<0) 
	{	push(list, new Int(-1));	
		if (d==-1) return list;
		d=-d;
	}
	if (d < 4)
	{	push(list, new Int(int(d)));
	}
	else if (d < Int::max)
	{	int n=int(d);
factorize:
		int maxi=int(sqrt(n));

		if (n % 2 == 0)
		{	push(list, new Int(2));
			n/=2;  goto factorize;
		}
		for (int i=3;  i <= maxi;  i+=2)
			if (n % i == 0)
			{	push(list, new Int(i));
				n/=i;  goto factorize;
			}	
		if (n!=1)
			push(list, new Int(n));
	}
	else if (d < Int::i64max/2)
	{	__int64 n=(__int64)(num->getnumber());	// don`t wanna loose precision
		if (n<0) 
			n=-n;
factorize64:
		int maxi=int(sqrt(double(n)));

		if ((n & 1) == 0)
		{	push(list, new Int(2));
			n/=2;  goto factorize64;
		}
		for (int i=3;  i <= maxi;  i+=2)
		{	if (n % i == 0)
			{	push(list, new Int(i));
				n/=i;  goto factorize64;
			}	
		}
		if (n!=1)
		{	if (n <= Int::max)
				  push(list, new Int(int(n)));
			else push(list, new Float(double(n)));
		}
	}else
		error(BAD_DATA, num); // number too big for exact representation
	
	return list;
}

double Node ::_min(double d) {error(BAD_DATA, this); return 0;}
double Int  ::_min(double d) {return int_ < d ?  int_ : d;}
double Int16::_min(double d) {return int_ < d ?  int_ : d;}
double Int8 ::_min(double d) {return int_ < d ?  int_ : d;}
double UInt8::_min(double d) {return int_ < d ?  int_ : d;}
double Float::_min(double d) {return float_ < d ?  float_ : d;}
double BigFloat::_min(double d) {return float_ < d ?  float_.to_double() : d;}
double Complexd::_min(double d) 
{	double f=Norm()->getnumber();
	return f < d ?  f : d;
}
double String::_min(double d) {return number()->_min(d);}
double Array::_min(double d)
{	NodePP dp=_data;
	for (int i=_length;  i;  i--)
		d = (*dp++)->_min(d);
	return d;
}
double IntArray::_min(double d)
{	intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int n=*dp++;
		if (n < d)
			d=n;
	}
	return d;
}
double Int16Array::_min(double d)
{	shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short n=*dp++;
		if (n < d)
			d=n;
	}
	return d;
}
double FloatArray::_min(double d)
{	doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double n=*dp++;
		if (n < d)
			d=n;
	}
	return d;
}
double Pair::_min(double d)
{	PairP list=this;
	while (list)
	{	d = list->first()->_min(d);
		next(list);
	}
	return d;
}
NodeP lmin(PairPP arg)
{
	if (exist(1) && !exist(2)) 
		return new Float(arg[1]->_min(Float::max_));
	double d=Float::max_;
	while (arg!=argStack)
		d = (*++arg)->_min(d);
	return new Float(d);
}

double Node ::_max(double d) {error(BAD_DATA, this); return 0;}
double Int  ::_max(double d) {return int_ > d ?  int_ : d;}
double Int16::_max(double d) {return int_ > d ?  int_ : d;}
double Int8 ::_max(double d) {return int_ > d ?  int_ : d;}
double UInt8::_max(double d) {return int_ > d ?  int_ : d;}
double Float::_max(double d) {return float_ > d ?  float_ : d;}
double BigFloat::_max(double d) {return float_ > d ?  float_.to_double() : d;}
double Complexd::_max(double d) 
{	double f=Norm()->getnumber();
	return f > d ?  f : d;
}
double String::_max(double d) {return number()->_max(d);}
double Array::_max(double d)
{	NodePP dp=_data;
	for (int i=_length;  i;  i--)
		d = (*dp++)->_max(d);
	return d;
}
double IntArray::_max(double d)
{	intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int n=*dp++;
		if (n > d)
			d=n;
	}
	return d;
}
double Int16Array::_max(double d)
{	shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short n=*dp++;
		if (n > d)
			d=n;
	}
	return d;
}
double FloatArray::_max(double d)
{	doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double n=*dp++;
		if (n > d)
			d=n;
	}
	return d;
}
double Pair::_max(double d)
{	PairP list=this;
	while (list)
	{	d = list->first()->_max(d);
		next(list);
	}
	return d;
}
NodeP lmax(PairPP arg)
{
	if (exist(1) && !exist(2)) 
		return new Float(arg[1]->_max(-Float::max_));
	double d=-Float::max_;
	while (arg!=argStack)
		d = (*++arg)->_max(d);
	return new Float(d);
}

NodeP Node::Norm() {error(BAD_DATA, this); return Unbound;}
NodeP Int::Norm()
{	if (int_ >=0) return this;

	else return new Int(-int_);
}
NodeP Int16::Norm()
{	if (int_ >=0) return this;
	else return new Int16(-int_);
}
NodeP Int8::Norm()
{	if (int_ >=0) return this;
	else return new Int8(-int_);
}
NodeP UInt8::Norm() {return this;}
NodeP Float::Norm()
{	if (float_ >=0) return this;
	else return new Float(-float_);
}
NodeP BigFloat::Norm()
{	if (float_ >=0) return this;
	else return new BigFloat(-float_);
}
NodeP Complexd::Norm()
{	return new Float(::sqrt(Float::sqr(complex_.real())+Float::sqr(complex_.imag())));
}
NodeP String::Norm()
{	return new Float(Float::abs(getnumber()));
}
NodeP Array::Norm()
{	double d=0;  NodePP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++)->Norm()->getnumber();
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}
NodeP IntArray::Norm()
{	double d=0;  intP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}
NodeP Int16Array::Norm()
{	double d=0;  shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	int h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}
NodeP FloatArray::Norm()
{	double d=0;  doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}
NodeP Pair::Norm()
{	double d=0;
	PairP list=this;
	while (list)
	{	double h = list->first()->Norm()->getnumber();
		if (throwing) 
			return Unbound;
		d+= h*h;
		next(list);
	}
	return new Float(::sqrt(d));
}

NodeP lNorm(PairPP arg)
{
	if (exist(1) && !exist(2)) 
		return arg[1]->Norm();
	double d=0;
	while (arg!=argStack)
	{	double h = (*++arg)->Norm()->getnumber();
		if (throwing) 
			return Unbound;
		d+= h*h;
	}
	return new Float(sqrt(d));
}

NodeP Node::maxNorm() {error(BAD_DATA, this); return Unbound;}
NodeP Int::maxNorm()
{	if (int_ >=0) return this;
	else return new Int(-int_);
}
NodeP Int16::maxNorm()
{	if (int_ >=0) return this;
	else return new Int16(-int_);
}
NodeP Int8::maxNorm()
{	if (int_ >=0) return this;
	else return new Int8(-int_);
}
NodeP UInt8::maxNorm() {return this;}
NodeP Float::maxNorm()
{	if (float_ >=0) return this;
	else return new Float(-float_);
}
NodeP BigFloat::maxNorm()
{	if (float_ >=0) return this;
	else return new BigFloat(-float_);
}
NodeP Complexd::maxNorm()
{	return Norm();
}
NodeP String::maxNorm()
{	return new Float(Float::abs(getnumber()));
}
NodeP Array::maxNorm()
{	double d=0;  NodePP dp=_data;
	for (int i=_length;  i;  i--)

	{	double h = (*dp++)->maxNorm()->getnumber();
		if (throwing) return Unbound;
		double ha=Float::abs(h);
		if (ha > d) d = ha;
	}
	return new Float(d);
}
NodeP IntArray::maxNorm()
{	int d=0;  intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int h = (*dp++);
		if (throwing) return Unbound;
		int ha=::abs(h);
		if (ha > d) d = ha;
	}
	return new Int(d);
}
NodeP Int16Array::maxNorm()
{	int d=0;  shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short h = (*dp++);
		if (throwing) return Unbound;
		short ha=::abs(h);
		if (ha > d) d = ha;
	}
	return new Int(d);
}
NodeP FloatArray::maxNorm()
{	double d=0;  doubleP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		if (throwing) return Unbound;
		double ha=Float::abs(h);
		if (ha > d) d = ha;
	}
	return new Float(d);
}

NodeP Pair::maxNorm()
{	double d=0;
	PairP list=this;
	while (list)
	{	double h = list->first()->maxNorm()->getnumber();
		if (throwing) return Unbound;
		double ha=Float::abs(h);
		if (ha > d) d = ha;
		next(list);
	}
	return new Float(d);
}

NodeP lmaxNorm(PairPP arg)
{
	if (exist(1) && !exist(2)) 
		return arg[1]->maxNorm();
	double d=0;
	while (arg!=argStack)
	{	double h = (*++arg)->maxNorm()->getnumber();
		if (throwing) return Unbound;
		double ha=Float::abs(h);
		if (ha > d) d = ha;

	}
	return new Float(d);
}

NodeP Node::resize(int newsize) {error(BAD_DATA, this); return Unbound;}
NodeP Array::resize(int newsize)
{	ArrayP r=new Array(newsize);
	NodePP d=data();
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	NodeP diff = d[i+1]->sub(d[i]),
						slope = diff->div(new Int(di));	// uahh!

					for (int ii=di-1;  ii;  ii--)
					{	r->_data[ni+ii] = d[i]->add( slope->mul(new Int(ii)) );	// ugly
					}
				}
				oi=ni;
			}
		}
	}
	return r;
}
NodeP IntArray::resize(int newsize)
{	IntArrayP r=new IntArray(newsize);
	intP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	int diff = d[i+1]-d[i];
					double slope = double(diff)/di;

					for (int ii=di-1;  ii;  ii--)


						r->_data[ni+ii] = int(d[i]+slope*ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=int(d[int(i)]+frac*diff+0.5);
			}
		}

	}
	return r;
}
NodeP Int16Array::resize(int newsize)
{	Int16ArrayP r=new Int16Array(newsize);
	shortP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	int diff = d[i+1]-d[i];
					double slope = double(diff)/di;

					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = short(d[i]+slope*ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=short(d[int(i)]+frac*diff+0.5);
			}
		}
	}
	return r;
}
NodeP FloatArray::resize(int newsize)
{	FloatArrayP r=new FloatArray(newsize);
	doubleP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	double diff = d[i+1]-d[i];
					double slope = diff/di;


					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = d[i]+slope*ii;
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=d[int(i)]+frac*diff;
			}
		}
	}
	return r;
}

NodeP lresize(PairPP arg)	// test:  resize {1 2 3} 7
{
	NodeP what=arg[1];
	IntP newsize= arg[2]->a_pos_int(2);
	if (throwing) return Unbound;
	if (what) return what->resize(newsize->int_);
	else return NIL;
}


NodeP Node::lowPassFilter(double weight) 
{	error(BAD_DATA, this); return Unbound;
}
NodeP FloatArray::lowPassFilter(double weight)
{
	FloatArrayP result=new FloatArray(length());
	doubleP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=median;
	}
	Float::reset_FPU();
	return result;
}
NodeP IntArray::lowPassFilter(double weight)
{
	IntArrayP result=new IntArray(length());
	intP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=int(median);
	}
	Float::reset_FPU();
	return result;
}
NodeP Int16Array::lowPassFilter(double weight)
{
	Int16ArrayP result=new Int16Array(length());
	shortP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=short(median);
	}
	Float::reset_FPU();
	return result;
}

NodeP llowPassFilter(PairPP arg)
{
	NodeP what=arg[1];
	double weight=arg[2]->a_float(2);
	if (throwing) return Unbound;
	if (what) return what->lowPassFilter(weight);
	else return NIL;
}

NodeP Node::saturateAbove(double limit) 
{	error(BAD_DATA, this); return Unbound;
}
NodeP FloatArray::saturateAbove(double limit) 
{
	FloatArrayP result=new FloatArray(length());
	doubleP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=*source++;
		*dest++= s>limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}
NodeP IntArray::saturateAbove(double limit) 
{
	IntArrayP result=new IntArray(length());
	intP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int s=*source++;
		*dest++= s>limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}
NodeP Int16Array::saturateAbove(double limit) 
{
	Int16ArrayP result=new Int16Array(length());
	shortP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	short s=*source++;
		*dest++= s>limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP lsaturateAbove(PairPP arg)
{
	double limit=arg[1]->a_float(2);
	NodeP what=arg[2];
	if (throwing) return Unbound;
	if (what) return what->saturateAbove(limit);
	else return NIL;
}


NodeP Node::saturateBelow(double limit) 
{	error(BAD_DATA, this); return Unbound;
}
NodeP FloatArray::saturateBelow(double limit) 
{
	FloatArrayP result=new FloatArray(length());
	doubleP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	double s=*source++;
		*dest++= s<limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}
NodeP IntArray::saturateBelow(double limit) 
{
	IntArrayP result=new IntArray(length());
	intP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int s=*source++;
		*dest++= s<limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}
NodeP Int16Array::saturateBelow(double limit) 
{
	Int16ArrayP result=new Int16Array(length());
	shortP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	short s=*source++;
		*dest++= s<limit? limit : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP lsaturateBelow(PairPP arg)
{
	double limit=arg[1]->a_float(2);
	NodeP what=arg[2];
	if (throwing) return Unbound;
	if (what) return what->saturateBelow(limit);
	else return NIL;
}


NodeP lreRandom(PairPP arg) 
{
	int seed=1;
	if (exist(1)) seed = arg[1]->a_int(1);
	if (!throwing) Int::seedMT(seed);
	return Unbound;
}

int Int::random(int max) 
{
	return max?  randomMT() % max  :  randomMT();
}

double Float::rnd()
{
	union DoubleStruct
	{	double f;  
		struct {unsigned int lo, hi;};
	} d;
	d.lo =   Int::randomMT();
	d.hi = (Int::randomMT() & 0xfffff) | 0x3ff00000;
	return d.f-1;
}

complexd Complexd::rnd()
{
	return complexd(Float::rnd(),Float::rnd());
}


NodeP lrnd(PairPP) {return new Float(Float::rnd());}

NodeP liSeq(PairPP arg)
{
	int
		a=arg[1]->a_int(1),
		b=arg[2]->a_int(2);

	if (!throwing)
	{	PairP list=0;
		if (a <= b)
			for (int i=b; i>=a; i--)
				list=new Pair(new Int(i), list);
		else
			for (int i=b; i<=a; i++)
				list=new Pair(new Int(i), list);
		
		return list;
	}
	return Unbound;
}

NodeP lrSeq(PairPP arg)
{
	NodeP
		a=arg[1]->a_number(1),
		b=arg[2]->a_number(2);
	int c=arg[3]->a_int(3);

	if (!throwing)

	{	if (!c)
			return 0;

		PairP list=0;

		bool iscomplex=false;
		complexd ac,bc;

		if (a->type==INT_)
			ac=IntP(a)->int_;
		if (a->type==FLOAT_)
			ac=FloatP(a)->float_;
		if (a->type==COMPLEX)
		{	ac=ComplexdP(a)->complex_;
			iscomplex=true;
		}
		if (b->type==INT_)
			bc=IntP(b)->int_;
		if (b->type==FLOAT_)
			bc=FloatP(b)->float_;
		if (b->type==COMPLEX)
		{	bc=ComplexdP(b)->complex_;
			iscomplex=true;
		}

		if (iscomplex)
		{	complexd v=bc, d;
			if (c-1)
				d=(v-ac)/complexd(c-1);
			else
				d=1;

			for (int i=1; i<=c; i++)
			{	list=new Pair(new Complexd(v), list);
				v-=d;
			}
		}else
		{	double v=bc.real(), d;
			if (c-1)
				d=(v-ac.real())/(c-1);
			else
				d=1;

			for (int i=1; i<=c; i++)
			{	list=new Pair(new Float(v), list);
				v-=d;
			}
		}
		return list;
	}
	return Unbound;
}

NodeP lrSeqFloatArray(PairPP arg)
{
	NodeP
		a=arg[1]->a_number(1),
		b=arg[2]->a_number(2);
	int c=arg[3]->a_int(3);

	if (!throwing)
	{	if (!c || c==1)
			return 0;

		double ac=0, bc=0;

		if (a->type==INT_)
			ac=IntP(a)->int_;
		if (a->type==FLOAT_)
			ac=FloatP(a)->float_;
		if (b->type==INT_)
			bc=IntP(b)->int_;
		if (b->type==FLOAT_)
			bc=FloatP(b)->float_;
		double v=ac, d=(v-bc)/(c-1);

		FloatArrayP result=new FloatArray(c);
		doubleP rd=result->data();

		for (int i=1; i<=c; i++)
		{	*rd++=v;
			v-=d;
		}
		return result;
	}
	return Unbound;
}

NodeP lmandelIterate(PairPP arg)
{
	int maxi=arg[3]->a_int(3);
	if throwing return Unbound;

	if (type_of(arg[1])==COMPLEX && type_of(arg[2])==COMPLEX)
	{	complexd 
			z=ComplexdP(arg[1])->complex_, 
			c=ComplexdP(arg[2])->complex_;
		int i=1;
		while (i < maxi)
		{	z*=z;
			z+=c;
			if (Float::sqr(z.real())+Float::sqr(z.imag()) > 4) break;
			i++;
		}
		return new Int(i);
	}
	else if (type_of(arg[1])==ARRAY && type_of(arg[2])==ARRAY)
	{	ArrayP za=ArrayP(arg[1]), ca=ArrayP(arg[2]);
		int l=za->length();
		if (l != ca->length())
		{	error(BAD_DATA, arg[2], 2);
			return Unbound;
		}
		NodePP zd=za->data(), cd=ca->data();
		IntArrayP result=new IntArray(l);
		intP r=result->data();
		while (l--)
		{	if (type_of(*zd)!=COMPLEX)
			{	error(BAD_DATA, arg[2], 2);
				return Unbound;
			}				
			if (type_of(*cd)!=COMPLEX)
			{	error(BAD_DATA, arg[3], 3);
				return Unbound;
			}				
			complexd 
				z=ComplexdP(*zd++)->complex_, 
				c=ComplexdP(*cd++)->complex_;
			int i=1;
			while (i < maxi)
			{	z*=z;
				z+=c;
				if (Float::sqr(z.real())+Float::sqr(z.imag()) > 4) break;
				i++;
			}
			*r++=i;
		}
		return result;
	}
	else
		error(BAD_DATA, arg[1], 1);
	return Unbound;
}

NodeP lCross(PairPP arg)
{
	if (arg[1]->islist())
	{	PairP 
			a=arg[1]->a_list(1),
			b=arg[2]->a_list(2);
		if (!throwing)
		{	if (a->count() != 3)
			{	error(BAD_DATA,a);
				return Unbound;
			}
			if (b->count() != 3)
			{	error(BAD_DATA, b, 2);
				return Unbound;
			}
			Point ap,bp;
			for (int i=0; i<3; i++, next(a), next(b))
			{	double 
					ad=a->first()->a_float(1),
					bd=b->first()->a_float(2);
				if (throwing)
				{	error(BAD_DATA, b, 2);
					return Unbound;
				}
				ap.a[i]=ad;
				bp.a[i]=bd;
			}
			Point cp=ap.X(bp);

			PairP o=0;
			o=new Pair(new Float(cp.z), o);
			o=new Pair(new Float(cp.y), o);
			o=new Pair(new Float(cp.x), o);
			return o;
		}
	}
	else
	{	ArrayP 
			a=arg[1]->a_array(1),
			b=arg[2]->a_array(2);

		if (!throwing)
		{	if (a->length() != 3)
			{	error(BAD_DATA,a);
				return Unbound;
			}
			if (b->length() != 3)
			{	error(BAD_DATA, b, 2);
				return Unbound;
			}
			Point ap,bp;
			for (int i=0; i<3; i++)
			{	double 
					ad=a->data()[i]->a_float(1),
					bd=b->data()[i]->a_float(2);
				if (throwing)
				{	error(BAD_DATA, b, 2);
					return Unbound;
				}
				ap.a[i]=ad;
				bp.a[i]=bd;
			}
			Point cp=ap.X(bp);

			ArrayP o=new Array(3);
			NodePP od=o->data();
			*od++=new Float(cp.x);
			*od++=new Float(cp.y);
			*od++=new Float(cp.z);
			return o;
		}
	}
	return Unbound;
}

double *invertMatrix(double *m, int n, NodeP arg)
{
	double 
		*imb=(double*)GC_::malloc(n*n*sizeof(double)), *im=imb;
	if (!imb)
		return 0;
	double det;

	switch (n)
	{	case 1:	det=*m;
					if (det==0)
					{	error(BAD_DATA, arg);
						break;
					}
					*im=1/det;
					_Determinant->setvar(new Float(det));
					break;

		case 2:	det=m[0+0]*m[1*2+1]-m[0+1]*m[1*2+0];
					if (det==0)
					{	error(BAD_DATA, arg);
						break;
					}
					*im++= m[1*2+1]/det;
					*im++=-m[0  +1]/det;
					*im++=-m[1*2+0]/det;
					*im++= m[0  +0]/det;
					_Determinant->setvar(new Float(det));
					break;
					
		case 3:
			{	double 
					d1=m[1*3+1]*m[2*3+2]-m[2*3+1]*m[1*3+2],
					d2=m[1*3+2]*m[2*3+0]-m[2*3+2]*m[1*3+0],
					d3=m[1*3+0]*m[2*3+1]-m[2*3+0]*m[1*3+1];

				det=m[0+0]*d1+m[0+1]*d2+m[0+2]*d3;
				if (det==0)
				{	error(BAD_DATA, arg);
					break;
				}
				*im++=d1/det;
				*im++=(m[2*3+1]*m[0*3+2]-m[0*3+1]*m[2*3+2])/det;
				*im++=(m[0*3+1]*m[1*3+2]-m[1*3+1]*m[0*3+2])/det;				
				*im++=d2/det;
				*im++=(m[2*3+2]*m[0*3+0]-m[0*3+2]*m[2*3+0])/det;
				*im++=(m[0*3+2]*m[1*3+0]-m[1*3+2]*m[0*3+0])/det;				
				*im++=d3/det;
				*im++=(m[2*3+0]*m[0*3+1]-m[0*3+0]*m[2*3+1])/det;
				*im++=(m[0*3+0]*m[1*3+1]-m[1*3+0]*m[0*3+1])/det;
				_Determinant->setvar(new Float(det));
				break;
			}
		default:
			{	int i;
				for (i=n*n; i>0; i--)
					*im++=*m++;	// copy the matrix (the original maybe used afterwards)
				im=imb;

				det=1;
				for (int l=0; l<n; l++)
				{	int ln=l*n;
					double x=im[ln+l];
					det*=x;
					if (det==0)
					{	error(BAD_DATA, arg);
						return imb;
					}
					im[ln+l]=1;
					for (i=0; i<n; i++)
						im[ln+i]=im[ln+i]/x;
					for (i=0; i<n; i++)
						if (i != l)
						{	int in=i*n;
							x=im[in+l];
							im[in+l]=0;
							for (int j=0; j<n; j++)
								im[in+j]=im[in+j]-x*im[ln+j];
						}
				}
			}				
	}
	return imb;
}

NodeP linvertMatrix(PairPP arg)
{
	ArrayP a=arg[1]->a_array(1);
	if (!throwing)
	{	int x=a->length();
		double 
			*mb=(double*)GC_::malloc(x*x*sizeof(double)), *m=mb;
		if (!mb)
			return Unbound;

		NodePP ad=a->data(), d=ad, de=ad+x-1;
		while (unsigned(d) <= unsigned(de))
		{	ArrayP b=(*d)->a_array(1);
			if (throwing) 
				return Unbound;
			if (b->length() != x) 
			{	error(BAD_DATA, a);
				return Unbound;
			}
			NodePP bd=b->data(), bde=bd+x-1;
			while (unsigned(bd) <= unsigned(bde))
			{	*m=(*bd)->a_float(1);	// argument checking first
				if (throwing)
					return Unbound;
				m++;
				bd++;
			}
			d++;
		}
		double *im=invertMatrix(mb,x,a);
		free(mb);

		if (throwing) 
			return Unbound;

		m=im;
		ArrayP o=new Array(x);
		NodePP od=o->data();
		for (int i=x; i>0; i--)
		{	*od=new Array(x);
			NodePP odd=ArrayP(*od)->data();
			for (int j=x; j>0; j--)
			{	*odd=new Float(*m++);
				odd++;
			}
			od++;
		}
		free(im);
		return o;
	}
	return Unbound;
}

#define x87_ov_code	0x0D

#ifdef WIN32
	#define FLOAT_ASM																		\
		_asm																					\
		{	_asm fnstsw ax _asm fnclex													\
			_asm test ax, x87_ov_code _asm jnz float_overflow					\
		}
#else
	#define FLOAT_ASM																		\
		int of;																				\
		asm("	fnstsw %%ax				\n													\
				fnclex					\n													\
				test %1,%%ax			\n													\
				jnz 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:"						\
				 : "=r" (of) : "g" (x87_ov_code));									\
		if (of) goto float_overflow;
#endif

#ifdef WIN32
	#define COMPLEX_ASM																	\
		_asm																					\
		{	_asm fnstsw ax _asm fnclex													\
			_asm test ax, x87_ov_code _asm jnz complex_overflow				\
		}
#else
	#define COMPLEX_ASM																	\
		int of;																				\
		asm("	fnstsw %%ax				\n													\
				fnclex					\n													\
				test %1,%%ax			\n													\
				jnz 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:"						\
				 : "=r" (of) : "g" (x87_ov_code));									\
		if (of) goto complex_overflow;
#endif

#ifdef WIN32
	#define INT_ASM	\
		_asm{_asm jo int_overflow}
#else
	#define INT_ASM	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int_overflow;
#endif

#ifdef WIN32
	#define INT_ASM2	\
		_asm{_asm jo int_overflow2}
#else
	#define INT_ASM2	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int_overflow2;
#endif

#ifdef WIN32
	#define INT_ASM3	\
		_asm{_asm jo int_overflow3}
#else
	#define INT_ASM3	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int_overflow3;
#endif

#ifdef WIN32
	#define INT_ASM4	\
		_asm{_asm jo int_overflow4}
#else
	#define INT_ASM4	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int_overflow4;
#endif

#ifdef WIN32
	#define INT16_ASM	\
		_asm{_asm jo int16_overflow}
#else
	#define INT16_ASM	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int16_overflow;
#endif

#ifdef WIN32
	#define INT16_ASM2	\
		_asm{_asm jo int16_overflow2}
#else
	#define INT16_ASM2	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int16_overflow2;
#endif

#ifdef WIN32
	#define INT16_ASM3	\
		_asm{_asm jo int16_overflow3}
#else
	#define INT16_ASM3	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int16_overflow3;
#endif

#ifdef WIN32
	#define INT8_ASM	\
		_asm{_asm jo int8_overflow}
#else
	#define INT8_ASM	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int8_overflow;
#endif

#ifdef WIN32
	#define INT8_ASM2	\
		_asm{_asm jo int8_overflow2}
#else
	#define INT8_ASM2	\
		int of;			\
		asm("jo 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:": "=r" (of));	\
		if (of) goto int8_overflow2;
#endif

#define TRY_FUNC(NAME, NUMBER)														\
{	double v;																				\
	if (true)																				\
	{	FUNC_##NAME(NUMBER);																\
		FLOAT_ASM																			\
	}else float_overflow:																\
	{	Float::reset_FPU();																\
		error(BAD_DATA, this);															\
		return Unbound;																	\
	}																							\
	return new Float(v);																	\
}

#define TRY_FUNCI(NAME, NUMBER)														\
{	int v;																					\
	if (true)																				\
	{	FUNC_i##NAME(NUMBER);															\
		INT_ASM																				\
	}else int_overflow:																	\
		TRY_FUNC(NAME, double(int_))													\
	return new Int(v);																	\
}

#define TRY_FUNCI16(NAME, NUMBER)													\
{	short v;																					\
	if (true)																				\
	{	FUNC_i16##NAME(NUMBER);															\
		INT16_ASM																			\
	}else int16_overflow:																\
		TRY_FUNCI(NAME, int(int_))														\
	return new Int(v);																	\
}

#define TRY_FUNCC(NAME, NUMBER)														\
{	complexd v;																				\
	if (true)																				\
	{	FUNC_c##NAME(NUMBER);															\
		FLOAT_ASM																			\
	}else float_overflow:																\
	{	Float::reset_FPU();																\
		error(BAD_DATA, this);															\
		return Unbound;																	\
	}																							\
	return new Complexd(v);																\
}

#define UNARY_OP(NAME, SYMBOLNODE)													\
NodeP Node::NAME() {error(BAD_DATA, this); return Unbound;}					\
NodeP Object::NAME()				 														\
{	if (!bit.permanent  ||  prim()->priority <= priority_prefix)			\
		return name()->NAME();															\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP String::NAME()																		\
{	NodeP n=number();																		\
	if (n != Unbound) return n->NAME();												\
	else return new Pair(SYMBOLNODE, this);										\
}																								\
NodeP Int::NAME()																			\
{	int v;																					\
	FUNC_i##NAME(int_);																	\
	INT_ASM																					\
	return new Int(v);																	\
int_overflow:																				\
	TRY_FUNC(NAME, double(int_))														\
}																								\
NodeP Int16::NAME()																		\
{	short v;																					\
	FUNC_i16##NAME(int_);																\
	INT16_ASM																				\
	return new Int16(v);																	\
int16_overflow:																			\
	TRY_FUNCI(NAME, int(int_))															\
}																								\
NodeP Int8::NAME()																		\
{	char v;																					\
	FUNC_i8##NAME(int_);																	\
	INT8_ASM																					\
	return new Int8(v);																	\
int8_overflow:																				\
	TRY_FUNCI16(NAME, short(int_))													\
}																								\
NodeP UInt8::NAME()																		\
{	char v;																					\
	FUNC_i8##NAME(int_);																	\
	INT8_ASM																					\
	return new UInt8(v);																	\
int8_overflow:																				\
	TRY_FUNCI16(NAME, short(int_))													\
}																								\
NodeP Float::NAME() {TRY_FUNC(NAME, float_);}									\
NodeP BigFloat::NAME()																	\
{	MAPM v;																					\
	if (true)																				\
	{	FUNC_b##NAME(float_);															\
		if (m_apm_error)																	\
			goto bigfloat_overflow;														\
	}else bigfloat_overflow:															\
	{	m_apm_reset_error();																\
		error(BAD_DATA, this);															\
		return Unbound;																	\
	}																							\
	return new BigFloat(v);																\
}																								\
NodeP Complexd::NAME() {TRY_FUNCC(NAME, complex_);}							\
NodeP Pair::NAME()																		\
{																								\
	PairP a=this, result=0, lastnode=0, tnode=0;									\
	while (a)																				\
	{	tnode = new Pair(a->first()->NAME(), 0);									\
		if (throwing) return Unbound;													\
		if (result)																			\
			lastnode->settail( tnode );												\
		else																					\
			result=tnode;																	\
		lastnode=tnode;																	\
		next(a);																				\
	}																							\
	if (!a) return result;																\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Array::NAME()																		\
{																								\
	NodePP d=data();																		\
	int i=length();																		\
	ArrayP result = new Array(i, origin());										\
	NodePP rd=result->data();															\
	while (i--)																				\
	{	if (*d)																				\
			*rd = (*d)->NAME();															\
		else																					\
		{	error(BAD_DATA, this);														\
			return Unbound;																\
		}																						\
		*d++;																					\
		result->check_oldyoung(*rd++);												\
		if (throwing) return Unbound;													\
	}																							\
	return result;																			\
}																								\
NodeP IntArray::NAME()																	\
{																								\
	intP d=_data;																			\
	int i=length();																		\
	IntArrayP result = new IntArray(i, origin());								\
	intP rd=result->_data;																\
	IntP element=new Int(0);															\
	NodeP res=0;																			\
	while (i--)																				\
	{	element->int_=(*d++);															\
		res = element->NAME();															\
		if (type_of(res) != INT_)														\
		{	error(BAD_DATA, this);														\
			return Unbound;																\
		}																						\
		(*rd++) = IntP(res)->int_;														\
		if (throwing) return Unbound;													\
	}																							\
	return result;																			\
}																								\
NodeP Int16Array::NAME()																\
{																								\
	shortP d=_data;																		\
	int i=length();																		\
	Int16ArrayP result = new Int16Array(i, origin());							\
	shortP rd=result->_data;															\
	Int16P element=new Int16(0);														\
	NodeP res=0;																			\
	while (i--)																				\
	{	element->int_=(*d++);															\
		res = element->NAME();															\
		if (type_of(res) != INT16)														\
		{	error(BAD_DATA, this);														\
			return Unbound;																\
		}																						\
		(*rd++) = Int16P(res)->int_;													\
		if (throwing) return Unbound;													\
	}																							\
	return result;																			\
}																								\
NodeP FloatArray::NAME()																\
{																								\
	doubleP d=_data;																		\
	int i=length();																		\
	FloatArrayP result = new FloatArray(i, origin());							\
	doubleP rd=result->_data;															\
	NodeP res=0;																			\
	while (i--)																				\
	{	double a=(*d++), v;																\
		if (true)																			\
		{	FUNC_##NAME(a);																\
			FLOAT_ASM																		\
		}else float_overflow:															\
		{	Float::reset_FPU();															\
			error(BAD_DATA, this);														\
			return Unbound;																\
		}																						\
		(*rd++) = v;																		\
		if (throwing) return Unbound;													\
	}																							\
	return result;																			\
}																								\
NodeP l##NAME(PairPP arg) {return arg[1]?  arg[1]->NAME() : 0;}

#define FUNC_BitNot(a)		goto float_overflow
#define FUNC_bBitNot(a)		goto bigfloat_overflow
#define FUNC_cBitNot(a)		goto float_overflow
#define FUNC_iBitNot(a)		v=~a
#define FUNC_i16BitNot(a)	v=~a
#define FUNC_i8BitNot(a)	v=~a

#define FUNC_random(a)		goto float_overflow
#define FUNC_brandom(a)		random()
#define FUNC_crandom(a)		goto float_overflow
#define FUNC_irandom(a)		v=Int::random(a)
#define FUNC_i16random(a)	v=short(Int::random(a))
#define FUNC_i8random(a)	v=char(Int::random(a))

#define FUNC_truncate(a)	if (a >= Int::min && a <= Int::max)	\
	v=int(a); else v=floor(a)
#define FUNC_btruncate(a)	v=floor(a)
#define FUNC_ctruncate(a)										\
	if (a.imag()==0 && Complexd::abs(a) <= Int::max)	\
		return new Int(int(a.real()));						\
	else v=complexd(floor(a.real()), floor(a.imag()))
#define FUNC_itruncate(a)		return this
#define FUNC_i16truncate(a)	return this
#define FUNC_i8truncate(a)		return this

#define FUNC__int(a)		if (a >= Int::min && a <= Int::max)	\
	return new Int(int(a)); else goto float_overflow
#define FUNC_b_int(a)	if (a >= Int::min && a <= Int::max)	\
	return new Int(a.to_int()); else goto bigfloat_overflow
#define FUNC_c_int(a)	if (Complexd::abs(a) <= Int::max)	\
	return new Int(int(abs(a))); else goto float_overflow
#define FUNC_i_int(a)	v=a
#define FUNC_i16_int(a)	return new Int(a)
#define FUNC_i8_int(a)	return new Int(a)

#define FUNC_toint16(a)		if (a >= Int16::min && a <= Int16::max)	\
	return new Int16(short(a)); else goto float_overflow
#define FUNC_btoint16(a)	if (a >= Int16::min && a <= Int16::max)	\
	return new Int16(a.to_int()); else goto bigfloat_overflow
#define FUNC_ctoint16(a)	if (Complexd::abs(a) <= Int16::max)	\
	return new Int16(short(abs(a))); else goto float_overflow
#define FUNC_itoint16(a)	if (a >= Int16::min && a <= Int16::max)	\
	return new Int16(short(a)); else goto int_overflow
#define FUNC_i16toint16(a)	v=a
#define FUNC_i8toint16(a)	return new Int16(a)

#define FUNC_toint8(a)	if (a >= Int8::min && a <= Int8::max)	\
	return new Int8(char(a)); else goto float_overflow
#define FUNC_btoint8(a)	if (a >= Int8::min && a <= Int8::max)	\
	return new Int8(a.to_int()); else goto bigfloat_overflow
#define FUNC_ctoint8(a)	if (Complexd::abs(a) <= Int8::max)	\
	return new Int8(char(abs(a))); else goto float_overflow
#define FUNC_itoint8(a)	if (a >= Int8::min && a <= Int8::max)	\
	return new Int8(char(a)); else goto int_overflow
#define FUNC_i16toint8(a)	if (a >= Int8::min && a <= Int8::max)	\
	return new Int8(char(a)); else goto int16_overflow
#define FUNC_i8toint8(a)	v=a

#define FUNC_touint8(a)	if (a >= 0 && a <= UInt8::max)	\
	return new UInt8((unsigned char)(a)); else goto float_overflow
#define FUNC_btouint8(a)	if (a >= 0 && a <= UInt8::max)	\
	return new UInt8(a.to_int()); else goto bigfloat_overflow
#define FUNC_ctouint8(a)	if (Complexd::abs(a) <= UInt8::max)	\
	return new UInt8((unsigned char)(abs(a))); else goto float_overflow
#define FUNC_itouint8(a)	if (a >= 0 && a <= UInt8::max)	\
	return new UInt8((unsigned char)(a)); else goto int_overflow
#define FUNC_i16touint8(a)	if (a >= 0 && a <= UInt8::max)	\
	return new UInt8((unsigned char)(a)); else goto int16_overflow
#define FUNC_i8touint8(a)	v=a


#define FUNC_round(a)		if (Float::abs(a) <= Int::max)	\
	return new Int(int(floor(a+0.5))); else v=floor(a+0.5)
#define FUNC_bround(a)		v=floor(a+0.5)
#define FUNC_cround(a)										\
	if (a.imag()==0 && Complexd::abs(a) <= Int::max)	\
		return new Int(int(floor(a.real()+0.5)));				\
	else v=complexd(floor(a.real()+0.5),floor(a.real()+0.5))
#define FUNC_iround(a)		return this
#define FUNC_i16round(a)	return this
#define FUNC_i8round(a)		return this

#define FUNC__float(a)		v=a;
#define FUNC_b_float(a)		return new Float(a.to_double());
#define FUNC_c_float(a)		return new Float(abs(a));
#define FUNC_i_float(a)		return new Float(a);
#define FUNC_i16_float(a)	return new Float(a);
#define FUNC_i8_float(a)	return new Float(a);

#define FUNC__bigfloat(a)		return new BigFloat(a);
#define FUNC_b_bigfloat(a)		return new BigFloat(a);
#define FUNC_c_bigfloat(a)		goto float_overflow;
#define FUNC_i_bigfloat(a)		return new BigFloat(a);
#define FUNC_i16_bigfloat(a)	return new BigFloat(a);
#define FUNC_i8_bigfloat(a)	return new BigFloat(a);

#define FUNC_Sin(a)		v=::sin(Float::Remainder(a,360)*Float::deg2rad)
#define FUNC_bSin(a)		v=::sin(BigFloat::Remainder(a,360)*BigFloat::deg2rad)
#define FUNC_cSin(a)		v=std::sin(Complexd::Remainder(a,360)*Float::deg2rad)
#define FUNC_iSin(a)		goto int_overflow
#define FUNC_i16Sin(a)	goto int_overflow
#define FUNC_i8Sin(a)	goto int_overflow
#define FUNC_Cos(a)		v=::cos(Float::Remainder(a,360)*Float::deg2rad)
#define FUNC_bCos(a)		v=::cos(BigFloat::Remainder(a,360)*BigFloat::deg2rad)
#define FUNC_cCos(a)		v=std::cos(Complexd::Remainder(a,360)*Float::deg2rad)
#define FUNC_iCos(a)		goto int_overflow
#define FUNC_i16Cos(a)	goto int_overflow
#define FUNC_i8Cos(a)	goto int_overflow
#define FUNC_Tan(a)		v=::tan(Float::Remainder(a,360)*Float::deg2rad)
#define FUNC_bTan(a)		v=::tan(BigFloat::Remainder(a,360)*BigFloat::deg2rad)
#define FUNC_cTan(a)		complexd c=Complexd::Remainder(a,360)*Float::deg2rad; \
	v=std::sin(c)/std::cos(c)
#define FUNC_iTan(a)		goto int_overflow
#define FUNC_i16Tan(a)	goto int_overflow
#define FUNC_i8Tan(a)	goto int_overflow

#define FUNC_ArcSin(a)		v=::asin(a)/Float::deg2rad
#define FUNC_bArcSin(a)		v=::asin(a)/BigFloat::deg2rad
#define FUNC_cArcSin(a)		goto float_overflow
#define FUNC_iArcSin(a)		goto int_overflow
#define FUNC_i16ArcSin(a)	goto int_overflow
#define FUNC_i8ArcSin(a)	goto int_overflow
#define FUNC_ArcCos(a)		v=::acos(a)/Float::deg2rad
#define FUNC_bArcCos(a)		v=::acos(a)/BigFloat::deg2rad
#define FUNC_cArcCos(a)		goto float_overflow
#define FUNC_iArcCos(a)		goto int_overflow
#define FUNC_i16ArcCos(a)	goto int_overflow
#define FUNC_i8ArcCos(a)	goto int_overflow

#define FUNC_radSin(a)		v=::sin(a)
#define FUNC_bradSin(a)		v=::sin(a)
#define FUNC_cradSin(a)		v=std::sin(a)
#define FUNC_iradSin(a)		goto int_overflow
#define FUNC_i16radSin(a)	goto int_overflow
#define FUNC_i8radSin(a)	goto int_overflow
#define FUNC_radCos(a)		v=::cos(a)
#define FUNC_bradCos(a)		v=::cos(a)
#define FUNC_cradCos(a)		v=std::cos(a)
#define FUNC_iradCos(a)		goto int_overflow
#define FUNC_i16radCos(a)	goto int_overflow
#define FUNC_i8radCos(a)	goto int_overflow
#define FUNC_radTan(a)		v=::tan(a)
#define FUNC_bradTan(a)		v=::tan(a)
#define FUNC_cradTan(a)		v=std::sin(a)/std::cos(a)
#define FUNC_iradTan(a)		goto int_overflow
#define FUNC_i16radTan(a)	goto int_overflow
#define FUNC_i8radTan(a)	goto int_overflow

#define FUNC_radArcSin(a)		v=::asin(a)
#define FUNC_bradArcSin(a)		v=::asin(a)
#define FUNC_cradArcSin(a)		goto float_overflow
#define FUNC_iradArcSin(a)		goto int_overflow
#define FUNC_i16radArcSin(a)	goto int_overflow
#define FUNC_i8radArcSin(a)	goto int_overflow
#define FUNC_radArcCos(a)		v=::acos(a)
#define FUNC_bradArcCos(a)		v=::acos(a)
#define FUNC_cradArcCos(a)		goto float_overflow
#define FUNC_iradArcCos(a)		goto int_overflow
#define FUNC_i16radArcCos(a)	goto int_overflow
#define FUNC_i8radArcCos(a)	goto int_overflow

#define FUNC_abs(a)		v=fabs(a)
#define FUNC_babs(a)		v=fabs(a)
#define FUNC_cabs(a)		return new Float(abs(a))
#define FUNC_iabs(a)		v=::abs(a)
#define FUNC_i16abs(a)	v=::abs(a)
#define FUNC_i8abs(a)	v=::abs(a)
#define FUNC_Sqr(a)		v=a*a
#define FUNC_bSqr(a)		v=a*a
#define FUNC_cSqr(a)		v=a*a
#define FUNC_iSqr(a)		v=a*a
#define FUNC_i16Sqr(a)	v=a*a
#define FUNC_i8Sqr(a)	v=a*a
#define FUNC_Sqrt(a)		if (a >= 0) v=::sqrt(a); \
	else return new Complexd(std::sqrt(complexd(a)));
#define FUNC_bSqrt(a)	if (a >= 0) v=::sqrt(a); else goto bigfloat_overflow
#define FUNC_cSqrt(a)	v=std::sqrt(a)
#define FUNC_iSqrt(a)	goto int_overflow
#define FUNC_i16Sqrt(a)	goto int_overflow
#define FUNC_i8Sqrt(a)	goto int_overflow
#define FUNC_exp(a)		v=::exp(a)
#define FUNC_bexp(a)		v=::exp(a)
#define FUNC_cexp(a)		v=std::exp(a)
#define FUNC_iexp(a)		goto int_overflow
#define FUNC_i16exp(a)	goto int_overflow
#define FUNC_i8exp(a)	goto int_overflow

#define FUNC_LN(a)		v=log(a)
#define FUNC_bLN(a)		v=log(a)
#define FUNC_cLN(a)		v=std::log(a)
#define FUNC_iLN(a)		goto int_overflow
#define FUNC_i16LN(a)	goto int_overflow
#define FUNC_i8LN(a)		goto int_overflow
#define FUNC_Log10(a)		v=::log10(a)
#define FUNC_bLog10(a)		v=::log10(a)
#define FUNC_cLog10(a)		v=std::log10(a)
#define FUNC_iLog10(a)		goto int_overflow

#define FUNC_i16Log10(a)	goto int_overflow
#define FUNC_i8Log10(a)		goto int_overflow

#define FUNC_real(a)		v=a
#define FUNC_breal(a)	v=a
#define FUNC_creal(a)	return new Float(a.real())
#define FUNC_ireal(a)	v=a
#define FUNC_i16real(a)	v=a
#define FUNC_i8real(a)	v=a
#define FUNC_imag(a)		v=0
#define FUNC_bimag(a)	v=0
#define FUNC_cimag(a)	return new Float(a.imag())
#define FUNC_iimag(a)	v=0
#define FUNC_i16imag(a)	v=0
#define FUNC_i8imag(a)	v=0
#define FUNC_conjugate(a)		v=a
#define FUNC_bconjugate(a)		v=a
#define FUNC_cconjugate(a)		v=complexd(a.real(),-a.imag())
#define FUNC_iconjugate(a)		v=a
#define FUNC_i16conjugate(a)	v=a
#define FUNC_i8conjugate(a)	v=a

#define FUNC_Signum(a)	if (a>0) return new Int(1); \
								else if (a<0) return new Int(-1); \
								else return new Int(0)
#define FUNC_bSignum(a)	FUNC_Signum(a)
#define FUNC_cSignum(a)	if (a.imag()!=0) goto float_overflow; \
								else if (a.real()>0) return new Int(1); \
								else if (a.real()<0) return new Int(-1); \
								else return new Int(0)
#define FUNC_iSignum(a)	a>0? v=1 \
									: a<0? v=-1 \
										: v=0
#define FUNC_i16Signum(a)	FUNC_iSignum(a)
#define FUNC_i8Signum(a)	FUNC_iSignum(a)


UNARY_OP(BitNot, _BitNot_)
UNARY_OP(random, _Random_)
UNARY_OP(_int, _Int_)
UNARY_OP(toint16, _Int16_)
UNARY_OP(toint8, _Int8_)
UNARY_OP(touint8, _UInt8_)
UNARY_OP(truncate, _Truncate_)
UNARY_OP(round, _Round_)
UNARY_OP(_float, _Float_)
UNARY_OP(_bigfloat, _BigFloat_)
UNARY_OP(Sin, _Sin_)
UNARY_OP(Cos, _Cos_)
UNARY_OP(Tan, _Tan_)
UNARY_OP(ArcSin, _ArcSin_)
UNARY_OP(ArcCos, _ArcCos_)
UNARY_OP(radSin, _radSin_)
UNARY_OP(radCos, _radCos_)
UNARY_OP(radTan, _radTan_)
UNARY_OP(radArcSin, _radArcSin_)
UNARY_OP(radArcCos, _radArcCos_)
UNARY_OP(abs, _Abs_)
UNARY_OP(Sqr, _Sqr_)
UNARY_OP(Sqrt, _Sqrt_)
UNARY_OP(exp, _Exp_)
UNARY_OP(LN, _LN_)
UNARY_OP(Log10, _Log10_)
UNARY_OP(real, _Real_)
UNARY_OP(imag, _Imag_)
UNARY_OP(conjugate, _Conjugate_)
UNARY_OP(Signum, _Signum_)

#ifdef WIN32
	#define BINARY_FLOAT_ASM(TYPE)													\
		_asm																					\
		{	_asm fnstsw ax _asm fnclex													\
			_asm test ax, x87_ov_code _asm jnz float_overflow##TYPE			\
		}
#else
	#define BINARY_FLOAT_ASM(TYPE)													\
		int of;																				\
		asm("	fnstsw %%ax				\n													\
				fnclex					\n													\
				test %1,%%ax			\n													\
				jnz 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:"						\
				 : "=r" (of) : "g" (x87_ov_code));									\
		if (of) goto float_overflow##TYPE;
#endif

#ifdef WIN32
	#define BINARY_COMPLEX_ASM(TYPE)													\
		_asm																					\
		{	_asm fnstsw ax _asm fnclex													\
			_asm test ax, x87_ov_code _asm jnz complex_overflow##TYPE		\
		}
#else
	#define BINARY_COMPLEX_ASM(TYPE)													\
		int of;																				\
		asm("	fnstsw %%ax				\n													\
				fnclex					\n													\
				test %1,%%ax			\n													\
				jnz 1f; mov $0,%0; jmp 2f; 1: mov $1,%0; 2:"						\
				 : "=r" (of) : "g" (x87_ov_code));									\
		if (of) goto complex_overflow##TYPE;
#endif

#define NR_if_(TYPE, NAME, NUMBER, ARG)											\
	if (arg->type==TYPE)																	\
	{	double f;																			\
		if (true)																			\
		{	f = FUNC_##NAME(NUMBER, ARG);												\
			BINARY_FLOAT_ASM(TYPE)														\
		}else float_overflow##TYPE:													\
		{	Float::reset_FPU();															\
			error(BAD_DATA, arg, 2);													\
			return Unbound;																\
		}																						\
		return new Float(f);																\
	}																							\

#define BNR_if_(TYPE, NAME, NUMBER, ARG)											\
	if (arg->type==TYPE)																	\
	{	MAPM f;																				\
		if (true)																			\
		{	f = FUNC_b##NAME(NUMBER, ARG);											\
			if (m_apm_error)																\
				goto bigfloat_overflow##TYPE;											\
		}else bigfloat_overflow##TYPE:												\
		{	m_apm_reset_error();															\
			error(BAD_DATA, arg, 2);													\
			return Unbound;																\
		}																						\
		return new BigFloat(f);															\
	}																							\

#define CNR_if_(TYPE, NAME, NUMBER, ARG)											\
	if (arg->type==TYPE)																	\
	{	complexd c;																			\
		if (true)																			\
		{	c = FUNC_c##NAME(NUMBER, ARG);											\
			BINARY_COMPLEX_ASM(TYPE)													\
		}else complex_overflow##TYPE:													\
		{	Float::reset_FPU();															\
			error(BAD_DATA, arg, 2);													\
			return Unbound;																\
		}																						\
		return new Complexd(c);															\
	}																							\

#define else_if_STRING(NAME, SYMBOLNODE)											\
	else if (arg->is.string )															\
	{	NodeP n=0;  n=arg->number();													\
		if (n != Unbound) return NAME(n);											\
		else return new List(this, SYMBOLNODE, arg, _ListEnd);				\
	}

#define else_if_OBJECT(NAME)															\
	else if (arg->type==OBJECT)														\
	{	ObjectP obj=ObjectP(arg);														\
		if (!ObjectP(obj)->bit.permanent												\
		||  obj->prim()->priority <= priority_prefix)							\
			return NAME(obj->name());													\
	}

#define else_if_LIST(NAME)																\
	else if (arg->is.list)																\
	{	NodeP a=this;																		\
		PairP b = PairP(arg);															\
		while (b)																			\
		{	a = a->NAME( b->first() );													\
			if (throwing) return Unbound;												\
			next(b);																			\
		}																						\
		if (!b) return a;																	\
		error(BAD_DATA, arg, 2); return Unbound;									\
	}

#define else_if_ARRAY(NAME)															\
	else if (arg->type==ARRAY)															\
	{	NodeP a=this;																		\
		NodePP d = ArrayP(arg)->data();												\
		int i=ArrayP(arg)->length();													\
		while (i--)																			\
		{	a = a->NAME(*d++);															\
			if (throwing) return Unbound;												\
		}																						\
		return a;																			\
	}

#define else_if_FLOATARRAY(NAME)														\
	else if (arg->type==FLOATARRAY)													\
	{	double f=getnumber();															\
		doubleP d = FloatArrayP(arg)->data();										\
		int i=FloatArrayP(arg)->length();											\
		while (i--)																			\
		{	if (true)																		\
			{	f = FUNC_##NAME(f, (*d++));											\
				BINARY_FLOAT_ASM(FloatArray)											\
			}else float_overflowFloatArray:											\
			{	Float::reset_FPU();														\
				error(BAD_DATA, arg, 2);												\
				return Unbound;															\
			}																					\
		}																						\
		return new Float(f);																\
	}

#define BINARY_OP(NAME, SYMBOLNODE, NEUTRAL)										\
NodeP Node::NAME(NodeP arg) {error(BAD_DATA, this); return Unbound;}		\
NodeP Object::NAME(NodeP arg) 														\
{	if (!bit.permanent  ||  prim()->priority <= priority_prefix)			\
		return name()->NAME(arg);														\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP String::NAME(NodeP arg)															\
{	NodeP n=number();																		\
	if (n != Unbound) return n->NAME(arg);											\
	else return new List(this, SYMBOLNODE, arg, _ListEnd);					\
}																								\
NodeP Int::NAME(NodeP arg)																\
{	if (arg)																					\
	{	if (arg->type==INT_)																\
		{	int iarg=IntP(arg)->int_;													\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM																		\
				return new Int(nval);													\
			}else																				\
	int_overflow:																			\
				NR_if_(INT_, NAME, double(int_), iarg)								\
		}																						\
		 NR_if_( FLOAT_, NAME,        int_ ,    FloatP(arg)->float_  )		\
		CNR_if_(COMPLEX, NAME, double(int_), ComplexdP(arg)->complex_)		\
		BNR_if_(BIGFLOAT,NAME,   MAPM(int_), BigFloatP(arg)->float_)		\
		else if (arg->type==INT16)														\
		{	short iarg=Int16P(arg)->int_;												\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM2																		\
				return new Int(nval);													\
			}else																				\
	int_overflow2:																			\
				NR_if_(INT16, NAME, double(int_), iarg)							\
		}																						\
		else if (arg->type==INT8)														\
		{	char iarg=Int8P(arg)->int_;												\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM3																		\
				return new Int(nval);													\
			}else																				\
	int_overflow3:																			\
				NR_if_(INT8, NAME, double(int_), iarg)								\
		}																						\
		else if (arg->type==UINT8)														\
		{	unsigned char iarg=UInt8P(arg)->int_;									\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM4																		\
				return new Int(nval);													\
			}else																				\
	int_overflow4:																			\
				NR_if_(UINT8, NAME, double(int_), iarg)							\
		}																						\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	}																							\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Int16::NAME(NodeP arg)															\
{	if (arg)																					\
	{	if (arg->type==INT_)																\
		{	int iarg=IntP(arg)->int_;													\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM																		\
				return new Int(nval);													\
			}else																				\
	int_overflow:																			\
				NR_if_(INT_, NAME, double(int_), iarg)								\
		}																						\
		 NR_if_( FLOAT_, NAME,        int_ ,   FloatP(arg)->float_  )		\
		CNR_if_(COMPLEX, NAME, double(int_), ComplexdP(arg)->complex_)		\
		BNR_if_(BIGFLOAT,NAME,   MAPM(int_), BigFloatP(arg)->float_)		\
		else if (arg->type==INT16)														\
		{	short iarg=Int16P(arg)->int_;												\
			if (true)																		\
			{	short nval = FUNC_i16##NAME(int_, iarg);							\
				INT16_ASM																	\
				return new Int16(nval);													\
			}else																				\
	int16_overflow:																		\
			{	int iarg=Int16P(arg)->int_;											\
				if (true)																	\
				{	int nval = FUNC_i##NAME(int_, iarg);							\
					INT_ASM2																	\
					return new Int(nval);												\
				}else																			\
	int_overflow2:																			\
					NR_if_(INT16, NAME, double(int_), iarg)						\
			}																					\
		}																						\
		else if (arg->type==INT8)														\
		{	char iarg=Int8P(arg)->int_;												\
			if (true)																		\
			{	short nval = FUNC_i16##NAME(int_, iarg);							\
				INT16_ASM2																	\
				return new Int16(nval);													\
			}else																				\
	int16_overflow2:																		\
			{	int iarg=Int8P(arg)->int_;												\
				if (true)																	\
				{	int nval = FUNC_i##NAME(int_, iarg);							\
					INT_ASM3																	\
					return new Int(nval);												\
				}else																			\
	int_overflow3:																			\
					NR_if_(INT8, NAME, double(int_), iarg)							\
			}																					\
		}																						\
		else if (arg->type==UINT8)														\
		{	unsigned char iarg=UInt8P(arg)->int_;									\
			if (true)																		\
			{	short nval = FUNC_i16##NAME(int_, iarg);							\
				INT16_ASM3																	\
				return new Int16(nval);													\
			}else																				\
	int16_overflow3:																		\
			{	int iarg=UInt8P(arg)->int_;											\
				if (true)																	\
				{	int nval = FUNC_i##NAME(int_, iarg);							\
					INT_ASM4																	\
					return new Int(nval);												\
				}else																			\
	int_overflow4:																			\
					NR_if_(UINT8, NAME, double(int_), iarg)						\
			}																					\
		}																						\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	}																							\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Int8::NAME(NodeP arg)															\
{	if (arg)																					\
	{	if (arg->type==INT_)																\
		{	int iarg=IntP(arg)->int_;													\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM																		\
				return new Int(nval);													\
			}else																				\
	int_overflow:																			\
				NR_if_(INT_, NAME, double(int_), iarg)								\
		}																						\
		 NR_if_( FLOAT_, NAME,        int_ ,   FloatP(arg)->float_  )		\
		CNR_if_(COMPLEX, NAME, double(int_), ComplexdP(arg)->complex_)		\
		BNR_if_(BIGFLOAT,NAME,   MAPM(int_), BigFloatP(arg)->float_)		\
		else if (arg->type==INT16)														\
		{	short iarg=Int16P(arg)->int_;												\
			if (true)																		\
			{	short nval = FUNC_i16##NAME(int_, iarg);							\
				INT16_ASM																	\
				return new Int16(nval);													\
			}else																				\
	int16_overflow:																		\
			{	int iarg=Int16P(arg)->int_;											\
				if (true)																	\
				{	int nval = FUNC_i##NAME(int_, iarg);							\
					INT_ASM2																	\
					return new Int(nval);												\
				}else																			\
	int_overflow2:																			\
					NR_if_(INT16, NAME, double(int_), iarg)						\
			}																					\
		}																						\
		else if (arg->type==INT8)														\
		{	char iarg=Int8P(arg)->int_;												\
			if (true)																		\
			{	char nval = FUNC_i8##NAME(int_, iarg);								\
				INT8_ASM																		\
				return new Int8(nval);													\
			}else																				\
	int8_overflow:																			\
			{	short iarg=Int8P(arg)->int_;											\
				if (true)																	\
				{	short nval = FUNC_i16##NAME(int_, iarg);						\
					INT16_ASM2																\
					return new Int16(nval);												\
				}else																			\
	int16_overflow2:																		\
				{	int iarg=Int8P(arg)->int_;											\
					if (true)																\
					{	int nval = FUNC_i##NAME(int_, iarg);						\
						INT_ASM3																\
						return new Int(nval);											\
					}else																		\
		int_overflow3:																		\
						NR_if_(INT8, NAME, double(int_), iarg)						\
				}																				\
			}																					\
		}																						\
		else if (arg->type==UINT8)														\
		{	unsigned char iarg=UInt8P(arg)->int_;									\
			if (true)																		\
			{	unsigned char nval = FUNC_i8##NAME(int_, iarg);					\
				INT8_ASM2																	\
				return new UInt8(nval);													\
			}else																				\
	int8_overflow2:																		\
			{	short iarg=UInt8P(arg)->int_;											\
				if (true)																	\
				{	short nval = FUNC_i16##NAME(int_, iarg);						\
					INT16_ASM3																\
					return new Int16(nval);												\
				}else																			\
	int16_overflow3:																		\
				{	int iarg=UInt8P(arg)->int_;										\
					if (true)																\
					{	int nval = FUNC_i##NAME(int_, iarg);						\
						INT_ASM4																\
						return new Int(nval);											\
					}else																		\
		int_overflow4:																		\
						NR_if_(UINT8, NAME, double(int_), iarg)					\
				}																				\
			}																					\
		}																						\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	}																							\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP UInt8::NAME(NodeP arg)															\
{	if (arg)																					\
	{	if (arg->type==INT_)																\
		{	int iarg=IntP(arg)->int_;													\
			if (true)																		\
			{	int nval = FUNC_i##NAME(int_, iarg);								\
				INT_ASM																		\
				return new Int(nval);													\
			}else																				\
	int_overflow:																			\
				NR_if_(INT_, NAME, double(int_), iarg)								\
		}																						\
		 NR_if_( FLOAT_, NAME,        int_ ,   FloatP(arg)->float_  )		\
		CNR_if_(COMPLEX, NAME, double(int_), ComplexdP(arg)->complex_)		\
		BNR_if_(BIGFLOAT,NAME,   MAPM(int_), BigFloatP(arg)->float_)		\
		else if (arg->type==INT16)														\
		{	short iarg=Int16P(arg)->int_;												\
			if (true)																		\
			{	short nval = FUNC_i16##NAME(int_, iarg);							\
				INT16_ASM																	\
				return new Int16(nval);													\
			}else																				\
	int16_overflow:																		\
			{	int iarg=Int16P(arg)->int_;											\
				if (true)																	\
				{	int nval = FUNC_i##NAME(int_, iarg);							\
					INT_ASM2																	\
					return new Int(nval);												\
				}else																			\
	int_overflow2:																			\
					NR_if_(INT16, NAME, double(int_), iarg)						\
			}																					\
		}																						\
		else if (arg->type==INT8)														\
		{	char iarg=Int8P(arg)->int_;												\
			if (true)																		\
			{	char nval = FUNC_i8##NAME(int_, iarg);								\
				INT8_ASM																		\
				return new Int8(nval);													\
			}else																				\
	int8_overflow:																			\
			{	short iarg=Int8P(arg)->int_;											\
				if (true)																	\
				{	short nval = FUNC_i16##NAME(int_, iarg);						\
					INT16_ASM2																\
					return new Int16(nval);												\
				}else																			\
	int16_overflow2:																		\
				{	int iarg=Int8P(arg)->int_;											\
					if (true)																\
					{	int nval = FUNC_i##NAME(int_, iarg);						\
						INT_ASM3																\
						return new Int(nval);											\
					}else																		\
		int_overflow3:																		\
						NR_if_(INT8, NAME, double(int_), iarg)						\
				}																				\
			}																					\
		}																						\
		else if (arg->type==UINT8)														\
		{	unsigned char iarg=UInt8P(arg)->int_;									\
			if (true)																		\
			{	unsigned char nval = FUNC_i8##NAME(int_, iarg);					\
				INT8_ASM2																	\
				return new UInt8(nval);													\
			}else																				\
	int8_overflow2:																		\
			{	short iarg=UInt8P(arg)->int_;											\
				if (true)																	\
				{	short nval = FUNC_i16##NAME(int_, iarg);						\
					INT16_ASM3																\
					return new Int16(nval);												\
				}else																			\
	int16_overflow3:																		\
				{	int iarg=UInt8P(arg)->int_;										\
					if (true)																\
					{	int nval = FUNC_i##NAME(int_, iarg);						\
						INT_ASM4																\
						return new Int(nval);											\
					}else																		\
		int_overflow4:																		\
						NR_if_(UINT8, NAME, double(int_), iarg)					\
				}																				\
			}																					\
		}																						\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	}																							\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Float::NAME(NodeP arg)															\
{	if (arg)																					\
		 NR_if_( INT_  , NAME, float_,     IntP(arg)->int_  )					\
		 NR_if_( FLOAT_, NAME, float_,   FloatP(arg)->float_)					\
		CNR_if_(COMPLEX, NAME, float_, ComplexdP(arg)->complex_)				\
		 NR_if_( INT16 , NAME, float_,   Int16P(arg)->int_  )					\
		 NR_if_( INT8  , NAME, float_,    Int8P(arg)->int_  )					\
		 NR_if_( UINT8 , NAME, float_,   UInt8P(arg)->int_  )					\
		BNR_if_(BIGFLOAT,NAME, float_,BigFloatP(arg)->float_)					\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP BigFloat::NAME(NodeP arg)														\
{	if (arg)																					\
		BNR_if_( INT_  , NAME, float_,     IntP(arg)->int_  )					\
		BNR_if_( FLOAT_, NAME, float_,MAPM(FloatP(arg)->float_))					\
		BNR_if_(BIGFLOAT,NAME, float_,BigFloatP(arg)->float_)					\
		BNR_if_( INT16 , NAME, float_,   Int16P(arg)->int_  )					\
		BNR_if_( INT8  , NAME, float_,    Int8P(arg)->int_  )					\
		BNR_if_( UINT8 , NAME, float_,   UInt8P(arg)->int_  )					\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Complexd::NAME(NodeP arg)														\
{	if (arg)																					\
		CNR_if_(  INT_ , NAME, complex_,   double(IntP(arg)->int_)   )		\
		CNR_if_( FLOAT_, NAME, complex_,        FloatP(arg)->float_  )		\
		CNR_if_(COMPLEX, NAME, complex_,     ComplexdP(arg)->complex_)		\
		CNR_if_( INT16 , NAME, complex_, double(Int16P(arg)->int_)   )		\
		CNR_if_( INT8  , NAME, complex_,  double(Int8P(arg)->int_)   )		\
		CNR_if_( UINT8 , NAME, complex_, double(UInt8P(arg)->int_)   )		\
		else_if_STRING(NAME, SYMBOLNODE)												\
		else_if_OBJECT(NAME)																\
		else_if_LIST  (NAME)																\
		else_if_ARRAY (NAME)																\
		else_if_FLOATARRAY (NAME)														\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Pair::NAME(NodeP arg)															\
{	PairP a=this, result=0, lastnode=0, tnode=0;									\
	if (arg->islist())																	\
	{	PairP b = PairP(arg);															\
		while (a && b)																		\
		{	tnode = new Pair(a->first()->NAME( b->first() ), 0);				\
			if (throwing) return Unbound;												\
			if (result)																		\
				lastnode->settail( tnode );											\
			else																				\
				result=tnode;																\
			lastnode=tnode;																\
			next(a);																			\
			next(b);																			\
		}																						\
		if (!a && !b) return result;													\
	}																							\
	else if (arg->type==ARRAY)															\
	{	NodePP bd = ArrayP(arg)->data();												\
		int ai = 0, bi = ArrayP(arg)->length();									\
		while (a)																			\
		{	NodeP r = a->first();														\
			int i = bi;																		\
			NodePP b=bd;																	\
			while (i--)																		\
			{	if ((*b)->islist())														\
					r = r->NAME((*b++)->item(ai));									\
				else error(BAD_DATA, this);											\
				if (throwing) return Unbound;											\
			}																					\
			tnode = new Pair(r, 0);														\
			if (throwing) return Unbound;												\
			if (result)																		\
				lastnode->settail( tnode );											\
			else																				\
				result=tnode;																\
			lastnode=tnode;																\
			next(a);	ai++;																	\
		}																						\
		 if (!a) return result;															\
	}																							\
	else if (arg->type==INT_ || arg->type==FLOAT_ || arg->type==COMPLEX	\
			|| arg->type==INT16|| arg->type==INT8 || arg->type==UINT8)		\
	{	while (a)																			\
		{	tnode = new Pair(a->first()->NAME(arg), 0);							\
			if (throwing) return Unbound;												\
			if (result)																		\
				lastnode->settail( tnode );											\
			else																				\
				result=tnode;																\
			lastnode=tnode;																\
			next(a);																			\
		}																						\
		 if (!a) return result;															\
	}																							\
	else_if_STRING(NAME, SYMBOLNODE)													\
	else_if_OBJECT(NAME)																	\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Array::NAME(NodeP arg)															\
{	if (!arg)																				\
	{	error(BAD_DATA, this); return Unbound;										\
	}																							\
	NodePP a=data();																		\
	if (arg->type==ARRAY)																\
	{	NodePP b = ArrayP(arg)->data();												\
		int i=length();																	\
		if (i==ArrayP(arg)->length())													\
		{	ArrayP result=0;  result = new Array(length(), origin());		\
			NodePP c=result->data();													\
			while (i--)																		\
			{	if (*a)																		\
					*c = (*a)->NAME(*b);													\
				else																			\
				{	error(BAD_DATA, this);												\
					return Unbound;														\
				}																				\
				a++, b++;																	\
				result->check_oldyoung(*c++);											\
				if (throwing) return Unbound;											\
			}																					\
			return result;																	\
		}																						\
	}																							\
	else if (arg->is.list																\
			|| arg->type==INT_ || arg->type==FLOAT_ || arg->type==COMPLEX	\
			|| arg->type==INT16|| arg->type==INT8 || arg->type==INT8)		\
	{	ArrayP result=0;  result = new Array(length(), origin());			\
		NodePP c=result->data();														\
		int i=length();																	\
		while (i--)																			\
		{	if (*a)																			\
				*c = (*a)->NAME(arg);													\
			else																				\
			{	error(BAD_DATA, this);													\
				return Unbound;															\
			}																					\
			a++;																				\
			result->check_oldyoung(*c++);												\
			if (throwing) return Unbound;												\
		}																						\
		return result;																		\
	}																							\
	else_if_STRING(NAME, SYMBOLNODE)													\
	else_if_OBJECT(NAME)																	\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP IntArray::NAME(NodeP arg)														\
{	if (!arg)																				\
	{	error(BAD_DATA, this); return Unbound;										\
	}																							\
	intP a=_data;																			\
	if (arg->type==INTARRAY)															\
	{	intP b = IntArrayP(arg)->_data;												\
		int l=length(), l2=IntArrayP(arg)->length();								\
		if (l2 < l) l=l2;																	\
		IntArrayP result=0;  result = new IntArray(length(), origin());	\
		intP c=result->_data;															\
		while (l--)																			\
			(*c++) = Int::clip(FUNC_i##NAME((*a++), (*b++)));					\
		return result;																		\
	}																							\
	else if (arg->type==INT_)															\
	{	IntArrayP result=0;  result = new IntArray(length(), origin());	\
		intP c=result->_data;															\
		int l=length(), i=IntP(arg)->int_;											\
		while (l--)																			\
			(*c++) = Int::clip(FUNC_i##NAME((*a++), i));							\
		return result;																		\
	}																							\
	else_if_STRING(NAME, SYMBOLNODE)													\
	else_if_OBJECT(NAME)																	\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP Int16Array::NAME(NodeP arg)													\
{	if (!arg)																				\
	{	error(BAD_DATA, this); return Unbound;										\
	}																							\
	shortP a=_data;																		\
	if (arg->type==INT16ARRAY)															\
	{	shortP b = Int16ArrayP(arg)->_data;											\
		int l=length(), l2=Int16ArrayP(arg)->length();							\
		if (l2 < l) l=l2;																	\
		Int16ArrayP result=0;															\
		result = new Int16Array(l, origin());										\
		shortP c=result->_data;															\
		while (l--)																			\
			(*c++) = Int16::clip(FUNC_i16##NAME((*a++), (*b++)));				\
		return result;																		\
	}																							\
	else if (arg->type==INT_)															\
	{	Int16ArrayP result=0;															\
		result = new Int16Array(length(), origin());								\
		shortP c=result->_data;															\
		int l=length(), i=IntP(arg)->int_;											\
		while (l--)																			\
			(*c++) = Int16::clip(FUNC_i##NAME((*a++), i));						\
		return result;																		\
	}																							\
	else if (arg->type==FLOAT_)														\
	{	FloatArrayP result=0;															\
		result = new FloatArray(length(), origin());								\
		doubleP c=result->data();														\
		int l=length();																	\
		double d=FloatP(arg)->float_;													\
		while (l--)																			\
		{	double f = FUNC_##NAME((*a++), d);										\
			(*c++) = f;																		\
		}																						\
		return result;																		\
	}																							\
	else_if_STRING(NAME, SYMBOLNODE)													\
	else_if_OBJECT(NAME)																	\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP FloatArray::NAME(NodeP arg)													\
{	if (!arg)																				\
	{	error(BAD_DATA, this); return Unbound;										\
	}																							\
	doubleP a=_data;																		\
	if (arg->type==FLOATARRAY)															\
	{	doubleP b = FloatArrayP(arg)->_data;										\
		int l=length(), l2=FloatArrayP(arg)->length();							\
		if (l2 < l) l=l2;																	\
		FloatArrayP result=0;															\
		result = new FloatArray(length(), origin());								\
		doubleP c=result->_data;														\
		while (l--)																			\
		{	double f = FUNC_##NAME((*a++), (*b++));								\
			(*c++) = f;																		\
		}																						\
		return result;																		\
	}																							\
	else if (arg->type==FLOAT_)														\
	{	FloatArrayP result=0;															\
		result = new FloatArray(length(), origin());								\
		doubleP c=result->_data;														\
		int l=length();																	\
		double d=FloatP(arg)->float_;													\
		while (l--)																			\
		{	double aa=*a++;																\
			{	double a=aa;																\
				double f = FUNC_##NAME(a, d);											\
				(*c++) = f;																	\
			}																					\
		}																						\
		return result;																		\
	}																							\
	else if (arg->type==INT_)															\
	{	FloatArrayP result=0;															\
		result = new FloatArray(length(), origin());								\
		doubleP c=result->_data;														\
		int l=length(), i=IntP(arg)->int_;											\
		while (l--)																			\
		{	double f = FUNC_##NAME((*a++), i);										\
			(*c++) = f;																		\
		}																						\
		return result;																		\
	}																							\
	else if (arg->type==INT16)															\
	{	FloatArrayP result=0;															\
		result = new FloatArray(length(), origin());								\
		doubleP c=result->_data;														\
		int l=length(), i=Int16P(arg)->int_;										\
		while (l--)																			\
		{	double f = FUNC_##NAME((*a++), i);										\
			(*c++) = f;																		\
		}																						\
		return result;																		\
	}																							\
	else_if_STRING(NAME, SYMBOLNODE)													\
	else_if_OBJECT(NAME)																	\
	error(BAD_DATA, this); return Unbound;											\
}																								\
NodeP l##NAME(PairPP arg)																\
{	if (!exist(3))																			\
		if (exist(2) && arg[1])															\
			return arg[1]->NAME(arg[2]);												\
		else if (exist(1))																\
			return NEUTRAL->NAME(arg[1]);	/* not really correct!!! */		\
																								\
	int nr=1;																				\
	NodeP a = (*++arg)->a_number(nr++);												\
	NodeType type=type_of(a);   														\
	char i8;	short i16;	int i;   double f;   complexd c;						\
	enum Mode {I8MODE, I16MODE, IMODE, FMODE, CMODE};							\
	Mode imode;																				\
																								\
	     if (type== INT16 ) {i16=  Int16P(a)->    int_; imode=I16MODE;}	\
	else if (type== INT8  ) {i8 =   Int8P(a)->    int_; imode= I8MODE;}	\
	else if (type== INT_  ) {i =     IntP(a)->    int_; imode=  IMODE;}	\
	else if (type== FLOAT_) {f =   FloatP(a)->  float_; imode=  FMODE;}	\
	else if (type==COMPLEX) {c =ComplexdP(a)->complex_; imode=  CMODE;}	\
	else {error(BAD_DATA, a); return Unbound;}									\
																								\
	while (arg!=argStack  &&  !throwing) 											\
	{																							\
		a = (*++arg)->a_number(nr++);													\
		if (throwing) return Unbound;													\
		char i8arg; short i16arg; int iarg; double farg; complexd carg;	\
																								\
		if (type_of(a)==INT8) 															\
		{	switch (imode)																	\
			{	case  I8MODE:  i8arg=IntP(a)->int_; break;						\
				case I16MODE: i16arg=IntP(a)->int_; break;						\
				case   IMODE: iarg = IntP(a)->int_; break;						\
				case   FMODE: farg = IntP(a)->int_; break;						\
				case   CMODE: carg = IntP(a)->int_; break;						\
			}																					\
		}																						\
		if (type_of(a)==INT16) 															\
		{	switch (imode)																	\
			{	case  I8MODE: i16=i8; imode=I16MODE; break;						\
				case I16MODE: i16arg=IntP(a)->int_; break;						\
				case   IMODE: iarg = IntP(a)->int_; break;						\
				case   FMODE: farg = IntP(a)->int_; break;						\
				case   CMODE: carg = IntP(a)->int_; break;						\
			}																					\
		}																						\
		else if (type_of(a)==INT_) 													\
		{	switch (imode)																	\
			{	case  I8MODE: i=i8;	imode=IMODE; break;							\
				case I16MODE: i=i16;	imode=IMODE; break;							\
				case   IMODE: iarg = IntP(a)->int_; break;						\
				case   FMODE: farg = IntP(a)->int_; break;						\
				case   CMODE: carg = IntP(a)->int_; break;						\
			}																					\
		}else if (type_of(a)==FLOAT_)													\
		{	switch (imode)																	\
			{	case  I8MODE: f=i8;	imode=FMODE; break;							\
				case I16MODE: f=i16;	imode=FMODE; break;							\
				case   IMODE: f=i;	imode=FMODE; break;							\
				default: ;																	\
			}																					\
			farg = FloatP(a)->float_;													\
		}																						\
		else if (type_of(a)==COMPLEX)													\
		{	switch (imode)																	\
			{	case  I8MODE: c=i8;	imode=CMODE; break;							\
				case I16MODE: c=i16;	imode=CMODE; break;							\
				case   IMODE: c=i;	imode=CMODE; break;							\
				case   FMODE: c=f;	imode=CMODE; break;							\
				default: ;																	\
			}																					\
			carg = ComplexdP(a)->complex_;											\
		}																						\
		switch (imode)																		\
		{	case I8MODE:																	\
				if (true) 																	\
				{	char nval = FUNC_i8##NAME(i8, i8arg);							\
					INT8_ASM	i8=nval;														\
				}else																			\
	int8_overflow:																			\
					if (true)																\
					{	imode=I16MODE;														\
						short nval = FUNC_i16##NAME(short(i8), short(i8arg));	\
						INT16_ASM	i16=nval;											\
					}else																		\
	int16_overflow:																		\
						if (true)															\
						{	imode=IMODE;													\
							int nval = FUNC_i##NAME(int(i8), int(i8arg));		\
							INT_ASM	i=nval;												\
						}else																	\
	int_overflow:	{imode=FMODE;  f = FUNC_##NAME(double(i8), i8arg);}	\
				break;																		\
																								\
			case I16MODE:																	\
				if (true) 																	\
				{	short nval = FUNC_i16##NAME(i16, i16arg);						\
					INT16_ASM2	i16=nval;												\
				}else																			\
	int16_overflow2:																		\
					if (true)																\
					{	imode=IMODE;														\
						int nval = FUNC_i##NAME(int(i16), int(i16arg));			\
						INT_ASM2	i=nval;													\
					}else																		\
	int_overflow2: {imode=FMODE;  f = FUNC_##NAME(double(i16), i16arg);}	\
				break;																		\
																								\
			case IMODE:																		\
				if (true) 																	\
				{	int nval = FUNC_i##NAME(i, iarg);								\
					INT_ASM3	i=nval;														\
				}else																			\
	int_overflow3: {imode=FMODE;  f = FUNC_##NAME(double(i), iarg);}		\
				break;																		\
																								\
			case FMODE:																		\
				if (true)																	\
				{	f = FUNC_##NAME(f, farg);											\
					FLOAT_ASM																\
				}else																			\
	float_overflow: {Float::reset_FPU();  error(BAD_DATA, a, nr-1);}		\
				break;																		\
																								\
			case CMODE:																		\
				if (true)																	\
				{	c = FUNC_c##NAME(c, carg);											\
					COMPLEX_ASM																\
				}else																			\
	complex_overflow: {Float::reset_FPU();  error(BAD_DATA, a, nr-1);}	\
				break;																		\
		}																						\
	}																							\
	if (!throwing) 																		\
	{	switch (imode)																		\
		{	case  I8MODE: return new     Int8(i);									\
			case I16MODE: return new    Int16(i);									\
			case   IMODE: return new      Int(i);									\
			case   FMODE: return new    Float(f);									\
			case   CMODE: return new Complexd(c);									\
		}																						\
	}																							\
	return Unbound;																		\
}

#define FUNC_add(a,b)		a+b
#define FUNC_badd(a,b)		a+b
#define FUNC_cadd(a,b)		a+b
#define FUNC_iadd(a,b)		a+b
#define FUNC_i16add(a,b)	a+b
#define FUNC_i8add(a,b)		a+b
#define FUNC_sub(a,b)		a-b
#define FUNC_csub(a,b)		a-b
#define FUNC_bsub(a,b)		a-b
#define FUNC_isub(a,b)		a-b
#define FUNC_i16sub(a,b)	a-b
#define FUNC_i8sub(a,b)		a-b
#define FUNC_mul(a,b)		a*b
#define FUNC_bmul(a,b)		a*b
#define FUNC_cmul(a,b)		a*b
#define FUNC_imul(a,b)		a*b
#define FUNC_i16mul(a,b)	a*b > Int16::max ? Int::max+1 : a*b
#define FUNC_i8mul(a,b)		a*b > Int8::max ? Int::max+1 : a*b
#define FUNC_div(a,b)		a/b
#define FUNC_bdiv(a,b)		a/b
#define FUNC_cdiv(a,b)		a/b
#define FUNC_idiv(a,b)		b? (a%b? Int::max+1 : a/b) : Int::max+1
#define FUNC_i16div(a,b)	b? (a%b? Int::max+1 : a/b) : Int::max+1
#define FUNC_i8div(a,b)		b? (a%b? Int::max+1 : a/b) : Int::max+1

#define FUNC_mod(a,b)		Float::mod(a, b)
#define FUNC_bmod(a,b)		BigFloat::mod(a, b)
#define FUNC_cmod(a,b)		Complexd::mod(a, b)
#define FUNC_imod(a,b)		Int::mod(a, b)
#define FUNC_i16mod(a,b)	Int::mod(a, b)
#define FUNC_i8mod(a,b)		Int::mod(a, b)
#define FUNC_Remainder(a,b)		Float::Remainder(a, b)
#define FUNC_bRemainder(a,b)		BigFloat::Remainder(a, b)
#define FUNC_cRemainder(a,b)		Complexd::Remainder(a, b)
#define FUNC_iRemainder(a,b)		Int::Remainder(a, b)
#define FUNC_i16Remainder(a,b)	Int::Remainder(a, b)
#define FUNC_i8Remainder(a,b)		Int::Remainder(a, b)

#define FUNC_aShift(a,b)		a*pow(2, b)
#define FUNC_baShift(a,b)		a*pow(2, b)
#define FUNC_caShift(a,b)		a*std::pow(double(2), complexd(b))
#define FUNC_iaShift(a,b)		b>=0? a<<b : int(unsigned(a)>>-b)
#define FUNC_i16aShift(a,b)		b>=0? a<<b : short(unsigned(a)>>-b)
#define FUNC_i8aShift(a,b)		b>=0? a<<b : char(unsigned(a)>>-b)
#define FUNC_lShift(a,b)	fabs(a)*pow(2, b)
#define FUNC_blShift(a,b)	fabs(a)*pow(2, b)
#define FUNC_clShift(a,b)										\
	complexd(Float::abs(complexd(a).real()),				\
				Float::abs(complexd(a).imag()))*std::pow(double(2), complexd(b))
#define FUNC_ilShift(a,b)	b>=0? a<<b : a>>-b
#define FUNC_i16lShift(a,b)	b>=0? a<<b : a>>-b
#define FUNC_i8lShift(a,b)	b>=0? a<<b : a>>-b

#define FUNC_BitOr(a,b)			Float::NaN
#define FUNC_bBitOr(a,b)		Float::NaN
#define FUNC_cBitOr(a,b)		Float::NaN
#define FUNC_iBitOr(a,b)		a|b
#define FUNC_i16BitOr(a,b)		a|b
#define FUNC_i8BitOr(a,b)		a|b
#define FUNC_BitAnd(a,b)		Float::NaN
#define FUNC_bBitAnd(a,b)		Float::NaN
#define FUNC_cBitAnd(a,b)		Float::NaN
#define FUNC_iBitAnd(a,b)		a&b
#define FUNC_i16BitAnd(a,b)		a&b
#define FUNC_i8BitAnd(a,b)		a&b
#define FUNC_BitXOr(a,b)		Float::NaN
#define FUNC_bBitXOr(a,b)		Float::NaN
#define FUNC_cBitXOr(a,b)		Float::NaN
#define FUNC_iBitXOr(a,b)		a^b
#define FUNC_i16BitXOr(a,b)	a^b
#define FUNC_i8BitXOr(a,b)		a^b

#define FUNC_Power(a,b)		f=pow(a, b);
#define FUNC_bPower(a,b)	0; if (a >= 0) f=pow(a, b); else Float::NaN
#define FUNC_cPower(a,b)	std::pow(a, complexd(b))
#define FUNC_iPower(a,b)	Int::max+1
#define FUNC_i16Power(a,b)	Int::max+1
#define FUNC_i8Power(a,b)	Int::max+1
#define FUNC_ArcTan(a,b)	atan2(b, a)/Float::deg2rad
#define FUNC_bArcTan(a,b)	atan2(b, a)/BigFloat::deg2rad
#define FUNC_cArcTan(a,b)	Float::NaN
#define FUNC_iArcTan(a,b)	Int::max+1
#define FUNC_i16ArcTan(a,b)	Int::max+1
#define FUNC_i8ArcTan(a,b)	Int::max+1
#define FUNC_radArcTan(a,b)	atan2(b, a)
#define FUNC_bradArcTan(a,b)	atan2(b, a)
#define FUNC_cradArcTan(a,b)	Float::NaN
#define FUNC_iradArcTan(a,b)	Int::max+1
#define FUNC_i16radArcTan(a,b)	Int::max+1
#define FUNC_i8radArcTan(a,b)	Int::max+1

BINARY_OP(add,  _Plus_Sign, _Zero)
BINARY_OP(sub, _Minus_Sign, _Zero)

BINARY_OP(mul,   _Mul_Sign, _One)
BINARY_OP(div,   _Div_Sign, _One)
BINARY_OP(mod,   _Mod_Sign, _Zero)
BINARY_OP(Remainder, _Remainder_, _Zero)

BINARY_OP(aShift, _AShift_, _One)
BINARY_OP(lShift, _LShift_, _One)
BINARY_OP(BitOr,   _BitOr_, _Zero)
BINARY_OP(BitAnd, _BitAnd_, _Minus_One)
BINARY_OP(BitXOr, _BitXOr_, _Zero)

BINARY_OP(Power, _Power_, _One)
BINARY_OP(ArcTan,    _ArcTan_, _Zero)
BINARY_OP(radArcTan, _RadArcTan_, _Zero)

#define _XFUNC_virtuals(NAME)															\
void Node::NAME(NodeP arg) {error(BAD_DATA, this);}							\
void String::NAME(NodeP arg) {error(BAD_DATA, this);}							\
void Object::NAME(NodeP arg) {var()->NAME(arg);}								\
void Int::NAME(NodeP arg)																\
{	if (type_of(arg)==INT_)	FUNC_i##NAME(int_, IntP(arg)->int_);			\
	else error(BAD_DATA, this);														\
}																								\
void Int16::NAME(NodeP arg)															\
{	if (type_of(arg)==INT16)	FUNC_i##NAME(int_, Int16P(arg)->int_);		\
	else error(BAD_DATA, this);														\
}																								\
void Int8::NAME(NodeP arg)																\
{	if (type_of(arg)==INT8)	FUNC_i##NAME(int_, Int8P(arg)->int_);			\
	else error(BAD_DATA, this);														\
}																								\
void UInt8::NAME(NodeP arg)															\
{	if (type_of(arg)==UINT8)	FUNC_i##NAME(int_, UInt8P(arg)->int_);		\
	else error(BAD_DATA, this);														\
}																								\
void Float::NAME(NodeP arg)															\
{	if (type_of(arg)==FLOAT_)															\
		FUNC_##NAME(float_, FloatP(arg)->float_);									\
	else if (type_of(arg)==INT_)														\
		FUNC_##NAME(float_, IntP(arg)->int_);										\
	else error(BAD_DATA, this);														\
}																								\
void BigFloat::NAME(NodeP arg)														\
{	if (type_of(arg)==FLOAT_)															\
		FUNC_b##NAME(float_, FloatP(arg)->float_);								\
	else if (type_of(arg)==INT_)														\
		FUNC_b##NAME(float_, IntP(arg)->int_);										\
	else if (type_of(arg)==BIGFLOAT)													\
		FUNC_b##NAME(float_, BigFloatP(arg)->float_);							\
	else error(BAD_DATA, this);														\
}																								\
void Complexd::NAME(NodeP arg)														\
{	if (type_of(arg)==COMPLEX)															\
		FUNC_c##NAME(complex_, ComplexdP(arg)->complex_);						\
	else if (type_of(arg)==FLOAT_)													\
		FUNC_c##NAME(complex_, FloatP(arg)->float_);								\
	else if (type_of(arg)==INT_)														\
		FUNC_c##NAME(complex_, IntP(arg)->int_);									\
	else error(BAD_DATA, this);														\
}																								\
void Pair::NAME(NodeP arg)																\
{	PairP a=this;																			\
	if (arg->islist())																	\
	{	PairP b=PairP(arg);																\
		while (a && b)																		\
		{	if (a->first())																\
				a->first()->NAME( b->first() );										\
			if (throwing) return;														\
			next(a); next(b);																\
		}																						\
		if (!a && !b) return;															\
	}																							\
	else																						\
	{	while (a)																			\
		{	if (a->first())																\
				a->first()->NAME(arg);													\
			if (throwing) return;														\
			next(a);																			\
		}																						\
		if (!a) return;																	\
	}																							\
	error(BAD_DATA, this);																\
}																								\
void Array::NAME(NodeP arg)															\
{	if (type_of(arg)==ARRAY)															\
	{	int i=length();																	\
		if (i==ArrayP(arg)->length())													\
		{	NodePP a=data(), b=ArrayP(arg)->data();								\
			while (i-- && *a)																\
			{	(*a++)->NAME(*b++);														\
				if (throwing) return;													\
			}																					\
		}																						\
		else																					\
			error(BAD_DATA, arg, 2);													\
	}																							\
	else																						\
	{	int i=length();																	\
		NodePP a=data();																	\
		while (i-- && *a && !throwing)												\
		{	(*a++)->NAME(arg);															\
			if (throwing) return;														\
		}																						\
	}																							\
	return;																					\
}																								\
void IntArray::NAME(NodeP arg)														\
{	if (type_of(arg)==INTARRAY)														\
	{	int i=length();																	\
		if (i==IntArrayP(arg)->length())												\
		{	intP a=_data, b=IntArrayP(arg)->_data;									\
			while (i-- && *a)																\
				FUNC_i##NAME((*a++), (*b++));											\
		}																						\
		else																					\
			error(BAD_DATA, arg, 2);													\
	}																							\
	else																						\
	{	if (type_of(arg)!=INT_)															\
		{	error(BAD_DATA, arg, 2);													\
			return;																			\
		}																						\
		int i=IntP(arg)->int_, l=length();											\
		intP a=_data;																		\
		while (l-- && *a)																	\
			FUNC_i##NAME((*a++), i);													\
	}																							\
	return;																					\
}																								\
void Int16Array::NAME(NodeP arg)														\
{	if (type_of(arg)==INT16ARRAY)														\
	{	int i=length();																	\
		if (i==Int16ArrayP(arg)->length())											\
		{	shortP a=_data, b=Int16ArrayP(arg)->_data;							\
			while (i-- && *a)																\
				FUNC_i##NAME((*a++), (*b++));											\
		}																						\
		else																					\
			error(BAD_DATA, arg, 2);													\
	}																							\
	else																						\
	{	if (type_of(arg)!=INT16)														\
		{	error(BAD_DATA, arg, 2);													\
			return;																			\
		}																						\
		short i=Int16P(arg)->int_, l=length();										\
		shortP a=_data;																	\
		while (l-- && *a)																	\
			FUNC_i##NAME((*a++), i);													\
	}																							\
	return;																					\
}																								\
void FloatArray::NAME(NodeP arg)														\
{	if (type_of(arg)==FLOATARRAY)														\
	{	int i=length();																	\
		if (i==FloatArrayP(arg)->length())											\
		{	doubleP a=_data, b=FloatArrayP(arg)->_data;							\
			while (i-- && *a)																\
				FUNC_##NAME((*a++), (*b++));											\
		}																						\
		else																					\
			error(BAD_DATA, arg, 2);													\
	}																							\
	else																						\
	{	if (type_of(arg)!=FLOAT_)														\
		{	error(BAD_DATA, arg, 2);													\
			return;																			\
		}																						\
		double i=FloatP(arg)->float_, l=length();									\
		doubleP a=_data;																	\
		while (l-- && *a)																	\
				FUNC_##NAME((*a++), i);													\
	}																							\
	return;																					\
}																								
//*
#define _XFUNC_primitive(NAME)														\
NodeP l##NAME(PairPP arg)																\
{	((ObjectP)(arg[1]))->var()->NAME( arg[2] );									\
	return Unbound;																		\
}
/*/
template <class NAME> NodeP _NAME(PairPP arg)																\
{	((ObjectP)(arg[1]))->var()->NAME( arg[2] );									
	return Unbound;																		
}
//*/

#define _XFUNC(NAME)	_XFUNC_virtuals(NAME)	_XFUNC_primitive(NAME)

#define FUNC_xCopy(a,b)	a=b
#define FUNC_bxCopy(a,b) a=b
#define FUNC_cxCopy(a,b) a=b
#define FUNC_ixCopy(a,b) a=b
#define FUNC_xAdd(a,b)	a+=b
#define FUNC_bxAdd(a,b)	a+=b
#define FUNC_cxAdd(a,b)	a+=b
#define FUNC_ixAdd(a,b)	a+=b
#define FUNC_xSub(a,b)	a-=b
#define FUNC_bxSub(a,b)	a-=b
#define FUNC_cxSub(a,b)	a-=b
#define FUNC_ixSub(a,b)	a-=b
#define FUNC_xMul(a,b)	a*=b
#define FUNC_bxMul(a,b)	a*=b
#define FUNC_cxMul(a,b)	a*=b
#define FUNC_ixMul(a,b)	a*=b
#define FUNC_xDiv(a,b)	a/=b
#define FUNC_bxDiv(a,b)	a/=b
#define FUNC_cxDiv(a,b)	a/=b
#define FUNC_ixDiv(a,b)	a/=b
#define FUNC_xMod(a,b)	a=Float::mod(a,b)
#define FUNC_bxMod(a,b)	a=BigFloat::mod(a,b)
#define FUNC_cxMod(a,b)	a=Complexd::mod(a,b)
#define FUNC_ixMod(a,b)	a=Int::mod(a,b)

_XFUNC(xAdd)  _XFUNC(xSub)  _XFUNC(xMul)  _XFUNC(xDiv)  _XFUNC(xMod)

//_XFUNC_virtuals(xCopy)

void Node::xCopy(NodeP arg) {error(BAD_DATA, this);}
void String::xCopy(NodeP arg)
{	if (arg->isstring())
	{	free();
		StringP a=StringP(arg);
		_length = a->length();
			_ptr = a->ptr();
		  _head = a->head();
		_head->refCount++;
	}
	else error(BAD_DATA, arg, 2);
}
void Object::xCopy(NodeP arg){var()->xCopy(arg);}
void Int::xCopy(NodeP arg)
{	if (type_of(arg) == INT_) int_ = IntP(arg)->int_;
	else error(BAD_DATA, this);
}
void Int16::xCopy(NodeP arg)
{	if (type_of(arg) == INT16) int_ = Int16P(arg)->int_;
	else error(BAD_DATA, this);
}
void Int8::xCopy(NodeP arg)
{	if (type_of(arg) == INT8) int_ = Int8P(arg)->int_;
	else error(BAD_DATA, this);
}
void UInt8::xCopy(NodeP arg)
{	if (type_of(arg) == UINT8) int_ = UInt8P(arg)->int_;
	else error(BAD_DATA, this);
}
void Float::xCopy(NodeP arg)
{	if (type_of(arg) == FLOAT_) float_ = FloatP(arg)->float_;
	else error(BAD_DATA, this);
}
void BigFloat::xCopy(NodeP arg)
{	if (type_of(arg) == BIGFLOAT) float_ = BigFloatP(arg)->float_;
	else error(BAD_DATA, this);
}
void Complexd::xCopy(NodeP arg)
{	if (type_of(arg) == COMPLEX) complex_ = ComplexdP(arg)->complex_;
	else error(BAD_DATA, this);
}
void Pair::xCopy(NodeP arg)
{	PairP a = this;

	if (arg->islist())
	{	PairP b = PairP(arg);
		while (a && b)
		{	if (a->first())
				a->first()->xCopy(b->first());
			if (throwing) return;
			next(a);
			next(b);
		}
		if (!a && !b) return;
	}else
	{	while (a)
		{	if (a->first())
				a->first()->xCopy(arg);
			if (throwing) return;
			next(a);
		}
		if (!a) return;
	}
	error(BAD_DATA, this);
}
void Array::xCopy(NodeP arg)
{	if (type_of(arg) == ARRAY)
	{	int i = length();
		if (i == ArrayP(arg)->length())
		{	NodePP a = data(), b = ArrayP(arg)->data();
			while (i-- && *a)
			{	(*a++)->xCopy(*b++);
				if (throwing) return;
			}
		}
		else
			error(BAD_DATA, arg, 2);
	}else
	{	int i = length();
		NodePP a = data();
		while (i-- && *a && !throwing)
		{	(*a++)->xCopy(arg);
			if (throwing) return;
		}
	}
	return;
}
void IntArray::xCopy(NodeP arg)
{	if (type_of(arg) == INTARRAY)
	{	int l = length();
		if (l == IntArrayP(arg)->length())
		{	intP a = _data, b = IntArrayP(arg)->_data;
			while (l-- && *a)
			{	(*a++) = (*b++);
				if (throwing) return;
			}
		}
		else
			error(BAD_DATA, arg, 2);
	}else
	{	if (type_of(arg)!=INT_)
		{	error(BAD_DATA, arg, 2);
			return;
		}
		int i=IntP(arg)->int_, l=length();
		intP a=_data;
		while (l-- && *a && !throwing)
			(*a++)=i;
	}
	return;
}
void Int16Array::xCopy(NodeP arg)
{	if (type_of(arg) == INT16ARRAY)
	{	int l = length();
		if (l == Int16ArrayP(arg)->length())
		{	shortP a = _data, b = Int16ArrayP(arg)->_data;
			while (l-- && *a)
			{	(*a++) = (*b++);
				if (throwing) return;
			}
		}
		else
			error(BAD_DATA, arg, 2);
	}else
	{	if (type_of(arg)!=INT16)
		{	error(BAD_DATA, arg, 2);
			return;
		}
		short i=Int16P(arg)->int_, l=length();
		shortP a=_data;
		while (l-- && *a && !throwing)
			(*a++)=i;
	}
	return;
}
void FloatArray::xCopy(NodeP arg)
{	if (type_of(arg) == FLOATARRAY)
	{	int l = length();
		if (l == FloatArrayP(arg)->length())
		{	doubleP a = _data, b = FloatArrayP(arg)->_data;
			while (l-- && *a)
			{	(*a++) = (*b++);
				if (throwing) return;
			}
		}
		else
			error(BAD_DATA, arg, 2);
	}else
	{	if (type_of(arg)!=FLOAT_)
		{	error(BAD_DATA, arg, 2);
			return;
		}
		double i=FloatP(arg)->float_, l=length();
		doubleP a=_data;
		while (l-- && *a && !throwing)
			(*a++)=i;
	}
	return;
}
NodeP lxCopy(PairPP arg)
{	NodeP name = arg[1]->a_name(1);
	if (!throwing) 
	{	ObjectP what = name->intern();
		what->var()->xCopy(arg[2]);
	}
	return Unbound;
}
