#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/html/helpctrl.h>

#include <stdio.h>
#include <ctype.h>

#include "Main.h"

#include "File.h"
#include "String.h"
#include "Pair.h"
#include "Procedur.h"
#include "Object.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Runparsd.h"
#include "Tree.h"
#include "Int.h"
#include "UInt8.h"

#include "Color.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Console.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

bool iswhitespace(char ch) {return ch==' ' || ch=='\t'|| ch=='\n'|| ch=='\0';}
bool isparen(char ch) {return ch=='(' || ch==')' || ch==';';}
//bool isunary(char ch) {return ch=='!';}
bool isdot  (char ch) {return ch=='.';}
bool issinglequote(char ch) {return ch=='\'';}
bool isassignmentop(char ch)
{	return
		ch=='=' || ch=='!' || ch=='<' || ch=='>'
	|| ch=='+' || ch=='-' || ch=='*' || ch=='/';
}
bool isinfix(char ch)
{	return       /* ch=='!' ||*/ ch=='^' 
	|| ch=='*' || ch=='/' || /*ch=='%' ||*/ ch=='+' || ch=='-'
	|| ch=='=' || ch=='<' || ch=='>';
}

bool Tree::dont_know_how_to;

PairP String::runparse_node(PairPP ndsptr, 
									 bool &was_assignment,
									 bool &was_dot) 
{
	NodeType wtype = type;
	   charP wptr  = ptr();
	     int wlen  = length(),	wcount=0;
	   HeadP whead = head();

	bool monadic_minus=false;

	PairP outline=0, tnode=0, lastnode=0;	// to construct the output
	
	while (wcount < wlen) 
	{	if (*wptr==';')		// semicolon comment
		{	*ndsptr=0;
			break;
		}
		if (*wptr=='"')		// quoted
		{	int tcount=0;
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  &&  !isparen(*wptr)) 
			{	if (wtype==BACKSLASHED  &&  backslash.is(*wptr))
					wtype=UNBOUND;    // flag for "\( case
				wptr++, wcount++, tcount++;
			}
			if (wtype==UNBOUND) 
			{	wtype=BACKSLASHED;
				tnode = new Line
				(	new Quote( 
						make(tptr, 0, tcount, wtype, 
							copy_noparity)->intern() ),
					0
				);
			}else
				tnode = new Line
				(	new Quote( make(tptr, whead, tcount, wtype, copy)->intern() ),
					0
				);
		}
		else if (*wptr==':')	// colon (asks for contents of a variable)
		{	int tcount=0;
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  &&  !isparen(*wptr)  &&  !isinfix(*wptr)  
			&& !isdot(*wptr) && !issinglequote(*wptr))
				wptr++, wcount++, tcount++;
			ObjectP obj=make(tptr, whead, tcount, wtype, copy)->intern();
			if (obj->proc_or_prim()!=Unbound)
				c_out << endL << "; Warning: the variable " << obj 
					<< " shadows a proc or primitive in line " << Line::nr << endL;
			tnode = new Line(new Colon(obj), 0);
		}
		else if (wcount==0  &&  *wptr=='-'	// minus sign with space before and no space after is unary
		&&  monadic_minus==false 
		&&  wcount+1 < wlen  &&  !iswhitespace(wptr[1])  &&  wptr[1] != '=') 
		{
			tnode = new Line( new Int(0), 0);
			monadic_minus=true;
		}
		else if (isparen(*wptr) || isinfix(*wptr) || *wptr=='!' 
		|| isdot(*wptr))	// paren or infix
		{	if (monadic_minus)
			{	tnode = new Line(	_Minus_Tight, 0 );
			}
			else if (isassignmentop(*wptr)  &&  wcount+1 < wlen  &&  wptr[1]=='=')	
			//  +=  -=  *=  /=  %=  ==  !=
			{
				if (*wptr != '=' &&  *wptr != '!' 
				&&  *wptr != '<' &&  *wptr != '>')	
					// not the comparsion ops, only assignments
					if (tnode)
					{	if (!tnode->tail() 
						&&  tnode->first() 
						&&  tnode->first()->type != COLON
						&& !(tnode->first()->type==OBJECT 
							&& ObjectP(tnode->first())->proc_or_prim() != Unbound))
							lastnode->setfirst( new Quote(lastnode->first()));
							// automatically quote the left arg
					}else
						was_assignment=true;	// 'x=' is the first thing in this expression; 
													// runparse has to notice that somehow
				tnode = new Line
					(	(new String(wptr, whead, 2, STRING, copy))->intern(),
						0
					);
				wptr++, wcount++;
			}
/*			else if (isunary(*wptr))	// Kludge for factorial operator '!'
			{	tnode = new Line
					(	(new String(wptr, whead, 1, STRING, copy))->intern(),
						new Int(0)
					);
			}
*/			else		// normal infix-ops + - * / = ect.
			{	if (*wptr=='=')	// assignment sign
				{	bool setItem=false, 
						setItem2D=false;
					if (outline)
						if (outline->tail()) 
							if (ObjectP(outline->second())==_Dot
							&& outline->tail()->tail())
							{	PairP list=outline->tail()->tail();
								while (list)
								{	if (ObjectP(list->first())==_Dot
									&& list->tail())
									{	setItem2D=true; break;
									}
									next(list);
								}
								if (!setItem2D)
									setItem=true;
							}
					if (setItem)
					// Kludge for name.item=value
					{	PairP 
							name=outline->first(),
							item=outline->tail()->tail();
						
						outline=new Line
							(	__SetItem, 
								item
							);
						lastnode=item;
						while (lastnode->tail())
							next(lastnode);
						tnode=new Quote(name);
					}
					else if (setItem2D)
					// even bigger KLUDGE for name.i.j=value
					{	PairP 
							name=outline->first(),
							item1=outline->tail()->tail(),
							item2=0,
							list=item1, lastlist=list;
						while (list && ObjectP(list->first())!=_Dot)
						{	lastlist=list;
							next(list);							
						}
						item2=list;
						if (item2)
							next(item2);

						lastlist->settail(item2);
						outline=
							new Line(_mdSetItem, 
							new Line(_List,
							item1));
						lastnode=item2;
						while (lastnode->tail())
							next(lastnode);
						tnode=new Quote(name);
					}
					else
					{	if (tnode)
						{	if (!tnode->tail() 
							&&   tnode->first() 
							&&   tnode->first()->type != COLON)
//							&& !(tnode->first()->type == OBJECT
//								&& ObjectP(tnode->first())->proc_or_prim() != Unbound))
								lastnode->setfirst( new Quote(lastnode->first()));	
								// automatically quote the left arg
						}else
							was_assignment=true;	// '=' is the first thing in this expression; 
														// runparse has to notice that somehow
						tnode = new Line
						(	(new String(wptr, whead, 1, STRING, copy))->intern(), 
							0
						);
					}
				}else if (*wptr=='.')
				{	tnode=new Line(_Dot, 0);
					was_dot=true;
				}else 
					tnode = new Line
					(	(new String(wptr, whead, 1, STRING, copy))->intern(), 
						0
					);
			}
			monadic_minus=false;
			wptr++, wcount++;
		}
		else if (issinglequote(*wptr))	// singlequote
		{	int tcount=0;	// code copied from "quoted" part
			charP tptr = ++wptr;
			wcount++;
			while (wcount < wlen  && !isparen(*wptr) && !isinfix(*wptr)) 
			{	if (wtype==BACKSLASHED  &&  backslash.is(*wptr))
					wtype=UNBOUND;    // flag for "\( case
				wptr++, wcount++, tcount++;
			}
			if (wtype==UNBOUND) 
			{	wtype=BACKSLASHED;
				tnode = new Line
					(	_Dot,
						new Line
						(	new Quote
							(	make(tptr, 0, tcount, wtype, 
									copy_noparity)->intern() 
							),
							0
						)
					);
			}else
				tnode = new Line
					(	_Dot,
						new Line
						(	new Quote
							(	make(tptr, whead, tcount, wtype, copy)->intern() 
							),
							0
						)
					);
		}
		else				// number or ?5
		{	int tcount=0, dotcount=0;
			charP tptr = wptr, dotptr=0;
			bool gotdot=false, goti=false;

			enum NumType
			{	digits_so_far, 
				just_saw_e,		// so minus can be next
				no_longer_eligible,	// even if an 'e' comes along
				question_mark,
				nothing_yet
			};
			NumType numtype = nothing_yet;

			if (*wptr=='?')	// turn ?5 to (? 5) 
			{	numtype = question_mark;
				wptr++, wcount++, tcount++;
			}
			while (wcount < wlen  
			&&  !isparen(*wptr) 
			&& (!(isinfix(*wptr) || issinglequote(*wptr))
				|| (numtype==just_saw_e  &&  (*wptr=='-' || *wptr=='+')))) 
			{
				if (numtype==just_saw_e  &&  (*wptr=='-' || *wptr=='+'))
				{	wptr++, wcount++, tcount++;
					continue;
				}
				if (numtype==nothing_yet  &&  is_digit(*wptr)) 
					numtype=digits_so_far;
				
				if (numtype==digits_so_far  &&  tcount > 0  
				&& (*wptr=='e' || *wptr=='E'))
					numtype = just_saw_e;
				else 
					if (!(	is_digit(*wptr) || (!gotdot && isdot(*wptr)) 
							|| (!goti && (*wptr=='i' || *wptr=='I'))) 
						/*|| numtype==just_saw_e*/)
						numtype = no_longer_eligible;

				if (isdot(*wptr)) 
					if (numtype==no_longer_eligible)
						break;
					else
					{	gotdot=true;
						dotcount=tcount;
						dotptr=wptr;
					}
				if (*wptr=='i' || *wptr=='I') 
					if (numtype==digits_so_far || numtype==just_saw_e)
					{	wptr++, wcount++, tcount++;
						goti=true;
						break;
					}
					else
						numtype = no_longer_eligible;//goti=true;
				wptr++, wcount++, tcount++;
			}
			if (numtype==question_mark  &&  tcount > 1)	// ?5 syntax
			{	PairP qmtnode = 
					new Line
					(	_Left_Paren, 
						new Line
						(	_Query,
							new Line
							(	make(tptr+1, whead, tcount-1, wtype, copy)->number(),
								0
							)
						)
					);
				if (!outline) 
					outline=qmtnode;
				else 
					lastnode->settail( qmtnode );
				lastnode = qmtnode->tail()->tail();

				tnode = new Line( _Right_Paren, 0 );
			}
			else if (numtype==digits_so_far  ||  numtype==just_saw_e  &&  tcount > 0) 
			{	tnode = new Line
					(	make(tptr, whead, tcount, wtype, copy)->number(),
						0
					);
			}else if (gotdot)
			{	tnode = 
				new Line(make(tptr, whead, dotcount, wtype, copy)->number(),
				new Line(_Dot,
				new Line(
				make(dotptr+1, whead, wptr-(dotptr+1), wtype, copy)->intern(),
					0)));
			}else
				tnode = new Line
					(	make(tptr, whead, tcount, wtype, copy)->intern(),
						0
					);
		}
		if (!outline) 
			outline=tnode;
		else
			lastnode->settail( tnode );
		lastnode=tnode;
	}
	return outline;
}

PairP Pair::runparse() 
{
	if (type_of(this)==RUNPARSED)
		return ((Runparsed*)this)->parsed();
	
	if (isnolist()) 
	{	error(BAD_DATA_UNREC, this);
		return 0;
	}
	PairP outline=0, tnode=0, lastnode=0, olastnode=0, ndlist=this;
	bool was_dot=false, owas_dot=false;

	while (ndlist) 
	{	NodeP nd = ndlist->first();

		if (ndlist->is.line)
		{	Line::nr=LineP(ndlist)->number;
			Line::fn=LineP(ndlist)->filename;
//			c_out << "runparse " << Line::fn << Line::nr << nd << endL;
		}
		next(ndlist);
		
		if (!nd || !nd->is.word)
			tnode = new Line(nd, 0);
		else 
		{	if (nd==_Minus_Tight) 
			{	tnode=new Line(nd, 0);
			}
			else if (!nd->numberP())
			{	bool was_assignment=false;

				tnode = nd->string()->
					runparse_node(&ndlist, was_assignment, was_dot);

				if ( owas_dot && olastnode)
					if (!olastnode->tail() && olastnode->first())
c_out << olastnode;

				if ( was_assignment && lastnode)
					if (!lastnode->tail() && lastnode->first())
						if (lastnode->first()->type != COLON
						&& lastnode->first()->type != QUOTE)
							lastnode->setfirst( new Quote(lastnode->first()));
			}else
				tnode = new Line(nd->number(), 0);
		}
		if (tnode) 
		{	if (!outline) 
				outline=tnode;
			else
				lastnode->settail( tnode );
			
			owas_dot=was_dot;
			was_dot=false;
			olastnode=lastnode;
			lastnode=tnode;
			while (lastnode->tail()) 
			{	next(lastnode);
				if (checkThrowing()) break;
			}
		}
		if (checkThrowing()) break;
	}
//	c_out << outline << endL;
	return outline;
}

NodeP lrunParse(PairPP arg) 
{
	NodeP s=arg[1];
	if (type_of(s)==ARRAY  &&  !throwing)	// arrays not yet allowed in runparse! array-to-list, then runparse!
	{	error(BAD_DATA, s);
		return Unbound;
	}
	PairP val=PairP(s);
	if (!throwing && s && !s->is.aggregate)
		val = StringP(s)->parser(true);

	if (!throwing) return val->runparse();
	return Unbound;
}

// Make a list of the next n expressions, where n is between min and max.
// Set args to immediately after the last expression.
 
PairP Procedure::gather_some_args(int min, int max, PairPP args, bool inparen, ObjectPP ifnode) 
{
	if (!*args
	|| (*args)->first() == NodeP(_Right_Paren)
	|| (type_of((*args)->first())==OBJECT
		&& type_of(ObjectP((*args)->first())->proc_or_prim())==INFIX) ) 
	{
		if (min > 0) 
			return _Not_Enough_node; //new Line(_Not_Enough_node, 0);
	}
	else if (max==0) 
	{	if (ifnode  &&  (*args)->first()->is.list)	// if -> ifelse kludge
		{	error(IF_WARNING, 0);
			*ifnode=_IfElse;

			PairP retval=0;
			retval = Pair::paren_expr(args, false);
			retval = retval->paren_infix(args, -1, inparen);

			StringP fn=Line::fn;
			int tmplinenr=Line::nr;

			PairP someArgs=gather_some_args(min, max, args, inparen, 0);


			Line::nr=tmplinenr;
			Line::fn=fn;
			return new Line(retval, someArgs);
		}
	}
	else 
	{	if (max < 0) max=0;   // negative max means unlimited

		if (!(*args)->first()
		|| ( (*args)->first() != NodeP(_Right_Paren)
			&& (type_of((*args)->first()) != OBJECT
				|| type_of( ObjectP((*args)->first())->proc_or_prim() ) != INFIX ) ) ) 
		{
			PairP retval=0;
			retval = Pair::paren_expr(args, false);
			retval = retval->paren_infix(args, -1, inparen);

			StringP fn=Line::fn;
			int tmplinenr=Line::nr;

			PairP someArgs=gather_some_args(min-1, max-1, args, inparen, ifnode);
			Line::nr=tmplinenr;
			Line::fn=fn;
			return new Line(retval, someArgs);
		}
	}
	return 0;
}

const char Primitive::ok_if_no_arg = -2;
	// Special value for pmin, means that it's OK 
	// if primitive name is on a line by itself 
	//even though defltargs=1 (ED, CO)

PairP Node::gather_args(PairPP args, bool inparen, ObjectPP ifnode)
{	error(BAD_DATA/*FATAL*/, this); return 0;
}
PairP Procedure::gather_args(PairPP args, bool inparen, ObjectPP ifnode)	// Gather the correct number of arguments to proc into a list. Set args to immediately after the last arg.
{	int
		argmin = inparen ?   min  :  std,
		argmax = inparen ?   max  :  std;

	return gather_some_args(argmin, argmax, args, inparen, ifnode);
}
PairP Primitive::gather_args(PairPP args, bool inparen, ObjectPP ifnode)	// Gather the correct number of arguments to proc into a list. Set args to immediately after the last arg.
{
	int argmin = inparen ?   min  :  std;

	if (argmin < 0)	// special form
	{	argStack << *args;
		PairP ret = PairP( call(argStack-1) );
		argStack.pop();
		return ret;
	}		
	if (min == Primitive::ok_if_no_arg)	
	// Kludge to allow EDIT and CO without input without paren
		argmin=0;

	int argmax = inparen ?   max  :  std;

	return Procedure::gather_some_args(argmin, argmax, args, inparen, ifnode);
}

PairP Pair::paren_expr(PairPP expr, bool inparen) 
// Parenthesize an expression. 
// Set expr to the Node after the first full expression.
{
	if (!*expr)
	{	if (inparen) error(PAREN_MISMATCH, 0);
		return *expr;
	}
	NodeP nd = (*expr)->first();
	if (!(*expr)->is.line)
		c_out << ";somewhere I forgot to change a new Pair to a new Line." << endL;
	else
	{	Line::nr=LineP(*expr)->number;
		Line::fn=LineP(*expr)->filename;
//		c_out << "paren_expr Line::nr=" << Line::nr << *expr << endL;
	}
	next(*expr);

	if (type_of(nd)==OBJECT  &&  !nd->numberP()) 
	{	ObjectP obj=ObjectP(nd);

		if (obj==_Left_Paren) 
		{	PairP tree=0;
			tree = paren_expr(expr, true);
			tree = tree->paren_infix(expr, -1, true);

			if (!*expr)
			{	error(PAREN_MISMATCH, 0);
			}
			else if (ObjectP((*expr)->first()) != _Right_Paren)
			{	if (!throwing) 
					error(TOO_MUCH, 0);	// throw the rest away

				for (int parens=0;  *expr;  next(*expr))
				{	if (ObjectP((*expr)->first())==_Left_Paren)
						parens++;
					else if (ObjectP((*expr)->first())==_Right_Paren)
						if (parens--==0) break;
				}
			}else next(*expr);
			return tree;
		}
		else if (obj==_Right_Paren) 
		{	error(UNEXPECTED_PAREN, 0);
			if (inparen) 
				push(*expr, obj);
			return 0;
		}
		else if (obj==_Minus_Sign) 
		{	push(*expr, _Minus_Tight);
			return PairP(new Int(0))->paren_infix(expr, -1, inparen);	// ugly
		}else 
		{
			StringP fn=Line::fn;
			int tmplinenr=Line::nr;

			if (obj->proc_or_prim()==Unbound	// only fix it if no proc or prim exists
			&& obj->var() != Unbound  &&  !throwing  &&  obj != _Null_Word )
			{	if (obj->proc_or_prim()!=Unbound)
					c_out << endL << "; Warning: the variable " << obj 
					<< " shadows a proc or primitive in line " << Line::nr << endL;
				return new Colon(obj);	// fix colonless variables
			}
			// it must be a procedure
			if ( obj->proc_or_prim()==Unbound  &&  obj != _Null_Word )
				obj->silent_load(environment.logolib.c_str()); // try <logolib>/<obj>.lg

			if ( obj->proc_or_prim()==Unbound  &&  obj != _Null_Word)
				obj->silent_load(0);    // try ./<obj>.lg

			NodeP proc_or_prim = obj->proc_or_prim();

			if (proc_or_prim==Unbound && !throwing) 
			{	Tree::dont_know_how_to=true;
				return new Line(obj, 0);
			}
			else if (type_of(proc_or_prim)==INFIX  &&  !throwing) 
			{	error(NOT_ENOUGH, obj);
				return new Line(obj, 0);
			}else 
			{	// Kludge follows to turn IF to IFELSE sometimes. (sometimes? Oh, oh!-)
				ObjectPP ifnode=0;
				if (obj==_If) 
					ifnode = &obj;

				NodeP ret = proc_or_prim->gather_args(expr, inparen, ifnode);

				if (ret != Unbound) 
				{	Line::nr=tmplinenr;
					Line::fn=fn;
					LineP l=new Line(obj, ret);
//					c_out << "paren_expr " << l << endL;
					return l;
				}
				else
					return PairP(Unbound);
			}
		}
	}
	else if (nd->islist())	// quoted list
		return new Quote(nd);
	else 
		return PairP(nd);
}

Priority Node::priority() {return priority_STOP;}

Priority Object::priority()	// Calculate the priority of a procedure.
{
	if (ObjectP(this)==_Minus_Tight) 
		return priority_tight_minus;

	NodeP p=proc_or_prim();

	if (p==Unbound  ||  p->type != INFIX)
		return priority_STOP;

	return PrimitiveP(p)->priority;
}

// Parenthesize an infix expression.
// <this> is the expression on the left (already parenthesized), 
// and <rest> is a pointer to the list starting with the infix procedure, if it's there.  
// Set rest to after the right end of the infix expression.

PairP Pair::paren_infix(PairPP rest, int old_priority, bool inparen)
{

	if ( !*rest) return this;	// this==left_arg

	NodeP infix_proc = NodeP((*rest)->first());

	if (!infix_proc) return this;

	int priority = infix_proc->priority();

	if (priority==priority_STOP  ||  priority <= old_priority) 
		return this;

	next(*rest);
	PairP retval=0;
	retval = paren_expr(rest, inparen);
	retval = retval->paren_infix(rest, priority, inparen);

	retval = new Line(infix_proc, 
					new Line(this, 
						new Line(retval, 0)));

	return retval->paren_infix(rest, old_priority, inparen);
}

PairP Pair::paren_line() 
// Fully parenthesize a complete line, i.e. transform it from a flat list to a tree.
{
	if (!this) return this;
	PairP line=this, 
		retval=0;	retval = paren_expr(&line, false);
//	c_out << "paren_line " << retval << endL;

	if (!throwing  &&  retval!=Unbound) 
	{	retval = retval->paren_infix(&line, -1, false);
		retval = new Line( retval, line->paren_line() );
//		c_out << "paren_line2 " << retval << endL;
	}
	return retval;
}

void Pair::make_runparse()
{
	((Runparsed*)this)->setparsed( runparse() );
	settype_RUNPARSED();
}

void Pair::make_tree() // Treeify a list of tokens (runparsed or not).
{
	if (!this
	|| (type==TREE  &&  ((Tree*)this)->generation()==Tree::the_generation) )
		return;

	if (type!=RUNPARSED)
		make_runparse();

	RunparsedP runp = (Runparsed*)this;
	Tree::dont_know_how_to=false;
	PairP tree=0;
	tree = runp->parsed()->paren_line();
//c_out << "make_tree:" << tree << endL;

	if (tree  &&  tree != Unbound) 
	{
		runp->settype_TREE();
		TreeP(runp)->settree(tree);

// Kludge! (temporarily commented out for making colonless local vars work)
		if (/*Tree::dont_know_how_to ||*/ throwing)
			TreeP(runp)->setgeneration( Unbound );
		else
			TreeP(runp)->setgeneration( Tree::the_generation );
	}
}

void Pair::make_tree_from_body() // Treeify a procedure`s body by appending the trees of the lines.
{
	Tree *body = (Tree*) this;

	if (!body
	||  (type==TREE  &&  body->generation()==Tree::the_generation) )
		return;

//	c_out << "B" << body << endL;
	settype_TREE();
	body->settree(NIL); 
	body->setgeneration(0);

	PairP end_ptr=0;

	for (PairP body_ptr=body;  body_ptr;  next(body_ptr)) 
	{
		PairP line = PairP(body_ptr->first());

		if (!line) continue;  // skip blank line

		this_line=LineP(line);
//		c_out << "maketfb " << line << endL;
		line->make_tree();
		if (line->type==TREE) 
		{	TreeP tree = TreeP(TreeP(line)->tree());
			tree->setline(body_ptr->first());
//			c_out << "E" << end_ptr << "T" << tree << endL;


			if (end_ptr)
			{	end_ptr->settail( tree );
/*
if (tree->circular())	// there`s a CIRCULAR-LIST-BUG somewhere in the parser !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
{	c_out << "CIRC";
	TreeP t=tree;
	for (int i=20;  i && t;  i--)
	{	if (t->first())	
			c_out << t->first() << endL;
		next(t);
	}
	console->getch();
	end_ptr->settail(0);	// really bad slow ugly KLUDGE!!!
}
*/
			}else
			{	body->settree(tree);
				body->setgeneration(Tree::the_generation);
			}
			if ( TreeP(body_ptr->first())->generation()==Unbound )
				body->setgeneration(Unbound);

			while (tree->tail())
				next(tree);

			end_ptr=tree;
		}else				// error while treeifying
		{	body->untreeify();
			return;
		}
	}
}
