#ifdef WIN32
	#include <crtdbg.h>
#endif
	
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <wx/string.h>

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Tree.h"
#include "Array.h"

#include "Error.h"
#include "Init.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

StringP File::line=0;

bool is_newline   (char c) {return c=='\n' || c=='\r';}
bool is_whitespace(char c) 
{	return c==' ' || c=='\t' || is_newline(c) || c=='\0';
}

StringP File::read(charP prompt) 
{
	static const int line_max=5000;
	static int len=line_max, end=0;

	static charP ender="\nEND";
	charP lookfor=ender;
	int dribbling, 
		vbar=0, paren=0, bracket=0, brace=0, pos=0, contin=1, insemi=0, raw=0;

	if (!strcmp(prompt, "RW"))	// called by readWord
	{	prompt="";
		contin=0;
	}
	if (!strcmp(prompt, "RAW")) // called by readRawLine
	{	prompt = "";
		contin = 0;
		raw = 1;
	}
	charmodeOff();
	dribbling=0;

	if (!line) 
	{	line = new String(line_max);
		if (!line) 
			return StringP(Unbound);

		end = line_max-1;
	}
	if (file()==stdin && *prompt) 
		if (interactive)
			printPrompt(prompt);

	if (file()==stdin) 
	{	File::inputBlocking++;
		erract_errtype=FATAL;
		Line::fn=Line::findFile(StringP(new String("unnamed")));
	}

	bool saw_end=false, saw_end_white=false;
	char c=0;
	NodeType this_type=STRING;
	Line::nr=Line::readnr;

	if (!setjmp(inputBlockingBuffer)) 
	{	c=read_getc();

		while (c!=EOF 
		&& (vbar || paren || bracket || brace || c !='\n' || File::toPending)
		&& !throwing) 
		{
			if (dribbling) dribbler.putc_(c);

			if (!raw  &&  c=='\\'  &&  (c=read_getc())!=EOF) 
			{	if (dribbling) 
					dribbler.putc_(c);

				c = backslash.encode(c);
				this_type=BACKSLASHED;
				
				if (c==backslash.encode('n') && file()==stdin) 
					if (interactive) 
						printPrompt("\\ ");
			}
			if (c!=EOF) 
			{	if (pos >= end)
				{	len += line_max;
					line = new String(line->ptr(), 0, len, STRING, String::copy);
					end = len-1;
				}
				line->ptr()[pos]=char(c);
				pos++;
			}
			if (raw)
			{	c=read_getc();
				continue;
			}

			if (saw_end)
			{	if (is_whitespace(c))
				{	//error(DEEPEND, deepend_proc_name);
					saw_end_white=true;
					pos-=5;
					if (pos < 0) 
						pos=0;
					break;
				}
				else saw_end=false;
			}
			else if (toupper(c) == *lookfor)
			{	lookfor++;
				if (*lookfor==0) saw_end=true;
			}else if (pos==1 && toupper(c)==lookfor[1])
			{	lookfor++; lookfor++;
				if (*lookfor==0) saw_end=true;
			}else
				lookfor=ender;
			
			if (c=='|') 
			{	vbar=!vbar;	this_type=VBAR_STRING;
			}
			else if (contin && !vbar && !insemi) 
			{	if (c=='(') 					paren++;
				else if (c==')' && paren)	paren--;
				else if (c=='[') 				bracket++;
				else if (c==']' && bracket)bracket--;
				else if (c=='{') 				brace++;
				else if (c=='}' && brace)	brace--;
				else if (c==';') 				insemi++;
			}
			if (this_type==STRING && strchr(backslash.chars, c))
				this_type=VBAR_STRING;

			if (// (vbar || paren ...) &&
					c=='\n') 
			{	insemi=0;	
				//if (file()==stdin &&  interactive) 
				//	zrd_print_prompt(vbar ?   "| " : "~ ");
			}
			while (!vbar && c=='~' && (c=read_getc())!=EOF) 
			{	while (c==' ' || c=='\t')
					c=read_getc();
				
				if (dribbling) dribbler.putc_(c);
				if (pos >= end)
				{	len += line_max;
					line = new String(line->ptr(), 0, len, STRING, String::copy);
					end = len-1;
				}
				line->ptr()[pos]=char(c);
				pos++;
				
				if (c=='\n' && file()==stdin) 
				{	insemi=0;	//if (interactive) zrd_print_prompt("~ ");
				}
			}
			if (c!=EOF) c=read_getc();
		}
	}
	if (saw_end && !saw_end_white) pos-=4;
	line->ptr()[pos]='\0';
	File::inputBlocking=0;

	if (dribbling) dribbler.putc_('\n');
	if (c==EOF && file()==stdin) 
	{
		if (interactive) 
			clearerr(stdin);
		printPrompt("\n");
	}
	if (pos==0) 
		return StringP(_Null_Word);		// so emptyp works
	
	return String::make(line->ptr(), 0, strlen(line->ptr()), 
		this_type, String::copy);
}

PairP parser_iterate(charPP inln, charP inlimit, String::HeadP inhead, bool semi, int endchar) 
{
	int windex=0, vbar=0;
	bool broken=false;

	char ch;
	charP wptr=0;
	static char terminate='\0';   // KLUDGE

	NodeType this_type=STRING;
	PairP tnode=0, lastnode=0, outline=0;
	do
	{	ch = **inln;			// get the current character and increase pointer

		if (!vbar && windex==0)
			wptr = *inln;

		if (++(*inln) >= inlimit)
			*inln = &terminate;
		
								 // skip through comments and line continuations
		while (!vbar && ( (semi && ch==';') || (ch=='~' && is_newline(**inln)) )) 
		{	while (ch=='~' && is_newline(**inln)) 
			{	Line::nr++;
				if (++(*inln) >= inlimit) 
					*inln = &terminate;
				ch = **inln;
				if (windex==0) 
					wptr = *inln;
				else 
				{	if (**inln==']' || **inln=='[' || **inln=='{' || **inln=='}') 
					{	ch=' ';
						break;
					}else 
						broken=true;
				}
				if (++(*inln) >= inlimit) 
					*inln = &terminate;
			}
			if (semi && ch==';') 
			{
				if (!is_newline(**inln))
					do{ ch = **inln;
						if (windex==0) 
							wptr=*inln;
						else 
							broken=true;

						if (++(*inln) >= inlimit)
							*inln = &terminate;
					}while (ch !='\0' && ch !='~' && !is_newline(**inln));

				if (ch !='\0' && ch !='~') 
				{	ch='\n';
					++(*inln);
				}
			}
		}
		if (backslash.is(ch))	// flag that this word will be of BACKSLASH_STRING type
			this_type=BACKSLASHED;
			
		if (ch=='|') 
		{	vbar = !vbar;
			this_type=VBAR_STRING;
			broken=true;	// so we'll copy the chars
		}else 
			if ( vbar 
			|| (!is_whitespace(ch) && ch !=']' &&	ch !='{' && ch !='}' && ch !='[') )
				windex++;

		if (vbar) continue;
		if (ch==endchar) break;

		switch (ch)
		{
		case ']':	error(UNEXPECTED_BRACKET, 0);	break;
		case '}':	error(UNEXPECTED_BRACE     , 0);    break;

		case '[':	// parse a new list
			{
				StringP fn=Line::fn;
				int tmplinenr=Line::nr;

				NodeP nd=parser_iterate(inln, inlimit, inhead, semi, ']');
				
				StringP fn2=Line::fn;
				int tmplinenr2=Line::nr;

				Line::nr=tmplinenr;
				Line::fn=fn;
				tnode = new Line( nd, 0 );
				Line::nr=tmplinenr2;
				Line::fn=fn2;

				if (**inln=='\0') ch='\0';
				break;
			}
		case '{': 
		{		PairP p=parser_iterate(inln, inlimit, inhead, semi, '}');
				if (!p)
				{	error(UNEXPECTED_BRACE, 0);    break;
				}
				tnode = new Line(p->array(), 0);
				if (**inln=='@') 
				{	int i=0, sign=1;
					(*inln)++;
					if (**inln=='-') 
					{	sign=-1;	(*inln)++;
					}
					while ((ch=**inln) >='0' && ch <='9')	// Ooops! No expressions as array origins allowed yet!
					{	i=(i*10) + ch - '0';
						(*inln)++;
					}
					ArrayP(tnode->first())->setorigin( sign*i );
				}
				if (**inln=='\0') ch='\0';
				break;
		}
		default:	// if this character or the next one will terminate string, make the word
				if (is_whitespace(ch) || **inln==']' || **inln=='[' || **inln=='{' || **inln=='}') 
				{	if (windex > 0) 
					{	if (broken==false)
							tnode = new Line
								(	String::make(wptr, inhead, windex, this_type, String::copy),
									0
								);
						else 
						{	tnode = new Line
								(	String::make
									(	wptr, 0, windex, this_type, 
									semi ?   String::copy_mend  :  String::copy_mend_nosemi 
									), 
									0
								);
							broken=false;
						}
						this_type=STRING;
						windex=0;
					}
					if (is_newline(ch))
					{	Line::nr++;
//						c_out << "parser_iterate " << Line::nr << " " << outline << endL;
					}
				}
		}
		if (tnode)	// put the word onto the end of the return list
		{	if (outline)
				lastnode->settail( tnode );
			else
				outline=tnode;
			
			lastnode=tnode;
			tnode=0;
		}
	}while (ch);
	return outline;
}

PairP String::parser(bool semi) 
{
	if (!this)
		return new Line(0,0);
	StringP str=0;	str = string();
	if (str==Unbound)
		return new Line(0,0);
	int slen = str->length();
	charP lnsav = str->ptr();

	if (!lnsav)
	{	c_out << this << endL; 
		return new Line(0, 0);
	}
	PairP ret=0;
	ret = parser_iterate(&lnsav, lnsav+slen, str->head(), semi, -1);
	return ret;
}

NodeP lparse(PairPP arg) 
{
	StringP s = arg[1]->a_string(1);
	if (!throwing) 
		return s->parser(false);
	return Unbound;
}
