//___________________________________________________________________
//
//                           Class Pen
//
//                          Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/glcanvas.h>
#include <math.h>
#include "Main.h"
#include "Pair.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "File.h"
#include "String.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#define RGBA(r,g,b,a)  ((unsigned int )(((unsigned char)(r)      \
							| ((unsigned short)((unsigned char)(g))<<8)) \
							|(((unsigned int  ) (unsigned char)(b))<<16)) \
							|(((unsigned int  ) (unsigned char)(a))<<24))

Color Pen::palette[16]=
{
	RGBA(  0,   0,   0, 255), /* black */
	RGBA(  0,   0, 255, 255), /* blue */
	RGBA(  0, 255,   0, 255), /* green */
	RGBA(  0, 255, 255, 255), /* cyan */
	RGBA(255,   0,   0, 255), /* red */
	RGBA(255,   0, 255, 255), /* magenta */
	RGBA(255, 255,   0, 255), /* yellow */
	RGBA(255, 255, 255, 255), /* white */
	RGBA(155,  96,  59, 255),    /* brown */
	RGBA(197, 136,  18, 255), /* tan */
	RGBA(100, 162,  64, 255), /* forest */
	RGBA(120, 187, 187, 255), /* aqua */
	RGBA(255, 149, 119, 255), /* salmon */
	RGBA(144, 113, 208, 255), /* purple */
	RGBA(255, 163,   0, 255), /* orange */
	RGBA(183, 183, 183, 255), /* gray */
};

Pen::Pen()
{
	down=true; 
	wide=false;
	roundLineEnds=true;
	size=1;
	mode=WIN_DOWN;
	internalMode=Paint; 
	color=rawcolor=floodcolor=RGBA(0,0,0,255);
	pattern=0xFFFFFFFF;
}

Pen::~Pen() {}

Pen &Pen::operator=(const Pen &b)
{
	down=b.down; wide=b.wide; roundLineEnds=b.roundLineEnds;
	internalMode=b.internalMode; mode=b.mode; 
	size=b.size; size_rotated=b.size_rotated;
	color=b.color; rawcolor=b.rawcolor; floodcolor=b.floodcolor;
	pattern=b.pattern;
	return *this;
}

void Pen::create(Color color_)
{
	rawcolor=color_;
	color=realcolor(color_);
	glColor4ubv((unsigned char*)&color);
//	glLineWidth(size_rotated.x);
}

void Pen::destroy() {}

Color Pen::rgb(double r, double g, double b)
{	int ir,ig,ib;
	ir=abs(int(r*0xff)); if (ir > 0xff) ir=0xff;	if (ir < 0) ir=0;
	ig=abs(int(g*0xff)); if (ig > 0xff) ig=0xff;	if (ig < 0) ig=0;
	ib=abs(int(b*0xff)); if (ib > 0xff) ib=0xff;	if (ib < 0) ib=0;
	return ir|(ig<<8)|(ib<<16)|0xff000000;
}

Color Pen::rgba(double r, double g, double b, double a)
{
	int ir,ig,ib,ia;
	ir=abs(int(r*0xff)); if (ir > 0xff) ir=0xff;	if (ir < 0) ir=0;
	ig=abs(int(g*0xff)); if (ig > 0xff) ig=0xff;	if (ig < 0) ig=0;
	ib=abs(int(b*0xff)); if (ib > 0xff) ib=0xff;	if (ib < 0) ib=0;
	ia=abs(int(a*0xff)); if (ia > 0xff) ia=0xff;	if (ia < 0) ia=0;
	return ir|(ig<<8)|(ib<<16)|(ia<<24); 
}

Color Pen::hsba(double h, double s, double b_, double a)
{
	double h2,L,x,y,z,r,g,b;
	h=fmod(h,360)/120;
	h2=fmod(h,1);
	if (h2 < 0.5)
	{	L=b_*(1+s*h2);
		s=s/(1+s*h2);	//s>0, h2>0 always so never s/0
		x=1+(-1-h2)*s;	y=1-h2*s;	z=1-(1-h2)*s;
	}else
	{	L=b_*(1+s*(1-h2));
		s=s/(1+s*(1-h2));
		y=1-h2*s;	z=1-(1-h2)*s;	x=1-(2-h2)*s;
	}
	x=x*L; 	y=y*L;	z=z*L;
	if (h < 1)
	{	b=x; r=y; g=z;
	}else	if (h < 2)
	{	r=x;	g=y;	b=z;
	}else
	{	g=x;	b=y;	r=z;
	}
	if (r < 0) r=0;
	if (g < 0) g=0;
	if (b < 0) b=0;
	if (r > 1) r=1;
	if (g > 1) g=1;
	if (b > 1) b=1;
	return rgba(r,g,b,a);
}

Color Pen::hsb(double h, double s, double b)
{	return hsba(h,s,b,1);
}

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

PairP Pen::rehsb(double R, double G, double B)
{
	double Min_, Max_, delta, H, S, L;
	Min_ = min(min(R,G),B);	// Min. value of RGB
	Max_ = max(max(R,G),B);	// Max. value of RGB
	delta = Max_-Min_;	// Delta RGB value
	
	L = Max_;

	if (delta==0)	// This is a gray, no chroma...
	{	H = 0;	// HSB results = 0  1
		S = 0;
	}else		// Chromatic data...
	{	double deltaR, deltaG, deltaB;
		deltaR = ((Max_-R)/6+delta/2)/delta;
		deltaG = ((Max_-G)/6+delta/2)/delta;
		deltaB = ((Max_-B)/6+delta/2)/delta;

		S = 1-Min_/Max_;

		if (R==Max_)
			H = deltaB-deltaG;
		else if (G==Max_)
			H = deltaR-deltaB+0.33333333;
		else if (B==Max_)
			H = deltaG-deltaR+0.66666666;
		if (H < 0) H += 1;
		if (H > 1) H -= 1;
	}
	H=H*360;
	return
		new Pair(new Float(H), 
		new Pair(new Float(S), 
		new Pair(new Float(L), 0)));
}

PairP Pen::rehsba(double R, double G, double B, double A)
{
	double Min_, Max_, delta, H, S, L;
	Min_ = min(min(R,G),B);	// Min. value of RGB
	Max_ = max(max(R,G),B);	// Max. value of RGB
	delta = Max_-Min_;	// Delta RGB value
	
	L = Max_;

	if (delta==0)	// This is a gray, no chroma...
	{	H = 0;	// HSB results = 0  1
		S = 0;
	}else		// Chromatic data...
	{	double deltaR, deltaG, deltaB;
		deltaR = ((Max_-R)/6+delta/2)/delta;
		deltaG = ((Max_-G)/6+delta/2)/delta;
		deltaB = ((Max_-B)/6+delta/2)/delta;

		S = 1-Min_/Max_;

		if (R==Max_)
			H = deltaB-deltaG;
		else if (G==Max_)
			H = deltaR-deltaB+0.33333333;
		else if (B==Max_)
			H = deltaG-deltaR+0.66666666;
		if (H < 0) H += 1;
		if (H > 1) H -= 1;
	}
	H=H*360;
	return
		new Pair(new Float(H), 
		new Pair(new Float(S), 
		new Pair(new Float(L), 
		new Pair(new Float(A), 0))));
}

Color Pen::rgb(int c) 
{	unsigned char i= c&8 ?  0xff : 0x7f;
	return
		 (c&1 ?    i<<16 : 0) 
		|(c&2 ?    i<< 8 : 0)
		|(c&4 ?    i     : 0)
		|(c&8 ? 0x404040 : 0)
		|0xff000000;
}

int Pen::getPattern(){return pattern;}
void Pen::setPattern(int pattern_){pattern=pattern_;}
	// sorry, no patterns yet!
void Pen::setListPattern(NodeP) {}

NodeP Pen::getNodePenPattern() 
{
	return new Pair
		(	new Int(pattern),
			0
		);
}
