//___________________________________________________________________
//
//                           Class Piper
//
//        Execute a process and redirect its output to stdout.
//                         single-threaded!
//___________________________________________________________________

#include <crtdbg.h>

#include <windows.h>
#include <conio.h>
#include <stdio.h>

#include <wx/string.h>

#include "Main.h"
#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Piper.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BOOL Piper::bRunThread;
HANDLE Piper::hChildProcess, Piper::hStdIn; // Handle to parents std input.
 
void Piper::displayError(charP pszAPI)
{
	LPVOID lpvMessageBuffer;
	CHAR szPrintBuffer[512];
//	DWORD nCharsWritten;
	
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
		0, GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR)&lpvMessageBuffer, 0, 0);

       wsprintf(szPrintBuffer,
         "ERROR: API    = %s.\n   error code = %d.\n   message    = %s.\n",
                pszAPI, GetLastError(), charP(lpvMessageBuffer));

//       WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), szPrintBuffer,
//                     lstrlen(szPrintBuffer), &nCharsWritten, NULL);
		c_out << szPrintBuffer;
		 
//	 _getch();
	LocalFree(lpvMessageBuffer);
//	ExitProcess(GetLastError());
}

void Piper::prepAndLaunchRedirectedChild(const char *commandLine,
						     HANDLE hChildStdOut, HANDLE hChildStdIn, HANDLE hChildStdErr)
{
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	
	// Set up the start up info struct.
	ZeroMemory(&si, sizeof(STARTUPINFO));
	si.cb=sizeof(STARTUPINFO);
	si.dwFlags=STARTF_USESTDHANDLES;// || STARTF_USESHOWWINDOW;
	si.hStdOutput=hChildStdOut;
	si.hStdInput   =hChildStdIn;
	si.hStdError  =hChildStdErr;
	//si.wShowWindow=SW_HIDE;
	
	// Launch the process that you want to redirect (in this case,
	// Child.exe). Make sure Child.exe is in the same directory as
	// redirect.c launch redirect from a command line to prevent location confusion.

	if (!CreateProcess(0, charP(commandLine), 0, 0,TRUE, CREATE_NEW_CONSOLE, 0, 0, &si, &pi))
		displayError("CreateProcess");	
	
	hChildProcess=pi.hProcess; // Set global child process handle to cause threads to exit.

	if (!CloseHandle(pi.hThread)) 
		displayError("CloseHandle"); // Close any unnecessary handles.
}

DWORD WINAPI Piper::getAndSendInputThread(LPVOID lpvThreadParam)
{
	char read_buff[0x10000];
	DWORD nBytesRead, nBytesWrote;
	HANDLE hPipeWrite=(HANDLE)lpvThreadParam;
	
	while (bRunThread) // Get input from our console and send it to child through the pipe.
	{	if(!ReadConsole(hStdIn, read_buff, 1, &nBytesRead, 0))
			displayError("ReadConsole");
		
		if (!WriteFile(hPipeWrite, read_buff, nBytesRead, &nBytesWrote, 0))
		{	if (GetLastError() == ERROR_NO_DATA)
				break; // Pipe was closed (normal exit path).
			else
				displayError("WriteFile");
		}
	}
	return 1;
}

void Piper::readAndHandleOutput(HANDLE hPipeRead)
{
	const unsigned bufferSize=0x10000;
	char buffer[bufferSize];
	DWORD nBytesRead;
//	DWORD nCharsWritten;
	NodeP line=0;
	PairP tnode=0;
	
Loop:
	{	if (!ReadFile(hPipeRead, buffer, sizeof(buffer), &nBytesRead, 0) || !nBytesRead)
		{	if (GetLastError() == ERROR_BROKEN_PIPE)
				goto end; // pipe done - normal exit path.
			else
				displayError("ReadFile"); // Something bad happened.
		}
		if (nBytesRead > 0  &&  nBytesRead < bufferSize)
		{	//buffer[nBytesRead]='\0';
//			print(stdout, "%t", buffer);

			unsigned start=0, end=0;
			while (start < nBytesRead)
			{	while (end < nBytesRead  &&  buffer[end] != '\n')
					end++;

//				buffer[end-1]='\n';
				buffer[end]='\0';

				if (wordmode)
					line = String::make(&buffer[start], 0, end-start, STRING, String::copy);
				else
					line = ( new String(&buffer[start]) )->parser(false);
				
				if (output==NIL) 
					tnode=output = new Pair(line, 0);
				else 
				{	tnode->settail( new Pair(line, 0) );
					next(tnode);
				}
				end++;
				start=end;
			}
		}
//		if (!WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), buffer, nBytesRead,&nCharsWritten, 0))
//			displayError("WriteConsole");
	}
	goto Loop;
end:
	;
}

Piper::Piper(const char *commandLine, bool wordmode_)
{
	output=0;
	wordmode=wordmode_;
	bRunThread=TRUE;
	hChildProcess=hStdIn=0;
	
	HANDLE 
		hOutputReadTmp,hOutputRead,hOutputWrite,
		hInputWriteTmp,hInputRead,hInputWrite,
		hErrorWrite,
		hThread;
	
	DWORD ThreadId;
	SECURITY_ATTRIBUTES sa;
	
	sa.nLength=sizeof(SECURITY_ATTRIBUTES); // Set up the security attributes struct.
	sa.lpSecurityDescriptor=0;
	sa.bInheritHandle=TRUE;
	
	if (!CreatePipe(&hOutputReadTmp, &hOutputWrite, &sa, 0))// Create the child output pipe.
		displayError("CreatePipe");

	// Create a duplicate of the output write handle for the std error
	// write handle. This is necessary in case the child application
	// closes one of its std output handles.
	if (!DuplicateHandle(GetCurrentProcess(), hOutputWrite,
		GetCurrentProcess(), &hErrorWrite, 0,
		TRUE, DUPLICATE_SAME_ACCESS))
		displayError("DuplicateHandle");
	
	if (!CreatePipe(&hInputRead, &hInputWriteTmp, &sa, 0))	// Create the child input pipe.
		displayError("CreatePipe");
		
	// Create new output read handle and the input write handles. Set
	// the Properties to FALSE. Otherwise, the child inherits the
	// properties and, as a result, non-closeable handles to the pipes
	// are created.
	if 
	(!DuplicateHandle(GetCurrentProcess(),hOutputReadTmp,
		GetCurrentProcess(),
		&hOutputRead, // Address of new handle.
		0, FALSE, // Make it uninheritable.
		DUPLICATE_SAME_ACCESS)
	)
		displayError("DupliateHandle");
	
	if 
	(!DuplicateHandle(GetCurrentProcess(),hInputWriteTmp,
		GetCurrentProcess(),
		&hInputWrite, // Address of new handle.
		0, FALSE, // Make it uninheritable.
		DUPLICATE_SAME_ACCESS)
	)
		displayError("DupliateHandle");
		
	// Close inheritable copies of the handles you do not want to be
	// inherited.
	if (!CloseHandle(hOutputReadTmp)) 
		displayError("CloseHandle");

	if (!CloseHandle(hInputWriteTmp)) 
		displayError("CloseHandle");
	
	// Get std input handle so you can close it and force the ReadFile to
	// fail when you want the input thread to exit.
	if ( (hStdIn=GetStdHandle(STD_INPUT_HANDLE))==INVALID_HANDLE_VALUE )
		displayError("GetStdHandle");
	
	prepAndLaunchRedirectedChild(commandLine, hOutputWrite,hInputRead,hErrorWrite);
	
	// Close pipe handles (do not continue to modify the parent).
	// You need to make sure that no handles to the write end of the
	// output pipe are maintained in this process or else the pipe will
	// not close when the child process exits and the ReadFile will hang.

	if (!CloseHandle(hOutputWrite)) 
		displayError("CloseHandle");

	if (!CloseHandle(hInputRead )) 
		displayError("CloseHandle");

	if (!CloseHandle(hErrorWrite)) 
		displayError("CloseHandle");
	
	// Launch the thread that gets the input and sends it to the child.
	hThread = CreateThread(0, 0, getAndSendInputThread, (LPVOID)hInputWrite, 0, &ThreadId);

	if (hThread==0) 
		displayError("CreateThread");

	readAndHandleOutput(hOutputRead); // Read the child's output. Redirection is complete!

	if (!CloseHandle(hStdIn)) 
		displayError("CloseHandle"); // Force the read on the input to return by closing the stdin handle.
	
	bRunThread=FALSE; // Tell the thread to exit and wait for thread to die.
	
	if (WaitForSingleObject(hThread,INFINITE)==WAIT_FAILED)
		displayError("WaitForSingleObject");
	
	if (!CloseHandle(hOutputRead)) 
		displayError("CloseHandle");

	if (!CloseHandle(hInputWrite)) 
		displayError("CloseHandle");
}
/**
main()
{
	Piper("ConSpawn c:/command.com /c dir");
	return 0;
}
/**/