#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/slider.h>
#include <wx/button.h>
#include <wx/radiobut.h>
#include <wx/checkbox.h>
#include <wx/stattext.h>
#include <wx/glcanvas.h>
#include <wx/tglbtn.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/sizer.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <GL/glu.h>

#include <m_apm.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Line.h"
#include "Tree.h"
#include "Procedur.h"
#include "Object.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "BigFloat.h"
#include "Point.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Table.h"
#include "Struct.h"
#include "TurtleNode.h"
#include "Bitmap.h"
#include "DL.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Frame.h"
#include "Graph.h"
#include "BoxSizer.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Color.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

void File::printIndents(FILEP strm)
{
	for (int i=0;  i < indent;  i++)
		print(strm, '\t');
}

void File::print(FILEP strm, unsigned char i)
{
	char ch, format[5], ca[15];
	charP cp;
	if (digits)
	{	sprintf(format, "%%%ii", digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::print(FILEP strm, short i)
{
	char ch, format[5], ca[15];
	charP cp;
	if (digits)
	{	sprintf(format, "%%%ii", digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::print(FILEP strm, int i)
{
	char ch, format[5], ca[15];
	charP cp;
	if (digits)
	{	sprintf(format, "%%%ii", digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%i", i);
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::printBin(FILEP strm, char i)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen--;
		if (File::stringlen >= 0)     // printing to string
			*File::stringptr++ = i;
		else
			error(FILE_ERROR, new String
				("memory block too small to print int8 to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	print(strm, i);
		return;
	}
	fwrite(&i, sizeof(char), 1, strm);
}

void File::printBin(FILEP strm, unsigned char i)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen--;
		if (File::stringlen >= 0)     // printing to string
			*File::stringptr++ = char(i);
		else
			error(FILE_ERROR, new String
				("memory block too small to print uint8 to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	print(strm, char(i));
		return;
	}
	fwrite(&i, sizeof(char), 1, strm);
}

void File::printBin(FILEP strm, short i)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen-=sizeof(short);
		if (File::stringlen >= 0)     // printing to string
		{	*shortP(File::stringptr) = i;
			File::stringptr+=sizeof(short);
		}
		else
			error(FILE_ERROR, new String
				("memory block too small to print int16 to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(&i);
		for (int l=sizeof(short); l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(&i, sizeof(short), 1, strm);
}

void File::printBin(FILEP strm, int i)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen-=sizeof(int);
		if (File::stringlen >= 0)     // printing to string
		{	*intP(File::stringptr) = i;
			File::stringptr+=sizeof(int);
		}
		else
			error(FILE_ERROR, new String
				("memory block too small to print int to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(&i);
		for (int l=sizeof(int); l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(&i, sizeof(int), 1, strm);
}

void File::printHex(FILEP strm, int i)
{
	char ch, format[5], ca[15];
	charP cp;
	if (digits)
	{	sprintf(format, "%%%ix", digits);
		sprintf(ca, format, i);
	}else
		sprintf(ca, "%x", i);
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::print(FILEP strm, double d)
{
	char ch, format[10], ca[200];
	charP cp;
	if (digits)
	{	sprintf(format, "%%%i.%ig", digits, precision);
		sprintf(ca, format, d);
	}else
	{	sprintf(format, "%%-.%ig", precision);
		sprintf(ca, format, d);
	}
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::printBin(FILEP strm, double d)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen-=sizeof(double);
		if (File::stringlen >= 0)     // printing to string
		{	*doubleP(File::stringptr) = d;
			File::stringptr+=sizeof(double);
		}
		else
			error(FILE_ERROR, new String
				("memory block too small to print float to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(&d);
		for (int l=sizeof(double); l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(&d, sizeof(double), 1, strm);
}

void File::print(FILEP strm, MAPM d)
{
	wxString s;
	if (d.is_integer())
	{	s.Alloc(d.IntegerStringSize());
		d.toIntegerString(charP(s.c_str()));
	}
	else
	{	s.Alloc(d.StringSize(-1));
		d.toString(charP(s.c_str()), -1);
	}
	char ch;
	charP cp=(charP)s.c_str();
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::print(FILEP strm, complexd c)
{
	char ch, format[20], ca[400];
	charP cp;
	if (digits)
	{	if (c.imag() >= 0)
			sprintf(format, "%%%i.%ig+%%%i.%igi", 
				digits, precision, digits, precision);
		else
			sprintf(format, "%%%i.%ig%%%i.%igi", 
				digits, precision, digits, precision);
		sprintf(ca, format, c.real(), c.imag());
	}else
	{	if (c.imag() >= 0)
			sprintf(format, "%%-.%ig+%%-.%igi", precision, precision);
		else
			sprintf(format, "%%-.%ig%%-.%igi", precision, precision);
		sprintf(ca, format, c.real(), c.imag());
	}
	cp=ca;
	while ((ch=*cp++) !='\0')
		print(strm, ch);
}

void File::printBin(FILEP strm, complexd c)
{
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen-=sizeof(complexd);
		if (File::stringlen >= 0)     // printing to string
		{	*complexdP(File::stringptr) = c;
			File::stringptr+=sizeof(complexd);
		}
		else
			error(FILE_ERROR, new String
				("memory block too small to print complex to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(&c);
		for (int l=sizeof(complexd); l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(&c, sizeof(complexd), 1, strm);
}

void File::print(FILEP strm, ArrayP array, int depth, int width)
{
	int i=0, wid,  length=array->length();
	NodePP pp=array->data();
						
	if (width < 0) 
		wid=length;
	else 
		wid= length > width ?   width  :  length;

	print(strm, '{');

	while (i < wid) 
	{	bool was_array;
		if (*pp && (*pp)->type==ARRAY)
		{	indent++;
			print(strm, '\t');	
			print(strm, *pp++, depth-1, width);	
			print(strm, endL);
			indent--;
			printIndents(strm);
			was_array=true;
		}else
		{	print(strm, *pp++, depth-1, width);
			was_array=false;
		}
		if (++i < length)
			if (!was_array) 
				print(strm, ' ');
	}
	if (wid < length) 
		print(strm, "...");
	print(strm, '}');

	if (/*backslashes && */(array->origin() != 1)) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", array->origin());
		print(strm, org);
	}
}

void File::printBin(FILEP strm, ArrayP array)
{
	int i=0, wid=array->length();
	NodePP pp=array->data();

	while (i++ < wid) 
		printBin(strm, *pp++);	
}

void File::print(FILEP strm, IntArrayP array, int width)
{
	int i=0, wid,  length=array->length();
	intP pp=array->data();
						
	if (width < 0) 
		wid=length;
	else 
		wid= length > width ?   width  :  length;

	print(strm, '{');

	while (i < wid) 
	{	print(strm, *pp++);	
		if (++i < length)
			print(strm, ' ');
	}
	if (wid < length) 
		print(strm, "...");
	print(strm, '}');

	if (array->origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", array->origin());
		print(strm, org);
	}
}

void File::printBin(FILEP strm, IntArrayP array)
{
	int wid=array->length();
	intP pp=array->data();
	if (!strm)
	{	if (File::stringlen > 0)     // printing to string
			File::stringlen-=sizeof(int)*wid;
		if (File::stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*intP(File::stringptr) = *pp++;
				File::stringptr+=sizeof(int);
			}
		else
			error(FILE_ERROR, new String
				("memory block too small to print intarray to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(array->data());
		for (int l=sizeof(int)*wid; l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(pp, sizeof(int), wid, strm);
}

void File::print(FILEP strm, Int16ArrayP array, int width)
{
	int i=0, wid,  length=array->length();
	shortP pp=array->data();
						
	if (width < 0) 
		wid=length;
	else 
		wid= length > width ?   width  :  length;

	print(strm, "{");

	while (i < wid) 
	{	print(strm, int(*pp++));	
		if (++i < length)
			print(strm, ' ');
	}
	if (wid < length) 
		print(strm, "...");
	print(strm, '}');

	if (array->origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", array->origin());
		print(strm, org);
	}
}

void File::printBin(FILEP strm, Int16ArrayP array)
{
	int wid=array->length();
	shortP pp=array->data();
	if (!strm)
	{	if (File::stringlen > 0)
			File::stringlen-=sizeof(short)*wid;
		if (File::stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*shortP(File::stringptr) = *pp++;
				File::stringptr+=sizeof(short);
			}
		else
			error(FILE_ERROR, new String
				("memory block too small to print int16array to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(array->data());
		for (int l=sizeof(int)*wid; l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(pp, sizeof(short), wid, strm);
}

void File::print(FILEP strm, FloatArrayP array, int width)
{
	int i=0, wid,  length=array->length();
	doubleP pp=array->data();
						
	if (width < 0) 
		wid=length;
	else 
		wid= length > width ?   width  :  length;


	print(strm, '{');

	while (i < wid) 
	{	print(strm, *pp++);	
		if (++i < length)
			print(strm, ' ');
	}
	if (wid < length) 
		print(strm, "...");
	print(strm, '}');

	if (array->origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", array->origin());
		print(strm, org);
	}
}

void File::printBin(FILEP strm, FloatArrayP array)
{
	int wid=array->length();
	doubleP pp=array->data();
	if (!strm)
	{	if (File::stringlen > 0)     // printing to string
			File::stringlen-=sizeof(double)*wid;
		if (File::stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*doubleP(File::stringptr) = *pp++;
				File::stringptr+=sizeof(double);
			}
		else
			error(FILE_ERROR, new String
				("memory block too small to print floatarray to"));
		return;
	}
	if (File::interactive && strm==stdout)
	{	charP cp=charP(array->data());
		for (int l=sizeof(double)*wid; l; l--)
			print(strm, *cp++);
		return;
	}
	fwrite(pp, sizeof(double), wid, strm);
}

//#define _DEBUG_LINES 1

void File::print(FILEP strm, PairP ndlist, int depth, int width) 
{
	NodeP arg=0;
	bool was_list;
	int wid=width;
	
	while (ndlist) 
	{	if (!ndlist->is.list) return;

		arg=ndlist->first();
		if (checkThrowing()) break;

		switch (ndlist->type)
		{
#ifdef _DEBUG
		case TODO:
			{	EvalCommand evco=ToDoNodeP(ndlist)->toDo.mnemonic;
				if (evco > 0  &&  evco <= end)
					print(strm, "%t", Evaluator::commandName[evco]);
				else print(strm, "unknown EvalCommand");
			}
			break;
#endif
		case NTFREE: print(strm, "NTFREE"); break;
		case PAIR:	if (arg->islist())
						{	indent++;
							print(strm, '\t');	
							print(strm, arg, depth, width);	
							print(strm, '\n');
							indent--;
							printIndents(strm);
							was_list=true;
							break;
						}
						print(strm, arg, depth, width);
						was_list=false;
						break;
#ifdef _DEBUG_LINES
		case TREE:	print(strm, '\n'); 
						print(strm, 'T'); 
						print(strm, TreeP(ndlist)->tree()); 

						print(strm, 't');
		case RUNPARSED:
		case LINE: 	//print(strm, LineP(ndlist)->filename);
						print(strm, LineP(ndlist)->number);
						if (arg->islist())
						{	indent++;
							print(strm, '\t');	
							print(strm, arg, depth, width);	
							print(strm, '\n');
							indent--;
							printIndents(strm);
							was_list=true;
						}
						else
						{	print(strm, arg, depth, width);	
							was_list=false;
						}
						break;
#endif
		default: 
			print(strm, arg, depth, width);
			was_list=false;
		}
		next(ndlist);
		if (ndlist) 
		{	if (!was_list) print(strm, ' ');
			if (--wid==0) 
			{	print(strm, "...");
				break;
			}
		}
	}
}

void File::printBin(FILEP strm, PairP ndlist) 
{
	NodeP arg=0;
	
	while (ndlist) 
	{	if (!ndlist->is.list) return;

		arg=ndlist->first();
		if (checkThrowing()) break;

		switch (ndlist->type)
		{
#ifdef _DEBUG
		case TODO:
			{	EvalCommand evco=ToDoNodeP(ndlist)->toDo.mnemonic;
				if (evco > 0  &&  evco <= end)

					print(strm, "%t", Evaluator::commandName[evco]);
				else print(strm, "unknown EvalCommand");
			}
			break;
#endif
		case NTFREE: print(strm, "NTFREE"); break;
		case PAIR:	if (arg->islist())
						{	printBin(strm, arg);	
							break;
						}
#ifdef _DEBUG_LINES
		case TREE:
		case RUNPARSED:
		case LINE: 	print(strm, LineP(ndlist)->number);
						printBin(strm, arg);
						break;
#endif
		default: 
			printBin(strm, arg);
		}
		next(ndlist);
	}
}

void File::print(FILEP strm, TableP table, int depth, int width)
{
	print(strm, table->contentsNameVar(), depth, width);
}

void File::printBin(FILEP strm, TableP table)
{
	printBin(strm, table->contentsNameVar());
}

void File::print(FILEP strm, StructP struct_, int depth, int width)
{
	print(strm, struct_->list(), depth, width);
}

void File::printBin(FILEP strm, StructP struct_)
{
	for (ObjectP obj=struct_->first();  obj;  obj=ObjectP(obj->proc()))
		printBin(strm, obj->var());
}

void File::print(FILEP strm, NodeP nd, int depth, int width) 
{
	if (depth==0) 
		print(strm, "...");
	else if (!nd) 
	{	print(strm, '[');
		print(strm, ']');
	}
	else if (nd==Unbound) 
		print(strm, "Unbound");
	else if (nd->is.primitive) 
	{	if (nd->type == PRIM)
			print(strm, "PRIM");
		else if (nd->type == NTFREE)
			print(strm, "NTFREE");
	}
	else if (nd->is.list) 
	{	print(strm, '[');
		print(strm, PairP(nd), depth-1, width);
		print(strm, ']');
/*		if (nd->type & (nt_RUNP | nt_TREE | nt_LINE) & ~PAIR)
			if (TreeP(nd)->tree())
			{	print(strm, "O{");
				print(strm, TreeP(nd)->tree(), depth-1, width);
				print(strm, "}");
			}
*/
	}else switch (nd->type)
	{
	case INT_       :	print(strm,       IntP(nd)->int_);			break;
	case FLOAT_     :	print(strm,     FloatP(nd)->float_);		break;
	case BIGFLOAT   :	print(strm,  BigFloatP(nd)->float_);		break;
	case COMPLEX    :	print(strm,  ComplexdP(nd)->complex_);		break;
	case INT16      :	print(strm,     Int16P(nd)->int_);			break;
	case INT8       :	print(strm,short(Int8P(nd)->int_));			break;
	case UINT8      :	print(strm,     UInt8P(nd)->int_);			break;
	case ARRAY      :	print(strm,     ArrayP(nd), depth, width);break;
	case INTARRAY   :	print(strm,  IntArrayP(nd), width);			break;
	case INT16ARRAY :	print(strm,Int16ArrayP(nd), width);			break;
	case FLOATARRAY :	print(strm,FloatArrayP(nd), width);			break;
	case QUOTE      :	print(strm, '\"'); 
							print(strm, PairP(nd)->first());				break;
	case OBJECT     :	print(strm,   ObjectP(nd)->name());			break;
	case PROC_      :	print(strm,ProcedureP(nd)->bodywords());	break;
	case FILE_      :	print(strm,     FileP(nd)->name());			break;
	case COLON      :	print(strm, ':'); 
							print(strm, PairP(nd)->first());				break;
	case TABLE      :	print(strm,    TableP(nd), depth, width);	break;
	case STRUCT     :	print(strm,   StructP(nd), depth, width);	break;
	case BITMAP_    :	print(strm, "Bitmap [");
							print(strm, int(BitmapP(nd)->width)); 
							print(strm, ' ');
							print(strm, int(BitmapP(nd)->height)); 
							print(strm, ']');									break;
	case TEXTURE    :	print(strm, "Texture");							break;
	case GRAPHIC    :	print(strm, "Graphic");							break;
	case DYNAMIC_LIB:	print(strm, 
							(wxString("DynamicLibrary ")
							+DynamicLibraryP(nd)->name()->ptr()));		break;
	case TURTLE_    :
						{	print(strm, "Turtle {"); 
							Turtle *t=TurtleNodeP(nd)->turtle;
							print(strm, t->x);	print(strm, ' ');
							print(strm, t->y);	print(strm, ' ');
							print(strm, t->z);	print(strm, '}');		break;
						}
	case SLIDER:	
						{	SliderNodeP s=SliderNodeP(nd);
							print(strm, s->wxstring());					break;
						}
	case GAUGE:	
						{	GaugeNodeP g=GaugeNodeP(nd);
							print(strm, g->wxstring());					break;
						}
	case INTCONTROL:
						{	IntControlNodeP i=IntControlNodeP(nd);
							print(strm, i->wxstring());					break;
						}
	case FLOATCONTROL:
						{	FloatControlNodeP f=FloatControlNodeP(nd);
							print(strm, f->wxstring());					break;
						}
	case BUTTON:	
						{	ButtonNodeP b=ButtonNodeP(nd);
							print(strm, b->wxstring());					break;
						}
	case TOGGLEBUTTON:	
						{	ToggleButtonNodeP t=ToggleButtonNodeP(nd);
							print(strm, t->wxstring());					break;
						}
	case CHECKBOX:	
						{	CheckBoxNodeP c=CheckBoxNodeP(nd);
							print(strm, c->wxstring());					break;
						}
	case RADIOBUTTON:	
						{	RadioButtonNodeP r=RadioButtonNodeP(nd);
							print(strm, r->wxstring());					break;
						}
	case STATICTEXT:	
						{	StaticTextNodeP s=StaticTextNodeP(nd);
							print(strm, s->wxstring());					break;
						}
	case TEXTCONTROL:
						{	TextControlNodeP t=TextControlNodeP(nd);
							print(strm, t->wxstring());					break;
						}
	case LISTBOX:
						{	ListBoxNodeP l=ListBoxNodeP(nd);
							print(strm, l->wxstring());					break;
						}
	case CHOICEBOX:
						{	ChoiceBoxNodeP c=ChoiceBoxNodeP(nd);
							print(strm, c->wxstring());					break;
						}
	case COMBOBOX:
						{	ComboBoxNodeP c=ComboBoxNodeP(nd);
							print(strm, c->wxstring());					break;
						}
	case LISTCONTROL:
						{	ListControlNodeP l=ListControlNodeP(nd);
							print(strm, l->wxstring());					break;
						}
	case FRAME:	
						{	FrameNodeP f=FrameNodeP(nd);
							print(strm, f->wxstring());					break;
						}
	case GRAPH:	
						{	GraphNodeP g=GraphNodeP(nd);
							print(strm, g->wxstring());					break;
						}
	case BOXSIZER:	
						{	BoxSizerNodeP b=BoxSizerNodeP(nd);
							print(strm, b->wxstring());					break;
						}
#ifdef _DEBUG
	case TODO:
	{	EvalCommand evco=ToDoNodeP(nd)->toDo.mnemonic;
		if (evco > 0  &&  evco <= end)
			print(strm, "%t", Evaluator::commandName[evco]);
		else print(strm, "unknown EvalCommand");
	}
	break;
#endif

	default:
		if (!nd->is.string)
			return;
		int wid, dots=0;
		StringP str = nd->string();	// Warning: This may create new nodes!
		charP cp = str->ptr();
		
		if (width < 0) 
			wid=str->length();
		else 
		{	wid= width < 10 ?   10  :  width;
			wid= wid < str->length() ?   wid  :  str->length();
		}
		if (wid < str->length()) 
			dots++;
		if ( ! str->is.backslashed)
		{	for (int i=0; i < wid; i++) 
				print(strm, *cp++);
		}
		else if (backslashes==false) 
			for (int i=0; i < wid; i++) 
				print(strm, backslash.decode(*cp++));
		else 
		{	int i;
			for (i=0; i < wid; i++) 
				if (backslash.is(cp[i])) 
					break;
			if (i < wid)		// word was in vbars
			{	if (strchr("\":", *cp)) 
				{	print(strm, *cp++);
					wid--;
				}
				print(strm, '|');
				for (i=0; i < wid; i++) 
					print(strm, backslash.decode(*cp++));
				print(strm, '|');
			}else
			{	for (i=0; i < wid; i++) 
				{	if (strchr(backslash.chars,(int)*cp)) 
						print(strm, '\\');
					print(strm, *cp++);
				}
			}
		}
		if (dots) print(strm, "...");
	}
}

void File::printBin(FILEP strm, NodeP nd) 
{
	if (!nd) 
		print(strm, int(0));
	else if (nd==Unbound) 
		print(strm, "Unbound");
	else if (nd->is.primitive) 
	{	if (nd->type == PRIM)
			print(strm, "PRIM");
		else if (nd->type == NTFREE)
			print(strm, "NTFREE");
	}
	else if (nd->is.list) 
		printBin(strm, PairP(nd));
	else switch (nd->type)
	{
	case INT_      : printBin(strm, IntP(nd)->int_);								break;
	case FLOAT_    : printBin(strm, FloatP(nd)->float_);							break;
	case COMPLEX   : printBin(strm, ComplexdP(nd)->complex_);					break;
	case INT16     : printBin(strm, Int16P(nd)->int_);								break;
	case INT8      : printBin(strm,  Int8P(nd)->int_);								break;
	case UINT8     : printBin(strm, UInt8P(nd)->int_);								break;
	case ARRAY     : printBin(strm, ArrayP(nd));										break;
	case INTARRAY  : printBin(strm, IntArrayP(nd));									break;
	case INT16ARRAY: printBin(strm, Int16ArrayP(nd));								break;
	case FLOATARRAY: printBin(strm, FloatArrayP(nd));								break;
	case QUOTE     : print(strm, '\"'); printBin(strm, PairP(nd)->first());	break;
	case OBJECT    : print(strm, ObjectP(nd)->name());								break;
	case FILE_     : print(strm, FileP(nd)->name());								break;
	case COLON     : print(strm, ':'); printBin(strm, PairP(nd)->first());	break;
	case TABLE     : printBin(strm, TableP(nd));										break;
	case STRUCT    : printBin(strm, StructP(nd));									break;

#ifdef _DEBUG
	case TODO:
	{	EvalCommand evco=ToDoNodeP(nd)->toDo.mnemonic;
		if (evco > 0  &&  evco <= end)
			print(strm, "%t", Evaluator::commandName[evco]);
		else print(strm, "unknown EvalCommand");
	}
	break;
#endif

	default:
		int wid, i;
		StringP str = nd->string();	// Warning: This may create new nodes!
		charP cp = str->ptr();
		
		wid=str->length();
		if ( ! str->is.backslashed)
		{	for (i=0; i < wid; i++) 
				print(strm, *cp++);
		}
		else if (backslashes==false) 
			for (i=0; i < wid; i++) 
				print(strm, backslash.decode(*cp++));
		else 
		{	for (i=0; i < wid; i++) 
				if (backslash.is(cp[i])) 
					break;
			if (i < wid)		// word was in vbars
			{	if (strchr("\":", *cp)) 
				{	print(strm, *cp++);
					wid--;
				}
				print(strm, '|');
				for (i=0; i < wid; i++) 
					print(strm, backslash.decode(*cp++));
				print(strm, '|');
			}else
			{	for (i=0; i < wid; i++) 
				{	if (strchr(backslash.chars,(int)*cp)) 
						print(strm, '\\');
					print(strm, *cp++);
				}
			}
		}
	}
}

void File::print(FILEP strm, NodeP nd) 
{
	print( strm, nd, 
		_PrintDepthLimit->find_limit(),
		_PrintWidthLimit->find_limit() );
}

void File::printHelper(FILEP strm, NodeP nd) 
{
	print( strm, PairP(nd), 
		_PrintDepthLimit->find_limit(),
		_PrintWidthLimit->find_limit() );
}

void File::printNoBracket(FILEP strm, NodeP nd) 
{
	if (nd->islist()) 
		printHelper(strm, nd);
	else 
		print(strm, nd);
}

void File::print(FILEP strm, PairPP arg) 
{
	int depth=_PrintDepthLimit->find_limit(),
		width=_PrintWidthLimit->find_limit();
	
	if (arg!=argStack) 
		print(strm, NodeP(*++arg), depth, width);

	while (arg!=argStack) 
	{	print(strm, ' ');
		print(strm, NodeP(*++arg), depth, width);
	}
}

void File::printStack(PairPP arg, bool space) 
{
	if (arg!=argStack) 
		if ((*++arg)->islist())
			printHelper(writer.file(), *arg);
		else
			print(writer.file(), *arg);

	while (arg!=argStack) 
	{	if (space) 
			writer << ' ';

		if ((*++arg)->islist())
			printHelper(writer.file(), *arg);
		else
			print(writer.file(), *arg);
	}
}

void File::printStackBin(PairPP arg) 
{
	if (arg!=argStack) 
		printBin(writer.file(), NodeP(*++arg));

	while (arg!=argStack) 
		printBin(writer.file(), NodeP(*++arg));
}

void _cdecl File::print(FILEP strm, const char* fmt, ...)	//VARARGS
{
	va_list ap;
	va_start(ap, fmt);
	char ch;
	while ((ch=*fmt++) !='\0') 
	{	if (ch=='%') 
		{	digits=0;
next_format_ch:
			ch=*fmt++;
			switch(ch)
			{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':digits=digits*10+ch-'0';
						goto next_format_ch;

			case 'p':{	NodeP nd=va_arg(ap, NodeP);	// print
							if (nd->islist())
								printHelper(strm, nd);
							else
								print(strm, nd);
							break;
						}
			case 's':print(strm, va_arg(ap, NodeP)); break;  // show			
			case 't':{	charP cp=va_arg(ap, charP);	// text
							while ((ch=*cp++) !='\0')
								print(strm, ch);
							break;
						}
			case 'i':print(strm, va_arg(ap, int));      break;  // int
			case 'f':print(strm, va_arg(ap, double));   break;  // double
#ifdef WIN32
			case 'c':print(strm, va_arg(ap, complexd)); break;  // complexd
#endif
			case 'x':printHex(strm, va_arg(ap, int));  break;  // int as hex
			default:	print(strm, '%'); print(strm, ch);
			}
		}else 
			print(strm, ch);
	}
	if (!strm) print(strm, '\0');
	va_end(ap);
}

File &File::operator<<(const   char c) {print(file(), c); return *this;}
File &File::operator<<(const    int i) {print(file(), i); return *this;}
File &File::operator<<(const double d) {print(file(), d); return *this;}
File &File::operator<<(const  NodeP n) {printNoBracket(file(), n); return *this;}
File &File::operator<<(const  LineP l) {printNoBracket(file(), NodeP(l)); return *this;}
File &File::operator<<(const QuoteP q) {print(file(), NodeP(q)); return *this;}
File &File::operator<<(const PairPP p) {print(file(), p); return *this;}
File &File::operator<<(const  char* s) {print(file(), s); return *this;}
File &File::operator<<(const   VarP var) 
	{print(file(), var->object); print(file(), var->val); return *this;}

NodeP lsetPrintPrecision(PairPP arg)
{
	int p=arg[1]->a_int(1);
	if (!throwing)
	{	if (p > Float::digits-10)
		{	error(BAD_DATA, arg[1]);
			return Unbound;
		}
		File::precision=p;
	}
	return Unbound;
}

NodeP lprint(PairPP arg) 
{
	File::printStack(arg, true);	writer << endL;
	return Unbound;
}

NodeP ltype(PairPP arg) 
{
	File::printStack(arg, 0);
	return Unbound;
}

NodeP ltypeBin(PairPP arg) 
{
	File::printStackBin(arg);
	return Unbound;
}

NodeP lshow(PairPP arg) 
{
	writer << arg << endL;
	return Unbound;
}
