//___________________________________________________________________
//
//                         Class Recorder
//
//                           Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/glcanvas.h>
#include <wx/txtstrm.h>
#include <wx/datstrm.h>
#include <wx/wfstream.h>
#include <wx/zipstrm.h>
#include <wx/ptr_scpd.h>

#include <GL/glu.h>
#include <stdio.h>
#include <math.h>

#include "Main.h"
#include "PtrToInt.h"
#include "IntToPtr.h"
#include "File.h"
#include "String.h"
#include "Pair.h"
#include "Int.h"
#include "UInt8.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Error.h"
#include "Init.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Font.h"
#include "Turtle.h"

#include "Recorder.h"
#include "Graph.h"
#include "LogoApp.h"
#include "Shadow.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//#define DEBUG_RECORDER_SAVE
//#define DEBUG_RECORDER_LOAD

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

wxDEFINE_SCOPED_PTR_TYPE(wxZipEntry);

//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, bool &b)
{
	b=istr.Read8()!=0;
	return istr;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, bool &b)
{
	b=istr.Read8()!=0;
	return istr;
}

wxTextOutputStream &operator<<(wxTextOutputStream &ostr, bool &b)
{
	ostr.Write8((unsigned char)(b));
	return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, bool &b)
{
	ostr.Write8((unsigned char)(b));
	return ostr;
}
//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, Point &p)
{
	return istr >> p.x >> p.y >> p.z;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, Point &p)
{
	return istr >> p.x >> p.y >> p.z;
}

wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Point &p)
{
	return ostr << p.x << ' ' << p.y << ' ' << p.z << ' ';
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Point &p)
{
	return ostr << p.x << p.y << p.z;
}
//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, Matrix &m)
{
	return istr >> m.x >> m.y >> m.z;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, Matrix &m)
{
	return istr >> m.x >> m.y >> m.z;
}
wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Matrix &m)
{
	return ostr << m.x << m.y << m.z;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Matrix &m)
{
	return ostr << m.x << m.y << m.z;
}
//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, Color &c)
{
	short r,g,b,a; 
	istr >> r >> g >> b >> a; 
	c.r = r & 0xff;
	c.g = g & 0xff;
	c.b = b & 0xff;
	c.a = a & 0xff;
	return istr;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, Color &c)
{
	return istr >> c.i;
}

wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Color &c)
{
	return ostr
		<< short(c.r) << ' ' 
		<< short(c.g) << ' ' 
		<< short(c.b) << ' ' 
		<< short(c.a) << ' ';
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Color &c)
{
	return ostr << c.i;
}

//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, Font_ &f)
{
	return istr >> f.sizex >> f.sizey 
		>> f.alignx >> f.aligny >> f.weight >> f.name; 
}
wxDataInputStream &operator>>(wxDataInputStream &istr, Font_ &f)
{
	return istr >> f.sizex >> f.sizey 
		>> f.alignx >> f.aligny >> f.weight >> f.name; 
}

wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Font_ &f)
{
	ostr 
		<< f.sizex << ' ' << f.sizey << ' ' 

		<< f.alignx << ' ' << f.aligny << ' ' 
		<< f.weight << ' ' << f.name << ' ';
	return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Font_ &f)
{
	ostr 
		<< f.sizex  << f.sizey
		<< f.alignx << f.aligny 
		<< f.weight << f.name;
	return ostr;
}
//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, Pen &p)
{
	istr >> p.down >> p.wide >> p.roundLineEnds
		>> (int&)p.internalMode >> (int&)p.mode
		>> p.size >> p.size_rotated
		>> p.color >> p.rawcolor >> p.floodcolor
		>> p.pattern; 
	return istr;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, Pen &p)
{
	istr >> p.down >> p.wide >> p.roundLineEnds
		>> (int&)p.internalMode >> (int&)p.mode
		>> p.size >> p.size_rotated
		>> p.color >> p.rawcolor >> p.floodcolor
		>> p.pattern; 
	return istr;
}

wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Pen &p)
{
	ostr << p.down << ' ' << p.wide << ' ' << p.roundLineEnds << ' '
		<< p.internalMode << ' ' << p.mode << ' ' 
		<< p.size << p.size_rotated
		<< p.color << p.rawcolor << p.floodcolor
		<< p.pattern << ' ';
	return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Pen &p)
{
	ostr << p.down << p.wide << p.roundLineEnds
		<< p.internalMode << p.mode
		<< p.size << p.size_rotated
		<< p.color << p.rawcolor << p.floodcolor
		<< p.pattern; 
	return ostr;
}
//___________________________________________________________________

wxTextInputStream &Turtle::read(wxTextInputStream &istr)
{
	istr >> (Point&)*this >> pen >> font 
		>> shown >> outOfBounds >> direct 
		>> inEraseMode >> cylinderLines 
		>> polyMode >> polyShown 
		>> tessMode >> tessShown
		>> surfaceMode
		>> (int&)mode >> heading
		>> polypt1 >> polypt2 >> texpt1 >> texpt2 >> texpos
		>> u >> hbase 
		>> tessWindingRule
		>> polyPointNr >> tessPointNr
		>> surfaceShown >> surfaceColumnNr
		>> size >> half_bottom 
		>> _sin >> _cos >> wrapDelta
		>> shoulder >> head >> polyv1;
//		>> refCount;
	return istr;
}
wxDataInputStream &Turtle::read(wxDataInputStream &istr)
{
	istr >> (Point&)*this >> pen >> font 
		>> shown >> outOfBounds >> direct 
		>> inEraseMode >> cylinderLines 
		>> polyMode >> polyShown 
		>> tessMode >> tessShown
		>> surfaceMode
		>> (int&)mode >> heading
		>> polypt1 >> polypt2 >> texpt1 >> texpt2 >> texpos
		>> u >> hbase 
		>> tessWindingRule
		>> polyPointNr >> tessPointNr
		>> surfaceShown >> surfaceColumnNr
		>> size >> half_bottom 
		>> _sin >> _cos >> wrapDelta
		>> shoulder >> head >> polyv1;
//		>> refCount;
	return istr;
}

wxTextOutputStream &Turtle::write(wxTextOutputStream &ostr)
{
#ifdef DEBUG_RECORDER_SAVE
#define D(x) #x << ' ' << x
	ostr << (Point&)*this << D(pen) << D(font) 
		<< D(shown) << ' ' << D(outOfBounds) << ' ' << D(direct) << ' ' 
		<< D(inEraseMode) << ' ' << D(cylinderLines) << ' ' 
		<< D(polyMode) << ' ' << D(polyShown) << ' ' 
		<< D(tessMode) << ' ' << D(tessShown) << ' '
		<< D(surfaceMode) << ' '
		<< D(mode) << ' ' << D(heading) << ' '
		<< D(polypt1) << D(polypt2) << D(texpt1) << D(texpt2) << D(texpos)
		<< D(u) << D(hbase) 
		<< D(tessWindingRule) << ' '
		<< D(polyPointNr) << ' ' << D(tessPointNr) << ' '
		<< D(surfaceShown) << ' ' << D(surfaceColumnNr) << ' '
		<< D(size) << ' ' << D(half_bottom) << ' ' 
		<< D(_sin) << ' ' << D(_cos) << ' ' << D(wrapDelta) << ' '
		<< D(shoulder) << D(head) << D(polyv1);
//		<< D(refCount) << ' ';
#else
	ostr << (Point&)*this << pen << font 
		<< shown << ' ' << outOfBounds << ' ' << direct << ' ' 
		<< inEraseMode << ' ' << cylinderLines << ' ' 
		<< polyMode << ' ' << polyShown << ' ' 
		<< tessMode << ' ' << tessShown << ' '
		<< surfaceMode << ' '
		<< mode << ' ' << heading << ' '
		<< polypt1 << polypt2 << texpt1 << texpt2 << texpos
		<< u << hbase 
		<< tessWindingRule << ' '
		<< polyPointNr << ' ' << tessPointNr << ' '
		<< surfaceShown << ' ' << surfaceColumnNr << ' '
		<< size << ' ' << half_bottom << ' ' 
		<< _sin << ' ' << _cos << ' ' << wrapDelta << ' '
		<< shoulder << head << polyv1;
//		<< refCount << ' ';
#endif
	return ostr;
}
wxDataOutputStream &Turtle::write(wxDataOutputStream &ostr)
{
	ostr << (Point&)*this << pen << font 
		<< shown << outOfBounds << direct 
		<< inEraseMode << cylinderLines 
		<< polyMode << polyShown 
		<< tessMode << tessShown
		<< surfaceMode
		<< mode << heading
		<< polypt1 << polypt2 << texpt1 << texpt2 << texpos
		<< u << hbase 
		<< tessWindingRule
		<< polyPointNr << tessPointNr
		<< surfaceShown << surfaceColumnNr
		<< size << half_bottom 
		<< _sin << _cos << wrapDelta
		<< shoulder << head << polyv1;
//		<< refCount;
	return ostr;
}
//___________________________________________________________________

wxTextInputStream &Turtle::readEssentials(wxTextInputStream &istr)
{
	istr >> (Point&)*this >> u >> heading >> shown;
	return istr;
}
wxDataInputStream &Turtle::readEssentials(wxDataInputStream &istr)
{
	istr >> (Point&)*this >> u >> heading >> shown;
	return istr;
}

wxTextOutputStream &Turtle::writeEssentials(wxTextOutputStream &ostr)
{
	ostr << (Point&)*this << u << heading << ' ' << shown << ' ';
	return ostr;
}
wxDataOutputStream &Turtle::writeEssentials(wxDataOutputStream &ostr)
{
	ostr << (Point&)*this << u << heading << shown;
	return ostr;
}
//___________________________________________________________________

const charP
	Command::penUpDownString[]={"Up", "Down"},

	Command::commandTypeString[]=  // for debugging
	{	"nop", 
		"Graphic", "EndGraphic", 
		"moveTo", "lineTo", "lineTo3d", 
		"setMatrix", "setMatrix4x4", "pushMatrix", "popMatrix",
		"setPixel",
		"PolyStart", "PolyPoint", "PolyEnd", 
		"glBegin", "TessPoint", 
		"QuadStripStart", "QuadStrip", "QuadStripEnd",
		"fill", "fillRect", "fillEllipse", 
		"Sphere", "Ellipsoid", "partialEllipsoid", "fillPie",
		"Perspective", "wrap", "mode",
		"setScreenColor", "Material", "Light", 
		"setPenColor", "setFloodColor", "setShadowColor", 
		"Shininess",
		"setPenDown",
		"setPolySmooth", "setLineSmooth", "setPointSmooth",
		"setDepthTest", "setLighting", "setDither", "setTextured",
		"setPenRLE", "setCyLines", "setShadows", "setPenMode",
		"setPenSize", "setPenPattern", "setDepthFunc", 
		"Texture", "TessWindingRule", 
		"Label", 
		"SetLightPos", 
		0
	};

wxTextInputStream &Command::read(wxTextInputStream &istr) 
{	return istr;
}
wxDataInputStream &Command::read(wxDataInputStream &istr) 
{	return istr;
}
wxTextOutputStream &Command::write(wxTextOutputStream &ostr) 
{	return ostr << type << endL;
}
wxDataOutputStream &Command::write(wxDataOutputStream &ostr) 
{	return ostr << type;
}
//___________________________________________________________________

wxTextInputStream &operator>>(wxTextInputStream &istr, CommandListIterator &i)
{	if (i) 
		return i->read(istr);
	else
		return istr;
}
wxDataInputStream &operator>>(wxDataInputStream &istr, CommandListIterator &i)
{	if (i) 
		return i->read(istr);
	else
		return istr;
}
wxTextOutputStream &operator<<(wxTextOutputStream &ostr, CommandListIterator &i)
{	if (i) 
		return i->write(ostr);
	else 
		return ostr;
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, CommandListIterator &i)
{	if (i) 
		return i->write(ostr);
	else 
		return ostr;
}
//___________________________________________________________________

const int Recorder::tableSize=1021;

PtrToInt *Recorder::ptrToInt;
IntToPtr *Recorder::intToPtr;

Recorder::Recorder(GraphP graph_) 
{
	refCount=1;
	recording=pendown=true; 
	inReplay=restartFlag=stop_play=stop_confirm=false; 
	action=0;
	lastMove=0;
	insave=inload=false;
	graph=graph_;
	screen=graph->screen;
	turtle=graph->turtle;
//	turtle->refCount++;
}

Recorder::~Recorder() 
{	
	clear();
/*	if (turtle->refCount > 0)
		turtle->refCount--;
	if (turtle->refCount==0)
	{
		turtle->destroy();
		delete turtle;
		turtle=0;
	}
*/
}

void Recorder::save(const char *name)
{
	wxFFileOutputStream output(name);
	wxZipOutputStream zip(output);
	zip.PutNextEntry(_T("graphics.dat"));
	wxDataOutputStream ostr(zip);
   tabs=0;
	insave=true;
	ptrToInt=new PtrToInt(tableSize);
	ostr << screen->eye << screen->center << screen->up;
	ostr << *this;
	graph->turtle->write(ostr);
	delete ptrToInt;
	insave=false;
}

void Recorder::load(const char *name)
{
	clear();
	wxFFileInputStream input(name);
	if (!input.Ok())
		return;
	
	wxZipInputStream zip(input);
	wxZipEntryPtr entry;
	if (entry.reset(zip.GetNextEntry()), entry.get() != 0)
	{	// access meta-data
		wxString name = entry->GetName();
		// read 'zip' to access the entry's data
		wxDataInputStream istr(zip);
		inload=true;
		intToPtr=new IntToPtr(tableSize);
		istr >> screen->eye >> screen->center >> screen->up;
		istr >> *this;
		graph->turtle->read(istr);
		delete intToPtr;
		inload=false;
	}
	playAgain();
}

void Recorder::saveText(const char *name)
{
	ptrToInt=new PtrToInt(tableSize);
	wxFFileOutputStream output(name);
	wxTextOutputStream ostr(output);
	ostr.SetDoubleFormat(wxT("%g"));
	tabs=0;
	for (int i=1; commandTypeString[i]; i++) 
		ostr << i << ' ' << commandTypeString[i] << endL;
	ostr << int(0) << endL;
	ostr << screen->eye << screen->center << screen->up;
	ostr << *this;
	graph->turtle->write(ostr);
	delete ptrToInt;
}

void Recorder::loadText(const char *name)
{
	intToPtr=new IntToPtr(tableSize);
	clear();
	wxFFileInputStream input(name);
	wxTextInputStream istr(input);
	int start=-1;
	while (!input.Eof() && start!=0)
	{	istr >> start;
		if (start)
			istr.ReadWord() << endL;
	}
	istr >> screen->eye >> screen->center >> screen->up;
	istr >> *this;
	graph->turtle->read(istr);
	delete intToPtr;
	playAgain();
}

wxTextOutputStream &operator <<(wxTextOutputStream &ostr, Recorder &recorder)
{	
	CommandListIterator i(&recorder);
	
	for (i=recorder.begin();  i;  i++)
	{
#ifdef DEBUG_RECORDER_SAVE
		for (int t=0; t<RECORDER.tabs; t++)
			ostr << '\t';
		ostr << Command::commandTypeString[i->type] << ' ';
#endif
		ostr << i;
/*		if (ostr.Fail()) 
		{	error(FILE_ERROR, new String(
				"File too big or something in Recorder::save"));
			return ostr;
		}
*/	}
#ifdef DEBUG_RECORDER_SAVE
	for (int t=0; t<RECORDER.tabs; t++)
		ostr << '\t';
	ostr << Command::commandTypeString[Command::ENDGRAPHIC] << ' ';
#endif
	ostr << Command::ENDGRAPHIC << endL;
	return ostr;
}

wxDataOutputStream &operator <<(wxDataOutputStream &ostr, Recorder &recorder)
{	
	CommandListIterator i(&recorder);
	
	for (i=recorder.begin();  i;  i++)
	{	ostr << i;
	}
	ostr << Command::ENDGRAPHIC;
	return ostr;
}

#define RECORDER_CASES \
case Command::GRAPHIC:			c=new R_Graphic;				break;\
case Command::ENDGRAPHIC:		break;									\
case Command::MOVETO:			c=new R_MoveTo;				break;\
case Command::LINETO:			c=new R_LineTo;				break;\
case Command::LINETO3D:			c=new R_LineTo3d;				break;\
case Command::SETMATRIX:		c=new R_SetMatrix;			break;\
case Command::SETMATRIX4X4:	c=new R_SetMatrix4x4;		break;\
case Command::PUSHMATRIX:		c=new R_PushMatrix;			break;\
case Command::POPMATRIX:		c=new R_PopMatrix;			break;\
case Command::SETPIXEL:			c=new R_SetPixel;				break;\
case Command::POLYSTART:		c=new R_PolyStart;			break;\
case Command::POLYPOINT:		c=new R_PolyPoint;			break;\
case Command::POLYEND:			c=new R_PolyEnd;				break;\
case Command::GLBEGIN:			c=new R_glBegin;				break;\
case Command::TESSPOINT:		c=new R_tessPoint;			break;\
case Command::QUADSTRIPSTART:	c=new R_QuadStripStart;		break;\
case Command::QUADSTRIP:		c=new R_QuadStrip;			break;\
case Command::QUADSTRIPEND:	c=new R_QuadStripEnd;		break;\
case Command::FILL:				c=new R_Fill;					break;\
case Command::FILLRECT:			c=new R_FillRect;				break;\
case Command::FILLELLIPSE:		c=new R_FillEllipse;			break;\
case Command::SPHERE:			c=new R_Sphere;				break;\
case Command::ELLIPSOID:		c=new R_Ellipsoid;			break;\
case Command::PARTIALELLIPSOID:	c=new R_partialEllipsoid;	break;\
case Command::FILLPIE:			c=new R_FillPie;				break;\
case Command::PERSPECTIVE:		c=new R_Perspective;			break;\
case Command::WRAP:				c=new R_wrap;					break;\
case Command::MODE:				c=new R_mode;					break;\
case Command::SETSCREENCOLOR:	c=new R_SetScreenColor;		break;\
case Command::MATERIAL:			c=new R_Material;				break;\
case Command::LIGHT:				c=new R_Light;					break;\
case Command::SETPENCOLOR:		c=new R_SetPenColor;			break;\
case Command::SETFLOODCOLOR:	c=new R_SetFloodColor;		break;\
case Command::SETSHADOWCOLOR:	c=new R_SetShadowColor;		break;\
case Command::SHININESS:		c=new R_Shininess;			break;\
case Command::SETPENDOWN: 		c=new R_SetPenDown;			break;\
case Command::SETPOLYSMOOTH:	c=new R_SetPolySmooth;		break;\
case Command::SETLINESMOOTH:	c=new R_SetLineSmooth;		break;\
case Command::SETPOINTSMOOTH:	c=new R_SetPointSmooth;		break;\
case Command::SETDEPTHTEST:	c=new R_SetDepthTest;		break;\
case Command::SETLIGHTING:		c=new R_SetLighting;			break;\
case Command::SETDITHER:		c=new R_SetDither;			break;\
case Command::SETTEXTURED:		c=new R_SetTextured;			break;\
case Command::SETPENRLE: 		c=new R_SetPenRLE;			break;\
case Command::SETCYLINES: 		c=new R_SetCyLines;			break;\
case Command::SETSHADOWS: 		c=new R_SetShadows;			break;\
case Command::SETPENMODE: 		c=new R_SetPenMode;			break;\
case Command::SETPENSIZE:		c=new R_SetPenSize;			break;\
case Command::SETPENPATTERN:	c=new R_SetPenPattern;		break;\
case Command::SETDEPTHFUNC:	c=new R_SetDepthFunc;		break;\
case Command::TEXTURE:			c=new R_Texture;				break;\
case Command::TESSWINDINGRULE:c=new R_TessWindingRule;	break;\
case Command::LABEL:				c=new R_Label;					break;\
case Command::SETLIGHTPOS:		c=new R_SetLightPos;			break;\

wxTextInputStream &operator >>(wxTextInputStream &istr, Recorder &recorder)
{	
	while (!istr.Eof())
	{	int i;
		istr >> i;
		if (!istr.CanRead()) 
		{	//error(FILE_ERROR, new String(
			//	"File bad format or some strange bug in Recorder::load"));
			recorder.playAgain();
			return istr;
		}
#ifdef DEBUG_RECORDER_LOAD
		for (int tab=0; tab<RECORDER.tabs; tab++)
			cout << '\t';
		if (i >=0 && i < Command::LASTCOMMAND)
			cout << Command::commandTypeString[i] << endL;
#endif
		Command::Type t=Command::Type(i);
		Command *c;
		switch(t)
		{
		RECORDER_CASES
		default: 
//			_pause(0);
					c=new Command;	// should never be reached if file format is correct!
					t=Command::NOP;
		}
		if (t==Command::ENDGRAPHIC)
			break;
		c->type=t;
		istr >> *c;
		recorder.push_back(c);
	}
	return istr;
}

wxDataInputStream &operator >>(wxDataInputStream &istr, Recorder &recorder)
{	
	while (!istr.Eof())
	{	int i;
		istr >> i;
		if (!istr.CanRead()) 
		{	//error(FILE_ERROR, new String(
			//	"File bad format or some strange bug in Recorder::load"));
			recorder.playAgain();
			return istr;
		}
#ifdef DEBUG_RECORDER_LOAD
		for (int tab=0; tab<RECORDER.tabs; tab++)
			cout << '\t';
		if (i >=0 && i < Command::LASTCOMMAND)
			cout << Command::commandTypeString[i] << endL;
#endif
		Command::Type t=Command::Type(i);
		Command *c;
		switch(t)
		{
		RECORDER_CASES
		default: 
//			_pause(0);
					c=new Command;	// should never be reached if file format is correct!
					t=Command::NOP;
		}
		if (t==Command::ENDGRAPHIC)
			break;
		c->type=t;
		istr >> *c;
		recorder.push_back(c);
	}
	return istr;
}

wxString Recorder::rgb(Color color)  // I hope to reimplement palettes when I find time
{
	std::stringstream sstr;
	sstr << color.r/255. << ' ' << color.g/255. << ' ' << color.b/255. << ' ';
	return sstr.str().c_str();
}

void Recorder::writePS(const char *name)
{
	wxFFileOutputStream output(name);
	wxTextOutputStream ostr(output);
	ostr.SetDoubleFormat(wxT("% 8.3f"));
	ostr<< "%!PS-Adobe-3.0 EPSF-3.0" << endL
		<< "%Title: " << name << endL
		<< "%%Creator: aUCBLogo 4.692" << endL
		<< "%%BoundingBox: " 
			<< (int)graph->turtle->screen->Min.x << ' '
			<< (int)graph->turtle->screen->Min.y << ' '
			<< (int)graph->turtle->screen->Max.x << ' '
			<< (int)graph->turtle->screen->Max.y << endL
		<< "%%EndComments" << endL
		<< "%%Page: " << name << endL
		<< "1 setlinecap 1 setlinejoin" << endL
		<< "/Bookman 24 selectfont" << endL
		<< endL
		<< "gsave" << endL
		<< rgb(graph->turtle->screen->color) << "setrgbcolor" << endL
		<< (int)graph->turtle->screen->Min.x << ' ' << (int)graph->turtle->screen->Min.y << " moveto" << endL
		<< (int)graph->turtle->screen->Max.x << ' ' << (int)graph->turtle->screen->Min.y << " lineto" << endL
		<< (int)graph->turtle->screen->Max.x << ' ' << (int)graph->turtle->screen->Max.y << " lineto" << endL
		<< (int)graph->turtle->screen->Min.x << ' ' << (int)graph->turtle->screen->Max.y << " lineto" << endL
		<< "   closepath fill" << endL

		<< "grestore" << endL
		<< endL;

	if (lastMove)
		ostr<< lastMove->point.x << ' ' << lastMove->point.y << " moveto" << endL;

	CommandListIterator i(this);

	for (i=begin();  i;  i++)
	{	i->writePS(ostr, *this);
/*			if (ostr.Fail()) 
		{	error(FILE_ERROR, new String(
				"writing failed in Recorder::writePS"));
			return Unbound;
		}
*/		}
	ostr<< "stroke" << endL
		<< "showpage" << endL
		<< "%%EOF" << endL;
}

DWORD WINAPI Recorder::playCommands(LPVOID lpv)
{
	Recorder &recorder=*(Recorder*) lpv;
	bool r=recorder.recording;
	recorder.recording=false;

	static Turtle turtle;
	turtle.shown=false;
	turtle.direct=false;	//comment out for debug

	turtle.create(&recorder, recorder.graph->screen);

	turtle.clearScreen();
	Point zero(0,0,0);
	turtle.moveTo(zero);

	Turtle *turtsav=recorder.turtle;
	recorder.turtle=&turtle;

	CommandListIterator i(&recorder);
	i=recorder.begin();
	while (i)
	{	i->play(recorder);
		++i;
		if (recorder.restartFlag)
		{	i=recorder.begin();
			recorder.restartFlag=false;
		}
		if (recorder.stop_play)
			break;
	}
	if (turtle.shadows)
	{
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
		ShadowObject *so=recorder.graph->shadowObject;
		so->castShadows(turtsav->lightPos, 1);
		glDisable(GL_CULL_FACE);
	}
	recorder.turtle=turtsav;
	turtle.setHeading(recorder.graph->turtle->heading);
	turtle.shown = recorder.graph->turtle->shown;
	turtle.draw();
	recorder.inReplay=false;
	turtle.direct = true;

	recorder.updateView(turtle);

	turtle.shown=false;
	turtle.destroy();

	recorder.recording=r;
	recorder.stop_confirm=true;
	return 0;
}

void Recorder::stop()
{
	if (!inReplay) return;
	stop_confirm=false;
	stop_play=true;
	while (!stop_confirm);
}

void Recorder::playAgain()
{
	if (inReplay) 
	{	restartFlag=true; return;
	}
	inReplay=true; // so this proc should not be entered twice 
					  // by multiple threads - when some exist - if there are some bugs
				    // - and they always exist - nothing serious is going to happen
	stop_play=false;

	playCommands(this);	// plain & correct, but freezing
//	CreateThread(0, 0, playCommands, 0, 0, 0);
		 // would be cooler, but is very buggy.
	   // I just didn`t like that freezing while I resized a window 
	  // containing a complex longish drawing.
	 // So some work had to be done on the Screen, Turtle & Recorder classes 
	// to make them reentrant.
}

void Recorder::updateView(Turtle &turtle)
{
	if (graph)
	{	wxPaintEvent pe;
		graph->OnPaint(pe);
	}
}
//___________________________________________________________________

R_Graphic::R_Graphic(Turtle &turtle, Recorder *recorder_, double scale_) 
{	type=GRAPHIC;
	recorder=recorder_;
	recorder->refCount++;
	recturtle=turtle;
	scale=scale_;
}
R_Graphic::~R_Graphic()
{	if (recorder && !--recorder->refCount)
	{	delete recorder;
		recorder=0;
	}
}
void R_Graphic::play(Recorder &rec) 
{
	bool rr=recorder->recording;
	recorder->recording=false;

	Turtle save=*rec.turtle;
	*rec.turtle=recturtle;
	glPushMatrix();
	rec.turtle->setTurtleMatrix();
	glScaled(scale, scale, scale);
	GLboolean normalized;
	normalized=glIsEnabled(GL_NORMALIZE);
	glEnable(GL_NORMALIZE);
	*rec.turtle=Point(0,0,0);
	rec.turtle->reset3d();
	rec.turtle->heading=0;

	recorder->screen=rec.screen;
	CommandListIterator i(recorder);
	i=recorder->begin();
	while (i)
	{	i->play(rec);
		++i;
	}
	if (!normalized)
		glDisable(GL_NORMALIZE);
	glPopMatrix();
	*rec.turtle=save;
	recorder->recording=rr;
}
wxTextInputStream  &R_Graphic::read(wxTextInputStream &istr) 
{	int nr;
	recturtle.readEssentials(istr);
	istr >> scale >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(graph);
		Recorder::intToPtr->insert(recorder);
#ifdef DEBUG_RECORDER_LOAD
		RECORDER.tabs++;
#endif
		istr >> *recorder;
#ifdef DEBUG_RECORDER_LOAD
		RECORDER.tabs--;
#endif
	}
	else
		recorder->refCount++;
	return istr;
}
wxDataInputStream  &R_Graphic::read(wxDataInputStream &istr) 
{	int nr;
	recturtle.readEssentials(istr);
	istr >> scale >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(graph);
		Recorder::intToPtr->insert(recorder);
#ifdef DEBUG_RECORDER_LOAD
		RECORDER.tabs++;
#endif
		istr >> *recorder;
#ifdef DEBUG_RECORDER_LOAD
		RECORDER.tabs--;
#endif
	}
	else
		recorder->refCount++;
	return istr;
}
void R_Graphic::setPtrs()
{
	CommandListIterator i(recorder);
	i=recorder->begin();
	while (i)
	{	i->setPtrs();
		i++;
	}
}
wxTextOutputStream &R_Graphic::write(wxTextOutputStream &ostr) 
{
	ostr << type << ' ';
	recturtle.writeEssentials(ostr);
	ostr << scale << ' ';
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
#ifdef DEBUG_RECORDER_SAVE
		RECORDER.tabs++;
#endif
		ostr << int(0) << endL 
			<< *recorder << endL;
#ifdef DEBUG_RECORDER_SAVE

		RECORDER.tabs--;
#endif
	}
	else
		ostr << nr << endL;
	return ostr;
}
wxDataOutputStream &R_Graphic::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	recturtle.writeEssentials(ostr);
	ostr << scale;
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
#ifdef DEBUG_RECORDER_SAVE
		RECORDER.tabs++;
#endif
		ostr << int(0) << *recorder;
#ifdef DEBUG_RECORDER_SAVE
		RECORDER.tabs--;
#endif
	}
	else
		ostr << nr;
	return ostr;
}
//___________________________________________________________________

R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder, 
						 Point point_) 
{	init(turtle, recorder, point_);
}
R_MoveTo::R_MoveTo(Turtle &turtle, Recorder &recorder, 
						 double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_MoveTo::play(Recorder &rec)
{	rec.turtle->moveTo(point);
	*rec.turtle=point;
}
wxTextInputStream  &R_MoveTo::read(wxTextInputStream &istr) 
{	return istr >> point;
}
wxDataInputStream  &R_MoveTo::read(wxDataInputStream &istr) 
{	return istr >> point;
}
wxTextOutputStream &R_MoveTo::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << point << endL;
}
wxDataOutputStream &R_MoveTo::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;
}
void R_MoveTo::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	ostr << point.x << ' ' << point.y << " moveto" << endL;
}
//___________________________________________________________________

R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder, 
						 Point point_) 
{	init(turtle, recorder, point_);
}
R_LineTo::R_LineTo(Turtle &turtle, Recorder &recorder, 
						 double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_LineTo::play(Recorder &rec) 
{	rec.turtle->u=u;
	rec.turtle->lineTo(point);	*rec.turtle=point;
}
wxTextInputStream  &R_LineTo::read(wxTextInputStream &istr) 
{	return istr >> point >> u;
}
wxDataInputStream  &R_LineTo::read(wxDataInputStream &istr) 
{	return istr >> point >> u;
}
wxTextOutputStream &R_LineTo::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << point << u << endL;
}
wxDataOutputStream &R_LineTo::write(wxDataOutputStream &ostr) 
{	return ostr << type << point << u;
}
void R_LineTo::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	if (recorder.pendown)
	{	ostr << point.x << ' ' << point.y << " lineto" << endL;
		recorder.lastMove=this;
		recorder.action++;
	}
}
//___________________________________________________________________

R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder) 
{	init(turtle, recorder, Point(turtle));
}
R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder, 
							  Point point_) 
{	init(turtle, recorder, point_);
}
R_LineTo3d::R_LineTo3d(Turtle &turtle, Recorder &recorder, 
							  double x, double y) 
{	init(turtle, recorder, Point(x,y,0));
}
void R_LineTo3d::init(Turtle &turtle, Recorder &recorder, 
							 Point point_)
{	type=LINETO3D;
	point=point_;
	u=turtle.u;
	recorder.lastMove=(R_MoveTo*)this;
}
void R_LineTo3d::play(Recorder &rec) 
{	rec.turtle->u=u;
	rec.turtle->lineTo(point);	*rec.turtle=point;
}
wxTextInputStream  &R_LineTo3d::read(wxTextInputStream &istr) 
{	return istr >> point >> u;
}
wxDataInputStream  &R_LineTo3d::read(wxDataInputStream &istr) 
{	return istr >> point >> u;
}
wxTextOutputStream &R_LineTo3d::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << point << u << endL;
}
wxDataOutputStream &R_LineTo3d::write(wxDataOutputStream &ostr) 
{	return ostr << type << point << u;
}
void R_LineTo3d::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	if (recorder.pendown)
	{	ostr << point.x << ' ' << point.y << " lineto" << endL;
		recorder.lastMove=this;
		recorder.action++;
	}
}
//___________________________________________________________________

void R_SetMatrix::play(Recorder &rec) 
{	rec.turtle->setMatrix(basex,basey,basez,trans);
}
wxTextInputStream  &R_SetMatrix::read(wxTextInputStream &istr) 
{	return istr >> basex >> basey >> basez >> trans;
}
wxDataInputStream  &R_SetMatrix::read(wxDataInputStream &istr) 
{	return istr >> basex >> basey >> basez >> trans;
}
wxTextOutputStream &R_SetMatrix::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' 
	<< basex << basey << basez << trans << endL;
}
wxDataOutputStream &R_SetMatrix::write(wxDataOutputStream &ostr) 
{	return ostr << type
	<< basex << basey << basez << trans;
}
void R_SetMatrix::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_SetMatrix4x4::play(Recorder &rec) 
{	glLoadMatrixd(m);
}
wxTextInputStream  &R_SetMatrix4x4::read(wxTextInputStream &istr) 
{	for (int i=0; i<16; i++)
		istr >> m[i];
	return istr;
}
wxDataInputStream  &R_SetMatrix4x4::read(wxDataInputStream &istr) 
{	for (int i=0; i<16; i++)
		istr >> m[i];
	return istr;
}
wxTextOutputStream &R_SetMatrix4x4::write(wxTextOutputStream &ostr) 
{	ostr << type << ' ';
	for (int i=0; i<16; i++)
		ostr << m[i] << ' ';
	ostr << endL;
	return ostr;
}
wxDataOutputStream &R_SetMatrix4x4::write(wxDataOutputStream &ostr) 
{	ostr << type;
	for (int i=0; i<16; i++)
		ostr << m[i];
	return ostr;
}
void R_SetMatrix4x4::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_PushMatrix::play(Recorder &rec) 
{	glPushMatrix();
}
void R_PushMatrix::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{} // still to do
//___________________________________________________________________

void R_PopMatrix::play(Recorder &rec) 
{	glPopMatrix();
}
void R_PopMatrix::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{} // still to do
//___________________________________________________________________

void R_SetPixel::play(Recorder &rec) 
{	rec.turtle->setPixel(point,color);
}
wxTextInputStream  &R_SetPixel::read(wxTextInputStream &istr) 
{	return istr >> point >> color;
}
wxDataInputStream  &R_SetPixel::read(wxDataInputStream &istr) 
{	return istr >> point >> color;
}
wxTextOutputStream &R_SetPixel::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << point << color << endL;
}
wxDataOutputStream &R_SetPixel::write(wxDataOutputStream &ostr) 
{	return ostr << type << point << color;
}
void R_SetPixel::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{} // still to do
//___________________________________________________________________

R_PolyStart::R_PolyStart(Turtle &turtle, Point &normal_, Point &p3_, Point &t3_)
{	type=POLYSTART; 
	normal=normal_;
	p1=turtle.polypt1;
	p2=turtle.polypt2;
	p3=p3_;
	t1=turtle.texpt1;
	t2=turtle.texpt2;
	t3=t3_;
}
void R_PolyStart::play(Recorder &rec) 
{
	glBegin(GL_POLYGON);
	glNormal3dv(doubleP(&normal));
	glTexCoord2dv(doubleP(&t1));
	glVertex3dv(doubleP(&p1));
	glTexCoord2dv(doubleP(&t2));
	glVertex3dv(doubleP(&p2));
	glTexCoord2dv(doubleP(&t3));
	glVertex3dv(doubleP(&p3));
}
wxTextInputStream  &R_PolyStart::read(wxTextInputStream &istr) 
{	return istr >> normal >> p1 >> p2 >> p3 >> t1 >> t2 >> t3;
}
wxDataInputStream  &R_PolyStart::read(wxDataInputStream &istr) 
{	return istr >> normal >> p1 >> p2 >> p3 >> t1 >> t2 >> t3;
}
wxTextOutputStream &R_PolyStart::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << normal
		<< p1 << p2 << p3
		<< t1 << t2 << t3
		<< endL;
}
wxDataOutputStream &R_PolyStart::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal
		<< p1 << p2 << p3
		<< t1 << t2 << t3;
}
void R_PolyStart::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________


void R_PolyPoint::play(Recorder &rec) 
{
	glTexCoord2dv(doubleP(&t));
	glVertex3dv(doubleP(&p));
	*rec.turtle=p;
}
wxTextInputStream  &R_PolyPoint::read(wxTextInputStream &istr) 
{	return istr >> p >> t;
}
wxDataInputStream  &R_PolyPoint::read(wxDataInputStream &istr) 
{	return istr >> p >> t;
}
wxTextOutputStream &R_PolyPoint::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << p << t << endL;
}
wxDataOutputStream &R_PolyPoint::write(wxDataOutputStream &ostr) 
{	return ostr << type << p << t;
}
void R_PolyPoint::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_PolyEnd::play(Recorder &rec) 
{
	glEnd();
}
void R_PolyEnd::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_glBegin::play(Recorder &rec) 
{
	glBegin(prim);
	glNormal3dv((double*)&normal);
}
wxTextInputStream  &R_glBegin::read(wxTextInputStream &istr) 
{	return istr >> prim >> normal;
}
wxDataInputStream  &R_glBegin::read(wxDataInputStream &istr) 
{	return istr >> prim >> normal;
}
wxTextOutputStream &R_glBegin::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << prim << ' ' << normal << endL;
}
wxDataOutputStream &R_glBegin::write(wxDataOutputStream &ostr) 
{	return ostr << type << prim << normal;
}
void R_glBegin::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_tessPoint::play(Recorder &rec) 
{
	glColor4ubv((GLubyte*)&c);
	glTexCoord2dv(doubleP(&t));
	glVertex3dv((double*)&p);
	*rec.turtle=p;
}
wxTextInputStream  &R_tessPoint::read(wxTextInputStream &istr) 
{	return istr >> p >> t >> c;
}
wxDataInputStream  &R_tessPoint::read(wxDataInputStream &istr) 
{	return istr >> p >> t >> c;
}
wxTextOutputStream &R_tessPoint::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << p << t << c << endL;
}
wxDataOutputStream &R_tessPoint::write(wxDataOutputStream &ostr) 
{	return ostr << type << p << t << c;
}
void R_tessPoint::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStripStart::play(Recorder &rec) 
{
	glBegin(GL_QUAD_STRIP);
	glNormal3dv(doubleP(&normal));
	glColor4ubv((unsigned char*)&c1);
	glTexCoord2dv(doubleP(&t1));
	glVertex3dv(doubleP(&p1));

	glNormal3dv(doubleP(&onormal));
	glColor4ubv((unsigned char*)&c2);
	glTexCoord2dv(doubleP(&t2));
	glVertex3dv(doubleP(&p2));
	
	glColor4ubv((unsigned char*)&c3);
	glTexCoord2dv(doubleP(&t3));
	glVertex3dv(doubleP(&p3));
	
	glColor4ubv((unsigned char*)&c4);
	glTexCoord2dv(doubleP(&t4));
	glVertex3dv(doubleP(&p4));
}
wxTextInputStream  &R_QuadStripStart::read(wxTextInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c1 >> p1 >> t1
		>> c2 >> p2 >> t2
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
wxDataInputStream  &R_QuadStripStart::read(wxDataInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c1 >> p1 >> t1
		>> c2 >> p2 >> t2
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
wxTextOutputStream &R_QuadStripStart::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << normal << onormal
		<< c1 << p1 << t1
		<< c2 << p2 << t2
		<< c3 << p3 << t3
		<< c4 << p4 << t4
		<< endL;
}
wxDataOutputStream &R_QuadStripStart::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal << onormal
		<< c1 << p1 << t1
		<< c2 << p2 << t2
		<< c3 << p3 << t3
		<< c4 << p4 << t4;
}
void R_QuadStripStart::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStrip::play(Recorder &rec) 
{
	glNormal3dv(doubleP(&normal));
	glColor4ubv((unsigned char*)&c3);
	glTexCoord2dv(doubleP(&t3));
	glVertex3dv(doubleP(&p3));

	glNormal3dv(doubleP(&onormal));
	glColor4ubv((unsigned char*)&c4);
	glTexCoord2dv(doubleP(&t4));
	glVertex3dv(doubleP(&p4));
}
wxTextInputStream  &R_QuadStrip::read(wxTextInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
wxDataInputStream  &R_QuadStrip::read(wxDataInputStream &istr) 
{	return istr >> normal >> onormal 
		>> c3 >> p3 >> t3
		>> c4 >> p4 >> t4;
}
wxTextOutputStream &R_QuadStrip::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << normal << onormal
		<< c3 << p3 << t3 
		<< c4 << p4 << t4 << endL;
}
wxDataOutputStream &R_QuadStrip::write(wxDataOutputStream &ostr) 
{	return ostr << type << normal << onormal
		<< c3 << p3 << t3 
		<< c4 << p4 << t4;
}
void R_QuadStrip::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{}
//___________________________________________________________________

void R_QuadStripEnd::play(Recorder &rec) 
{
	glEnd();
}
//___________________________________________________________________

wxTextInputStream  &R_bool::read(wxTextInputStream &istr) 
{	return istr >> on;
}
wxDataInputStream  &R_bool::read(wxDataInputStream &istr) 
{	return istr >> on;
}
wxTextOutputStream &R_bool::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << on << endL;
}
wxDataOutputStream &R_bool::write(wxDataOutputStream &ostr) 
{	return ostr << type << on;
}
//___________________________________________________________________

void R_Fill::play(Recorder &rec) 
{	rec.turtle->fill(on);
}
void R_Fill::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	ostr
		<< "closepath fill" << endL
		<< endL
		<< "newpath" << endL
		<< recorder.lastMove->point.x 
		<< recorder.lastMove->point.y << " moveto" << endL;
	recorder.action=0;
}
//___________________________________________________________________

R_FillRect::R_FillRect(Turtle &turtle, Point p1_, Point p2_)
{	type=FILLRECT; 
	p1=p1_;	
	p2=p2_;
	ori=turtle.orientation();
}
void R_FillRect::play(Recorder &rec) 
{	if (rec.turtle->screen->perspective)
		rec.turtle->setOrientation(ori);
	else
		rec.turtle->setHeading(ori.z);
	rec.turtle->fillRect(p1,p2);
}
wxTextInputStream  &R_FillRect::read(wxTextInputStream &istr) 
{	return istr >> p1 >> p2 >> ori;
}
wxDataInputStream  &R_FillRect::read(wxDataInputStream &istr) 
{	return istr >> p1 >> p2 >> ori;
}
wxTextOutputStream &R_FillRect::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << p1 << p2 << ori << endL;
}
wxDataOutputStream &R_FillRect::write(wxDataOutputStream &ostr) 
{	return ostr << type << p1 << p2 << ori;
}
void R_FillRect::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_FillEllipse::R_FillEllipse(Turtle &turtle, Point radius_)
{	type=FILLELLIPSE; 
	radius=radius_;
	ori=turtle.orientation();
}
void R_FillEllipse::play(Recorder &rec) 
{	if (rec.screen->perspective)
		rec.turtle->setOrientation(ori);
	else
		rec.turtle->setHeading(ori.z);
	rec.turtle->fillEllipse(radius);
}
wxTextInputStream  &R_FillEllipse::read(wxTextInputStream &istr) 
{	return istr >> radius >> ori;
}
wxDataInputStream  &R_FillEllipse::read(wxDataInputStream &istr) 
{	return istr >> radius >> ori;
}
wxTextOutputStream &R_FillEllipse::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << radius << ori << endL;
}
wxDataOutputStream &R_FillEllipse::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << ori;
}
void R_FillEllipse::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_Sphere::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->sphere(radius, slices, stacks);
}
wxTextInputStream  &R_Sphere::read(wxTextInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
wxDataInputStream  &R_Sphere::read(wxDataInputStream &istr) 
{	return istr >> radius >> slices >> stacks  >> ori;
}
wxTextOutputStream &R_Sphere::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< ' ' << slices << ' ' << stacks << ' ' << ori	<< endL;
}
wxDataOutputStream &R_Sphere::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << slices << stacks << ori;
}
void R_Sphere::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_Ellipsoid::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->ellipsoid(radius, slices, stacks);
}
wxTextInputStream  &R_Ellipsoid::read(wxTextInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
wxDataInputStream  &R_Ellipsoid::read(wxDataInputStream &istr) 
{	return istr >> radius >> slices >> stacks >> ori;
}
wxTextOutputStream &R_Ellipsoid::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< slices << ' ' << stacks << ' ' << ori << endL;
}
wxDataOutputStream &R_Ellipsoid::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius << slices << stacks << ori;
}
void R_Ellipsoid::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_partialEllipsoid::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->partialEllipsoid(radius, 
		slicesstart, slicesend, slices, 
		stacksstart, stacksend, stacks);
}
wxTextInputStream  &R_partialEllipsoid::read(wxTextInputStream &istr) 
{	return istr >> radius 
		>> slicesstart >> slicesend >> slices 
		>> stacksstart >> stacksend >> stacks >> ori;
}
wxDataInputStream  &R_partialEllipsoid::read(wxDataInputStream &istr) 
{	return istr >> radius 
		>> slicesstart >> slicesend >> slices 
		>> stacksstart >> stacksend >> stacks >> ori;
}
wxTextOutputStream &R_partialEllipsoid::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << radius 
		<< slicesstart << ' ' << slicesend << ' ' << slices << ' ' 
		<< stacksstart << ' ' << stacksend << ' ' << stacks << ' ' << ori << endL;
}
wxDataOutputStream &R_partialEllipsoid::write(wxDataOutputStream &ostr) 
{	return ostr << type << radius 
		<< slicesstart << slicesend << slices 
		<< stacksstart << stacksend << stacks << ori;
}
void R_partialEllipsoid::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

R_FillPie::R_FillPie(Turtle &turtle, double angle_, Point radius_)
{	type=FILLPIE; 
	heading=turtle.heading;
	angle=angle_;
	radius=radius_;
}
void R_FillPie::play(Recorder &rec) 
{	rec.turtle->setHeading(heading);
	rec.turtle->fillPie(angle, radius);
}
wxTextInputStream  &R_FillPie::read(wxTextInputStream &istr) 
{	return istr >> heading >> angle >> radius;
}
wxDataInputStream  &R_FillPie::read(wxDataInputStream &istr) 
{	return istr >> heading >> angle >> radius;
}
wxTextOutputStream &R_FillPie::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << heading << ' '
		<< angle << ' ' << radius << endL;
}
wxDataOutputStream &R_FillPie::write(wxDataOutputStream &ostr) 
{	return ostr << type << heading << angle << radius;
}
void R_FillPie::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_Perspective::play(Recorder &rec)
{
	rec.screen->perspective=true;
//	rec.graph->setFrustum(rec.screen->winsizex, rec.screen->winsizey);
	rec.graph->setFrustum(rec.graph->GetSize().x, rec.graph->GetSize().y);

	rec.screen->setEye();
	rec.screen->getMatrices();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

	rec.turtle->reset3d();
	rec.turtle->mode=Turtle::windowmode;
}
//___________________________________________________________________

R_SetEye::R_SetEye(Point &eye_, Point &center_, Point &up_):
	eye(eye_), center(center_), up(up_)
{	type=SETEYE;
}
void R_SetEye::play(Recorder &rec) 
{	rec.screen->eye=eye;
	rec.screen->center=center;
	rec.screen->up=up;
	rec.screen->setEye();
	rec.screen->getMatrices();
}
wxTextInputStream  &R_SetEye::read(wxTextInputStream &istr) 
{	return istr >> eye >> center >> up;
}
wxDataInputStream  &R_SetEye::read(wxDataInputStream &istr) 
{	return istr >> eye >> center >> up;
}
wxTextOutputStream &R_SetEye::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << eye << center << up << endL;
}
wxDataOutputStream &R_SetEye::write(wxDataOutputStream &ostr) 
{	return ostr << type << eye << center << up;
}
void R_SetEye::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	// still to do
}
//___________________________________________________________________

void R_wrap::play(Recorder &rec)
{	rec.turtle->wrap();
}
//___________________________________________________________________

R_mode::R_mode(Turtle &turtle)
{	type=MODE;
	mode=turtle.mode;
}

void R_mode::play(Recorder &rec)
{	rec.turtle->draw();
	rec.turtle->mode=mode;
	rec.turtle->draw();
}
wxTextInputStream  &R_mode::read(wxTextInputStream &istr) 
{	int &m=(int&)mode;
	return istr >> m;
}
wxDataInputStream  &R_mode::read(wxDataInputStream &istr) 
{	int &m=(int&)mode;
	return istr >> m;
}
wxTextOutputStream &R_mode::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << mode << endL;
}
wxDataOutputStream &R_mode::write(wxDataOutputStream &ostr) 
{	return ostr << type << mode;
}
//___________________________________________________________________

wxTextInputStream  &R_Color::read(wxTextInputStream &istr) 
{	return istr >> color;
}
wxDataInputStream  &R_Color::read(wxDataInputStream &istr) 

{	return istr >> color;
}
wxTextOutputStream &R_Color::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << color << endL;
}
wxDataOutputStream &R_Color::write(wxDataOutputStream &ostr) 
{	return ostr << type << color;
}
//___________________________________________________________________

void R_SetScreenColor::play(Recorder &rec)
{	rec.turtle->screen->color=color; 
	rec.turtle->setScreenColor(color);
}
//___________________________________________________________________

void R_Material::play(Recorder &rec)
{
	float cf[4]=
	{	float(color.r)/255,
		float(color.g)/255,
		float(color.b)/255,
		float(color.a)/255
	};
	glMaterialfv(GL_FRONT_AND_BACK, gltype, cf);
}
wxTextInputStream  &R_Material::read(wxTextInputStream &istr) 
{	return istr >> gltype >> color;
}
wxDataInputStream  &R_Material::read(wxDataInputStream &istr) 
{	return istr >> gltype >> color;
}
wxTextOutputStream &R_Material::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << gltype << ' ' << color << endL;
}
wxDataOutputStream &R_Material::write(wxDataOutputStream &ostr) 
{	return ostr << type << gltype << color;
}
//___________________________________________________________________

void R_Light::play(Recorder &rec)
{
	float cf[4]=
	{	float(color.r)/255,
		float(color.g)/255,
		float(color.b)/255,
		float(color.a)/255
	};
	glLightfv(GL_LIGHT0, gltype, cf);
}
wxTextInputStream  &R_Light::read(wxTextInputStream &istr) 
{	return istr >> gltype >> color;
}
wxDataInputStream  &R_Light::read(wxDataInputStream &istr) 
{	return istr >> gltype >> color;
}
wxTextOutputStream &R_Light::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << gltype << ' ' << color << endL;
}
wxDataOutputStream &R_Light::write(wxDataOutputStream &ostr) 
{	return ostr << type << gltype << color;
}
//___________________________________________________________________

R_SetPenColor::R_SetPenColor(Turtle &turtle) 
{	type=SETPENCOLOR; color=turtle.pen.color;
}
void R_SetPenColor::play(Recorder &rec) 
{	rec.turtle->setColor(color);
}
void R_SetPenColor::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << recorder.rgb(color) << "setrgbcolor" << endL;
}
//___________________________________________________________________

R_SetFloodColor::R_SetFloodColor(Turtle &turtle) 
{	type=SETFLOODCOLOR; color=turtle.pen.floodcolor;
}
void R_SetFloodColor::play(Recorder &rec) 
{	rec.turtle->pen.floodcolor=color;
}
void R_SetFloodColor::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << recorder.rgb(color) << "setfloodcolor" << endL;
}
//___________________________________________________________________

R_SetShadowColor::R_SetShadowColor(Turtle &turtle) 
{	type=SETSHADOWCOLOR; color=turtle.shadowcolor;
}
void R_SetShadowColor::play(Recorder &rec) 
{	rec.turtle->shadowcolor=color;
}
void R_SetShadowColor::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{
}
//___________________________________________________________________

void R_Shininess::play(Recorder &rec)
{
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, s);
}
wxTextInputStream  &R_Shininess::read(wxTextInputStream &istr) 
{	return istr >> s;
}
wxDataInputStream  &R_Shininess::read(wxDataInputStream &istr) 
{	return istr >> s;
}
wxTextOutputStream &R_Shininess::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << s << ' ' << endL;
}
wxDataOutputStream &R_Shininess::write(wxDataOutputStream &ostr) 
{	return ostr << type << s;
}
//___________________________________________________________________

R_SetPenDown::R_SetPenDown(Turtle &turtle) 
{	type=SETPENDOWN; on=turtle.pen.down;
}
void R_SetPenDown::play(Recorder &rec) 
{	rec.turtle->pen.down=on;
}
void R_SetPenDown::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{	recorder.pendown=on;
}
//___________________________________________________________________

void R_SetPolySmooth::play(Recorder &rec)
{	if (on)
		glEnable(GL_POLYGON_SMOOTH);
	else

		glDisable(GL_POLYGON_SMOOTH);
}
void R_SetPolySmooth::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetLineSmooth::play(Recorder &rec)
{	if (on)
		glEnable(GL_LINE_SMOOTH);
	else
		glDisable(GL_LINE_SMOOTH);
}
void R_SetLineSmooth::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetPointSmooth::play(Recorder &rec)
{	if (on)
		glEnable(GL_POINT_SMOOTH);
	else
		glDisable(GL_POINT_SMOOTH);
}
void R_SetPointSmooth::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetDepthTest::play(Recorder &rec)
{	if (on)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);
}
void R_SetDepthTest::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetLighting::play(Recorder &rec)
{	if (on)
		glEnable(GL_LIGHTING);
	else
		glDisable(GL_LIGHTING);
}
void R_SetLighting::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetDither::play(Recorder &rec)
{	if (on)
		glEnable(GL_DITHER);
	else
		glDisable(GL_DITHER);
}
void R_SetDither::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

void R_SetTextured::play(Recorder &rec)
{	if (on)
	{	glEnable(GL_TEXTURE_2D); rec.screen->textured=true; 

	}else
	{	glDisable(GL_TEXTURE_2D); rec.screen->textured=false; 
	}
}
void R_SetTextured::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_SetPenRLE::R_SetPenRLE(Turtle &turtle) 
{	type=SETPENRLE; on=turtle.pen.roundLineEnds;
}
void R_SetPenRLE::play(Recorder &rec) 
{	rec.turtle->pen.roundLineEnds=on;
}
void R_SetPenRLE::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_SetCyLines::R_SetCyLines(Turtle &turtle) 
{	type=SETCYLINES; on=turtle.cylinderLines;
}
void R_SetCyLines::play(Recorder &rec) 
{	rec.turtle->cylinderLines=on;
}
void R_SetCyLines::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_SetShadows::R_SetShadows(Turtle &turtle) 
{	type=SETSHADOWS; on=turtle.shadows;
}
void R_SetShadows::play(Recorder &rec) 
{	rec.turtle->shadows=on;
}
void R_SetShadows::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_SetFogEnabled::R_SetFogEnabled(bool fog) 
{	type=SETCYLINES; on=fog;
}
void R_SetFogEnabled::play(Recorder &rec) 
{	if (on)
		glEnable(GL_FOG);
	else
		glDisable(GL_FOG);
}
void R_SetFogEnabled::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{}
//___________________________________________________________________

R_SetPenMode::R_SetPenMode(Turtle &turtle) 
{	type=SETPENMODE; 
	penMode=turtle.pen.mode; 
	internalMode=turtle.pen.internalMode;
}
void R_SetPenMode::play(Recorder &rec) 
{	rec.turtle->setPenMode(penMode); 
	rec.turtle->pen.internalMode=internalMode;
}
wxTextInputStream  &R_SetPenMode::read(wxTextInputStream &istr) 
{	int &pm=(int&)penMode, &im=(int&)internalMode;
	return istr >> pm >> im;
}
wxDataInputStream  &R_SetPenMode::read(wxDataInputStream &istr) 
{	int &pm=(int&)penMode, &im=(int&)internalMode;
	return istr >> pm >> im;
}
wxTextOutputStream &R_SetPenMode::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' 
		<< penMode << ' ' << internalMode << endL;
}
wxDataOutputStream &R_SetPenMode::write(wxDataOutputStream &ostr) 
{	return ostr << type
		<< penMode << internalMode;
}
//___________________________________________________________________

R_SetPenSize::R_SetPenSize(Turtle &turtle) 
{	type=SETPENSIZE; penSize=turtle.pen.size;
}
void R_SetPenSize::play(Recorder &rec) 
{	rec.turtle->setPenSize(penSize);
}
wxTextInputStream  &R_SetPenSize::read(wxTextInputStream &istr) 
{	return istr >> penSize;
}
wxDataInputStream  &R_SetPenSize::read(wxDataInputStream &istr) 
{	return istr >> penSize;
}
wxTextOutputStream &R_SetPenSize::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << penSize << endL;
}
wxDataOutputStream &R_SetPenSize::write(wxDataOutputStream &ostr) 
{	return ostr << type << penSize;
}
void R_SetPenSize::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	if (recorder.action) 
	{	ostr<< "stroke" << endL
			<< recorder.lastMove->point.x 
			<< recorder.lastMove->point.y << " moveto" << endL;
		recorder.action=0;
	}
	ostr << penSize.x << " setlinewidth" << endL;
}
//___________________________________________________________________

R_SetPenPattern::R_SetPenPattern(Turtle &turtle) 
{	type=SETPENPATTERN; 
	penPattern=turtle.pen.getPattern();
}
void R_SetPenPattern::play(Recorder &rec) 
{	rec.turtle->pen.setPattern(penPattern);
}
wxTextInputStream  &R_SetPenPattern::read(wxTextInputStream &istr) 
{	istr >> penPattern;
	return istr;
}
wxDataInputStream  &R_SetPenPattern::read(wxDataInputStream &istr) 
{	istr >> penPattern;
	return istr;
}
wxTextOutputStream &R_SetPenPattern::write(wxTextOutputStream &ostr) 
{	ostr << type << ' ' << penPattern << endL;
	return ostr;
}
wxDataOutputStream &R_SetPenPattern::write(wxDataOutputStream &ostr) 
{	ostr << type << penPattern;
	return ostr;
}
//___________________________________________________________________

R_SetDepthFunc::R_SetDepthFunc(int func_) 
{	type=SETDEPTHFUNC; 
	func=func_;
}
void R_SetDepthFunc::play(Recorder &rec) 
{	glDepthFunc(GL_NEVER+func);
}
wxTextInputStream  &R_SetDepthFunc::read(wxTextInputStream &istr) 
{	istr >> func;
	return istr;
}
wxDataInputStream  &R_SetDepthFunc::read(wxDataInputStream &istr) 
{	istr >> func;
	return istr;
}
wxTextOutputStream &R_SetDepthFunc::write(wxTextOutputStream &ostr) 
{	ostr << type << ' ' << func << endL;
	return ostr;
}
wxDataOutputStream &R_SetDepthFunc::write(wxDataOutputStream &ostr) 
{	ostr << type << func;
	return ostr;
}
//___________________________________________________________________

int R_Texture::maxnr=0;
PairP R_Texture::textures=0;

R_Texture::R_Texture(TextureP tex_):
	x(tex_->_bitmap->winx), y(tex_->_bitmap->winy), 
	texi(tex_->_texnr), pixels(tex_->_bitmap->pixels),
	tex(tex_)
{	type=TEXTURE;
	set=true;
}
R_Texture::~R_Texture()
{
}
void R_Texture::play(Recorder &rec) 
{
	glBindTexture(GL_TEXTURE_2D, tex->_texnr);
	maxnr=max(maxnr, tex->_texnr);
	glEnable(GL_TEXTURE_2D);
	rec.screen->textured=true;
}
void R_Texture::setPtrs()
{
	if (!set)
	{	TextureP t = (Texture*)GC_::intToPtr->find(int(tex));
		tex=t;
		textures=new Pair(tex, textures);
		set=true;
	}
}
wxTextInputStream  &R_Texture::read(wxTextInputStream &istr) 
{	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	tex = (Texture*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	istr >> x >> y;
			int size=sizeof(Color)*x*y;
			pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	x=y=size=0;
				tex=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			BitmapP bm=new Bitmap(x,y, x,y, pixels);
			tex=new Texture(bm);
			textures=new Pair(tex, textures);
			set=true;
			RECORDER.intToPtr->insert(&*tex);
		}
	}
	else
	{	if (nr)
		{	tex = (Texture*)nr;
			set=false;
		}
		else
		{	istr >> (int&)tex;
			set=false;
		}
	}
	return istr;
}
wxDataInputStream  &R_Texture::read(wxDataInputStream &istr) 
{	int nr;
	istr >> nr;
	if (RECORDER.inload)
	{	if (nr)
		{	tex = (Texture*)RECORDER.intToPtr->find(nr);
			set=true;
		}
		else
		{	istr >> x >> y;
			int size=sizeof(Color)*x*y;
			pixels=ColorP(GC_::malloc(size));
			if (!pixels)
			{	x=y=size=0;
				tex=0;
				return istr;
			}
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				istr >> *c++;
			BitmapP bm=new Bitmap(x,y, x,y, pixels);
			tex=new Texture(bm);
			textures=new Pair(tex, textures);
			set=true;
			RECORDER.intToPtr->insert(&*tex);
		}
	}
	else
	{	if (nr)
		{	tex = (Texture*)nr;
			set=false;
		}
		else
		{	istr >> (int&)tex;
			set=false;
		}
	}
	return istr;
}
wxTextOutputStream &R_Texture::write(wxTextOutputStream &ostr) 
{	ostr << type << ' ';
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(tex);
		if (nr)
			ostr << nr << ' ';
		else
		{	RECORDER.ptrToInt->insert(&*tex);
			nr=RECORDER.ptrToInt->Nr();
			x=tex->_bitmap->winx;
			y=tex->_bitmap->winy;
			texi=tex->_texnr;
			pixels=tex->_bitmap->pixels;
			ostr << int(0) << ' ' << x << ' ' << y << ' ';
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*tex);
			tex->write(ostr);
		}		
	}	
	return ostr;
	return ostr;
}
wxDataOutputStream &R_Texture::write(wxDataOutputStream &ostr) 
{	ostr << type;
	if (RECORDER.insave)
	{	int nr = RECORDER.ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	RECORDER.ptrToInt->insert(&*tex);
			nr=RECORDER.ptrToInt->Nr();
			x=tex->_bitmap->winx;
			y=tex->_bitmap->winy;
			texi=tex->_texnr;
			pixels=tex->_bitmap->pixels;
			ostr << int(0) << x << y;
			ColorP c=pixels;
			for (int i=x*y; i; i--)
				ostr << *c++;
		}
	}
	else
	{	int nr = GC_::ptrToInt->find(tex);
		if (nr)
			ostr << nr;
		else
		{	ostr << int(0);
			GC_::ptrToInt->insert(&*tex);
			tex->write(ostr);
		}		
	}	
	return ostr;
}
//___________________________________________________________________

void R_TessWindingRule::play(Recorder &rec) 
{	rec.turtle->tessWindingRule=rule;
}
wxTextInputStream  &R_TessWindingRule::read(wxTextInputStream &istr) 
{	return istr >> rule;
}
wxDataInputStream  &R_TessWindingRule::read(wxDataInputStream &istr) 
{	return istr >> rule;
}
wxTextOutputStream &R_TessWindingRule::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << rule << endL;
}
wxDataOutputStream &R_TessWindingRule::write(wxDataOutputStream &ostr) 
{	return ostr << type << rule;
}
//___________________________________________________________________

R_Label::R_Label(Turtle &turtle, wxString str_) 
{	type=LABEL; str=str_; 
	ori=Point(
		turtle.roll(), 
		turtle.pitch(), 
		turtle.getHeading());
	font=turtle.font;
}
void R_Label::play(Recorder &rec) 
{	rec.turtle->setOrientation(ori);
	rec.turtle->font=font;
	rec.turtle->Label(str); rec.turtle->moveTo();
}
wxTextInputStream  &R_Label::read(wxTextInputStream &istr) 
{	return istr >> str >> ori >> font;
}
wxDataInputStream  &R_Label::read(wxDataInputStream &istr) 
{	return istr >> str >> ori >> font;
}
wxTextOutputStream &R_Label::write(wxTextOutputStream &ostr) 
{	ostr << type << ' ' << str << ' ' << ori << ' ' << font << endL;
	return ostr;
}
wxDataOutputStream &R_Label::write(wxDataOutputStream &ostr) 
{	return ostr << type << str << ori << font;
}
void R_Label::writePS(wxTextOutputStream &ostr, Recorder &recorder)
{	for (unsigned int i=0;  i < str.length();  i++)
	{	char ch=str[i];
		if (ch=='(' || ch==')' || ch=='\\')
			ostr << "\\";
		ostr << ch;

	}
	ostr<< ") show grestore" << endL
		<< recorder.lastMove->point.x 
		<< recorder.lastMove->point.y << " moveto" << endL;
}
//___________________________________________________________________

void R_SetLightPos::play(Recorder &rec) 
{	float p[4]={point.x,point.y,point.z,0};
	glLightfv(GL_LIGHT0, GL_POSITION, p);
}
wxTextInputStream  &R_SetLightPos::read(wxTextInputStream &istr) 
{	return istr >> point;
}
wxDataInputStream  &R_SetLightPos::read(wxDataInputStream &istr) 
{	return istr >> point;
}
wxTextOutputStream &R_SetLightPos::write(wxTextOutputStream &ostr) 
{	return ostr << type << ' ' << point << endL;
}
wxDataOutputStream &R_SetLightPos::write(wxDataOutputStream &ostr) 
{	return ostr << type << point;
}
void R_SetLightPos::writePS(wxTextOutputStream &ostr, Recorder &recorder) 
{} // still to do
