//___________________________________________________________________
//
//                           Class  Slider
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/slider.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Eval.h"
#include "Stack.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"

#include "Event.h"
#include "Slider.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Slider, wxSlider)
	EVT_SCROLL	(Slider::OnScroll	)
END_EVENT_TABLE()

int Slider::defaultStyle=wxSL_VERTICAL | wxSL_AUTOTICKS | wxSL_LABELS;

Slider::Slider(SliderNodeP sliderNode_, int style):
	sliderNode(sliderNode_)
{
	wxWindow *parent=console;
	if (sliderNode->parentFrame)
		parent=sliderNode->parentFrame->frame;
	else if (sliderNode->parentGraph)
		parent=sliderNode->parentGraph->graph;
	Create(parent, -1, 
			1,1,1, 


			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			"Slider");
}

Slider::Slider(SliderNodeP sliderNode_, 
			wxWindow* parent, wxWindowID id, 
			int value, int minValue, int maxValue, 
			PairP onScroll_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxSlider(parent, id, 
			value, minValue, maxValue, 
			pos, size, style, 
			validator, name),
onScroll(onScroll_),
sliderNode(sliderNode_)
{
	label = new wxStaticText(this, -1, name, 
		wxDefaultPosition, wxDefaultSize, 
		0, "staticText");
}

Slider::~Slider()
{
	if (sliderNode)
		sliderNode->slider=0;
}

void Slider::OnScroll(wxScrollEvent &event)
{
	Event::fifo << new EventSlider(this);
	event.Skip();
}

wxString Slider::wxstring()
{
	wxString str;
	str << "Slider " << GetMin() << ' ' << GetValue() << ' ' << GetMax() << ' ' 
		<< '[' << GetName() << "] ";
	if (onScroll)
		str << onScroll->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &Slider::read(wxDataInputStream &istr) 
{	
	int value, min, max;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> name >> (int&)onScroll
		>> pos.x >> pos.y >> size.x >> size.y;

	SetRange(min, max);
	SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	label=new wxStaticText(this, -1, name, 
		wxDefaultPosition, wxDefaultSize, 
		0, "staticText");
	return istr;
}
wxDataOutputStream &Slider::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetMin() << GetValue() << GetMax() << GetName() 
		<< GC_::ptrToInt->find(onScroll)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP SliderNode::onScroll()
{
	if (slider)
		return slider->onScroll;
	return 0;
}

void SliderNode::setonScroll(NodeP nd) 
{
	if (slider)
	{	slider->onScroll = PairP(nd);  check_oldyoung(nd);
	}
}

void SliderNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void SliderNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

SliderNode::SliderNode(): 
	Node(SLIDER), slider(0), parentFrame(0), parentGraph(0) {}

SliderNode::SliderNode(FrameNodeP parentFrame_, 
							  GraphNodeP parentGraph_): 
	Node(SLIDER), slider(0), 
	parentFrame(parentFrame_), 
	parentGraph(parentGraph_)  
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void SliderNode::free() 
{	
	if (slider) 
	{	EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tSlider:
				if (((EventSlider*)ev)->slider==slider)
					((EventSlider*)ev)->slider=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		slider->sliderNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		slider->Destroy(); 
		slider=0;
	}
}
void SliderNode::inc_gc() {fifo << onScroll();}
int SliderNode::youngs() {return gmark(onScroll());}
int SliderNode::youngsPtr() {return gmarkPtr(onScroll());}
void SliderNode::check_oldyoungs() {check_oldyoung(onScroll());}

wxDataInputStream &SliderNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		slider=new Slider(SliderNodeP(this), style);
		slider->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &SliderNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (slider)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph) 
			<< style;
		slider->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void SliderNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (slider)
	{	setonScroll((Pair*)GC_::intToPtr->find(int(slider->onScroll)));
	}
}

void SliderNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	SliderP s=slider, 
		s2=new Slider(SliderNodeP(this),
			p, -1, s->GetValue(), s->GetMin(), s->GetMax(),
			s->onScroll, s->GetPosition(), s->GetSize(), style, 
			wxDefaultValidator, s->GetName());
	slider=s2;
	setonScroll(s->onScroll);
	delete s;
	slider=s2;
}

StringP SliderNode::string()
{
	if (slider)
		return new String(slider->wxstring());
	return _Null_Word->name();
}

wxString SliderNode::wxstring()
{
	if (slider)
		return slider->wxstring();
	return wxEmptyString;
}

NodeP lSlider(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=0;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}else
		parentGraphNode=theApp->mainFrame->graphNode;
		
	wxString name=_T("Slider");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int minValue;
	{	int v=arg[3]->a_int(3);
		if throwing return Unbound;
		minValue=v;	
	}
	int value;
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		value=v;	
	}
	int maxValue;
	{	int v=arg[5]->a_int(5);
		if throwing return Unbound;
		maxValue=v;	
	}
	PairP onScroll=0;
	if (exist(6))
	{	PairP os=arg[6]->a_runnable(6);
		if throwing return Unbound;
		onScroll=os;
	}
	long style = Slider::defaultStyle;
	if (exist(7))
	{	int s=arg[7]->a_int(7);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(8))
	{	PairP p=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(9))
	{	PairP s=arg[9]->a_int_vector(9);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	SliderNodeP sliderNode=new SliderNode(
		parentFrameNode, 
		parentGraphNode);
	SliderP slider=new Slider(sliderNode,
			parent, id, value, minValue, maxValue, 
			onScroll, pos, size, style, wxDefaultValidator, name);
	sliderNode->slider=slider;
	sliderNode->style=style;
	return sliderNode;
}

NodeP lSliderValue(PairPP arg)
{
	if (exist(1))
	{	SliderNodeP slider=arg[1]->a_slider(1);
		if throwing return Unbound;
		return new Int(slider->slider->GetValue());
	}
	else
		return new Int(Event::sliderValue);
}

NodeP lSliderSetValue(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
	slider->slider->SetValue(value);
	return Unbound;
}

NodeP lSliderSetRange(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int min=arg[2]->a_int(2);
	if throwing return Unbound;
	int max=arg[3]->a_int(3);
	if throwing return Unbound;
	slider->slider->SetRange(min, max);
	return Unbound;
}

NodeP lSliderOnScroll(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	slider->setonScroll(function);
	return Unbound;
}

NodeP lSliderEnable(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	slider->slider->Enable(state!=0);
	return Unbound;
}

NodeP lSliderDestroy(PairPP arg)
{
	SliderNodeP slider=arg[1]->a_slider(1);
	if throwing return Unbound;
	{	FrameNodeP p=slider->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(slider));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=slider->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(slider));
			p->check_oldyoung(p->children);
		}
	}
	slider->slider->Destroy();
	slider->slider=0;
	return Unbound;
}
