//___________________________________________________________________
//
//                         TextControl Header
//___________________________________________________________________


extern Function 
	lTextControl, lTextControlDestroy,
	lTextControlValue, lTextControlSetValue, 
	lTextControlWrite, lTextControlAppend,
	lTextControlSetInsertionPointEnd,
	lTextControlCursor, lTextControlSetCursor,
	lTextControlInsertMode, lTextControlOverwriteMode,
	lTextControlSetBackgroundColor, lTextControlSetColor,
	lTextControlSetFontSize, lTextControlSetFontName,
	lTextControlSetFontStyle, lTextControlSetFontWeight,
	lTextControlOnChar, lTextControlOnKeyDown, lTextControlOnKeyUp,
	lTextControlOnChange, lTextControlOnEnter,
	lTextControlEnable;

class wxStaticText;
class MyTextCtrl;

class TextControl: public wxControl
{
	public:
		wxStaticText *label;
		MyTextCtrl *text;
		wxSize bestSize;
		wxString value;

		PairP onChange, onEnter,
			onChar, onKeyUp, onKeyDown;
		TextControlNodeP textControlNode;
		bool insertMode;
		unsigned int writeBufferSize;
		wxString writeBuffer;

		TextControl(TextControlNodeP textControlNode_, int style);
		TextControl(TextControlNodeP textControlNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString &value, 
			PairP onChange_=PairP(0),
			PairP onEnter_=PairP(0),
			const wxPoint& pos = wxDefaultPosition, 
			const wxSize& size = wxDefaultSize, 
			long style=0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = "TextControl");

		bool create(wxWindow* parent, wxWindowID id, 
			const wxString &value, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name);

		~TextControl();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		virtual void SetValue(wxString &value);
		virtual wxString GetValue() const;

		void 
			OnClose	(wxCloseEvent &event),
			OnChange	(wxCommandEvent &event),
			OnEnter	(wxCommandEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		void flush(), toggleInsertMode(), 
			setInsertMode(), setOverwriteMode();
		NodeP Cursor(), setCursor(PairP args);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef TextControl *TextControlP;

class TextControlNode: public Node
{
	public:
		TextControlP textControl;
		FrameNodeP parentFrame;
		GraphNodeP parentGraph;
		int style;

		static const TextControlNode aTextControlNode;

		operator_new;
		PairP onChange(), onEnter(), 
			onChar(), onKeyDown(), onKeyUp();
		void
			setonChange	(NodeP function),
			setonEnter	(NodeP function),
			setonChar	(NodeP function), 
			setonKeyDown(NodeP function),
			setonKeyUp	(NodeP function);

		TextControlNode();
		TextControlNode(FrameNodeP parentFrame_, GraphNodeP parentGraph_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();
		
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		StringP _typeof();

		void setParentFrame(FrameNodeP frame);
		void setParentGraph(GraphNodeP graph);
};
