//___________________________________________________________________
//
//                          Class Turtle
//___________________________________________________________________

class Font_;
class wxTextInputStream;
class wxDataInputStream;
class wxTextOutputStream;
class wxDataOutputStream;
class Recorder;

struct PointPointColor
{	Point p, t;
	Color c;
};

typedef PointPointColor *PointPointColorP;

class Screen_;

class Turtle: public Point
{
	public:
		enum Mode {wrapmode, fencemode, windowmode};

		Pen pen;
		Font_ font;
		bool shown, outOfBounds, direct, inEraseMode, 
			cylinderLines, 
			polyMode, polyShown, 
			tessMode, tessShown, surfaceMode, shadows;

		Mode mode;
		double heading;
		Point polypt1,polypt2, normal, texpt1, texpt2, texpos, lightPos;
		Color polyc1, polyc2, shadowcolor;
		Matrix u, hbase;
		Screen_ *screen;
		Recorder *recorder;
		unsigned int tessWindingRule;

//		int refCount;
		int polyPointNr, tessPointNr;
		GLUtesselator *tess;
		
		Turtle(bool makeOld=true);
		~Turtle();

		Turtle &operator=(const Turtle &b);
		Turtle &operator=(const Point &b);

		void 
			create(Recorder *recorder_, Screen_ *screen_), destroy(),

			clearScreen(), setScreenColor(Color c), plainXOrPen(),

			prepareToDraw(),
			
			setColor(Color c),
			setPenSize(Point size),
			setPenSize(double sizeX, double sizeY),
			setPenMode(Pen::WinMode newmode),

			penErase(), penDown(), penReverse(), restorePen(),

			reset3d(),
			moveTo(),
			moveTo(Point target), lineTo(Point target),
			moveTo(double x, double y), lineTo(double x, double y),
			polyStart(),
			polyEnd(),
			tessStart(),
			tessDraw(),
			tessContour(),
			tessEnd(),
			surfaceStart(),
			surfaceColumn(),
			surfaceColumnSh(),
			surfaceEnd(),

			pix(int w,int h, Color color),
			floodFillX(int x, int y, Color fill, Color rim),
			floodFillXrecursiv(int x, int y, Color fill, Color rim),

			floodFillX2(int x, int y, Color oldColor, Color newColor),
			floodFillX2recursiv(int x, int y, Color oldColor, Color newColor),

			fill(bool surface), 

			fillRect(Point a, Point b),
			fillEllipse(Point radius), 
			sphere(double radius, int slices=0, int stacks=0),
			ellipsoid(Point radius, int slices=0, int stacks=0),
			partialEllipsoid(Point radius, 
				double slicestart, double sliceend, int slices,
				double stackstart, double stackend, int stacks),
			partialEllipsoidSh(Point radius, 
									 double slicestart, double sliceend, int slices,
									 double stackstart, double stackend, int stacks),
			fillPie(double angle, Point radius),

			getInWindow(),
			draw(), show(), hide(),

			setHeading(double angle), 
			     right(double angle), 
			   upPitch(double angle), 
			 rightRoll(double angle),
			     spinX(double angle),
			     spinY(double angle),
			     spinZ(double angle),
			   setRoll(double angle),
			  setPitch(double angle),
			setOrientation(Point o),

			forward(double d),

			setPos(Point target),
			setPos(double x, double y, double z),
			setPos(NodeP xnode, NodeP ynode, NodeP znode),
			_setPos(Point target),
			_setPos(double x, double y, double z),
			_setPos(NodeP xnode, NodeP ynode, NodeP znode),
			setSpherePos(double r, double phi, double theta),
			setCylinderPos(double r, double phi, double y),
			home(),

			setMatrix(PairP basex,PairP basey,PairP basez,PairP trans),
			setMatrix(Point basex,Point basey,Point basez,Point trans),
			setTurtleMatrix(),
			wrap(), arc(double angle, double radius),
			EllipseArc(double angle, double minorRadius, 
							double majorRadius, double startAngle),
			Label(wxString str), 
			setLabelSize(double x, double y),
			setLabelFont(wxString face),
			setLabelWeight(int weight),
			setLabelAlign(int alignx, int aligny),
			setPixel(Point p, Color c);
		
		wxString getLabelFont();
		double towards_helper(Point end), distance(Point target),
			getHeading(), pitch(), roll();
		NodeP matrix(), getNodePenMode();

		Point getWindowPos();
		Point orientation();

		int Pixel();
		Point LabelSize(wxString str);

		void
			bitPaste(BitmapP b),
			scroll(int width, int height, double sx, double sy);

		BitmapP bitCopy(double w, double h);
		Point	scrollCalibrate(double sx, double sy);

		wxTextInputStream &readEssentials(wxTextInputStream &istr);
		wxDataInputStream &readEssentials(wxDataInputStream &istr);
		wxTextOutputStream &writeEssentials(wxTextOutputStream &ostr);
		wxDataOutputStream &writeEssentials(wxDataOutputStream &ostr);

		wxTextInputStream &read(wxTextInputStream &istr);
		wxDataInputStream &read(wxDataInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

	private:
		bool surfaceShown;
		int surfaceColumnNr;
		double size, half_bottom, _sin, _cos, wrapDelta;
		Point shoulder, head, polyv1;
		ColorFiFo 
			*colorFiFo, *colorFiFo2, 
			*colorFiFo00, *colorFiFo01, *colorFiFo02;
		PointFiFo 
			*surfaceFiFo, *surfaceFiFo2, *surfaceFiFo3, 
			*surfaceFiFo00, *surfaceFiFo01, *surfaceFiFo02,
			*tessFiFo;

		bool 
			wrapLeft (double d, Point end),
			wrapUp   (double d, Point end),
			wrapRight(double d, Point end),
			wrapDown (double d, Point end);

		void 
			sinCos(), drawRaw(), drawp(),
			checkMinX(), checkMaxX(), checkMinY(), checkMaxY();

		static int tess_nPoints, tess_prim;
		static Point tessNormal;
		static void
			__stdcall tessglBegin(GLenum prim),
			__stdcall tessglVertex(void *data),
			__stdcall tessglEnd(),
			__stdcall tessglCombine(PointP p, PointPointColorP d[4], 
							GLfloat w[4], void **out),
			__stdcall tessglError(GLenum errorCode);

	friend Function lTurtleMatrix, lsetTurtleMatrix;
};
